#include "objectCurve.h"
#include "QuaternionSet.h"
#include "VisibleForm.h"

//global declarations
MathObject	*mo2;
char		obj_qstr[128];
//static		GLUquadricObj *qobjCYL, *qobjSPH;
static		quaternion	*q0;
static		int flag_quat_firsttime;
static		int flip;
static		int t_pnts = 10;
static		int t_stps = 10;
static		GLuint sphere_disp_list;

/* Material structures */
static const GLfloat no_mat1[]					= {0.0f,0.0f,0.0f,1.0f};
static const GLfloat mat_ambient1[]				= {0.2f,0.2f,0.6f,1.0f};
static const GLfloat mat_ambient_color1[]		= {0.5f,0.5f,0.5f,1.0f};
static const GLfloat mat_diffuse1[]				= {0.95f,0.95f,0.1f,1.0f};
static const GLfloat mat_specular1[]			= {0.1f,0.1f,0.1f,1.0f};
static const GLfloat no_shininess1[]			= {0.0f};
static const GLfloat low_shininess1[]			= {1.0f};
static const GLfloat high_shininess1[]			= {10.0f};
static const GLfloat mat_emission1[]			= {0.3f,0.2f,0.2f,0.0f};

static const GLfloat no_mat2[]					= {0.0f,0.0f,0.0f,1.0f};
static const GLfloat mat_ambient2[]				= {0.8f,0.55f,0.75f,1.0f}; 
static const GLfloat mat_ambient_color2[]		= {0.5f,0.5f,0.5f,1.0f};
static const GLfloat mat_diffuse2[]				= {0.65f,0.2f,0.45f,1.0f};
static const GLfloat mat_specular2[]			= {0.5f,0.5f,0.5f,1.0f};
static const GLfloat no_shininess2[]			= {0.0f};
static const GLfloat low_shininess2[]			= {5.0f};
static const GLfloat high_shininess2[]			= {65.0f};
static const GLfloat mat_emission2[]			= {0.3f,0.2f,0.2f,0.0f};


static const GLfloat no_mat3[]					= {0.0f,0.0f,0.0f,1.0f};
static const GLfloat mat_ambient3[]				= {0.6f,0.2f,0.0f,1.0f}; 
static const GLfloat mat_ambient_color3[]		= {0.5f,0.5f,0.5f,1.0f};
static const GLfloat mat_diffuse3[]				= {0.9f,0.7f,0.1f,1.0f};
static const GLfloat mat_specular3[]			= {0.5f,0.5f,0.5f,1.0f};
static const GLfloat no_shininess3[]			= {0.0f};
static const GLfloat low_shininess3[]			= {5.0f};
static const GLfloat high_shininess3[]			= {65.0f};
static const GLfloat mat_emission3[]			= {0.3f,0.2f,0.2f,0.0f};

static const GLfloat no_mat4[]					= {0.0f,0.0f,0.0f,1.0f};
static const GLfloat mat_ambient4[]				= {0.2f,0.0f,0.6f,1.0f}; 
static const GLfloat mat_ambient_color4[]		= {0.5f,0.5f,0.5f,1.0f};
static const GLfloat mat_diffuse4[]				= {0.0f,0.95f,0.95f,1.0f};
static const GLfloat mat_specular4[]			= {0.5f,0.5f,0.5f,1.0f};
static const GLfloat no_shininess4[]			= {0.0f};
static const GLfloat low_shininess4[]			= {5.0f};
static const GLfloat high_shininess4[]			= {65.0f};
static const GLfloat mat_emission4[]			= {0.3f,0.2f,0.2f,0.0f};

//### INITIALIZATIONS
void initializeApplication 						( MathObject *mathobjRef )
{
//	int	i =	0;
    q0							= (quaternion*) my_malloc(sizeof(quaternion));
    flag_quat_firsttime			= 1;
    flip						= 0;

	mo2	= mathobjRef;							//global: object reference
	mo2->whichCurve				= OBJKNOT;	//global: curve id
    mo2->whoCommandsMainWin		= MAIN_WINDOW;	//global: window id
    
    mo2->flag_quat_movingquat	= 0;			
    mo2->flag_multi_lines       = 0;            //global: single or multiple lines
    mo2->flag_multi_lines_first_time = 1;

	//@ GLOBAL VIEWING OPTIONS
    mo2->showAxes				= 0;			//object: axes
	mo2->showCurve				= 1;			//object: curve    
	if(mo2->objectID==OBJSURFACE)				//object: surface
		mo2->showEdge			= 1;
	else
		mo2->showEdge			= 0;
	mo2->showFace				= 1;			//object: face    
	if(mo2->objectID==OBJSURFACE)				//object: frames
		mo2->showFrame			= 1;
	else
		mo2->showFrame			= 0;
    mo2->showLabels             = 0;			//object: labels
	if(mo2->objectID==OBJSURFACE)				//object: normals
		mo2->showNormals		= 1;
	else
		mo2->showNormals		= 0;	
	mo2->showOnScreenHelp		= 1;			//on screen text
	mo2->showPalette			= 0;			//palette
    mo2->tubeThickness			= 0.06;			//object: tube thickness
    mo2->qmaptubeThickness      = 0.025;			//quatmap: tube thickness
	mo2->showTexture			= 0;			//object: texture
	mo2->showColorCoded			= 0;
    mo2->showTSpace				= 0;			//tangent space
    mo2->whichFrame				= FRENETSERRET;	//global: frame id
    mo2->selectFrame_FS			= 1;			//fs frame flag
    mo2->selectFrame_PT			= 0;			//pt frame flag
    mo2->selectFrame_GR			= 0;			//gr frame flag
    mo2->qfont					= GLUT_BITMAP_HELVETICA_10; //global font

	//@ QUATERNION VIEWING OPTIONS
    mo2->numOfRings             = 4;			//quatmap: # of quaternion rings
	mo2->qshowAxes				= 0;			//quatmap: axes flag
    mo2->qshowCurve             = 1;			//quatmap: curve flag
    mo2->qshowFace				= 1;			//quatmap: face flag
	if(mo2->objectID==OBJCURVE)
	    mo2->qshowEdge			= 0;
	else 	if(mo2->objectID==OBJSURFACE)
		mo2->qshowEdge			= 1;
	mo2->qshowFrame				= 0;			//quatmap: frame flag
	if(mo2->objectID==OBJCURVE)
		mo2->qshowNormals		= 0;
	else 	if(mo2->objectID==OBJSURFACE)
		mo2->qshowNormals		= 1;
	mo2->qshowOptimPoints		= 0;
    mo2->qshowPoints			= 1;			//quatmap: curve-points flag
    mo2->qshowSphere			= 0;			//quatmap: unit-sphere flag
    mo2->qwhichQuatComponent	= SHOWXYZ;		//quatmap: quaternion-component
    mo2->qshowReflectedPoints	= 0;			//quatmap: reflect-map flag
    mo2->qshowQuatRings			= 0;			//quatmap: quaternion-rings flag
    mo2->trackingRing			= -1;			//quatmap: track ring flag

	//@ RINGS
    mo2->startTheta             = 0.0;			//quatmap::rings: pt-start angle
	mo2->rshowCurve				= 1;			//quatmap::rings: curve flag
	mo2->rshowFace				= 1;			//quatmap::rings: face flag
	mo2->rshowEdge				= 0;			//quatmap::rings: edge flag
	mo2->rshowSurfaceNormals	= 0;			//quatmap: surface-normal flag
	mo2->rshowSurfacePoints		= 0;
	mo2->rquatRingAngle			= 0.0;			//quatmap::rings: ?
	mo2->rtrackQuatAngle		= 0.0;			//quatmap::rings: ?
	mo2->rshowColorMode			= CODE_SOLID_COLOR; //quatmap: surface color-mode

	//@ CURVE params
    mo2->showControlPoints		= 1;			//editable-curve: ctrl-points flag
    mo2->whichSpline			= CATROM;		//editable-curve: spline-type flag
    mo2->param_spline_increment = 0.02;			//editable-curve: spline-tesselation

	//@ KNOT params 
    mo2->showGap				= 0;			//torus-knot: gap-at-end flag
    mo2->m						= 2;			//torus-knot: param m
    mo2->n						= 3;			//torus-knot: param n
    mo2->a						= 0.30;			//torus-knot: param a
    mo2->b						= 0.35;			//torus-knot: param b
    mo2->r						= 0.80;			//torus-knot: param r
    mo2->lock_a_b				= 0;			//torus-knot: lock a-b flag
    mo2->flag_update_track_ring = 1;			//torus-knot: update ring flag

	//@ SURFACE params
	mo2->u						= 20;			//surface-patch: param u
	mo2->v						= 20;			//surface-patch: param v
	mo2->param_a				= 0.35;			//surface-patch: param a
	mo2->param_b				= 0.35;			//surface-patch: param b
	mo2->param_r				= 0.8;			//surface-patch: param r
	mo2->surfObjectID			= 0;			//surface-patch: surface-type
	mo2->surfNumObjects			= 3;			//surface-patch: global # of objects

	//@ OPTIMIZATION STUFF
	mo2->curveQuatAngle_PT		= 0.0;
	mo2->curveQuatAngle			= 0.0;
	mo2->curveQuatDistance		= 0.0;
	mo2->curveQuatAngle_CNT		= 0.0;
	mo2->curveQuatDistance_CNT	= 0.0;
	mo2->opt_EPSILON			= 0.001;
	mo2->opt_ERROR				= 0.0;
	mo2->opt_INDEX				= 1;
	mo2->updatedTHETA			= 0.0;
	mo2->flag_opt_first_time	= 1;
	mo2->flag_opt_only_once		= 1;
	mo2->flag_opt_continue		= 1;
	mo2->index_q0				= mo2->opt_INDEX-1;
	mo2->index_q2				= mo2->opt_INDEX+1;
	mo2->fixedindex				= 0;

    mo2->pt_flag_first_time     = 1;

	//@ COLORS
    mo2->fgcolor[0]				= 0.9;
    mo2->fgcolor[1]				= 0.65;
    mo2->fgcolor[2]				= 0.1;
    /* set light matrices (ambient, diffuse, specular) */
    mo2->mata[0] = 0.2; mo2->mata[1] = 0.2; mo2->mata[2] = 0.2; mo2->mata[3] = 0.2;
    mo2->matd[0] = 0.0; mo2->matd[1] = 0.7; mo2->matd[2] = 0.3; mo2->matd[3] = 1.0;
    mo2->mats[0] = 1.0; mo2->mats[1] = 1.0; mo2->mats[2] = 1.0; mo2->mats[3] = 1.0;
	mo2->shininess = 30.0;
	mo2->specular  = 0.1;

	//@ TEXTURE parameters
	mo2->texParam				= 1.0;

	//@ BOOLEAN FLAGS
	mo2->flag_keypress_ALT		= 0;
	mo2->flag_keypress_CTRL		= 0;
	mo2->flag_keypress_SHIFT	= 0;

	//@ PALETTE
    mo2->showColorCoded         = CODE_SOLID_COLOR;
	mo2->whichPalette			= 0.0;
	mo2->scale = mo2->obj_scale = 1.0;

}

//*********************************************//
//
//*********************************************//
void initCurve 									( void )
{
    int	i=0;
	int	j=0;
	//int	k=0;
	//@ init all lists
    for(i=0; i<MAX_POINTS; i++)	{
		
        mo2->numPoints                  = 0;
        mo2->numCtrlPoints              = 0;

        mo2->pntList[i].p3.d[0]         = 0.0;
        mo2->pntList[i].p3.d[1]         = 0.0;
        mo2->pntList[i].p3.d[2]         = 0.0;

        mo2->ctrlPntList[i].p3.d[0]     = 0.0;
        mo2->ctrlPntList[i].p3.d[1]     = 0.0;
        mo2->ctrlPntList[i].p3.d[2]     = 0.0;

        mo2->TangList[i].p3.d[0]        = 0.0;
        mo2->TangList[i].p3.d[1]        = 0.0;
        mo2->TangList[i].p3.d[2]        = 0.0;

        mo2->qTangList[i].p3.d[0]       = 0.0;
        mo2->qTangList[i].p3.d[1]       = 0.0;
        mo2->qTangList[i].p3.d[2]       = 0.0;

        mo2->qdTangList[i].p3.d[0]        = 0.0;
        mo2->qdTangList[i].p3.d[1]        = 0.0;
        mo2->qdTangList[i].p3.d[2]        = 0.0;

        mo2->qdTangList[i].p3.d[0]       = 0.0;
        mo2->qdTangList[i].p3.d[1]       = 0.0;
        mo2->qdTangList[i].p3.d[2]       = 0.0;

		mo2->basisRList[i].p3.d[0]      = 0.0;
        mo2->basisRList[i].p3.d[1]      = 0.0;
        mo2->basisRList[i].p3.d[2]      = 0.0;

        mo2->basisSList[i].p3.d[0]      = 0.0;
        mo2->basisSList[i].p3.d[1]      = 0.0;
        mo2->basisSList[i].p3.d[2]      = 0.0;

        mo2->refquatList[i].r =			0.0;
        mo2->refquatList[i].i =			0.0;
        mo2->refquatList[i].j =			0.0;
        mo2->refquatList[i].k =			0.0;

        mo2->ptQuatList[i].r =			0.0;
        mo2->ptQuatList[i].i =			0.0;
        mo2->ptQuatList[i].j =			0.0;
        mo2->ptQuatList[i].k =			0.0;

        mo2->fsQuatList[i].r =			0.0;
        mo2->fsQuatList[i].i =			0.0;
        mo2->fsQuatList[i].j =			0.0;
        mo2->fsQuatList[i].k =			0.0;

        mo2->grQuatList[i].r =			0.0;
        mo2->grQuatList[i].i =			0.0;
        mo2->grQuatList[i].j =			0.0;
        mo2->grQuatList[i].k =			0.0;

//		mo2->opQuatList[i].r =			1.0;
//      mo2->opQuatList[i].i =			0.0;
//      mo2->opQuatList[i].j =			0.0;
//      mo2->opQuatList[i].k =			0.0;

//		mo2->opupdQuatList[i].r =		1.0;
//      mo2->opupdQuatList[i].i =		0.0;
//      mo2->opupdQuatList[i].j =		0.0;
//      mo2->opupdQuatList[i].k =		0.0;

		mo2->ptRBasis[i].p3.d[0] =		0.0;
        mo2->ptRBasis[i].p3.d[1] =		0.0;
        mo2->ptRBasis[i].p3.d[2] =		0.0;

        mo2->ptSBasis[i].p3.d[0] =		0.0;
        mo2->ptSBasis[i].p3.d[1] =		0.0;
        mo2->ptSBasis[i].p3.d[2] =		0.0;

        mo2->fsRBasis[i].p3.d[0] =		0.0;
        mo2->fsRBasis[i].p3.d[1] =		0.0;
        mo2->fsRBasis[i].p3.d[2] =		0.0;

        mo2->fsSBasis[i].p3.d[0] =		0.0;
        mo2->fsSBasis[i].p3.d[1] =		0.0;
        mo2->fsSBasis[i].p3.d[2] =		0.0;

        mo2->grRBasis[i].p3.d[0] =		0.0;
        mo2->grRBasis[i].p3.d[1] =		0.0;
        mo2->grRBasis[i].p3.d[2] =		0.0;

        mo2->grSBasis[i].p3.d[0] =		0.0;
        mo2->grSBasis[i].p3.d[1] =		0.0;
        mo2->grSBasis[i].p3.d[2] =		0.0;

    }

	if(mo2->objectID==OBJSURFACE) {

		for(i=0; i<(mo2->TORUS_PNTS); i++) {
			for(j=0; j<(mo2->TORUS_STPS); j++) {
				mo2->tpnts[i][j][0] = 0.0;
				mo2->tpnts[i][j][1] = 0.0;
				mo2->tpnts[i][j][2] = 0.0;

				mo2->upnts[i][j][0] = 0.0;
				mo2->upnts[i][j][1] = 0.0;
				mo2->upnts[i][j][2] = 0.0;
			}
		}
	}

}
//*********************************************//
//
//*********************************************//
void initKnot 									( void )
{
    mo2->m = 2;
    mo2->n = 3;
    mo2->a = 0.3;
    mo2->b = 0.5;
    mo2->r = 0.8;
    mo2->increment = 0.05;
}
//*********************************************//
// MANAGE Data structure
//*********************************************//
void copyQuatList								( void )	
{
    int	i =	0;
    int limit = 0;

    if(mo2->flag_curve_closed)
        limit = mo2->numPoints;
    else
        limit = mo2->numPoints-2;

    for(i=0; i<=limit; i++) {
        if(mo2->whichFrame == FRENETSERRET) {
            mo2->refquatList[i].r = -mo2->fsQuatList[i].r;
            mo2->refquatList[i].i = -mo2->fsQuatList[i].i;
            mo2->refquatList[i].j = -mo2->fsQuatList[i].j;
            mo2->refquatList[i].k = -mo2->fsQuatList[i].k;
        }
        else if(mo2->whichFrame == PARALLELTRANSPORT) {
            mo2->refquatList[i].r = -mo2->ptQuatList[i].r;
            mo2->refquatList[i].i = -mo2->ptQuatList[i].i;
            mo2->refquatList[i].j = -mo2->ptQuatList[i].j;
            mo2->refquatList[i].k = -mo2->ptQuatList[i].k;
        }
        else if(mo2->whichFrame == GEODESICREFERENCE) {
            mo2->refquatList[i].r = -mo2->grQuatList[i].r;
            mo2->refquatList[i].i = -mo2->grQuatList[i].i;
            mo2->refquatList[i].j = -mo2->grQuatList[i].j;
            mo2->refquatList[i].k = -mo2->grQuatList[i].k;
        }
    }
}
void copyQuatList_OPT							( void ) 
{
    int	i =	0;
    int limit = 0;
    //double angle = 0.0;

    if(mo2->flag_curve_closed)
        limit = mo2->numPoints;
    else
        limit = mo2->numPoints-2;

    for(i=0; i<=limit; i++) {
        if(mo2->whichFrame == FRENETSERRET) {
            mo2->opQuatList[i].r = mo2->fsQuatList[i].r;
            mo2->opQuatList[i].i = mo2->fsQuatList[i].i;
            mo2->opQuatList[i].j = mo2->fsQuatList[i].j;
            mo2->opQuatList[i].k = mo2->fsQuatList[i].k;
            //findQuat_Current( i, &(mo2->opQuatList[i]), &angle );
            //mo2->opQuatList[i].theta = angle;
        }
        else if(mo2->whichFrame == PARALLELTRANSPORT) {
            mo2->opQuatList[i].r = mo2->ptQuatList[i].r;
            mo2->opQuatList[i].i = mo2->ptQuatList[i].i;
            mo2->opQuatList[i].j = mo2->ptQuatList[i].j;
            mo2->opQuatList[i].k = mo2->ptQuatList[i].k;
            //findQuat_Current( i, &(mo2->opQuatList[i]), &angle );
            //mo2->opQuatList[i].theta = angle;
        }
        else if(mo2->whichFrame == GEODESICREFERENCE) {
            mo2->opQuatList[i].r = mo2->grQuatList[i].r;
            mo2->opQuatList[i].i = mo2->grQuatList[i].i;
            mo2->opQuatList[i].j = mo2->grQuatList[i].j;
            mo2->opQuatList[i].k = mo2->grQuatList[i].k;
            //findQuat_Current( i, &(mo2->opQuatList[i]), &angle );
            //mo2->opQuatList[i].theta = angle;
        }
    }
}
//*********************************************//
//
//*********************************************//
void copyQuatToPointList						( int for_which_frame )
{
    int		i		=	0;
    int		limit	= 0;
    double	qvec[3];

    if(mo2->flag_curve_closed)
        limit = mo2->l_numPoints;
    else
        limit = mo2->numPoints-2;

    for(i=0; i<=limit; i++) {
        if( for_which_frame == FRENETSERRET ) {
            getQuatComp( mo2->qwhichQuatComponent, &(mo2->fsQuatList[i]), qvec );
            mo2->qpntList[i].p3.d[0] = qvec[0];
            mo2->qpntList[i].p3.d[1] = qvec[1];
            mo2->qpntList[i].p3.d[2] = qvec[2];
        }
        else if( for_which_frame == PARALLELTRANSPORT ){
            getQuatComp( mo2->qwhichQuatComponent, &(mo2->ptQuatList[i]), qvec );
            mo2->qpntList[i].p3.d[0] = qvec[0];
            mo2->qpntList[i].p3.d[1] = qvec[1];
            mo2->qpntList[i].p3.d[2] = qvec[2];
        }
        else if( for_which_frame == GEODESICREFERENCE ){
            getQuatComp( mo2->qwhichQuatComponent, &(mo2->grQuatList[i]), qvec );
            mo2->qpntList[i].p3.d[0] = qvec[0];
            mo2->qpntList[i].p3.d[1] = qvec[1];
            mo2->qpntList[i].p3.d[2] = qvec[2];
        }
        else if( for_which_frame == OPTIMIZATION ){
            getQuatComp( mo2->qwhichQuatComponent, &(mo2->opQuatList[i]), qvec );
            mo2->qpntList[i].p3.d[0] = qvec[0];
            mo2->qpntList[i].p3.d[1] = qvec[1];
            mo2->qpntList[i].p3.d[2] = qvec[2];
        }
    }


}

//*********************************************//
//
//*********************************************//
void copyReflectedQuatToPointList				( void )
{
    int		i		=	0;
    int		limit	= 0;
    double	qvec[3];

    if(mo2->flag_curve_closed)
        limit = mo2->numPoints;
    else
        limit = mo2->numPoints-2;

    for(i=0; i<limit; i++) {
        getQuatComp( mo2->qwhichQuatComponent, &(mo2->refquatList[i]), qvec );
        mo2->qpntList[i].p3.d[0] = qvec[0];
        mo2->qpntList[i].p3.d[1] = qvec[1];
        mo2->qpntList[i].p3.d[2] = qvec[2];
    }
}
/********************************************************************************
* Date   : Sat Oct 29 12:04:49 EST 2005
* Method : void copyLineList_toPointList ( int LineIndex )
* Args   : index of line from the set of multiple-lines
* Return : none
* Purpose: copy the list of vertices of current line into global pntList[] array
********************************************************************************/
void copyLineList_toPointList                   ( int LineIndex )
{
    int     i = 0;
    int     inx = LineIndex;
    int     limit = mo2->lineList[0][inx]->nvertices;
    //printf("\t Line index = %d\n",LineIndex);
    //printf("\t Total points = %d\n",limit);
    for(i=0; i<limit; i++) {
        mo2->pntList[i].p3.d[0] = mo2->lineList[0][inx]->vlist[i].p3.d[0];
        mo2->pntList[i].p3.d[1] = mo2->lineList[0][inx]->vlist[i].p3.d[1];
        mo2->pntList[i].p3.d[2] = mo2->lineList[0][inx]->vlist[i].p3.d[2];
        //printf("\t vertex[%d] = %f, %f, %f\n",i,mo2->pntList[i].p3.d[0],mo2->pntList[i].p3.d[1],mo2->pntList[i].p3.d[2]);
    }
}
//*********************************************//
//
//*********************************************//
void getQuatComp								( int whichQuatComponent, quaternion* q, double* qvec ) {
    switch(whichQuatComponent) {

        case SHOWXYZ:
            qvec[0] = q->i;
            qvec[1] = q->j;
            qvec[2] = q->k;
            break;

        case SHOWWYZ:
            qvec[0] = q->r;
            qvec[1] = q->j;
            qvec[2] = q->k;
            break;
		
        case SHOWWXY:
            qvec[0] = q->r;
            qvec[1] = q->i;
            qvec[2] = q->j;
            break;

        case SHOWWXZ:
            qvec[0] = q->r;
            qvec[1] = q->i;
            qvec[2] = q->k;
            break;

        default:
            break;
    }	
}

//*********************************************//
//
//*********************************************//
void getSegmentColor							( int i, double *ccolor )	{

    switch(i) {
        case 0:
            ccolor[0] =	0.2; ccolor[1]	= 0.55; ccolor[2] = 0.35;
            break;
        case 1:
            ccolor[0] =	0.75; ccolor[1]	= ccolor[2]	= 0.2;
            break;
        case 2:
            ccolor[1] =	0.75; ccolor[0]	= ccolor[2]	= 0.2;
            break;
        case 3:
            ccolor[2] =	0.75; ccolor[1]	= ccolor[0]	= 0.2;
            break;
        case 4:
            ccolor[0] =	0.0; ccolor[1] = ccolor[2] = 0.75;
            break;
        case 5:
            ccolor[2] =	0.0; ccolor[1] = ccolor[0] = 0.75;
            break;
        case 6:
            ccolor[1] =	0.0; ccolor[0] = ccolor[2] = 0.75;
            break;
        case 7:
            ccolor[0] =	ccolor[1] =	ccolor[2] =	0.2;
            break;
        case 8:
            ccolor[0] =	ccolor[1] =	ccolor[2] =	0.4;
            break;
        case 9:
            ccolor[0] =	ccolor[1] =	ccolor[2] =	0.6;
            break;
        case 10:
            ccolor[0] =	ccolor[1] =	ccolor[2] =	1.0;
            break;
        case 11:
            ccolor[2] =	0.0; ccolor[1] = ccolor[0] = 1.0;
            break;
        case 12:
            ccolor[2] =	0.2; ccolor[1] = 0.5; ccolor[0]	= 1.0; 
            break;

        default:
            ccolor[0] =	ccolor[1] =	ccolor[2] =	1.0;
            break;
    }
}

//*********************************************//
//
//*********************************************//
void stext										( char* string ) {
    char* p;
    for(p=string;*p;p++) {
        glutBitmapCharacter((void*)mo2->qfont,*p);
    }
}

//### CURVE-CONSTRUCTION
void computeSplineEquation						( void ) 
{
    int    i	 = 0;
    double t	 = 0.0;
    int    count = 0;
    int    step	 = 0;
    
    Vertex P0,P1,P2,P3;
    Vertex *pt;

    pt = (Vertex*) malloc(sizeof(Vertex));

    if(mo2->whichSpline!=NOSPLINE) {

            //@ Define spline parameter
        if(mo2->whichSpline==BEZIER)
            step=3;
        else if(mo2->whichSpline==CATROM || mo2->whichSpline==BSPLINE)
            step=1;
	    
        mo2->numPoints = 0;

            //@ Compute spline points and derivatives
        for(i=0; i<mo2->numCtrlPoints-3; i+=step) {
            t=0;
            P0.p3.d[0] = mo2->ctrlPntList[i].p3.d[0];
            P0.p3.d[1] = mo2->ctrlPntList[i].p3.d[1];
            P0.p3.d[2] = mo2->ctrlPntList[i].p3.d[2];
	        
            P1.p3.d[0] = mo2->ctrlPntList[i+1].p3.d[0];
            P1.p3.d[1] = mo2->ctrlPntList[i+1].p3.d[1];
            P1.p3.d[2] = mo2->ctrlPntList[i+1].p3.d[2];

            P2.p3.d[0] = mo2->ctrlPntList[i+2].p3.d[0];
            P2.p3.d[1] = mo2->ctrlPntList[i+2].p3.d[1];
            P2.p3.d[2] = mo2->ctrlPntList[i+2].p3.d[2];

            P3.p3.d[0] = mo2->ctrlPntList[i+3].p3.d[0];
            P3.p3.d[1] = mo2->ctrlPntList[i+3].p3.d[1];
            P3.p3.d[2] = mo2->ctrlPntList[i+3].p3.d[2];

            while(t<=1.0) {

                    // compute interpolated spline-point
                P( P0, P1, P2, P3, pt, t, mo2->whichSpline );
                mo2->pntList[count].p3.d[0] = pt->p3.d[0];
                mo2->pntList[count].p3.d[1] = pt->p3.d[1];
                mo2->pntList[count].p3.d[2] = pt->p3.d[2];

                    // compute tangent-first-derivative
                    //computeSplineFirstDerivative(P0,P1,P2,P3,pt,t,whichSpline);
                    //TangList[count].p3.d[0] = pt->p3.d[0];
                    //TangList[count].p3.d[1] = pt->p3.d[1];
                    //TangList[count].p3.d[2] = pt->p3.d[2];

                    // compute tangent-second-derivative
                    //computeSplineSecondDerivative(P0,P1,P2,P3,pt,t,whichSpline);
                    //dTangList[count].p3.d[0] = pt->p3.d[0];
                    //dTangList[count].p3.d[1] = pt->p3.d[1];
                    //dTangList[count].p3.d[2] = pt->p3.d[2];

                t = t + mo2->param_spline_increment;
                count++;
            }

                //// compute tangent-first-derivative on last point on curve
                //computeSplineFirstDerivative(P0,P1,P2,P3,pt,1.0,whichSpline);
                //TangList[count].p3.d[0] = pt->p3.d[0];
                //TangList[count].p3.d[1] = pt->p3.d[1];
                //TangList[count].p3.d[2] = pt->p3.d[2];

                //// compute tangent-second-derivative on last point on curve
                //computeSplineSecondDerivative(P0,P1,P2,P3,pt,1.0,whichSpline);
                //dTangList[count].p3.d[0] = pt->p3.d[0];
                //dTangList[count].p3.d[1] = pt->p3.d[1];
                //dTangList[count].p3.d[2] = pt->p3.d[2];
        }

        if(mo2->numCtrlPoints > 3) {
                // add last point to curve
            if(mo2->whichSpline==BEZIER) {
                mo2->pntList[count].p3.d[0] = P3.p3.d[0];
                mo2->pntList[count].p3.d[1] = P3.p3.d[1];
                mo2->pntList[count].p3.d[2] = P3.p3.d[2];
                count++;
            }
            else if(mo2->whichSpline==CATROM) {
                mo2->pntList[count].p3.d[0] = P2.p3.d[0];
                mo2->pntList[count].p3.d[1] = P2.p3.d[1];
                mo2->pntList[count].p3.d[2] = P2.p3.d[2];
                count++;
            }
        }
        mo2->numPoints = count;
    }
    else {
        step = 1;
            //@ Compute spline points and derivatives
        for(i=0; i<mo2->numCtrlPoints-1; i+=step) {
            if(i>0)
                count--;

            t=0.0;
            P0.p3.d[0] = mo2->ctrlPntList[i].p3.d[0];
            P0.p3.d[1] = mo2->ctrlPntList[i].p3.d[1];
            P0.p3.d[2] = mo2->ctrlPntList[i].p3.d[2];
	        
            P1.p3.d[0] = mo2->ctrlPntList[i+1].p3.d[0];
            P1.p3.d[1] = mo2->ctrlPntList[i+1].p3.d[1];
            P1.p3.d[2] = mo2->ctrlPntList[i+1].p3.d[2];	

            while(t<=1.0) {

                pt->p3.d[0] = P1.p3.d[0]*t + (1-t)*P0.p3.d[0];
                pt->p3.d[1] = P1.p3.d[1]*t + (1-t)*P0.p3.d[1];
                pt->p3.d[2] = P1.p3.d[2]*t + (1-t)*P0.p3.d[2];

                mo2->pntList[count].p3.d[0] = pt->p3.d[0];
                mo2->pntList[count].p3.d[1] = pt->p3.d[1];
                mo2->pntList[count].p3.d[2] = pt->p3.d[2];
                t = t + mo2->param_spline_increment;
                count++;
            }
        }
        mo2->numPoints = count;
    }
}

//*********************************************//
//
//*********************************************//
void P											( Vertex P0, Vertex P1, Vertex P2, Vertex P3 , Vertex *P, double t, int currentSpline )
{
    if(currentSpline==BEZIER) {

            /* Parametric equation of Bezier Spline */
            /*
              P(t) = (1-t)^3.P0 + 3.t.(1-t)^2.P1 + 3.t^2.(1-t).P2 + t^3.P3
            */
        P->p3.d[0] = pow((1-t),3.0)*P0.p3.d[0] + 3*t*pow((1-t),2.0)*P1.p3.d[0] + 3*pow(t,2.0)*(1-t)*P2.p3.d[0] + pow(t,3.0)*P3.p3.d[0];
        P->p3.d[1] = pow((1-t),3.0)*P0.p3.d[1] + 3*t*pow((1-t),2.0)*P1.p3.d[1] + 3*pow(t,2.0)*(1-t)*P2.p3.d[1] + pow(t,3.0)*P3.p3.d[1];
        P->p3.d[2] = pow((1-t),3.0)*P0.p3.d[2] + 3*t*pow((1-t),2.0)*P1.p3.d[2] + 3*pow(t,2.0)*(1-t)*P2.p3.d[2] + pow(t,3.0)*P3.p3.d[2];
    }
    else if(currentSpline==CATROM) {
        
            /* Parametric equation of Catmull-Rom Spline */
            /*
              P(t) = 0.5*( -t.(1-t)^3.P0 + (2-5t^2+3t^3)*P1 + t(1+4t-3t^2).P2 + -t^2.(1-t).P3 );
            */
        P->p3.d[0] = 0.5*(-1*t*pow((1-t),3.0)*P0.p3.d[0] + (2 - 5.0*pow(t,2.0) + 3.0*pow(t,3.0))*P1.p3.d[0] + t*(1 + 4.0*t - 3.0*pow(t,2.0))*P2.p3.d[0] + -1*pow(t,2.0)*(1-t)*P3.p3.d[0]);
        P->p3.d[1] = 0.5*(-1*t*pow((1-t),3.0)*P0.p3.d[1] + (2 - 5.0*pow(t,2.0) + 3.0*pow(t,3.0))*P1.p3.d[1] + t*(1 + 4.0*t - 3.0*pow(t,2.0))*P2.p3.d[1] + -1*pow(t,2.0)*(1-t)*P3.p3.d[1]);
        P->p3.d[2] = 0.5*(-1*t*pow((1-t),3.0)*P0.p3.d[2] + (2 - 5.0*pow(t,2.0) + 3.0*pow(t,3.0))*P1.p3.d[2] + t*(1 + 4.0*t - 3.0*pow(t,2.0))*P2.p3.d[2] + -1*pow(t,2.0)*(1-t)*P3.p3.d[2]);

    }
    else if(currentSpline==BSPLINE) {
        
            /* Parametric equation of B-Spline */
            /*
              P(t) = (1.0/6.0)*( (1-t)^3.P0 + (3t^3 - 6t^2 + 4).P1 + (-3t^3 + 3t^2 + 3t +1).P2 + t^3.P3 );
            */
        P->p3.d[0] = (1.0/6.0)*( pow((1-t),3.0)*P0.p3.d[0] + (3.0*pow(t,3.0) - 6.0*pow(t,2.0) + 4.0)*P1.p3.d[0] + (-3.0*pow(t,3.0) + 3.0*pow(t,2.0) + 3.0*t + 1.0)*P2.p3.d[0] + pow(t,3.0)*P3.p3.d[0]);
        P->p3.d[1] = (1.0/6.0)*( pow((1-t),3.0)*P0.p3.d[1] + (3.0*pow(t,3.0) - 6.0*pow(t,2.0) + 4.0)*P1.p3.d[1] + (-3.0*pow(t,3.0) + 3.0*pow(t,2.0) + 3.0*t + 1.0)*P2.p3.d[1] + pow(t,3.0)*P3.p3.d[1]);
        P->p3.d[2] = (1.0/6.0)*( pow((1-t),3.0)*P0.p3.d[2] + (3.0*pow(t,3.0) - 6.0*pow(t,2.0) + 4.0)*P1.p3.d[2] + (-3.0*pow(t,3.0) + 3.0*pow(t,2.0) + 3.0*t + 1.0)*P2.p3.d[2] + pow(t,3.0)*P3.p3.d[2]);
        
    }
    else if(currentSpline==NOSPLINE) { //LERP
		
    }
}
//*********************************************//
//
//*********************************************//
void constructKnot 								( void )
{
    double	t	= 0.0;
    int		i	= 0;
    int		mod	= 0;
    double	div	= 1.0;
    double	inc	= 0.0;
    int		m =	mo2->m;
    int		n =	mo2->n;
    double	a =	mo2->a;
    double	b =	mo2->b;
    double	r =	mo2->r;
    double	maxAngle = 0.0;
    double	numPntsFactor = 0.06; //factor controls number of points [decrease to increase points]
	
	/* // Equation of Torus	Knot
///////////////////////////////////
torusknot[a_,b_,c_][p_,q_][t_] :=
{(a	+ b*Cos[q*t])*Cos[p*t],
(a + b*Cos[q*t])*Sin[p*t],
c*Sin[q*t]}
    //////////////////////////////////
    */

            // WARNING:
            // date: 01-17-2005
            // for : solving the double	covering problem
            // :( needs	the	prime numbers to make sure the cases are accounted for.	But
            // this	handles	only a few cases. A	prime-number generater would make it generic.
	if(m==n)
            div=(double)m;

	if(m>n)	{
            mod	= m%n;
            if(mod==0) {
                div=(double)n;
            }
            else {
                if(m%2==0 && n%2==0	&& mod%2==0){
                    if((m%mod)%2==0)
                    {
                        if(m%4==0 && n%4==0)
                            div=(double)4;
                        else
                            div=(double)2;
                    }
                    else 
                        div=(double)mod;
                }

                if(m%3==0 && n%3==0	&& mod%3==0){
                    if((m%mod)%3==0)
                    {
                        if(m%6==0 && n%6==0)
                            div=(double)6;
                        else
                            div=(double)3;
                    }			
                    else
                        div=(double)mod;
                }
			
                if(m%5==0 && n%5==0	&& mod%5==0){
                    if((n%mod)%5==0)
                        div=(double)5;
                    else
                        div=(double)mod;
                }

                if(m%7==0 && n%7==0	&& mod%7==0){
                    if((n%mod)%7==0)
                        div=(double)7;
                    else
                        div=(double)mod;
                }
            }
		
	}
	
	else {
            mod	= n%m;
            if(mod==0) {
                div=(double)m;
            }
            else {
                if(m%2==0 && n%2==0	&& mod%2==0){
                    if((m%mod)%2==0)
                    {
                        if(m%4==0 && n%4==0)
                            div=(double)4;
                        else
                            div=(double)2;
                    }
                    else
                        div=(double)mod;
                }
                if(m%3==0 && n%3==0	&& mod%3==0){
                    if((m%mod)%3==0)
                    {
                        if(m%6==0 && n%6==0)
                            div=(double)6;
                        else
                            div=(double)3;
                    }			
                    else
                        div=(double)mod;
                }
                if(m%5==0 && n%5==0	&& mod%5==0){
                    if((m%mod)%5==0)
                        div=(double)5;
                    else
                        div=(double)mod;
                }
                if(m%7==0 && n%7==0	&& mod%7==0){
                    if((m%mod)%7==0)
                        div=(double)7;
                    else
                        div=(double)mod;
                }
            }
	}
            //@@

	inc	= Max(numPntsFactor,mo2->increment/Max(m,n));
	
	mo2->numPoints = 0;
	if(div>0) {
            maxAngle = 2.0;
            while(t<=maxAngle*M_PI/div) {
			
                mo2->pntList[i].p3.d[0]	= (r + a*cos(n*t))*cos(m*t);
                mo2->pntList[i].p3.d[1]	= (r + a*cos(n*t))*sin(m*t);
                mo2->pntList[i].p3.d[2]	= b*sin( n*t );
                    //fprintf(stderr,"pntList[%d] =	%f,	%f,	%f\n",i, mo2->pntList[i].p3.d[0],mo2->pntList[i].p3.d[1],mo2->pntList[i].p3.d[2]);
                i++;
                t =	t+inc;
            }

			mo2->pntList[i].p3.d[0] = mo2->pntList[0].p3.d[0];
			mo2->pntList[i].p3.d[1] = mo2->pntList[0].p3.d[1];
			mo2->pntList[i].p3.d[2] = mo2->pntList[0].p3.d[2];
			i++;

            mo2->numPoints = i-1;

			mo2->pntList[i].p3.d[0] = mo2->pntList[1].p3.d[0];
			mo2->pntList[i].p3.d[1] = mo2->pntList[1].p3.d[1];
			mo2->pntList[i].p3.d[2] = mo2->pntList[1].p3.d[2];
			i++;
			mo2->pntList[i].p3.d[0] = mo2->pntList[2].p3.d[0];
			mo2->pntList[i].p3.d[1] = mo2->pntList[2].p3.d[1];
			mo2->pntList[i].p3.d[2] = mo2->pntList[2].p3.d[2];
			i++;
			mo2->pntList[i].p3.d[0] = mo2->pntList[3].p3.d[0];
			mo2->pntList[i].p3.d[1] = mo2->pntList[3].p3.d[1];
			mo2->pntList[i].p3.d[2] = mo2->pntList[3].p3.d[2];
			i++;
			//mo2->pntList[i].p3.d[0] = mo2->pntList[4].p3.d[0];
			//mo2->pntList[i].p3.d[1] = mo2->pntList[4].p3.d[1];
			//mo2->pntList[i].p3.d[2] = mo2->pntList[4].p3.d[2];
			//i++;
			//mo2->pntList[i].p3.d[0] = mo2->pntList[5].p3.d[0];
			//mo2->pntList[i].p3.d[1] = mo2->pntList[5].p3.d[1];
			//mo2->pntList[i].p3.d[2] = mo2->pntList[5].p3.d[2];
			//i++;
			mo2->l_numPoints = i;

#ifdef _DBG_
			//printf("inside obj::computeObject() > mo2->numPoints = %d\n",mo2->numPoints);
#endif
		//mo2->numPoints = 10;
	}
}

//### COMPUTATIONS
void makeTangent 								( void ) 
{
	int		i = 0;
	int		limit = 0;
	double	p1[3];
	double	p2[3];				//points on	spline	  
	double	len = 0.0;
	double	v[3];

	if(mo2->flag_curve_closed)
		limit = mo2->l_numPoints-1;
	else
		limit = mo2->numPoints-1;


        //////////////// BUILD Tangents	/////////////////////////////////
	for ( i=0; i<(limit); i++ ) {

                //asign vector-1
            
        if(i==limit+1) { //@ to calculate (n-th)-tangent
			p1[0] =	mo2->pntList[0].p3.d[0];
			p1[1] =	mo2->pntList[0].p3.d[1];
			p1[2] =	mo2->pntList[0].p3.d[2];                
        }
        else { //@ for all other tangents: i<(n-1)
			p1[0] =	mo2->pntList[i].p3.d[0];
			p1[1] =	mo2->pntList[i].p3.d[1];
			p1[2] =	mo2->pntList[i].p3.d[2];
		}

        //asign vector-2
        if(i==(limit)) { //@ to calculate (n-1)-th tangent
            p2[0] =	mo2->pntList[0].p3.d[0];
            p2[1] =	mo2->pntList[0].p3.d[1];
            p2[2] =	mo2->pntList[0].p3.d[2];
        }
        else if (i==(limit+1)) { //@ to calculate (n-th)-tangent
            p2[0] =	mo2->pntList[1].p3.d[0];
            p2[1] =	mo2->pntList[1].p3.d[1];
            p2[2] =	mo2->pntList[1].p3.d[2];
        }
        else { //@ for all other tangents: i<(n-1)
            p2[0] =	mo2->pntList[i+1].p3.d[0];
            p2[1] =	mo2->pntList[i+1].p3.d[1];
            p2[2] =	mo2->pntList[i+1].p3.d[2];
        }

        v[0] = p2[0]-p1[0];
        v[1] = p2[1]-p1[1];
        v[2] = p2[2]-p1[2];


        len	= length3D(v);

        v[0] = v[0]/len;
        v[1] = v[1]/len;
        v[2] = v[2]/len;

        mo2->TangList[i].p3.d[0] = v[0];
        mo2->TangList[i].p3.d[1] = v[1];
        mo2->TangList[i].p3.d[2] = v[2];
                /*if( i == limit ) {
                    //printf("[%d] Tang = %f, %f ,%f\n",i,v[0],v[1],v[2]);
                glPushMatrix();
                glColor3f(0,0,0);
                    
                glTranslatef(mo2->pntList[i].p3.d[0],mo2->pntList[i].p3.d[1],mo2->pntList[i].p3.d[2]);
                glutSolidSphere(0.03,8,8);
                glBegin(GL_LINES);

                glVertex3f(0,0,0);

                glVertex3f(v[0],v[1],v[2]);
                glEnd();
                    
                glPopMatrix();
                                    
                }*/
                
	}
}

//*********************************************//
//
//*********************************************//
void makeTangentDerivatives						( void ) 
{
	int		i = 0;
	int		limit = 0;
	double	t = 0;
	double	len	 = 0.0;
	
	double T0[3];
	double T1[3];
	double TT[3];

	if(mo2->flag_curve_closed)
		limit = mo2->l_numPoints-2;
	else
		limit = mo2->numPoints-1;

	for(i=0; i<(limit); i++) {

            t=0;
            if(i==(limit+1)) { //@ to calculate (n-th)-tangent

                T0[0] =	mo2->TangList[0].p3.d[0];
                T0[1] =	mo2->TangList[0].p3.d[1];
                T0[2] =	mo2->TangList[0].p3.d[2];
    
            }
            else { //@ for all other tangents: i<(n-1)
                
                T0[0] =	mo2->TangList[i].p3.d[0];
                T0[1] =	mo2->TangList[i].p3.d[1];
                T0[2] =	mo2->TangList[i].p3.d[2];
            }
            
            if(i==(limit)) { //@ to calculate (n-1)-th tangent
                T1[0] =	mo2->TangList[0].p3.d[0];
                T1[1] =	mo2->TangList[0].p3.d[1];
                T1[2] =	mo2->TangList[0].p3.d[2];
            }
            else if(i==(limit+1)) { //@ to calculate (n-th)-tangent
                T1[0] =	mo2->TangList[1].p3.d[0];
                T1[1] =	mo2->TangList[1].p3.d[1];
                T1[2] =	mo2->TangList[1].p3.d[2];
            }
                /*
            else if(i==(limit+2)) {
                T1[0] =	mo2->TangList[2].p3.d[0];
                T1[1] =	mo2->TangList[2].p3.d[1];
                T1[2] =	mo2->TangList[2].p3.d[2];
                }*/
            else { //@ for all other tangents: i<(n-1)
                T1[0] =	mo2->TangList[i+1].p3.d[0];
                T1[1] =	mo2->TangList[i+1].p3.d[1];
                T1[2] =	mo2->TangList[i+1].p3.d[2];
            }
            minus3D(T1,T0,TT);

            len	= length3D(TT);
		
            TT[0] =	TT[0]/len;
            TT[1] =	TT[1]/len;
            TT[2] =	TT[2]/len;

            mo2->dTangList[i].p3.d[0] =	TT[0];
            mo2->dTangList[i].p3.d[1] =	TT[1];
            mo2->dTangList[i].p3.d[2] =	TT[2];
		    //printf("[%d] dTang = %f, %f ,%f\n",i,TT[0],TT[1],TT[2]);

                //if( i >limit-3 ) {
                //    //printf("[%d]dTang = %f, %f ,%f\n",i,TT[0],TT[1],TT[2]);
                //    glPushMatrix();
                //    glColor3f(0,0,0);
                //        
                //    glTranslatef(mo2->pntList[i].p3.d[0],mo2->pntList[i].p3.d[1],mo2->pntList[i].p3.d[2]);
                //    glutSolidSphere(0.03,8,8);
                //    glBegin(GL_LINES);

                //    glVertex3f(0,0,0);

                //    glVertex3f(TT[0],TT[1],TT[2]);
                //    glEnd();
                //        
                //    glPopMatrix();
                //}
	}
}

//*********************************************//
//
//*********************************************//
void makeCurve_Curvature						( void )
{
	int i = 0;
	int limit = 0;
	double cross_dT_dTT[3];
	double dT[3];
	double dTT[3];
	double norm_cross_dT_dTT = 0.0;
	double norm_T = 0.0;
	double curvature = 0.0;
	double s = 0.0;

	mo2->gmax_curve_curvature = -1.0;

	if(mo2->flag_curve_closed==1)
		limit = mo2->l_numPoints;
	else
		limit = mo2->numPoints-1;
    if(mo2->whichCurve==OBJCURVELIST)
        limit = mo2->numPoints;

	for(i=0;i<limit-1;i++) {
		dT[0] = mo2->TangList[i].p3.d[0];
		dT[1] = mo2->TangList[i].p3.d[1];
		dT[2] = mo2->TangList[i].p3.d[2];

		dTT[0] = mo2->dTangList[i].p3.d[0];
		dTT[1] = mo2->dTangList[i].p3.d[1];
		dTT[2] = mo2->dTangList[i].p3.d[2];

		cross3D(dT, dTT, cross_dT_dTT);
		norm_cross_dT_dTT = length3D(cross_dT_dTT);
		norm_T = length3D(dT);

		curvature = ( norm_cross_dT_dTT )/(norm_T*norm_T*norm_T);
		//printf("[%d] curvature = %f\n",i,curvature);
		mo2->CurvList[i] = curvature;
		if(curvature > mo2->gmax_curve_curvature)
			mo2->gmax_curve_curvature = curvature;
		if(mo2->showColorCoded == CODE_CURVATURE) {
			s = 255.99/(1.25*mo2->scale);
			mo2->pntList[i].color = (int) ((curvature*100.0 - 99.0)*s);
			if(mo2->pntList[i].color > 255)
				mo2->pntList[i].color = 255;
			mo2->pntList[i].color = 255 - mo2->pntList[i].color;
			//printf("mo2->pntList[i].color = %d\n",mo2->pntList[i].color);
		}
	}
	//printf("MAX Curvature = %f\n",mo2->gmax_curve_curvature);
}
//*********************************************//
//
//*********************************************//
void makeCurve_Torsion							( void )
{
	int i = 0;
	int limit = 0;
	double cross_dT_dTT[3];
	double dot_dT_dTT_dTTT = 0;
	double dT[3];
	double dTT0[3];
	double dTT1[3];
	double dTTT[3];
	double norm_cross_dT_dTT = 0.0;
	double norm_TTT=0.0;
	double torsion = 0.0;
	double s = 0;

	if(mo2->flag_curve_closed)
		limit = mo2->l_numPoints;
	else
		limit = mo2->numPoints-1;


	for(i=0;i<limit-2;i++) {
		
		//t'
		dT[0] = mo2->TangList[i].p3.d[0];
		dT[1] = mo2->TangList[i].p3.d[1];
		dT[2] = mo2->TangList[i].p3.d[2];

		//t''(0)
		dTT0[0] = mo2->dTangList[i].p3.d[0];
		dTT0[1] = mo2->dTangList[i].p3.d[1];
		dTT0[2] = mo2->dTangList[i].p3.d[2];

		//t''(1)
		dTT1[0] = mo2->dTangList[i+1].p3.d[0];
		dTT1[1] = mo2->dTangList[i+1].p3.d[1];
		dTT1[2] = mo2->dTangList[i+1].p3.d[2];

		//t'''
		minus3D(dTT1, dTT0, dTTT);
		norm_TTT = length3D(dTTT);

		cross3D(dT,dTT0,cross_dT_dTT);
		norm_cross_dT_dTT = length3D(cross_dT_dTT);
		
		dot_dT_dTT_dTTT = dot3D( cross_dT_dTT, dTTT );

		torsion = ( dot_dT_dTT_dTTT ) / ( norm_cross_dT_dTT*norm_cross_dT_dTT );
		//printf("[%d] torsion = %f\n",i,torsion);
		mo2->TorsnList[i] = torsion;

		if(mo2->showColorCoded == CODE_TORSION) {
			s = 255.99/(2.0*mo2->scale);
			if(torsion>0.0) {
				mo2->pntList[i].color = (int) ((torsion*10.0)*s);
			}
			else
				mo2->pntList[i].color = (int) ((-torsion*10.0)*s);
			if(mo2->pntList[i].color > 255)
				mo2->pntList[i].color = 255;
			//printf("mo2->pntList[i].color = %d\n",mo2->pntList[i].color);
		}
	}
}
//*********************************************//
//
//*********************************************//
void computeFrenetSerretFrame 					( void ) 
{
	int		i = 0;						
	int		limit = mo2->l_numPoints;			// total points to compute for
	double	normB;							// lenght of various vectors
	double	b[3];							// dot and cross product results
	double	n[3];							// vector for cross	product	of Binormal	and	Tangent
	double	t[3];							// vector for Tangent
	double	tt[3];							// vector for first	derivative of Tangent

	//for asigning quaternions
	double		mat[16];					// 4x4 matrix
	double		dotPrevCurrBinormal = 0.0;	// dot product
	double		qdotProd_PQ = 0.0;			// quaternion dot product
	double		prevb[3];					// previous binormal
	quaternion	*q, *p, *r;					// quaternion
	
	//init quaternions
	p =	(quaternion*) my_malloc(sizeof(quaternion));
	q =	(quaternion*) my_malloc(sizeof(quaternion));
	r =	(quaternion*) my_malloc(sizeof(quaternion));
	
    mo2->markFrameCounter =	0;
	mo2->markQuatCurveCounter = 0;
    
    if(mo2->objectID == OBJCURVE && mo2->whichCurve == OBJKNOT)
        limit = mo2->l_numPoints;
    else if(mo2->objectID == OBJCURVE && mo2->whichCurve == OBJCURVELIST)
        limit = mo2->numPoints;
        
	flip = 0;

	if(limit>2) {

            for(i=0; i<=limit+2; i++) {
                    //printf("[%d]\n",i);
                    //### Compute Frenet Frame ###
                t[0] = mo2->TangList[i].p3.d[0];
                t[1] = mo2->TangList[i].p3.d[1];
                t[2] = mo2->TangList[i].p3.d[2];
                    //printf("norm(t) = %f\n",length3D(t));

                tt[0] =	mo2->dTangList[i].p3.d[0];
                tt[1] =	mo2->dTangList[i].p3.d[1];
                tt[2] =	mo2->dTangList[i].p3.d[2];

                    //obtain cross product of tangent and derivative of	tangent
                cross3D(t, tt, b);

                    //obtain length	of vector cross_vt
                normB = length3D(b);
			
                    //normalize	vector cross_vt
                b[0] =	b[0]/normB;
                b[1] =	b[1]/normB;
                b[2] =	b[2]/normB;

                    //save cross_vt into	vector list
                mo2->fsSBasis[i].p3.d[0] = b[0]; //NOTE! changed from mo2->fsRBasis[i]...
                mo2->fsSBasis[i].p3.d[1] = b[1];
                mo2->fsSBasis[i].p3.d[2] = b[2];
                    //printf("norm(b) = %f\n",length3D(b));

                    //obtain cross product of Binormal and Tangent
                cross3D(b, t, n);
                    //printf("norm(n) = %f\n",length3D(n));
                    //save cross_BV	into vector	list
                mo2->fsRBasis[i].p3.d[0] = n[0]; //NOTE! changed from mo2->fsSBasis[i]...
                mo2->fsRBasis[i].p3.d[1] = n[1];
                mo2->fsRBasis[i].p3.d[2] = n[2];

                //### Assign Quaternions ###
                //obtain 4x4 matrix
                mat[0] = n[0];
                mat[1] = n[1];
                mat[2] = n[2];

                mat[4] = b[0];
                mat[5] = b[1];
                mat[6] = b[2];

                mat[8] = t[0];
                mat[9] = t[1];
                mat[10]= t[2];		

                mat[3]	= mat[7] = mat[11] = mat[12] = mat[13] = mat[14] = 0.0;
                mat[15]	= 1.0;

                ExtractQuaternionFromMat( mat, q );
                quatNormalize( q );	
			
                    //# Check to keep the quaternions with same sign
                    //# copy the first quaternion into a global variable
                    //# compare against that first quaternion (now this
                    //# is tricky! as the 'i' component may not change 
                    //# and still flip occurs. Set 'flip' accordingly.
                    //# At end of frame computations flip quaternions if
                    //# flip is set.
                if(flag_quat_firsttime) {
                    quatCopy( q, q0 );
                    flag_quat_firsttime = 0;
                }
                if(i==0) {
                    if( (q0->i * q->i) < 0.0 ) {
                        flip = 1;
                    }
                }


                if(i>0) {
                    dotPrevCurrBinormal = dot3D(prevb, b);
                    if(dotPrevCurrBinormal<=0.6) {
                        mo2->markFrame[mo2->markFrameCounter] = i-1;
                        mo2->markFrameCounter++;
                        mo2->markQuatCurve[mo2->markQuatCurveCounter] = i-1;
                        mo2->markQuatCurveCounter++;
                    }
                    qdotProd_PQ = quatQuatDotProduct( p, q );
                    if(qdotProd_PQ < -0.9) {
                        r->r = -q->r;
                        r->i = -q->i;
                        r->j = -q->j;
                        r->k = -q->k;
                        quatCopy(r, q);				
                            //mo2->markQuatCurve[mo2->markQuatCurveCounter] = i-1;
                            //mo2->markQuatCurveCounter++;
                   }
                }
                    //save quaternions
                mo2->fsQuatList[i].r = q->r;
                mo2->fsQuatList[i].i = q->i;
                mo2->fsQuatList[i].j = q->j;
                mo2->fsQuatList[i].k = q->k;

                    //copy current-binormal into previous-binormal
                prevb[0] = b[0];
                prevb[1] = b[1];
                prevb[2] = b[2];
                    //copy current quaternion into previous one
                quatCopy( q, p );

            }//$FOR(i..)
	}//$IF(limit>2)

            //printf("flip = %d\n",flip);
	if(flip) {
            for(i=0;i<=(limit+1);i++) {
                quatCopy( &(mo2->fsQuatList[i]), p );
                r->r = -p->r;
                r->i = -p->i;
                r->j = -p->j;
                r->k = -p->k;
                quatCopy(r, &(mo2->fsQuatList[i]));				
            }
            flip = 0;
	}
    //printf("total points = %d, mo2->quatCurvecounter = %d\n",i,mo2->markQuatCurveCounter);
//        for(i=0;i<mo2->markQuatCurveCounter;i++)
//            printf("markQuatCurve[%d] = %d\n",mo2->markQuatCurveCounter, mo2->markQuatCurve[i] );

    //### DONE ###
	my_free(q);
	my_free(p);
	my_free(r);
  
}

//*********************************************//
//
//*********************************************//
void computeFrenetSerretFrame_forSingleFrame	( void ) 
{
	int		i = 0;						
	int		limit = 1;						// total points to compute for
	double	normB;							// lenght of various vectors
	double	b[3];							// dot and cross product results
	double	n[3];							// vector for cross	product	of Binormal	and	Tangent
	double	t[3];							// vector for Tangent
	double	tt[3];							// vector for first	derivative of Tangent

	//for asigning quaternions
	double		mat[16];					// 4x4 matrix
	double		dotPrevCurrBinormal = 0.0;	// dot product
	double		qdotProd_PQ = 0.0;			// quaternion dot product
	double		prevb[3];					// previous binormal
	quaternion	*q, *p, *r;					// quaternion
	
	//init quaternions
	p =	(quaternion*) my_malloc(sizeof(quaternion));
	q =	(quaternion*) my_malloc(sizeof(quaternion));
	r =	(quaternion*) my_malloc(sizeof(quaternion));
	
            //mo2->markFrameCounter =	0;
            //mo2->markQuatCurveCounter = 0;


	for(i=0; i<limit; i++) {

		//### Compute Frenet Frame ###
		t[0] = mo2->TangList[i].p3.d[0];
		t[1] = mo2->TangList[i].p3.d[1];
		t[2] = mo2->TangList[i].p3.d[2];

		tt[0] =	mo2->dTangList[i].p3.d[0];
		tt[1] =	mo2->dTangList[i].p3.d[1];
		tt[2] =	mo2->dTangList[i].p3.d[2];
	
		//obtain cross product of tangent and derivative of	tangent
		cross3D(t, tt, b);

		//obtain length	of vector cross_vt
		normB = length3D(b);
		
		//normalize	vector cross_vt
		b[0] =	b[0]/normB;
		b[1] =	b[1]/normB;
		b[2] =	b[2]/normB;

		//save cross_vt into	vector list
		mo2->fsSBasis[i].p3.d[0] = b[0]; //NOTE! changed from mo2->fsRBasis[i]...
        mo2->fsSBasis[i].p3.d[1] = b[1];
        mo2->fsSBasis[i].p3.d[2] = b[2];
		//printf("norm(b) = %f\n",length3D(b));

		//obtain cross product of Binormal and Tangent
		cross3D(b, t, n);
		//printf("norm(n) = %f\n",length3D(n));
		//save cross_BV	into vector	list
		mo2->fsRBasis[i].p3.d[0] = n[0]; //NOTE! changed from mo2->fsSBasis[i]...
        mo2->fsRBasis[i].p3.d[1] = n[1];
        mo2->fsRBasis[i].p3.d[2] = n[2];

		//### Assign Quaternions ###
		//obtain 4x4 matrix
		mat[0] = n[0];
		mat[1] = n[1];
		mat[2] = n[2];

		mat[4] = b[0];
		mat[5] = b[1];
		mat[6] = b[2];

		mat[8] = t[0];
		mat[9] = t[1];
		mat[10]= t[2];		

		mat[3]	= mat[7] = mat[11] = mat[12] = mat[13] = mat[14] = 0.0;
		mat[15]	= 1.0;

		ExtractQuaternionFromMat( mat, q );
		quatNormalize( q );	
		
		if(i>0) {
			dotPrevCurrBinormal = dot3D(prevb, b);
			if(dotPrevCurrBinormal<=0.6) {
				mo2->markFrame[mo2->markFrameCounter] = i-1;
				mo2->markFrameCounter++;
				mo2->markQuatCurve[mo2->markQuatCurveCounter] = i-1;
				mo2->markQuatCurveCounter++;
			}
			qdotProd_PQ = quatQuatDotProduct( p, q );
			if(qdotProd_PQ < -0.9) {
				r->r = -q->r;
				r->i = -q->i;
				r->j = -q->j;
				r->k = -q->k;
				quatCopy(r, q);				
			}
		}
		//save quaternions
		mo2->fsQuatList[i].r = q->r;
		mo2->fsQuatList[i].i = q->i;
		mo2->fsQuatList[i].j = q->j;
		mo2->fsQuatList[i].k = q->k;

		//copy current-binormal into previous-binormal
		prevb[0] = b[0];
		prevb[1] = b[1];
		prevb[2] = b[2];
		//copy current quaternion into previous one
		quatCopy( q, p );

		//# Check to keep the quaternions with same sign
		//# copy the first quaternion into a global variable
		//# compare against that first quaternion (now this
		//# is tricky! as the 'i' component may not change 
		//# and still flip occurs. Set 'flip' accordingly.
		//# At end of frame computations flip quaternions if
		//# flip is set.
		if(flag_quat_firsttime) {
			quatCopy( q, q0 );
			flag_quat_firsttime = 0;
		}
		if(i==0) {
			//printf("[%d] q = %f, %f, %f, %f\n", i, q->r, q->i, q->j, q->k );
			//printf("[%d] q0= %f, %f, %f, %f\n", i, q0->r, q0->i, q0->j, q0->k );
			if( (q0->i * q->i) < 0.0 ) { //is this better? if( q0->i*q->i || q0->j*q->j || q0->k*q->k) 
				flip = 1;
			}
		}

	}//$ FOR(i..)
	my_free(q);
	my_free(p);
	my_free(r);
}

//*********************************************//
//
//*********************************************//
void computeParallelTransportFrame 				( void ) 
{
	int		i		= 0;	// index into points array
	int		limit	= mo2->l_numPoints;	// local variable for number of	points for frame calculation
	double	t[3];			// arbitrary vector
	double	v[3];			// vector for Tangent
	double	s[3];			// vector for third	orthonormal	component
	double	v0[3];			// vector for initial normal vector
	double	len;			// vector lenght
	double	theta	= 0.0;	// PT initial frame	angle
	double	t1[3];			// vector for Tangent
	double	t2[3];			// vector for Tangent
	double	cross_T1T2[3];	// vector for tangent-tangent cross	product
	double	angle = mo2->startTheta;
	//double	mat[16];
	//double	dotPQ	= 0.0;
	quaternion *p,*q,*r, *u, *prev;

	//initialize
	p =	(quaternion*) malloc(sizeof(quaternion));
	q =	(quaternion*) malloc(sizeof(quaternion));
	r =	(quaternion*) malloc(sizeof(quaternion));
	u =	(quaternion*) malloc(sizeof(quaternion));
	prev = (quaternion*) malloc(sizeof(quaternion));

	t[0] = 0.0;
	t[1] = 0.0;
	t[2] = 1.0;

    if(mo2->objectID == OBJCURVE && mo2->whichCurve == OBJKNOT)
        limit = mo2->l_numPoints;
    else if(mo2->objectID == OBJCURVE && mo2->whichCurve == OBJCURVELIST)
        limit = mo2->numPoints;


	computeFrenetSerretFrame_forSingleFrame();
	//@ Make first frame same as FRENET  ///////////////////////////
	//v0[0] = mo2->fsRBasis[0].p3.d[0];
	//v0[1] = mo2->fsRBasis[0].p3.d[1];
	//v0[2] = mo2->fsRBasis[0].p3.d[2];
	v0[0] = mo2->fsRBasis[0].p3.d[0];
	v0[1] = mo2->fsRBasis[0].p3.d[1];
	v0[2] = mo2->fsRBasis[0].p3.d[2];

	//get first tangent
	v[0] = mo2->TangList[0].p3.d[0];
	v[1] = mo2->TangList[0].p3.d[1];
	v[2] = mo2->TangList[0].p3.d[2];

	//@ Build first PT-Frame ///////////////////////////
	rotate3D_3(v0, v, angle);
	len=length3D(v0);
	v0[0] /= len;
	v0[1] /= len;
	v0[2] /= len;

	//copy v0 into basisRList
	mo2->ptRBasis[0].p3.d[0] = v0[0];
	mo2->ptRBasis[0].p3.d[1] = v0[1];
	mo2->ptRBasis[0].p3.d[2] = v0[2];

	//compute third component S
	cross3D(v,v0,s);	//#@changed from cross3D(v0,v,s) 
	mo2->ptSBasis[0].p3.d[0] = s[0];
	mo2->ptSBasis[0].p3.d[1] = s[1];
	mo2->ptSBasis[0].p3.d[2] = s[2];

	//compute quaternion affected by rotation by angle
	prev->r = cos(angle/2.0);
	prev->i = v[0]*sin(angle/2.0);
	prev->j = v[1]*sin(angle/2.0);
	prev->k = v[2]*sin(angle/2.0);

	quatQuatProduct( prev, &(mo2->fsQuatList[0]), u);
	quatCopy( u, &(mo2->ptQuatList[0]) );
	
	//@ Calculate for all points
	for(i=0;i<=(limit+2);i++) {

		t1[0] =	mo2->TangList[i].p3.d[0];
		t1[1] =	mo2->TangList[i].p3.d[1];
		t1[2] =	mo2->TangList[i].p3.d[2];

		if(i<limit) {
			t2[0] =	mo2->TangList[i+1].p3.d[0];
			t2[1] =	mo2->TangList[i+1].p3.d[1];
			t2[2] =	mo2->TangList[i+1].p3.d[2];
		}
		else {
			t1[0] =	mo2->TangList[0].p3.d[0];
			t1[1] =	mo2->TangList[0].p3.d[1];
			t1[2] =	mo2->TangList[0].p3.d[2];
		}

		cross3D(t1,t2, cross_T1T2);
		len	= length3D(cross_T1T2);
		//printf("[%d] cross_T1T2 = %f, %f, %f\tlen = %f\n",i,cross_T1T2[0],cross_T1T2[1],cross_T1T2[2],len );

		//get current quaternion
		p->r = mo2->ptQuatList[i].r;
		p->i = mo2->ptQuatList[i].i;
		p->j = mo2->ptQuatList[i].j;
		p->k = mo2->ptQuatList[i].k;
		//if(i>=0 && i<3) {
		//	printf("q[%d] = %f, %f, %f, %f\n",i, mo2->ptQuatList[i].r, mo2->ptQuatList[i].i, mo2->ptQuatList[i].j, mo2->ptQuatList[i].k);
		//}
		//if( !mo2->flag_quat_movingquat && (q0->r*mo2->ptQuatList[1].r < 0.0) ) {
		//	flip = 1;
		//}

		if(len == 0.00) {

			//copy current basisR into next	basisRList
			mo2->ptRBasis[i+1].p3.d[0] = mo2->ptRBasis[i].p3.d[0];
			mo2->ptRBasis[i+1].p3.d[1] = mo2->ptRBasis[i].p3.d[1];
			mo2->ptRBasis[i+1].p3.d[2] = mo2->ptRBasis[i].p3.d[2];
			
			//copy basisRList into v0
			v0[0] =	mo2->ptRBasis[i+1].p3.d[0];
			v0[1] =	mo2->ptRBasis[i+1].p3.d[1];
			v0[2] =	mo2->ptRBasis[i+1].p3.d[2];

			//assign next quaternion
			quatCopy (p, &(mo2->ptQuatList[i+1]) );
			//quatCopy( p, u ); 
		}
		else {

			cross_T1T2[0] /= len;
			cross_T1T2[1] /= len;
			cross_T1T2[2] /= len;

			theta =	acos(dot3D(t1,t2));
			//printf("[%d] theta = %f\n",i,theta);

			rotate3D_3(v0, cross_T1T2, theta);
			
			//copy v0 into basisRList
			mo2->ptRBasis[i+1].p3.d[0] = v0[0];
			mo2->ptRBasis[i+1].p3.d[1] = v0[1];
			mo2->ptRBasis[i+1].p3.d[2] = v0[2];

			//compute next quaternion
			q->r = cos(theta/2.0);
			q->i = cross_T1T2[0]*sin(theta/2.0);
			q->j = cross_T1T2[1]*sin(theta/2.0);
			q->k = cross_T1T2[2]*sin(theta/2.0);
			quatQuatProduct	(q,	p, u);
			quatNormalize(u);

		}

		//asign next quaternion
		mo2->ptQuatList[i+1].r = u->r;
		mo2->ptQuatList[i+1].i = u->i;
		mo2->ptQuatList[i+1].j = u->j;
		mo2->ptQuatList[i+1].k = u->k;

		cross3D(t2,v0,s); //!changed from cross3D(v0,t2,s)
		len	= length3D(s);
		s[0] /=	len;
		s[1] /=	len;
		s[2] /=	len;

		// copy	s into basisSList
		mo2->ptSBasis[i+1].p3.d[0] = s[0];
		mo2->ptSBasis[i+1].p3.d[1] = s[1];
		mo2->ptSBasis[i+1].p3.d[2] = s[2];

	}//$ FOR(i=0,..)
	
	//printf("flip = %d\n",flip);
	if(flip) {
		for(i=0;i<=(limit+1);i++) {
			quatCopy( &(mo2->ptQuatList[i]), p );
			r->r = -p->r;
			r->i = -p->i;
			r->j = -p->j;
			r->k = -p->k;
			quatCopy(r, &(mo2->ptQuatList[i]));				
		}
		flip = 0;
	}
	my_free(q);
	my_free(p);
	my_free(r);
}

//*********************************************//
//
//*********************************************//
void computeGeodesicReferenceFrame 				( void )
{
	int		i		= 0;				// index into points array
	int		limit	= mo2->l_numPoints;	// local variable for number of	points for frame calculation
	double	z[3]	= {0.0, 0.0, 1.0};	// z-hat: reference axis to be tilted into tangent/normals
	double	v[3];						// vector for Tangent
	double	n[3];						// vector for Normal
	double	s[3];						// vector for third	orthonormal	component
	double	cross_ZV[3];				// vector for tangent-tangent cross	product
	double	theta	= 0.0;				// PT initial frame	angle
	double	len		= 0.0;				// vector norm (length)
	quaternion *q;						// quaternion variables

	//initialize
	q =	(quaternion*) malloc(sizeof(quaternion));

    if(mo2->objectID == OBJCURVE && mo2->whichCurve == OBJKNOT)
        limit = mo2->l_numPoints;
    else if(mo2->objectID == OBJCURVE && mo2->whichCurve == OBJCURVELIST)
        limit = mo2->numPoints;

	//@ Compute GEODESIC-REFERENCE FRAME for all points ////////////
	for(i=0; i<limit; i++) {
		
		v[0] = mo2->TangList[i].p3.d[0];
		v[1] = mo2->TangList[i].p3.d[1];
		v[2] = mo2->TangList[i].p3.d[2];

		//#NOTE: have to handle the case when v-hat = -(z-hat)
		//obtain axis 'cross_ZV' for rotation of reference-frame 'z-hat'
		cross3D( z, v, cross_ZV );	
		theta = acos( dot3D( z, v ) );

		//normalize cross_ZV
		len = length3D( cross_ZV );
		cross_ZV[0] /= len; 
		cross_ZV[1] /= len; 
		cross_ZV[2] /= len; 

		cross3D( cross_ZV, v, n );
		len = length3D( n );
		n[0] /= len;
		n[1] /= len;
		n[2] /= len;

		mo2->grRBasis[i].p3.d[0] = n[0];
		mo2->grRBasis[i].p3.d[1] = n[1];
		mo2->grRBasis[i].p3.d[2] = n[2];

		cross3D( v, n, s );
		len = length3D( s );
		s[0] /= len;
		s[1] /= len;
		s[2] /= len;

		mo2->grSBasis[i].p3.d[0] = s[0];
		mo2->grSBasis[i].p3.d[1] = s[1];
		mo2->grSBasis[i].p3.d[2] = s[2];

		q->r =			   cos( theta/2.0 );
		q->i = cross_ZV[0]*sin( theta/2.0 );
		q->j = cross_ZV[1]*sin( theta/2.0 );
		q->k = cross_ZV[2]*sin( theta/2.0 );
		quatNormalize(q);

		//asign quaternion
		quatCopy( q, &(mo2->grQuatList[i]) );
	
	}//$ FOR(i=1,..)
	my_free(q);

}
//*********************************************//
//
//*********************************************//
//NOTE: Don't think I'll use these!
//*********************************************//
//
//*********************************************//
void makeTangent_forQuatMap 					( void	) 
{
    int		i = 0;
    int		limit = 0;
    double	p1[3];
    double	p2[3];				//points on	spline	  
    double	len = 0.0;
    double	v[3];
    double	qvec[3];

    if(mo2->flag_curve_closed)
        limit = mo2->l_numPoints;
    else
        limit = mo2->numPoints-1;

    if(mo2->whichCurve == OBJCURVELIST)
        limit = mo2->numPoints-4;

	//////////////// BUILD Tangents	/////////////////////////////////
    for	( i=0; i<=(limit+1); i++ ) {

        if( mo2->flag_quat_map_current_frame == FRENETSERRET ){
            getQuatComp( mo2->qwhichQuatComponent, &(mo2->fsQuatList[i]), qvec );
            mo2->qpntList[i].p3.d[0] = qvec[0];
            mo2->qpntList[i].p3.d[1] = qvec[1];
            mo2->qpntList[i].p3.d[2] = qvec[2];
        }
        else if( mo2->flag_quat_map_current_frame == PARALLELTRANSPORT ){
            getQuatComp( mo2->qwhichQuatComponent, &(mo2->ptQuatList[i]), qvec );
            mo2->qpntList[i].p3.d[0] = qvec[0];
            mo2->qpntList[i].p3.d[1] = qvec[1];
            mo2->qpntList[i].p3.d[2] = qvec[2];
        }
        else if( mo2->flag_quat_map_current_frame == GEODESICREFERENCE ){
            getQuatComp( mo2->qwhichQuatComponent, &(mo2->grQuatList[i]), qvec );
            mo2->qpntList[i].p3.d[0] = qvec[0];
            mo2->qpntList[i].p3.d[1] = qvec[1];
            mo2->qpntList[i].p3.d[2] = qvec[2];
        }
        else if( mo2->flag_quat_map_current_frame == REFLECTEDQUATLIST ) {
            getQuatComp( mo2->qwhichQuatComponent, &(mo2->refquatList[i]), qvec );
            mo2->qpntList[i].p3.d[0] = qvec[0];
            mo2->qpntList[i].p3.d[1] = qvec[1];
            mo2->qpntList[i].p3.d[2] = qvec[2];
        }
        
        p1[0] =	mo2->qpntList[i].p3.d[0];
        p1[1] =	mo2->qpntList[i].p3.d[1];
        p1[2] =	mo2->qpntList[i].p3.d[2];

        if(i==(limit)) {
            p2[0] =	mo2->qpntList[i+1].p3.d[0];
            p2[1] =	mo2->qpntList[i+1].p3.d[1];
            p2[2] =	mo2->qpntList[i+1].p3.d[2];
        }
        else {
            p2[0] =	mo2->qpntList[i+1].p3.d[0];
            p2[1] =	mo2->qpntList[i+1].p3.d[1];
            p2[2] =	mo2->qpntList[i+1].p3.d[2];
        }
            //printf("[%d] Point2 = %f, %f ,%f\n",i,p2[0],p2[1],p2[2]);

        v[0] = p2[0]-p1[0];
        v[1] = p2[1]-p1[1];
        v[2] = p2[2]-p1[2];
	
        len	= length3D(v);
	
        v[0] = v[0]/len;
        v[1] = v[1]/len;
        v[2] = v[2]/len;

        mo2->qTangList[i].p3.d[0] = v[0];
        mo2->qTangList[i].p3.d[1] = v[1];
        mo2->qTangList[i].p3.d[2] = v[2];
        //printf("[%d] Tang = %f, %f ,%f\n",i,v[0],v[1],v[2]);
                
            /*
        if( i <= limit ) {
                //printf("[%d] Tang = %f, %f ,%f\n",i,v[0],v[1],v[2]);
            glPushMatrix();
            glColor3f(0,0,0);
            
            glTranslatef(mo2->qpntList[i].p3.d[0],mo2->qpntList[i].p3.d[1],mo2->qpntList[i].p3.d[2]);
            glutSolidSphere(0.015,8,8);
            glBegin(GL_LINES);

            glVertex3f(0,0,0);

            glVertex3f(v[0],v[1],v[2]);
            glEnd();
                    
            glPopMatrix();
                    
                    
        }*/
    }
    
}

//*********************************************//
//
//*********************************************//
void makeTangentDerivatives_forQuatMap			( void ) 
{
	int		i = 0;
	int		limit = 0;
	double	t = 0;
	double	len	 = 0.0;
	
	double T0[3];
	double T1[3];
	double TT[3];

    if(mo2->flag_curve_closed)
        limit = mo2->l_numPoints-1;
    else
        limit = mo2->numPoints-2;

    if(mo2->whichCurve == OBJCURVELIST)
        limit = mo2->numPoints-4;

	for(i=0; i<=limit; i++) {

		t=0;
		T0[0] =	mo2->qTangList[i].p3.d[0];
		T0[1] =	mo2->qTangList[i].p3.d[1];
		T0[2] =	mo2->qTangList[i].p3.d[2];
		
		if(i==(limit)) {
			T1[0] =	mo2->qTangList[0].p3.d[0];
			T1[1] =	mo2->qTangList[0].p3.d[1];
			T1[2] =	mo2->qTangList[0].p3.d[2];
		}
		else {
			T1[0] =	mo2->qTangList[i+1].p3.d[0];
			T1[1] =	mo2->qTangList[i+1].p3.d[1];
			T1[2] =	mo2->qTangList[i+1].p3.d[2];
		}
		//else if(i==(limit+1)) {
		//	T1[0] =	mo2->qTangList[1].p3.d[0];
		//	T1[1] =	mo2->qTangList[1].p3.d[1];
		//	T1[2] =	mo2->qTangList[1].p3.d[2];
		//}

		minus3D(T1,T0,TT);

		len	= length3D(TT);
		
		TT[0] =	TT[0]/len;
		TT[1] =	TT[1]/len;
		TT[2] =	TT[2]/len;

		mo2->qdTangList[i].p3.d[0] =	TT[0];
		mo2->qdTangList[i].p3.d[1] =	TT[1];
		mo2->qdTangList[i].p3.d[2] =	TT[2];
		//printf("[%d] dTang = %f, %f ,%f\n",i,TT[0],TT[1],TT[2]);
	}
}

//### RENDERINGS: curve
void drawPoint 									( double *point, float size, int name, GLenum mode)
{
	glPointSize(size);
	if(mode == GL_SELECT){
		glLoadName(name);
	}
	glBegin(GL_POINTS);
		glVertex3f( point[0], point[1], point[2] );
	glEnd();
}
//*********************************************//
//
//*********************************************//
void drawPoints 								( GLenum mode )
{
    int i;
	float size=10.0;
    
    glPushAttrib(GL_ALL_ATTRIB_BITS);
    glDisable(GL_LIGHTING);
    for(i=0; i<mo2->numCtrlPoints; i++){
		if(i == mo2->selected_point){
			glColor3f(1.0f,0.0f,0.0f);
			size = 25;
		}
		else {
			glColor3f(0.0f,0.0f,0.0f);
		}
		drawPoint( (mo2->ctrlPntList[i].p3.d), size, i, mode);
		size = 10.0;
    }
    glPopAttrib();
}
//*********************************************//
//
//*********************************************//
void drawCurveShadow 							( void )
{
	int i=0;
	int	limit = -1;
	float size = 1.25f;
	if(mo2->objectID == OBJCURVE && mo2->whichCurve==OBJSPLINE)
		limit = mo2->numPoints;

	glPushAttrib( GL_ALL_ATTRIB_BITS );
	glDisable( GL_LIGHTING );
	glLineWidth(size);
	glColor3f(0.3f,0.3f,0.3f);
	glPushMatrix();
	glTranslatef(0.0f,-1.0f,0.0f);
	glBegin( GL_LINE_STRIP );
	for(i=0;i<limit;i++) {
		//glVertex3f( mo2->pntList[i].p3.d[0], mo2->pntList[i].p3.d[1], mo2->pntList[i].p3.d[2] );
		glVertex3f( mo2->pntList[i].p3.d[0], mo2->pntList[i].p3.d[1], 0.0 );
	}
	glEnd();
	glDisable(GL_LIGHTING);
	
	for(i=0;i<limit;i++) {
		glPushMatrix();
			//glTranslatef( mo2->pntList[i].p3.d[0], mo2->pntList[i].p3.d[1], mo2->pntList[i].p3.d[2] );
			glTranslatef( mo2->pntList[i].p3.d[0], mo2->pntList[i].p3.d[1], 0.0 );
			glutSolidSphere(0.015,8,8);
		glPopMatrix();
	}
	
	glPopMatrix();
	
	glPushMatrix();
	glTranslatef(0.0f,0.0f,-0.25f);
	glBegin( GL_LINE_STRIP );
	for(i=0;i<limit;i++) {
		glVertex3f( mo2->pntList[i].p3.d[0], 0.0, mo2->pntList[i].p3.d[2] );
	}
	glEnd();
	glDisable(GL_LIGHTING);
	
	for(i=0;i<limit;i++) {
		glPushMatrix();
			glTranslatef( mo2->pntList[i].p3.d[0], 0.0, mo2->pntList[i].p3.d[2] );
			glutSolidSphere(0.015,8,8);
		glPopMatrix();
	}
	
	glPopMatrix();
	glPopAttrib();
	
}

//*********************************************//
//
//*********************************************//
void drawCurve 									( void )
{
	int i=0;
	int	limit = -1;
	float size = 1.25f;
	if(mo2->objectID == OBJCURVE && mo2->whichCurve==OBJSPLINE)
		limit = mo2->numPoints;

	glPushAttrib( GL_ALL_ATTRIB_BITS );
		glDisable( GL_LIGHTING );
		glLineWidth(size);
		glColor3f(0.0f,0.0f,0.0f);
		glBegin( GL_LINE_STRIP );
		for(i=0;i<limit;i++) {
			glVertex3f( mo2->pntList[i].p3.d[0], mo2->pntList[i].p3.d[1], mo2->pntList[i].p3.d[2] );
		}
		glEnd();
		glEnable(GL_LIGHTING);
		glColor3f(0.0f,0.6f,0.6f);
		
		for(i=0;i<limit;i++) {
			glPushMatrix();
				glTranslatef( mo2->pntList[i].p3.d[0], mo2->pntList[i].p3.d[1], mo2->pntList[i].p3.d[2] );
				glutSolidSphere(0.02,8,8);
			glPopMatrix();
		}
	glPopAttrib();

}
//*********************************************//
//
//*********************************************//
void drawKnottedCurve 							( void ) 
{
	int		i = 0;
	float	size = 2.0f;
	int		limit = mo2->l_numPoints-3; 
	int		numslice = 4;
	int		numstacks = 4;
	double	radius = 0.015;

    // generate display lists for spheres
    sphere_disp_list = glGenLists(1);
	glNewList(sphere_disp_list, GL_COMPILE);
        glColor3f(0.8f,0.0f,0.0f);
		glutSolidSphere(radius,numslice,numstacks);
	glEndList();

    GLuint  sphere_disp_list2 = glGenLists(1);
	glNewList(sphere_disp_list2, GL_COMPILE);
		glutSolidSphere(radius+.005,numslice+4,numstacks+4);
	glEndList();


    glPushAttrib(GL_LIGHTING_BIT | GL_LINE_BIT | GL_COLOR_BUFFER_BIT );
		glDisable(GL_LIGHTING);
		glLineWidth(size);
   		glColor3f(0.7f,0.1f,0.1f);
		glBegin(GL_LINE_STRIP);
		for(i=0; i<limit; i++) {
			glVertex3f( mo2->pntList[i].p3.d[0], mo2->pntList[i].p3.d[1], mo2->pntList[i].p3.d[2] );
		}
		glEnd();
		if(!mo2->showGap) {
			glBegin(GL_LINE_STRIP);
				glVertex3f( mo2->pntList[limit-1].p3.d[0], mo2->pntList[limit-1].p3.d[1], mo2->pntList[limit-1].p3.d[2] );
				glVertex3f( mo2->pntList[0].p3.d[0], mo2->pntList[0].p3.d[1], mo2->pntList[0].p3.d[2] );
			glEnd();
		}
		glEnable(GL_LIGHTING);
		glEnable(GL_COLOR_MATERIAL);
        // first point
        limit = 0;
   		glPushMatrix();
			glTranslatef(mo2->pntList[0].p3.d[0], mo2->pntList[0].p3.d[1], mo2->pntList[0].p3.d[2] );
            glColor3f(0.7f,0.7f,0.1f);
            glCallList(sphere_disp_list2);
		glPopMatrix();
		for(i=1; i<limit-1; i+=1) {
			glPushMatrix();
				glTranslatef(mo2->pntList[i].p3.d[0], mo2->pntList[i].p3.d[1], mo2->pntList[i].p3.d[2] );
                glCallList(sphere_disp_list);
			glPopMatrix();
		}
        // last point
        limit = mo2->numPoints - 2;
   		glPushMatrix();
			glTranslatef(mo2->pntList[limit].p3.d[0], mo2->pntList[limit].p3.d[1], mo2->pntList[limit].p3.d[2] );
            glColor3f(0.1f,0.7f,0.1f);
            glCallList(sphere_disp_list2);
		glPopMatrix();

	glPopAttrib();                    

	glDeleteLists(sphere_disp_list,1);
}
//*********************************************//
//
//*********************************************//
void drawTangentsDTangents 						( void )
{
	int		i = 0;
	int		limit = 0;
	double	scale = 0.4;
	float	size = 1.5f;

	if(mo2->flag_curve_closed)
		limit = mo2->numPoints-1; //@ use mo2->l_numPoints-1 for all the frames
	else
		limit = mo2->numPoints-2;
        
	glPushAttrib( GL_LIGHTING_BIT );
	glDisable( GL_LIGHTING );
	glLineWidth( size );
	for(i=0;i<limit;i++) {
        //if(i<mo2->numPoints) {
		glPushMatrix();
			glTranslatef( mo2->pntList[i].p3.d[0], mo2->pntList[i].p3.d[1], mo2->pntList[i].p3.d[2] );
			glScalef(scale,scale,scale);
			glColor3f(0.0f,0.65f,0.85f);
			glBegin(GL_LINES);
				glVertex3f(0.0f,0.0f,0.0f);
				glVertex3f(mo2->TangList[i].p3.d[0],mo2->TangList[i].p3.d[1],mo2->TangList[i].p3.d[2]);
			glEnd();
			glColor3f(0.65f,0.0f,0.85f);
			glBegin(GL_LINES);
				glVertex3f(0.0f,0.0f,0.0f);
				glVertex3f(mo2->dTangList[i].p3.d[0],mo2->dTangList[i].p3.d[1],mo2->dTangList[i].p3.d[2]);
			glEnd();
		glPopMatrix();
        //  }
        //    else {
        //    glPushMatrix();
        //		glTranslatef( mo2->pntList[i].p3.d[0], mo2->pntList[i].p3.d[1], mo2->pntList[i].p3.d[2] );
        //		//glScalef(scale,scale,scale);
        //		glColor3f(0.0f,0.65f,0.85f);
        //		glBegin(GL_LINES);
        //			glVertex3f(0.0f,0.0f,0.0f);
        //			glVertex3f(mo2->TangList[i].p3.d[0],mo2->TangList[i].p3.d[1],mo2->TangList[i].p3.d[2]);
        //		glEnd();
        //		glColor3f(0.65f,0.0f,0.85f);
        //		glBegin(GL_LINES);
        //			glVertex3f(0.0f,0.0f,0.0f);
        //			glVertex3f(mo2->dTangList[i].p3.d[0],mo2->dTangList[i].p3.d[1],mo2->dTangList[i].p3.d[2]);
        //		glEnd();
        //	glPopMatrix();
        //    }
    }
	glPopAttrib();
}
//*********************************************//
//
//*********************************************//
void drawFrames 								( void )
{
	int		i = 0;
	int		limit = 0;
	double	scale = 0.33;
	float	size = 1.5f;

	if(mo2->flag_curve_closed)
		limit = mo2->numPoints-1; //@ use mo2->l_numPoints-1 for all the frames
	else
		limit = mo2->numPoints-1;

    if(mo2->whichCurve == OBJCURVELIST)
        limit = mo2->numPoints;

	glPushAttrib( GL_LIGHTING_BIT );
	glDisable( GL_LIGHTING );
	glLineWidth( size );

	if(mo2->whichFrame == FRENETSERRET) {

		for(i=0;i<limit;i++) {
            if(i<mo2->numPoints) {
			glPushMatrix();
				glTranslatef( mo2->pntList[i].p3.d[0], mo2->pntList[i].p3.d[1], mo2->pntList[i].p3.d[2] );
				glScalef(scale,scale,scale);
				glColor3f(1.0f,0.0f,0.0f);
				glBegin(GL_LINES);
					glVertex3f(0.0f,0.0f,0.0f);
					glVertex3f(mo2->TangList[i].p3.d[0],mo2->TangList[i].p3.d[1],mo2->TangList[i].p3.d[2]);
				glEnd();
				glColor3f(0.0f,1.0f,0.0f);
				glBegin(GL_LINES);
					glVertex3f(0.0f,0.0f,0.0f);
					glVertex3f(mo2->fsRBasis[i].p3.d[0],mo2->fsRBasis[i].p3.d[1],mo2->fsRBasis[i].p3.d[2]);
				glEnd();
				glColor3f(0.0f,0.0f,1.0f);
				glBegin(GL_LINES);
					glVertex3f(0.0f,0.0f,0.0f);
					glVertex3f(mo2->fsSBasis[i].p3.d[0],mo2->fsSBasis[i].p3.d[1],mo2->fsSBasis[i].p3.d[2]);
				glEnd();
			glPopMatrix();
            }
            else {
            glPushMatrix();
				glTranslatef( mo2->pntList[i].p3.d[0], mo2->pntList[i].p3.d[1], mo2->pntList[i].p3.d[2] );
				//glScalef(scale,scale,scale);
				glColor3f(1.0f,0.0f,0.0f);
				glBegin(GL_LINES);
					glVertex3f(0.0f,0.0f,0.0f);
					glVertex3f(mo2->TangList[i].p3.d[0],mo2->TangList[i].p3.d[1],mo2->TangList[i].p3.d[2]);
				glEnd();
				glColor3f(0.0f,1.0f,0.0f);
				glBegin(GL_LINES);
					glVertex3f(0.0f,0.0f,0.0f);
					glVertex3f(mo2->fsRBasis[i].p3.d[0],mo2->fsRBasis[i].p3.d[1],mo2->fsRBasis[i].p3.d[2]);
				glEnd();
				glColor3f(0.0f,0.0f,1.0f);
				glBegin(GL_LINES);
					glVertex3f(0.0f,0.0f,0.0f);
					glVertex3f(mo2->fsSBasis[i].p3.d[0],mo2->fsSBasis[i].p3.d[1],mo2->fsSBasis[i].p3.d[2]);
				glEnd();
			glPopMatrix();
            }
		}	
	}//$ IF(FS)
	else if(mo2->whichFrame == PARALLELTRANSPORT) {
		for(i=0;i<limit;i++) {
			glPushMatrix();
				glTranslatef( mo2->pntList[i].p3.d[0], mo2->pntList[i].p3.d[1], mo2->pntList[i].p3.d[2] );
				glScalef(scale,scale,scale);
				glColor3f(1.0f,0.0f,0.0f);
				glBegin(GL_LINES);
					glVertex3f(0.0f,0.0f,0.0f);
					glVertex3f(mo2->TangList[i].p3.d[0],mo2->TangList[i].p3.d[1],mo2->TangList[i].p3.d[2]);
				glEnd();
				glColor3f(0.0f,1.0f,0.0f);
				glBegin(GL_LINES);
					glVertex3f(0.0f,0.0f,0.0f);
					glVertex3f(mo2->ptRBasis[i].p3.d[0],mo2->ptRBasis[i].p3.d[1],mo2->ptRBasis[i].p3.d[2]);
				glEnd();
				glColor3f(0.0f,0.0f,1.0f);
				glBegin(GL_LINES);
					glVertex3f(0.0f,0.0f,0.0f);
					glVertex3f(mo2->ptSBasis[i].p3.d[0],mo2->ptSBasis[i].p3.d[1],mo2->ptSBasis[i].p3.d[2]);
				glEnd();
			glPopMatrix();
		}	
	}//$ IF(PT)
	else if(mo2->whichFrame == GEODESICREFERENCE) {
		for(i=0;i<limit;i++) {
			glPushMatrix();
				glTranslatef( mo2->pntList[i].p3.d[0], mo2->pntList[i].p3.d[1], mo2->pntList[i].p3.d[2] );
				glScalef(scale,scale,scale);
				glColor3f(1.0f,0.0f,0.0f);
				glBegin(GL_LINES);
					glVertex3f(0.0f,0.0f,0.0f);
					glVertex3f(mo2->TangList[i].p3.d[0],mo2->TangList[i].p3.d[1],mo2->TangList[i].p3.d[2]);
				glEnd();
				glColor3f(0.0f,1.0f,0.0f);
				glBegin(GL_LINES);
					glVertex3f(0.0f,0.0f,0.0f);
					glVertex3f(mo2->grRBasis[i].p3.d[0],mo2->grRBasis[i].p3.d[1],mo2->grRBasis[i].p3.d[2]);
				glEnd();
				glColor3f(0.0f,0.0f,1.0f);
				glBegin(GL_LINES);
					glVertex3f(0.0f,0.0f,0.0f);
					glVertex3f(mo2->grSBasis[i].p3.d[0],mo2->grSBasis[i].p3.d[1],mo2->grSBasis[i].p3.d[2]);
				glEnd();
			glPopMatrix();
		}	
	}//$IF(GR)
	glPopAttrib();
}
void drawFrames_forOptimization					( void )
{
	int		i = 0;
	int		limit = 0;
	//double	scale = 0.33;
	float	size = 1.5f;
    double  v1[3], v2[3], v3[3];
    float	matf[16];

	if(mo2->flag_curve_closed)
		limit = mo2->numPoints-1; //@ use mo2->l_numPoints-1 for all the frames
	else
		limit = mo2->numPoints-1;

	glPushAttrib( GL_LIGHTING_BIT );
	glDisable( GL_LIGHTING );
	glLineWidth( size );

    i = 0;
    v1[0] = mo2->TangList[i].p3.d[0];
    v1[1] = mo2->TangList[i].p3.d[1];
    v1[2] = mo2->TangList[i].p3.d[2];

    v2[0] = mo2->fsRBasis[i].p3.d[0];
    v2[1] = mo2->fsRBasis[i].p3.d[1];
    v2[2] = mo2->fsRBasis[i].p3.d[2];

    v3[0] = mo2->fsSBasis[i].p3.d[0];
    v3[1] = mo2->fsSBasis[i].p3.d[1];
    v3[2] = mo2->fsSBasis[i].p3.d[2];

        for(i=0;i<limit;i++) {
            
            quatToRotationMatrixf( &(mo2->opQuatList[i]), matf );
            glPushMatrix();
                //glMultMatrixf(matf);
                glPushMatrix();
                    glTranslatef( mo2->pntList[i].p3.d[0], mo2->pntList[i].p3.d[1], mo2->pntList[i].p3.d[2] );
                    glScalef(0.2,0.2,0.2);
					glColor3f(0.0f,0.3f,0.9f);
                    glLineWidth(3.0f);
					glBegin(GL_LINES);
					glVertex3f(0.0f,0.0f,0.0f);
					glVertex3f(matf[0],matf[1],matf[2]);
					glEnd();
                glPopMatrix();
            glPopMatrix();
		}	

	glPopAttrib();
}

//*********************************************//
//
//*********************************************//
void drawLabels 								( void ) 
{
    int i = 0;
    int limit = 0;

    if(mo2->flag_curve_closed)
        limit = mo2->numPoints;
    else
        limit = mo2->numPoints-2;

    glPushAttrib( GL_ALL_ATTRIB_BITS );
    glDisable(GL_LIGHTING);
    glColor3f(0.2f,0.2f,0.2f);
    for(i=0;i<limit;i+=10) {
		glPushMatrix();
		glTranslatef(mo2->pntList[i].p3.d[0],mo2->pntList[i].p3.d[1],mo2->pntList[i].p3.d[2]);
		sprintf(obj_qstr,"%d",i);
		glRasterPos2f(0.025,0.025);
		mo2->qfont = GLUT_BITMAP_HELVETICA_10;
		stext(obj_qstr);
		glPopMatrix();
    }
    glPopAttrib();
}

//*********************************************//
//
//*********************************************//
void constructFramedCurve						( double radius ) 
{
	int		i				= 0;
	int		j				= 0;
	int		start			= 0;
	int		renderpoints	= 0;
	double	theta			= 0.0;
	double	dtheta			= 0.0;
	double	n_radius		= 2.0*radius;
	int		radial_steps	= mo2->radial_steps;
	double	rlist[3];
	double	slist[3];
	
	//@ Initializations
	dtheta = 2.0*M_PI/mo2->radial_steps;
	renderpoints = mo2->l_numPoints-1;
//#ifdef _DBG_
//	//printf("inside obj::constructFramedCurve()\n");
//#endif
	if(mo2->whichCurve==OBJSPLINE) {
		if( mo2->numCtrlPoints>3) 
			renderpoints = mo2->numPoints-1;
		else
			renderpoints = 0;
	}
    else if (mo2->whichCurve == OBJCURVELIST) {
        renderpoints = mo2->numPoints;
        //printf("\t Render points = %d\n",renderpoints);
    }

	//@ FRAMES
	for	( i=start; i<renderpoints; i++ ) {

			theta= 0.0;
			//printf("%d > %f,%f,%f\n",i,mo2->pntList[i].p3.d[0],mo2->pntList[i].p3.d[1],mo2->pntList[i].p3.d[2]);
			switch(mo2->whichFrame) {

				case FRENETSERRET:
					rlist[0] = mo2->fsRBasis[i].p3.d[0];
					rlist[1] = mo2->fsRBasis[i].p3.d[1];
					rlist[2] = mo2->fsRBasis[i].p3.d[2];
					slist[0] = mo2->fsSBasis[i].p3.d[0];
					slist[1] = mo2->fsSBasis[i].p3.d[1];
					slist[2] = mo2->fsSBasis[i].p3.d[2];
					//slist[0] = mo2->fsRBasis[i].p3.d[0];
					//slist[1] = mo2->fsRBasis[i].p3.d[1];
					//slist[2] = mo2->fsRBasis[i].p3.d[2];
					//rlist[0] = mo2->fsSBasis[i].p3.d[0];
					//rlist[1] = mo2->fsSBasis[i].p3.d[1];
					//rlist[2] = mo2->fsSBasis[i].p3.d[2];
					break;
				case PARALLELTRANSPORT:
					rlist[0] = mo2->ptRBasis[i].p3.d[0];
					rlist[1] = mo2->ptRBasis[i].p3.d[1];
					rlist[2] = mo2->ptRBasis[i].p3.d[2];
					slist[0] = mo2->ptSBasis[i].p3.d[0];
					slist[1] = mo2->ptSBasis[i].p3.d[1];
					slist[2] = mo2->ptSBasis[i].p3.d[2];
					break;
				case GEODESICREFERENCE:
					rlist[0] = mo2->grRBasis[i].p3.d[0];
					rlist[1] = mo2->grRBasis[i].p3.d[1];
					rlist[2] = mo2->grRBasis[i].p3.d[2];
					slist[0] = mo2->grSBasis[i].p3.d[0];
					slist[1] = mo2->grSBasis[i].p3.d[1];
					slist[2] = mo2->grSBasis[i].p3.d[2];
					//rlist[0] = mo2->ptRBasis[i].p3.d[0];
					//rlist[1] = mo2->ptRBasis[i].p3.d[1];
					//rlist[2] = mo2->ptRBasis[i].p3.d[2];
					//slist[0] = mo2->ptSBasis[i].p3.d[0];
					//slist[1] = mo2->ptSBasis[i].p3.d[1];
					//slist[2] = mo2->ptSBasis[i].p3.d[2];					
					break;

				default:
					break;

			}
			for	( j=0; j<radial_steps; j++ )	{
				
				//printf("[%d][%d] \nrlist = %f, %f, %f\nslist = %f, %f, %f\npntList=  = %f, %f, %f\n",i,j,rlist[0],rlist[1],rlist[2],slist[0],slist[1],slist[2],
				//	mo2->pntList[i].p3.d[0],mo2->pntList[i].p3.d[1],mo2->pntList[i].p3.d[2]);
				mo2->rpnts[i][j][0]	= mo2->pntList[i].p3.d[0] +	radius*rlist[0]*cos(theta) + radius*slist[0]*sin(theta);
				mo2->rpnts[i][j][1]	= mo2->pntList[i].p3.d[1] +	radius*rlist[1]*cos(theta) + radius*slist[1]*sin(theta);
				mo2->rpnts[i][j][2]	= mo2->pntList[i].p3.d[2] +	radius*rlist[2]*cos(theta) + radius*slist[2]*sin(theta);
				
				mo2->npnts[i][j][0]	= n_radius*rlist[0]*cos(theta)+n_radius*slist[0]*sin(theta);
				mo2->npnts[i][j][1]	= n_radius*rlist[1]*cos(theta)+n_radius*slist[1]*sin(theta);
				mo2->npnts[i][j][2]	= n_radius*rlist[2]*cos(theta)+n_radius*slist[2]*sin(theta);
				
				theta =	theta +	dtheta;
			}
	}
}
//*********************************************//
//
//*********************************************//
void renderFramedCurve 							( double	radius ) 
{
	int		i				= 0;
	int		j				= 0;
	int		p				= 0;
	int		start			= 0;
	int		renderpoints	= 0;
	int		radial_steps	= mo2->radial_steps;
	double	t				= mo2->texParam;
	double	l_radius = radius;
	l_radius = 0.0;
	//@ Initializations
	renderpoints = mo2->numPoints-1;
	if(mo2->whichCurve==OBJSPLINE) {
		if( mo2->numCtrlPoints>3) 
			renderpoints = mo2->numPoints-1;
		else
			renderpoints = 0;
	}
    else if(mo2->whichCurve==OBJCURVELIST) {
        renderpoints = mo2->numPoints;
        //printf("\t Render points = %d (renderFramed...)\n",renderpoints);
    }

	//@ EDGES
	if(mo2->showEdge) {
		glPushAttrib( GL_LIGHTING_BIT );
		glDisable(GL_LIGHTING);
		//glEnable(GL_COLOR_MATERIAL);
		
		if(mo2->showFace)
			glLineWidth(0.75);
		else
			glLineWidth(1.5);
		
		// ring 
		for(i=start; i<(renderpoints); i++) {
			
			glBegin(GL_LINE_STRIP);
			glColor3f(0.0,0.0,0.0);
			for(j=0; j<=radial_steps; j++) {
				if(j<(radial_steps))
					glVertex3d(mo2->rpnts[i][j][0],mo2->rpnts[i][j][1],mo2->rpnts[i][j][2]);
			else
				glVertex3d(mo2->rpnts[i][0][0],mo2->rpnts[i][0][1],mo2->rpnts[i][0][2]);
			}
			glEnd();
		}
		
		///////////////////////// longitudinal lines /////////////////////////
		glBegin(GL_LINES);
		glColor3f(0.0,0.0,0.0);
		for(i=start+1;i<(renderpoints); i++) {
			if(i<(renderpoints)) {
				for(j=0;j<radial_steps;j++) {
					glVertex3d(mo2->rpnts[i][j][0],	mo2->rpnts[i][j][1], mo2->rpnts[i][j][2]); 
					glVertex3d(mo2->rpnts[i-1][j][0],mo2->rpnts[i-1][j][1],mo2->rpnts[i-1][j][2]);
					if(j<(radial_steps-1)){
						glVertex3d(mo2->rpnts[i-1][j+1][0],mo2->rpnts[i-1][j+1][1],mo2->rpnts[i-1][j+1][2]);
						glVertex3d(mo2->rpnts[i][j+1][0],mo2->rpnts[i][j+1][1],mo2->rpnts[i][j+1][2]);
					}
					else {
						glVertex3d(mo2->rpnts[i-1][j-radial_steps+1][0],	mo2->rpnts[i-1][j-radial_steps+1][1],mo2->rpnts[i-1][j-radial_steps+1][2]	);
						glVertex3d(mo2->rpnts[i][j-radial_steps+1][0],	mo2->rpnts[i][j-radial_steps+1][1],	mo2->rpnts[i][j-radial_steps+1][2]);
					}
				}
			}
		}
		
        // draw	last section
		if(!mo2->showGap && mo2->flag_curve_closed)	{
			i=0;
			p=renderpoints-1;
			for(j=0;j<radial_steps;j++) {
				glVertex3d(mo2->rpnts[i][j][0],	mo2->rpnts[i][j][1], mo2->rpnts[i][j][2]); 
				glVertex3d(mo2->rpnts[p][j][0],	mo2->rpnts[p][j][1], mo2->rpnts[p][j][2]);
				if(j<(radial_steps-1)){
					glVertex3d(mo2->rpnts[p][j+1][0], mo2->rpnts[p][j+1][1], mo2->rpnts[p][j+1][2]);
					glVertex3d(mo2->rpnts[i][j+1][0], mo2->rpnts[i][j+1][1], mo2->rpnts[i][j+1][2]);
				}
				else {
					glVertex3d(mo2->rpnts[p][j-radial_steps+1][0], mo2->rpnts[p][j-radial_steps+1][1], mo2->rpnts[p][j-radial_steps+1][2] );
					glVertex3d(mo2->rpnts[i][j-radial_steps+1][0], mo2->rpnts[i][j-radial_steps+1][1], mo2->rpnts[i][j-radial_steps+1][2]);
				}
			}
		}
		glEnd();
        glPopAttrib();
                
	}		
	//@ NORMALS
	if(mo2->showNormals) {
			///////////////////////// normals /////////////////////////
		glPushAttrib( GL_ALL_ATTRIB_BITS );
		glDisable(GL_LIGHTING);
		glLineWidth(0.5);
		glColor3f(0.95f,0.95f,0.0f);
		for(i=start; i<renderpoints; i++) {
			//printf("%d > %f,%f,%f\n",i,mo2->fstlist[i].p3.d[0],mo2->fstlist[i].p3.d[1],mo2->fstlist[i].p3.d[2]);

			for(j=0; j<radial_steps;	j++) {
				glPushMatrix();
				glTranslatef(mo2->rpnts[i][j][0], mo2->rpnts[i][j][1], mo2->rpnts[i][j][2]);
				glBegin(GL_LINES);
					glVertex3d(0.0,0.0,0.0);
						glVertex3d(mo2->npnts[i][j][0],	mo2->npnts[i][j][1], mo2->npnts[i][j][2]);
					glEnd();
				glPopMatrix();
			}
			
		}
		glPopAttrib();
	}

	//@ SURFACE
	if(mo2->showFace) {
		///////////////////////// quad faces /////////////////////////
		glEnable(GL_COLOR_MATERIAL);
		glEnable(GL_LIGHTING);
		glEnable(GL_POLYGON_OFFSET_FILL);
		glPolygonOffset(1.0, 1.0);

			if(mo2->showTexture) {
                glEnable(GL_TEXTURE_2D);
				glBindTexture(GL_TEXTURE_2D, mo2->texName[0]);
			}

			glBegin( GL_QUADS );
			for(i=start+1;i<(renderpoints); i++) {
				if(i==mo2->trackingRing)
					glColor3f(0.9f,0.7f,0.0f);
				else if(mo2->showColorCoded != CODE_SOLID_COLOR)
                    //CODE_CURVATURE || mo2->showColorCoded==CODE_TORSION  || 
                    //mo2->showColorCoded==CODE_QUAT_TURN_ANGLE || mo2->showColorCoded==CODE_QUAT_DOTPROD )
					glColor4fv(mo2->palette[mo2->pntList[i].color]);
				else
					glColor3f(mo2->fgcolor[0],mo2->fgcolor[1],mo2->fgcolor[2]);
				

				for(j=0;j<radial_steps;j++) {
					
					//glTexCoord2f(1.0, 1.0);
					glTexCoord2f(t, t);
					glNormal3d(mo2->npnts[i][j][0],	mo2->npnts[i][j][1], mo2->npnts[i][j][2]);
					glVertex3d(mo2->rpnts[i][j][0],	mo2->rpnts[i][j][1], mo2->rpnts[i][j][2]);
					//printf("normals = %f, %f, %f\n",mo2->npnts[i][j][0],mo2->npnts[i][j][1],mo2->npnts[i][j][2]);
					
					//glTexCoord2f(1.0, 0.0);
					glTexCoord2f(t, 0.0);
					glNormal3d(mo2->npnts[i-1][j][0], mo2->npnts[i-1][j][1], mo2->npnts[i-1][j][2]);
					glVertex3d(mo2->rpnts[i-1][j][0], mo2->rpnts[i-1][j][1], mo2->rpnts[i-1][j][2]);
					if(j<(radial_steps-1)){
						glTexCoord2f(0.0, 0.0);
						glNormal3d(mo2->npnts[i-1][j+1][0],	mo2->npnts[i-1][j+1][1], mo2->npnts[i-1][j+1][2]);
						glVertex3d(mo2->rpnts[i-1][j+1][0],	mo2->rpnts[i-1][j+1][1], mo2->rpnts[i-1][j+1][2]);
						
						//glTexCoord2f(0.0, 1.0);
						glTexCoord2f(0.0,	t);
						glNormal3d(mo2->npnts[i][j+1][0],mo2->npnts[i][j+1][1],mo2->npnts[i][j+1][2]);
						glVertex3d(mo2->rpnts[i][j+1][0],mo2->rpnts[i][j+1][1],mo2->rpnts[i][j+1][2]);
					}
					else {
						glTexCoord2f(0.0, 0.0);
						glNormal3d(mo2->npnts[i-1][j-radial_steps+1][0],	mo2->npnts[i-1][j-radial_steps+1][1], mo2->npnts[i-1][j-radial_steps+1][2] );
						glVertex3d(mo2->rpnts[i-1][j-radial_steps+1][0],	mo2->rpnts[i-1][j-radial_steps+1][1], mo2->rpnts[i-1][j-radial_steps+1][2] );
						
						//glTexCoord2f(0.0, 1.0);
						glTexCoord2f(0.0,	t);
						glNormal3d(mo2->npnts[i][j-radial_steps+1][0], mo2->npnts[i][j-radial_steps+1][1], mo2->npnts[i][j-radial_steps+1][2]);
						glVertex3d(mo2->rpnts[i][j-radial_steps+1][0], mo2->rpnts[i][j-radial_steps+1][1], mo2->rpnts[i][j-radial_steps+1][2]);
					}
				}
			}
			
			// draw	last section
			if(!mo2->showGap && mo2->flag_curve_closed) {
				i=0;
				p=renderpoints-1;
				if(i==mo2->trackingRing)
					glColor3f(0.9f,0.7f,0.0f);
                else if( mo2->showColorCoded != CODE_SOLID_COLOR ) {
                    //CODE_CURVATURE || mo2->showColorCoded==CODE_TORSION || 
                    //mo2->showColorCoded==CODE_QUAT_TURN_ANGLE || mo2->showColorCoded==CODE_QUAT_DOTPROD ) {
					glColor4fv(mo2->palette[mo2->pntList[i].color]);
                    // printf("mo2->pntList[i].color = %d\n",mo2->pntList[i].color);
                    }
				else
					glColor3f(mo2->fgcolor[0],mo2->fgcolor[1],mo2->fgcolor[2]);
				for(j=0;j<radial_steps;j++) {
					
					//glTexCoord2f(1.0,	1.0);
					glTexCoord2f(t, t);
					glNormal3d(mo2->npnts[i][j][0],	mo2->npnts[i][j][1], mo2->npnts[i][j][2]);
					glVertex3d(mo2->rpnts[i][j][0],	mo2->rpnts[i][j][1], mo2->rpnts[i][j][2]);
					
					//glTexCoord2f(1.0,	0.0);
					glTexCoord2f(t, 0.0);
					glNormal3d(mo2->npnts[p][j][0],	mo2->npnts[p][j][1], mo2->npnts[p][j][2]);
					glVertex3d(mo2->rpnts[p][j][0],	mo2->rpnts[p][j][1], mo2->rpnts[p][j][2]);
					if(j<(radial_steps-1)){
						glTexCoord2f(0.0,	0.0);
						glNormal3d(mo2->npnts[p][j+1][0], mo2->npnts[p][j+1][1], mo2->npnts[p][j+1][2]);
						glVertex3d(mo2->rpnts[p][j+1][0], mo2->rpnts[p][j+1][1], mo2->rpnts[p][j+1][2]);
						
						//glTexCoord2f(0.0,	1.0);
						glTexCoord2f(0.0,	t);
						glNormal3d(mo2->npnts[i][j+1][0],mo2->npnts[i][j+1][1],mo2->npnts[i][j+1][2]);
						glVertex3d(mo2->rpnts[i][j+1][0],mo2->rpnts[i][j+1][1],mo2->rpnts[i][j+1][2]);
					}
					else {
						glTexCoord2f(0.0,	0.0);
						glNormal3d(mo2->npnts[p][j-radial_steps+1][0], mo2->npnts[p][j-radial_steps+1][1], mo2->npnts[p][j-radial_steps+1][2] );
						glVertex3d(mo2->rpnts[p][j-radial_steps+1][0], mo2->rpnts[p][j-radial_steps+1][1], mo2->rpnts[p][j-radial_steps+1][2] );
						
						//glTexCoord2f(0.0,	1.0);
						glTexCoord2f(0.0,	t);
						glNormal3d(mo2->npnts[i][j-radial_steps+1][0], mo2->npnts[i][j-radial_steps+1][1], mo2->npnts[i][j-radial_steps+1][2]);
						glVertex3d(mo2->rpnts[i][j-radial_steps+1][0], mo2->rpnts[i][j-radial_steps+1][1], mo2->rpnts[i][j-radial_steps+1][2]);
					}
				}//end for()
			}//end if()
			glEnd();
			
			
 			glDisable(GL_TEXTURE_2D);
	}//end if()
}

//### RENDERINGS: quaternion space
void drawQuaternionUnitSphere					( void )
{
    //draw unit sphere
    if(mo2->qshowSphere) {
		glPushAttrib(GL_LINE_BIT);
        glLineWidth(0.5f);
        glDisable(GL_LIGHTING);
        glColor3f(0.5f,0.5f,0.5f);
        glutWireSphere(1.0,16,16);
		glPopAttrib();
    }
}

//*********************************************//
//
//*********************************************//
void drawQuaternionCurve 						( void )
{
	int		i = 0;
	int		limit = 0;
	double	qvec[3];
	float	size = 3.0f;
    //double  qfgcolor[3];
	double	radius = 0.03;
	//double	cubesize = 0.03;
	int		numslice = 6;
	int		numstacks = 6;

	sphere_disp_list = glGenLists(1);
	glNewList(sphere_disp_list, GL_COMPILE);
		glutSolidSphere(0.02,numslice,numstacks);
	glEndList();

    
	if(mo2->objectID == OBJCURVE) {
		if(mo2->flag_curve_closed)
			limit = mo2->l_numPoints-3;
		else
			limit = mo2->numPoints-2;
	}
	//else if(mo2->objectID == OBJSURFACE) {
	//	limit = t_pnts*t_stps;
	//}

	//FRENET-SERRET
	if(mo2->selectFrame_FS) {
		glPushAttrib(GL_LIGHTING_BIT | GL_POINT_BIT | GL_LINE_BIT);
		glEnable(GL_LIGHTING);
		glEnable(GL_COLOR_MATERIAL);
		glLineWidth(size);
		//glPointSize(5.0f);
		radius = 0.02;
		glColor3f(0.0f,0.0f,0.8f);

		for(i=0; i<limit; i+=1) {
			getQuatComp(mo2->qwhichQuatComponent, &(mo2->fsQuatList[i]), qvec);
			//glBegin(GL_POINTS);
			//	glVertex3dv(qvec);
			//glEnd();
			glPushMatrix();
                if(mo2->whichCurve != OBJCURVELIST) {
	    			glTranslatef( qvec[0], qvec[1], qvec[2] );
		    		glCallList(sphere_disp_list);
                }
				if(mo2->showLabels) {
					glPushAttrib(GL_LIGHTING_BIT);
						glDisable(GL_LIGHTING);
						sprintf(obj_qstr,"%d",i);
						glRasterPos2f(0.025,0.025);
						mo2->qfont = GLUT_BITMAP_HELVETICA_10;
						stext(obj_qstr);
					glPopAttrib();				
				}
			glPopMatrix();
		}

		//curve
		glColor3f(0.0f,0.0f,0.5f);
		glDisable(GL_LIGHTING);
		glBegin(GL_LINE_STRIP);
		for(i=1; i<limit; i++) {
			getQuatComp(mo2->qwhichQuatComponent, &(mo2->fsQuatList[i]), qvec);
			glVertex3f( qvec[0], qvec[1], qvec[2] );
		}
		glEnd();
		glPopAttrib();
	}

	//PARALLEL TRANSPORT
	if(mo2->selectFrame_PT) {
	
		glPushAttrib(GL_LIGHTING_BIT | GL_POINT_BIT | GL_LINE_BIT);
			glEnable(GL_LIGHTING);
			glEnable(GL_COLOR_MATERIAL);
			glLineWidth(size);
			//glPointSize(5.0f);
			glColor3f(0.0f,0.6f,0.0f);
			for(i=0; i<limit; i+=1) {
			
				getQuatComp(mo2->qwhichQuatComponent, &(mo2->ptQuatList[i]), qvec);
					glPushMatrix();
						glTranslatef( qvec[0], qvec[1], qvec[2] );
						glCallList(sphere_disp_list);
						if(mo2->showLabels) {
							glPushAttrib(GL_LIGHTING_BIT);
								glDisable(GL_LIGHTING);
								sprintf(obj_qstr,"%d",i);
								glRasterPos2f(0.025,0.025);
								mo2->qfont = GLUT_BITMAP_HELVETICA_10;
								stext(obj_qstr);
							glPopAttrib();				
						}
					glPopMatrix();
			}

			//curve
			
			glDisable(GL_LIGHTING);
			glBegin(GL_LINE_STRIP);
			for(i=0; i<limit; i++) {
				getQuatComp(mo2->qwhichQuatComponent, &(mo2->ptQuatList[i]), qvec);
				glVertex3f( qvec[0], qvec[1], qvec[2] );
			}
			glEnd();
		glPopAttrib();
	}

	//GEODESIC REFERENCE
	if(mo2->selectFrame_GR) {

		glPushAttrib(GL_LIGHTING_BIT | GL_POINT_BIT | GL_LINE_BIT);
		glEnable(GL_LIGHTING);
		glEnable(GL_COLOR_MATERIAL);
		glLineWidth(size);
		//glPointSize(5.0f);
		glColor3f(0.85f,0.0f,0.0f);
		for(i=0; i<limit; i+=1) {
			getQuatComp(mo2->qwhichQuatComponent, &(mo2->grQuatList[i]), qvec);
			//glBegin(GL_POINTS);
			//	glVertex3dv(qvec);
			//glEnd();
			glPushMatrix();
				glTranslatef( qvec[0], qvec[1], qvec[2] );
				glCallList(sphere_disp_list);
				//glutSolidSphere(radius,numslice,numstacks);
				//glutSolidCube(cubesize);
				if(mo2->showLabels) {
					glPushAttrib(GL_LIGHTING_BIT);
						glDisable(GL_LIGHTING);
						sprintf(obj_qstr,"%d",i);
						glRasterPos2f(0.025,0.025);
						mo2->qfont = GLUT_BITMAP_HELVETICA_10;
						stext(obj_qstr);
					glPopAttrib();				
				}
			glPopMatrix();
		}

		//curve
		glDisable(GL_LIGHTING);
		glBegin(GL_LINE_STRIP);
		for(i=0; i<limit; i++) {
			getQuatComp(mo2->qwhichQuatComponent, &(mo2->grQuatList[i]), qvec);
			glVertex3f( qvec[0], qvec[1], qvec[2] );
		}
		glEnd();
		glPopAttrib();
	}
     
	glDeleteLists(sphere_disp_list,1);
}


//*********************************************//
//
//*********************************************//
void drawQuaternionCurve_OPT					( void )
{
	int		i = 0;
	int		limit = 0;
	double	qvec[3];
	float	size = 2.25f;
    //double  qfgcolor[3];
	double	radius = 0.0;
    
	if(mo2->objectID == OBJCURVE) {
		if(mo2->flag_curve_closed)
			limit = mo2->numPoints;
		else
			limit = mo2->numPoints-2;
	}

	glPushAttrib(GL_LIGHTING_BIT | GL_LINE_BIT | GL_LINE_BIT);
		glEnable(GL_LIGHTING);
		glEnable(GL_COLOR_MATERIAL);
		glLineWidth(size);
		radius = 0.0225;
        //qfgcolor[0] = 0.5f; qfgcolor[1] = 0.3f; qfgcolor[2] = 0.0f;
		glColor3f(0.9f,0.7f,0.0f);
		for(i=0; i< limit; i+=1) {
		
			//if(i==0) 
			//	glColor3f(0.3f,0.3f,0.0f);
			//else if(i%5==0 && mo2->showLabels)
			//	glColor3f(qfgcolor[0]-0.0,qfgcolor[1]-0.0,qfgcolor[2]-0.0);
			//else if(i==mo2->opt_INDEX)
			//	glColor3f(0.9f,0.8f,0.0f);
			//else
			//	glColor3f(qfgcolor[0],qfgcolor[1],qfgcolor[2]);
			//
			//if(i==mo2->index_q0)
			//	glColor3f(0.0f,0.0f,1.0f);
			//if(i==mo2->index_q1)
			//	glColor3f(1.0f,1.0f,1.0f);
			//if(i==mo2->index_q2)
			//	glColor3f(0.0f,1.0f,0.0f);
			if(i==mo2->numPoints-mo2->opt_INDEX)
				glColor3f(0.7f,0.0f,0.0f);
			else
				glColor3f(0.3f,0.0f,0.6f);
			getQuatComp(mo2->qwhichQuatComponent, &(mo2->opQuatList[i]), qvec);
            radius = 0.0175;

            glPushMatrix();
				glTranslatef( qvec[0], qvec[1], qvec[2] );
				glutSolidSphere(radius,16,16);
				if(mo2->showLabels && i%5==0) {
					glPushAttrib(GL_LIGHTING_BIT);
						glDisable(GL_LIGHTING);
						sprintf(obj_qstr,"%d",i);
						glRasterPos2f(0.025,0.025);
						mo2->qfont = GLUT_BITMAP_HELVETICA_10;
						stext(obj_qstr);
					glPopAttrib();				
				}
			glPopMatrix();
		}

		//curve
		glDisable(GL_LIGHTING);
		glLineWidth(3.0f);
		glColor3f(0.8f,0.8f,0.0f);
		glBegin(GL_LINE_STRIP);
		for(i=0; i<limit; i++) {
			getQuatComp(mo2->qwhichQuatComponent, &(mo2->opQuatList[i]), qvec);
			glVertex3f( qvec[0], qvec[1], qvec[2] );
		}
		glEnd();
	glPopAttrib();
}

//*********************************************//
//
//*********************************************//
void drawQuaternionReflectedCurve				( void )
{
    int i = 0;
    int limit = 0;
    double qvec[3];
    float size = 3.0f;
    
    if(mo2->flag_curve_closed)
        limit = mo2->numPoints;
    else
        limit = mo2->numPoints-2;

	//@ Draw Reflected Quaternion curve
    glPushAttrib(GL_LIGHTING_BIT | GL_LINE_BIT);
    glEnable(GL_LIGHTING);
    glEnable(GL_COLOR_MATERIAL);
    glLineWidth(size);

    //@ points
    for(i=0; i<limit; i++) {
		
        if(i==0) 
            glColor3f(0.9f,0.7f,0.0f);
        else if(i%5==0 && mo2->showLabels)
            glColor3f(0.5f,0.0f,0.5f);
        else if(i==limit-1)
            glColor3f(0.0f, 1.0f, 0.0f);
        else
            glColor3f(0.6f,0.0f,0.6f);
        getQuatComp(mo2->qwhichQuatComponent, &(mo2->refquatList[i]), qvec);

        glPushMatrix();
        //glTranslatef( qvec[0], qvec[1], qvec[2] );
        //glutSolidSphere(0.02,16,16);
        if(mo2->showLabels && i%5==0) {
            glPushAttrib(GL_LIGHTING_BIT);
            glDisable(GL_LIGHTING);
            glColor3f(1.0f,0.7f,1.0f);
            sprintf(obj_qstr,"%d",i);
            glRasterPos2f(0.025,0.025);
            mo2->qfont = GLUT_BITMAP_HELVETICA_10;
            stext(obj_qstr);
            glPopAttrib();				
        }
        glPopMatrix();
    }

    //@ curve
    glColor3f(0.8f,0.3f,0.8f);
    glDisable(GL_LIGHTING);
    glBegin(GL_LINE_STRIP);
    for(i=0; i<limit; i++) {
        getQuatComp(mo2->qwhichQuatComponent, &(mo2->refquatList[i]), qvec);
        glVertex3f( qvec[0], qvec[1], qvec[2] );
    }
    glEnd();
    glPopAttrib();
	
}


//*********************************************//
//
//*********************************************//
void drawQuaternionPoints						( void )
{
	//int		i = 0;
	int		limit = 0;
	if(mo2->flag_curve_closed)
		limit = mo2->numPoints;
	else
		limit = mo2->numPoints-2;

	glPushAttrib( GL_ALL_ATTRIB_BITS );
	if( mo2->flag_quat_map_current_frame == FRENETSERRET )
		glColor3f(0.0f, 0.4f, 0.4f);
	else if( mo2->flag_quat_map_current_frame == PARALLELTRANSPORT )
		glColor3f(0.0f, 0.4f, 0.0f);
	else if( mo2->flag_quat_map_current_frame == GEODESICREFERENCE )
		glColor3f(0.4f, 0.0f, 0.0f);

	//for(i=0;i<limit;i++) {
	//	glPushMatrix();
	//		glTranslatef( mo2->qpntList[i].p3.d[0], mo2->qpntList[i].p3.d[1], mo2->qpntList[i].p3.d[2] );
	//		glutSolidSphere(0.01,16,16);
	//	glPopMatrix();
	//}
	//glPopAttrib();	
	//glPushAttrib(GL_ALL_ATTRIB_BITS);
	//glColor3f(0.0f,0.0f,0.0f);
	//for(i=0;i<limit;i++) {
	//	glPushMatrix();
	//		glTranslatef( mo2->qpntList[i].p3.d[0], mo2->qpntList[i].p3.d[1], mo2->qpntList[i].p3.d[2] );
	//		glScalef(scale,scale,scale);
	//		glBegin( GL_LINES );
	//			glVertex3f(0.0f,0.0f,0.0f);
	//			glVertex3f(mo2->qTangList[i].p3.d[0], mo2->qTangList[i].p3.d[1], mo2->qTangList[i].p3.d[2]);
	//		glEnd();
	//	glPopMatrix();
	//}
	//glPopAttrib();	
	//glPushAttrib(GL_ALL_ATTRIB_BITS);
	//glColor3f(1.0f,1.0f,1.0f);
	//for(i=0;i<limit;i++) {
	//	glPushMatrix();
	//		glTranslatef( mo2->qpntList[i].p3.d[0], mo2->qpntList[i].p3.d[1], mo2->qpntList[i].p3.d[2] );
	//		glScalef(scale,scale,scale);
	//		glBegin( GL_LINES );
	//			glVertex3f(0.0f,0.0f,0.0f);
	//			glVertex3f(mo2->qTangList[i].p3.d[0], mo2->qTangList[i].p3.d[1], mo2->qTangList[i].p3.d[2]);
	//		glEnd();
	//	glPopMatrix();
	//}
	//glPopAttrib();	
}

//*********************************************//
//
//*********************************************//
void computeFrenetSerretFrame_forSingleFrame_forQuatMap ( void ) 
{
	int		i = 0;						
	int		limit = 1;							// total points to compute for
	double	normB;								// lenght of various vectors
	double	b[3];								// dot and cross product results
	double	n[3];								// vector for cross	product	of Binormal	and	Tangent
	double	t[3];								// vector for Tangent
	double	tt[3];								// vector for first	derivative of Tangent

	//for asigning quaternions
	double		mat[16];						// 4x4 matrix
	quaternion	*q, *p, *r;						// quaternion
	
	//init quaternions
	p =	(quaternion*) my_malloc(sizeof(quaternion));
	q =	(quaternion*) my_malloc(sizeof(quaternion));
	r =	(quaternion*) my_malloc(sizeof(quaternion));
	
	for(i=0; i<limit; i++) {

		//### Compute Frenet Frame ###
		t[0] = mo2->qTangList[i].p3.d[0];
		t[1] = mo2->qTangList[i].p3.d[1];
		t[2] = mo2->qTangList[i].p3.d[2];

		tt[0] =	mo2->qdTangList[i].p3.d[0];
		tt[1] =	mo2->qdTangList[i].p3.d[1];
		tt[2] =	mo2->qdTangList[i].p3.d[2];
	
		//obtain cross product of tangent and derivative of	tangent
		cross3D(t, tt, b);

		//obtain length	of vector cross_vt
		normB = length3D(b);
		
		//normalize	vector cross_vt
		b[0] =	b[0]/normB;
		b[1] =	b[1]/normB;
		b[2] =	b[2]/normB;

		//save cross_vt into	vector list
		mo2->fsSBasis2[i].p3.d[0] = b[0];//NOTE! changed from mo2->fsRBasis2[i]...
		mo2->fsSBasis2[i].p3.d[1] = b[1];
		mo2->fsSBasis2[i].p3.d[2] = b[2];
		//printf("norm(b) = %f\n",length3D(b));

		//obtain cross product of Binormal and Tangent
		cross3D(b, t, n);
		//printf("norm(n) = %f\n",length3D(n));
		//save cross_BV	into vector	list
		mo2->fsRBasis2[i].p3.d[0] = n[0];//NOTE! changed from mo2->fsSBasis2[i]...
		mo2->fsRBasis2[i].p3.d[1] = n[1];
		mo2->fsRBasis2[i].p3.d[2] = n[2];

		//### Assign Quaternions ###
		//obtain 4x4 matrix
		mat[0] = n[0];
		mat[1] = n[1];
		mat[2] = n[2];

		mat[4] = b[0];
		mat[5] = b[1];
		mat[6] = b[2];

		mat[8] = t[0];
		mat[9] = t[1];
		mat[10]= t[2];		

		mat[3]	= mat[7] = mat[11] = mat[12] = mat[13] = mat[14] = 0.0;
		mat[15]	= 1.0;

		ExtractQuaternionFromMat( mat, q );
		quatNormalize( q );	
		
		//save quaternions
		mo2->fsQuatList2[i].r = q->r;
		mo2->fsQuatList2[i].i = q->i;
		mo2->fsQuatList2[i].j = q->j;
		mo2->fsQuatList2[i].k = q->k;

	}//$ FOR(i..)
	my_free(q);
	my_free(p);
	my_free(r);
}

//*********************************************//
//
//*********************************************//
void computeParallelTransportFrame_forQuatMap	( void ) 
{
	int		i		= 0;						// index into points array
	int		limit	= mo2->l_numPoints;			// local variable for number of	points for frame calculation
	double	t[3];								// arbitrary vector
	double	v[3];								// vector for Tangent
	double	s[3];								// vector for third	orthonormal	component
	double	v0[3];								// vector for initial normal vector
	double	len;								// vector lenght
	double	theta	= 0.0;						// PT initial frame	angle
	double	t1[3];								// vector for Tangent
	double	t2[3];								// vector for Tangent
	double	cross_T1T2[3];						// vector for tangent-tangent cross	product
	double	angle2 = 0.0;						// mo2->startTheta2;
	quaternion *p,*q,*r, *u, *prev;				// quaternions

	//initialize
	p =	(quaternion*) malloc(sizeof(quaternion));
	q =	(quaternion*) malloc(sizeof(quaternion));
	r =	(quaternion*) malloc(sizeof(quaternion));
	u =	(quaternion*) malloc(sizeof(quaternion));
	prev = (quaternion*) malloc(sizeof(quaternion));

	t[0] = 0.0;
	t[1] = 0.0;
	t[2] = 1.0;

	computeFrenetSerretFrame_forSingleFrame_forQuatMap();
    
    if(mo2->objectID == OBJCURVE && mo2->whichCurve == OBJKNOT)
        limit = mo2->l_numPoints;
    else if(mo2->objectID == OBJCURVE && mo2->whichCurve == OBJCURVELIST)
        limit = mo2->numPoints;

	//@ Make first frame same as FRENET  ///////////////////////////
	v0[0] = mo2->fsRBasis2[0].p3.d[0];
	v0[1] = mo2->fsRBasis2[0].p3.d[1];
	v0[2] = mo2->fsRBasis2[0].p3.d[2];

	//get first tangent
	v[0] = mo2->qTangList[0].p3.d[0];
	v[1] = mo2->qTangList[0].p3.d[1];
	v[2] = mo2->qTangList[0].p3.d[2];

	////@ Build first PT-Frame /////////////////////////////////////
	rotate3D_3(v0, v, angle2);
	len=length3D(v0);
	v0[0] /= len;
	v0[1] /= len;
	v0[2] /= len;

	//copy v0 into basisRList
	mo2->ptRBasis2[0].p3.d[0] = v0[0];
	mo2->ptRBasis2[0].p3.d[1] = v0[1];
	mo2->ptRBasis2[0].p3.d[2] = v0[2];

	//compute third component S 
	cross3D(v,v0,s);	
	mo2->ptSBasis2[0].p3.d[0] = s[0];
	mo2->ptSBasis2[0].p3.d[1] = s[1];
	mo2->ptSBasis2[0].p3.d[2] = s[2];

	//compute quaternion affected by rotation by angle
	prev->r = cos(angle2/2.0);
	prev->i = v[0]*sin(angle2/2.0);
	prev->j = v[1]*sin(angle2/2.0);
	prev->k = v[2]*sin(angle2/2.0);

	quatQuatProduct( prev, &(mo2->fsQuatList2[0]), u);
	quatCopy( u, &(mo2->ptQuatList2[0]) );

	//@ Calculate for all points
	for(i=0;i<=(limit+1);i++) {

		t1[0] =	mo2->qTangList[i].p3.d[0];
		t1[1] =	mo2->qTangList[i].p3.d[1];
		t1[2] =	mo2->qTangList[i].p3.d[2];

		if(i<limit) {
			t2[0] =	mo2->qTangList[i+1].p3.d[0];
			t2[1] =	mo2->qTangList[i+1].p3.d[1];
			t2[2] =	mo2->qTangList[i+1].p3.d[2];
		}
		else {
			t1[0] =	mo2->qTangList[0].p3.d[0];
			t1[1] =	mo2->qTangList[0].p3.d[1];
			t1[2] =	mo2->qTangList[0].p3.d[2];
		}

		cross3D(t1,t2, cross_T1T2);
		len	= length3D(cross_T1T2);
		//printf("[%d] cross_T1T2 = %f, %f, %f\tlen = %f\n",i,cross_T1T2[0],cross_T1T2[1],cross_T1T2[2],len );

		//get current quaternion
		p->r = mo2->ptQuatList2[i].r;
		p->i = mo2->ptQuatList2[i].i;
		p->j = mo2->ptQuatList2[i].j;
		p->k = mo2->ptQuatList2[i].k;

		if(len == 0.00) {

			//copy current basisR into next	basisRList
			mo2->ptRBasis2[i+1].p3.d[0] = mo2->ptRBasis2[i].p3.d[0];
			mo2->ptRBasis2[i+1].p3.d[1] = mo2->ptRBasis2[i].p3.d[1];
			mo2->ptRBasis2[i+1].p3.d[2] = mo2->ptRBasis2[i].p3.d[2];
			
			//copy basisRList into v0
			v0[0] =	mo2->ptRBasis2[i+1].p3.d[0];
			v0[1] =	mo2->ptRBasis2[i+1].p3.d[1];
			v0[2] =	mo2->ptRBasis2[i+1].p3.d[2];

			//assign next quaternion
			quatCopy (p, &(mo2->ptQuatList2[i+1]) );
			//quatCopy( p, u ); 
		}
		else {

			cross_T1T2[0] /= len;
			cross_T1T2[1] /= len;
			cross_T1T2[2] /= len;

			theta =	acos(dot3D(t1,t2));
			//printf("[%d] theta = %f\n",i,theta);

			rotate3D_3(v0, cross_T1T2, theta);
			
			//copy v0 into basisRList
			mo2->ptRBasis2[i+1].p3.d[0] = v0[0];
			mo2->ptRBasis2[i+1].p3.d[1] = v0[1];
			mo2->ptRBasis2[i+1].p3.d[2] = v0[2];

			//compute next quaternion
			q->r = cos(theta/2.0);
			q->i = cross_T1T2[0]*sin(theta/2.0);
			q->j = cross_T1T2[1]*sin(theta/2.0);
			q->k = cross_T1T2[2]*sin(theta/2.0);
			quatQuatProduct	(q,	p, u);
			quatNormalize(u);

		}

		//asign next quaternion
		mo2->ptQuatList2[i+1].r = u->r;
		mo2->ptQuatList2[i+1].i = u->i;
		mo2->ptQuatList2[i+1].j = u->j;
		mo2->ptQuatList2[i+1].k = u->k;

		cross3D(t2,v0,s);//!changed from cross3D(v0,t2,s)
		len	= length3D(s);
		s[0] /=	len;
		s[1] /=	len;
		s[2] /=	len;

		// copy	s into basisSList
		mo2->ptSBasis2[i+1].p3.d[0] = s[0];
		mo2->ptSBasis2[i+1].p3.d[1] = s[1];
		mo2->ptSBasis2[i+1].p3.d[2] = s[2];
		
	}//$ FOR(i=0,..)
	
	//NOTE: Don't know if this occurs here
	//printf("flip = %d\n",flip);
	//if(flip) {
	//	for(i=0;i<=(limit+1);i++) {
	//		quatCopy( &(mo2->ptQuatList[i]), p );
	//		r->r = -p->r;
	//		r->i = -p->i;
	//		r->j = -p->j;
	//		r->k = -p->k;
	//		quatCopy(r, &(mo2->ptQuatList[i]));				
	//	}
	//	flip = 0;
	//}
	//#DONE
	my_free(q);
	my_free(p);
	my_free(r);
	my_free(prev);
	my_free(u);
}


//*********************************************//
//
//*********************************************//
void drawFrames_forQuatMap						( void )
{
	int		i = 0;
	int		limit = 0;
	double	scale = 0.25;
	float	size = 1.5f;

	if(mo2->flag_curve_closed)
		limit = mo2->l_numPoints-2;
	else
		limit = mo2->numPoints-2;

    if(mo2->whichCurve == OBJCURVELIST) {
        limit = mo2->numPoints - 3;
        scale = 0.1;
    }

	glPushAttrib( GL_LIGHTING_BIT | GL_LINE_BIT );
	glDisable( GL_LIGHTING );
	glLineWidth( size );

	if(mo2->flag_quat_map_current_frame_for_QuatMap == FRENETSERRET) {

		for(i=0;i<=limit+1;i++) {
			glPushMatrix();
				glTranslatef( mo2->qpntList[i].p3.d[0], mo2->qpntList[i].p3.d[1], mo2->qpntList[i].p3.d[2] );
				glScalef(scale,scale,scale);
				glColor3f(1.0f,0.0f,0.0f);
				glBegin(GL_LINES);
					glVertex3f(0.0f,0.0f,0.0f);
					glVertex3f(mo2->qTangList[i].p3.d[0],mo2->qTangList[i].p3.d[1],mo2->qTangList[i].p3.d[2]);
				glEnd();
				glColor3f(0.0f,1.0f,0.0f);
				glBegin(GL_LINES);
					glVertex3f(0.0f,0.0f,0.0f);
					glVertex3f(mo2->fsRBasis2[i].p3.d[0],mo2->fsRBasis2[i].p3.d[1],mo2->fsRBasis2[i].p3.d[2]);
				glEnd();
				glColor3f(0.0f,0.0f,1.0f);
				glBegin(GL_LINES);
					glVertex3f(0.0f,0.0f,0.0f);
					glVertex3f(mo2->fsSBasis2[i].p3.d[0],mo2->fsSBasis2[i].p3.d[1],mo2->fsSBasis2[i].p3.d[2]);
				glEnd();
			glPopMatrix();
		}	
	}//$ IF(FS)
	else if(mo2->flag_quat_map_current_frame_for_QuatMap == PARALLELTRANSPORT) {
		for(i=0;i<limit;i++) {
			glPushMatrix();
				glTranslatef( mo2->qpntList[i].p3.d[0], mo2->qpntList[i].p3.d[1], mo2->qpntList[i].p3.d[2] );
				glScalef(scale,scale,scale);
				glColor3f(1.0f,0.0f,0.0f);
				glBegin(GL_LINES);
					glVertex3f(0.0f,0.0f,0.0f);
					glVertex3f(mo2->qTangList[i].p3.d[0],mo2->qTangList[i].p3.d[1],mo2->qTangList[i].p3.d[2]);
				glEnd();
				glColor3f(0.0f,1.0f,0.0f);
				glBegin(GL_LINES);
					glVertex3f(0.0f,0.0f,0.0f);
					glVertex3f(mo2->ptRBasis2[i].p3.d[0],mo2->ptRBasis2[i].p3.d[1],mo2->ptRBasis2[i].p3.d[2]);
				glEnd();
				glColor3f(0.0f,0.0f,1.0f);
				glBegin(GL_LINES);
					glVertex3f(0.0f,0.0f,0.0f);
					glVertex3f(mo2->ptSBasis2[i].p3.d[0],mo2->ptSBasis2[i].p3.d[1],mo2->ptSBasis2[i].p3.d[2]);
				glEnd();
			glPopMatrix();
		}	
	}//$ IF(PT)
	//else if(mo2->flag_quat_map_current_frame == GEODESICREFERENCE) {
	//	for(i=0;i<=limit+1;i++) {
	//		glPushMatrix();
	//			glTranslatef( mo2->pntList[i].p3.d[0], mo2->pntList[i].p3.d[1], mo2->pntList[i].p3.d[2] );
	//			glScalef(scale,scale,scale);
	//			glColor3f(1.0f,0.0f,0.0f);
	//			glBegin(GL_LINES);
	//				glVertex3f(0.0f,0.0f,0.0f);
	//				glVertex3f(mo2->TangList[i].p3.d[0],mo2->TangList[i].p3.d[1],mo2->TangList[i].p3.d[2]);
	//			glEnd();
	//			glColor3f(0.0f,1.0f,0.0f);
	//			glBegin(GL_LINES);
	//				glVertex3f(0.0f,0.0f,0.0f);
	//				glVertex3f(mo2->grRBasis2[i].p3.d[0],mo2->grRBasis2[i].p3.d[1],mo2->grRBasis2[i].p3.d[2]);
	//			glEnd();
	//			glColor3f(0.0f,0.0f,1.0f);
	//			glBegin(GL_LINES);
	//				glVertex3f(0.0f,0.0f,0.0f);
	//				glVertex3f(mo2->grSBasis2[i].p3.d[0],mo2->grSBasis2[i].p3.d[1],mo2->grSBasis2[i].p3.d[2]);
	//			glEnd();
	//		glPopMatrix();
	//	}	
	//}//$IF(GR)

	glPopAttrib();
}
//*********************************************//
//
//*********************************************//
void renderFramedCurve_forQuatMap				( double radius ) 
{
	int		i		= 0;
	int		j		= 0;
	int		start	= 0;
	int		renderpoints = 0;
	double	theta		 = 0.0;
	double	dtheta		 = 0.0;
	double	n_radius	 = 2.0*radius;
	int		radial_steps	 = mo2->radial_steps;
	double	rlist[3], slist[3];
	
    //@ Initializations
	dtheta = 2.0*M_PI/mo2->radial_steps;
	if(mo2->flag_curve_closed)
		renderpoints = mo2->numPoints;
	else
		renderpoints = mo2->numPoints-2;

	if(mo2->selectFrame_GR)
		renderpoints = renderpoints-1;

    if(mo2->whichCurve == OBJCURVELIST)
        renderpoints = mo2->numPoints-4;

    //@ FRAMES
	for	( i=start; i<renderpoints; i++ ) {

            theta= 0.0;

            //printf("%d > %f,%f,%f\n",i,mo2->qpntList[i].p3.d[0],mo2->pntList[i].p3.d[1],mo2->pntList[i].p3.d[2]);
            switch(mo2->flag_quat_map_current_frame_for_QuatMap) {

                case FRENETSERRET:
                    rlist[0] = mo2->fsRBasis2[i].p3.d[0];
                    rlist[1] = mo2->fsRBasis2[i].p3.d[1];
                    rlist[2] = mo2->fsRBasis2[i].p3.d[2];
                    slist[0] = mo2->fsSBasis2[i].p3.d[0];
                    slist[1] = mo2->fsSBasis2[i].p3.d[1];
                    slist[2] = mo2->fsSBasis2[i].p3.d[2];
                    break;
                case PARALLELTRANSPORT:
                    rlist[0] = mo2->ptRBasis2[i].p3.d[0];
                    rlist[1] = mo2->ptRBasis2[i].p3.d[1];
                    rlist[2] = mo2->ptRBasis2[i].p3.d[2];
                    slist[0] = mo2->ptSBasis2[i].p3.d[0];
                    slist[1] = mo2->ptSBasis2[i].p3.d[1];
                    slist[2] = mo2->ptSBasis2[i].p3.d[2];
                    break;
                        //case GEODESICREFERENCE:
                        //	rlist[0] = mo2->grRBasis2[i].p3.d[0];
                        //	rlist[1] = mo2->grRBasis2[i].p3.d[1];
                        //	rlist[2] = mo2->grRBasis2[i].p3.d[2];
                        //	slist[0] = mo2->grSBasis2[i].p3.d[0];
                        //	slist[1] = mo2->grSBasis2[i].p3.d[1];
                        //	slist[2] = mo2->grSBasis2[i].p3.d[2];
                        //	break;

                default:
                    break;

            }
            for	( j=0; j<radial_steps; j++ )	{
				
                //printf("[%d][%d] \nrlist = %f, %f, %f\nslist = %f, %f, %f\npntList=  = %f, %f, %f\n",i,j,
				//	rlist[0],rlist[1],rlist[2],slist[0],slist[1],slist[2],
                //	mo2->pntList[i].p3.d[0],mo2->pntList[i].p3.d[1],mo2->pntList[i].p3.d[2]);
                mo2->rpnts[i][j][0]	= mo2->qpntList[i].p3.d[0] +	radius*rlist[0]*cos(theta) + radius*slist[0]*sin(theta);
                mo2->rpnts[i][j][1]	= mo2->qpntList[i].p3.d[1] +	radius*rlist[1]*cos(theta) + radius*slist[1]*sin(theta);
                mo2->rpnts[i][j][2]	= mo2->qpntList[i].p3.d[2] +	radius*rlist[2]*cos(theta) + radius*slist[2]*sin(theta);
				
                mo2->npnts[i][j][0]	= n_radius*rlist[0]*cos(theta)+n_radius*slist[0]*sin(theta);
                mo2->npnts[i][j][1]	= n_radius*rlist[1]*cos(theta)+n_radius*slist[1]*sin(theta);
                mo2->npnts[i][j][2]	= n_radius*rlist[2]*cos(theta)+n_radius*slist[2]*sin(theta);
				
                theta =	theta +	dtheta;
            }
	}
    
	//@ EDGES
	if(mo2->qshowEdge) {
        glPushAttrib( GL_LIGHTING_BIT | GL_LINE_BIT );
        glDisable(GL_LIGHTING);
	
        glLineWidth(2.0f);
	
		// draw RINGS around points
        for(i=start; i<(renderpoints); i++) {
		
            glBegin(GL_LINE_STRIP);
            glColor3f(0.0,0.0,0.0);
            for(j=0; j<=radial_steps; j++) {
                if(j<(radial_steps))
                    glVertex3d(mo2->rpnts[i][j][0],mo2->rpnts[i][j][1],mo2->rpnts[i][j][2]);
                else
                    glVertex3d(mo2->rpnts[i][0][0],mo2->rpnts[i][0][1],mo2->rpnts[i][0][2]);
            }
            glEnd();
        }
	
		// draw LONGITUDINAL-LINES through rings
        glBegin(GL_LINES);
        glColor3f(0.0,0.0,0.0);
        for(i=start+1;i<(renderpoints); i++) {
            if(i<(renderpoints)) {
                for(j=0;j<radial_steps;j++) {
                    glVertex3d(mo2->rpnts[i][j][0],	mo2->rpnts[i][j][1], mo2->rpnts[i][j][2]); 
                    glVertex3d(mo2->rpnts[i-1][j][0],mo2->rpnts[i-1][j][1],mo2->rpnts[i-1][j][2]);
                    if(j<(radial_steps-1)){
                        glVertex3d(mo2->rpnts[i-1][j+1][0],mo2->rpnts[i-1][j+1][1],mo2->rpnts[i-1][j+1][2]);
                        glVertex3d(mo2->rpnts[i][j+1][0],mo2->rpnts[i][j+1][1],mo2->rpnts[i][j+1][2]);
                    }
                    else {
                        glVertex3d(mo2->rpnts[i-1][j-radial_steps+1][0],	mo2->rpnts[i-1][j-radial_steps+1][1],mo2->rpnts[i-1][j-radial_steps+1][2]	);
                        glVertex3d(mo2->rpnts[i][j-radial_steps+1][0],	mo2->rpnts[i][j-radial_steps+1][1],	mo2->rpnts[i][j-radial_steps+1][2]);
                    }
                }
            }
        }
        glEnd();
        glPopAttrib();
            
	}//$ IF(qshowEdge..)
    
	//@ NORMALS
	if(mo2->qshowNormals) {
        glPushAttrib( GL_LIGHTING_BIT | GL_LINE_BIT );
        glDisable(GL_LIGHTING);
        glLineWidth(0.5);
        for(i=start; i<renderpoints; i++) {
                //printf("%d > %f,%f,%f\n",i,mo2->fstlist[i].p3.d[0],mo2->fstlist[i].p3.d[1],mo2->fstlist[i].p3.d[2]);

            for(j=0; j<radial_steps;	j++) {
                glPushMatrix();
                glTranslatef(mo2->rpnts[i][j][0], mo2->rpnts[i][j][1], mo2->rpnts[i][j][2]);
                glBegin(GL_LINES);
                glColor3f(0.95f,0.95f,0.0f);
                glVertex3d(0.0,0.0,0.0);
                glVertex3d(mo2->npnts[i][j][0],	mo2->npnts[i][j][1], mo2->npnts[i][j][2]);
                glEnd();
                glPopMatrix();
            }		
        }
        glPopAttrib();
	}//$ IF(qshowNormals..)

    //@ SURFACE
	if(mo2->qshowFace) {

		glPushAttrib( GL_ALL_ATTRIB_BITS );
			
            glEnable(GL_COLOR_MATERIAL);
			
            //glEnable(GL_NORMALIZE);
            glEnable(GL_LIGHTING);
            glDisable( GL_CULL_FACE );

			
            glBegin( GL_QUADS );
            for(i=start+1;i<(renderpoints); i++) {
                if(i==(start+1))
                    glColor3f(0.9,0.1,0.2);
                else
                    glColor3f(mo2->qfgcolor[0],mo2->qfgcolor[1],mo2->qfgcolor[2]);
                for(j=0;j<radial_steps;j++) {
					
                        //gltexcoord2f(1.0,	1.0);
                        //gltexcoord2f(t, t);
                    glNormal3d(mo2->npnts[i][j][0],	mo2->npnts[i][j][1], mo2->npnts[i][j][2]);
                    glVertex3d(mo2->rpnts[i][j][0],	mo2->rpnts[i][j][1], mo2->rpnts[i][j][2]);
					
                        //gltexcoord2f(1.0,	0.0);
                        //gltexcoord2f(t, 0.0);
                    glNormal3d(mo2->npnts[i-1][j][0], mo2->npnts[i-1][j][1], mo2->npnts[i-1][j][2]);
                    glVertex3d(mo2->rpnts[i-1][j][0], mo2->rpnts[i-1][j][1], mo2->rpnts[i-1][j][2]);
                    if(j<(radial_steps-1)){
                            //gltexcoord2f(0.0,	0.0);
                        glNormal3d(mo2->npnts[i-1][j+1][0],	mo2->npnts[i-1][j+1][1], mo2->npnts[i-1][j+1][2]);
                        glVertex3d(mo2->rpnts[i-1][j+1][0],	mo2->rpnts[i-1][j+1][1], mo2->rpnts[i-1][j+1][2]);
						
                            //gltexcoord2f(0.0,	1.0);
                            //gltexcoord2f(0.0,	t);
                        glNormal3d(mo2->npnts[i][j+1][0],mo2->npnts[i][j+1][1],mo2->npnts[i][j+1][2]);
                        glVertex3d(mo2->rpnts[i][j+1][0],mo2->rpnts[i][j+1][1],mo2->rpnts[i][j+1][2]);
                    }
                    else {
                            //gltexcoord2f(0.0,	0.0);
                        glNormal3d(mo2->npnts[i-1][j-radial_steps+1][0],	mo2->npnts[i-1][j-radial_steps+1][1], mo2->npnts[i-1][j-radial_steps+1][2] );
                        glVertex3d(mo2->rpnts[i-1][j-radial_steps+1][0],	mo2->rpnts[i-1][j-radial_steps+1][1], mo2->rpnts[i-1][j-radial_steps+1][2] );
						
                            //gltexcoord2f(0.0,	1.0);
                            //gltexcoord2f(0.0,	t);
                        glNormal3d(mo2->npnts[i][j-radial_steps+1][0], mo2->npnts[i][j-radial_steps+1][1], mo2->npnts[i][j-radial_steps+1][2]);
                        glVertex3d(mo2->rpnts[i][j-radial_steps+1][0], mo2->rpnts[i][j-radial_steps+1][1], mo2->rpnts[i][j-radial_steps+1][2]);
                    }
                }
            }
            glEnd();

            glDisable(GL_TEXTURE_2D);
		glPopAttrib();			  
	}//$ IF(qshowFace..)
}

//*********************************************//
//
//*********************************************//
void constructFramedCurve_forQuatMap			( double radius ) 
{
	int		i				= 0;
	int		j				= 0;
	int		start			= 0;
	int		renderpoints	= 0;
	double	theta			= 0.0;
	double	dtheta			= 0.0;
	double	n_radius		= 2.0*radius;
	int		radial_steps	= mo2->radial_steps;
	double	rlist[3];
	double	slist[3];
	
	//@ Initializations
	dtheta = 2.0*M_PI/mo2->radial_steps;
	
#ifdef _DBG_
	////printf("inside obj::constructFramedCurve()\n");
#endif
	if(mo2->flag_curve_closed)
		renderpoints = mo2->l_numPoints;
	else
		renderpoints = mo2->numPoints-2;
    if(mo2->whichCurve == OBJCURVELIST)
        renderpoints = mo2->numPoints-3;
	//@ FRAMES
	for	( i=start; i<renderpoints; i++ ) {
		theta= 0.0;
		switch(mo2->flag_quat_map_current_frame_for_QuatMap) {

			case FRENETSERRET:
				rlist[0] = mo2->fsRBasis2[i].p3.d[0];
				rlist[1] = mo2->fsRBasis2[i].p3.d[1];
				rlist[2] = mo2->fsRBasis2[i].p3.d[2];
				slist[0] = mo2->fsSBasis2[i].p3.d[0];
				slist[1] = mo2->fsSBasis2[i].p3.d[1];
				slist[2] = mo2->fsSBasis2[i].p3.d[2];
				break;
			case PARALLELTRANSPORT:
				rlist[0] = mo2->ptRBasis2[i].p3.d[0];
				rlist[1] = mo2->ptRBasis2[i].p3.d[1];
				rlist[2] = mo2->ptRBasis2[i].p3.d[2];
				slist[0] = mo2->ptSBasis2[i].p3.d[0];
				slist[1] = mo2->ptSBasis2[i].p3.d[1];
				slist[2] = mo2->ptSBasis2[i].p3.d[2];
				break;
			//case GEODESICREFERENCE:
			//	rlist[0] = mo2->grRBasis[i].p3.d[0];
			//	rlist[1] = mo2->grRBasis[i].p3.d[1];
			//	rlist[2] = mo2->grRBasis[i].p3.d[2];
			//	slist[0] = mo2->grSBasis[i].p3.d[0];
			//	slist[1] = mo2->grSBasis[i].p3.d[1];
			//	slist[2] = mo2->grSBasis[i].p3.d[2];
			//	break;
			default:
				break;
		}
		for	( j=0; j<radial_steps; j++ )	{
			mo2->qrpnts[i][j][0]	= mo2->qpntList[i].p3.d[0] +	radius*rlist[0]*cos(theta) + radius*slist[0]*sin(theta);
			mo2->qrpnts[i][j][1]	= mo2->qpntList[i].p3.d[1] +	radius*rlist[1]*cos(theta) + radius*slist[1]*sin(theta);
			mo2->qrpnts[i][j][2]	= mo2->qpntList[i].p3.d[2] +	radius*rlist[2]*cos(theta) + radius*slist[2]*sin(theta);
			//printf(" [%d][%d]pnt = %f, %f, %f\n",i,j,mo2->qrpnts[i][j][0], mo2->qrpnts[i][j][1], mo2->qrpnts[i][j][2]);

			mo2->qnpnts[i][j][0]	= n_radius*rlist[0]*cos(theta)+n_radius*slist[0]*sin(theta);
			mo2->qnpnts[i][j][1]	= n_radius*rlist[1]*cos(theta)+n_radius*slist[1]*sin(theta);
			mo2->qnpnts[i][j][2]	= n_radius*rlist[2]*cos(theta)+n_radius*slist[2]*sin(theta);
			
			theta =	theta +	dtheta;
		}
	}
}
void renderFramedQuaternionCurve_forFSMap		( void ) 
{
	int			i				= 0;
	int			j				= 0;
	int			h				= 0;
	int			start			= 0;
	int			end				= 0;
	int			renderpoints	= 0;
	int			radial_steps	= mo2->radial_steps;
	
    //@ Initializations
	if(mo2->flag_curve_closed)
		renderpoints = mo2->l_numPoints-5;
	else
		renderpoints = mo2->numPoints-2;

    if(mo2->whichCurve == OBJCURVELIST)
        renderpoints = mo2->numPoints - 3;

	//@ EDGES
    if(mo2->qshowEdge) {
        glPushAttrib(GL_LIGHTING_BIT | GL_LINE_BIT);
        glDisable(GL_LIGHTING);
	
        glLineWidth(0.5f);
        //rings
        glColor3f(0,0,0);
        start =	0;
        if(mo2->markQuatCurveCounter==	0)
            end = mo2->l_index_end;
            //end	= mo2->l_numPoints-2;
        else
            end	= mo2->markQuatCurve[0];

        for(h=start; h<=mo2->markQuatCurveCounter; h++)	{
            if(h==mo2->markQuatCurveCounter-1)
                end-=3;
            for(i=start+1; i<end;	i++) {
                glBegin(GL_LINE_STRIP);
                for(j=0; j<=radial_steps; j++) {
                    if(j<(radial_steps))
                        glVertex3d(mo2->qrpnts[i][j][0],mo2->qrpnts[i][j][1],mo2->qrpnts[i][j][2]);
                    else
                        glVertex3d(mo2->qrpnts[i][0][0],mo2->qrpnts[i][0][1],mo2->qrpnts[i][0][2]);
                }
                glEnd();
            }

            start =	mo2->markQuatCurve[h]+1;
            end	= mo2->markQuatCurve[h+1]-1;
            if(h==(mo2->markQuatCurveCounter)-1) {
                end	= mo2->numPoints-2;//#@changed from ..numPoints-1;
            }
        }
    
        //latitudinal lines
        start =	0;
        if(mo2->markQuatCurveCounter==	0)
            end = mo2->l_index_end;
            //end	= mo2->l_numPoints-2;
        else
            end	= mo2->markQuatCurve[0];

        for(h=start; h<=mo2->markQuatCurveCounter; h++)	{

            if(h==mo2->markQuatCurveCounter-1)
                end-=3;
            glBegin(GL_LINES);
        
            for(i=start+2; i<end;	i++) {

				for(j=0;j<radial_steps;j++) {
                    glVertex3d(mo2->qrpnts[i][j][0],	mo2->qrpnts[i][j][1],	mo2->qrpnts[i][j][2]); 
                    glVertex3d(mo2->qrpnts[i-1][j][0],	mo2->qrpnts[i-1][j][1],	mo2->qrpnts[i-1][j][2]);
                    if(j<(radial_steps-1)){
                        glVertex3d(mo2->qrpnts[i-1][j+1][0], mo2->qrpnts[i-1][j+1][1],	mo2->qrpnts[i-1][j+1][2]);
                        glVertex3d(mo2->qrpnts[i][j+1][0],	 mo2->qrpnts[i][j+1][1],		mo2->qrpnts[i][j+1][2]);
                    }
                    else {
                        glVertex3d(mo2->qrpnts[i-1][j-radial_steps+1][0], mo2->qrpnts[i-1][j-radial_steps+1][1], mo2->qrpnts[i-1][j-radial_steps+1][2] );
                        glVertex3d(mo2->qrpnts[i][j-radial_steps+1][0],	  mo2->qrpnts[i][j-radial_steps+1][1],	 mo2->qrpnts[i][j-radial_steps+1][2]);
                    }
                }
            }
            glEnd();

            start =	mo2->markQuatCurve[h]+1;
            end	= mo2->markQuatCurve[h+1]-1;
            if(h==(mo2->markQuatCurveCounter)-1) {
                end	= mo2->numPoints-2;//#@changed from ..numPoints-1;
            }
        }
        glPopAttrib();
    
    }//$ IF(qshowEdge..)

    
	//@ NORMALS
    if(mo2->qshowNormals) {
        
        glPushAttrib( GL_LIGHTING_BIT |	GL_LINE_BIT );
        glDisable(GL_LIGHTING);
        glLineWidth(2.0);
        glLineWidth(1.0f);
        
        start = 0;
        for(i=start; i<renderpoints; i++) {
            
            for(j=0; j<radial_steps;	j++) {
                glPushMatrix();
                glTranslatef(mo2->qrpnts[i][j][0], mo2->qrpnts[i][j][1], mo2->qrpnts[i][j][2]);
                glBegin(GL_LINES);
                glColor3f(0.95f,0.95f,0.0f);
                glVertex3d(0.0,0.0,0.0);
                glVertex3d(mo2->qnpnts[i][j][0],	mo2->qnpnts[i][j][1], mo2->qnpnts[i][j][2]);
                glEnd();
                glPopMatrix();
            }
        }
        
        glPopAttrib();
    }//$ IF(qshowNormals..)
 
	// @FACES
    if(mo2->qshowFace) {
		
		glEnable(GL_COLOR_MATERIAL);
        glEnable(GL_LIGHTING);

		glColor3f(mo2->qfgcolor[0],mo2->qfgcolor[1],mo2->qfgcolor[2]);
		glEnable(GL_POLYGON_OFFSET_FILL);
		glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);
		glPolygonOffset(1.0f,1.0f);

        start =	0;
        if(mo2->markQuatCurveCounter == 0) 
            end = mo2->l_index_end;
            //end	= mo2->l_numPoints-2;
        else
            end	= mo2->markQuatCurve[0];

        for(h=0; h<=mo2->markQuatCurveCounter; h++)	{
            //--------------------------->
            if(h==mo2->markQuatCurveCounter-1)
                end-=3;
            glBegin(GL_QUADS);
            for(i =	start+2; i<end; i++) {
				if(i==(start+1))
					glColor3f(0.8,0.0,0.0);
				else
					glColor3f(mo2->qfgcolor[0],mo2->qfgcolor[1],mo2->qfgcolor[2]);
                for(j=0;j<radial_steps;j++) {
                    glNormal3d(mo2->qnpnts[i][j][0],	mo2->qnpnts[i][j][1],	mo2->qnpnts[i][j][2]);
                    glVertex3d(mo2->qrpnts[i][j][0],	mo2->qrpnts[i][j][1],	mo2->qrpnts[i][j][2]);
                    glNormal3d(mo2->qnpnts[i-1][j][0],	mo2->qnpnts[i-1][j][1], mo2->qnpnts[i-1][j][2]);
                    glVertex3d(mo2->qrpnts[i-1][j][0],	mo2->qrpnts[i-1][j][1], mo2->qrpnts[i-1][j][2]);
                    if(j<(radial_steps-1)){
                        glNormal3d(mo2->qnpnts[i-1][j+1][0],	mo2->qnpnts[i-1][j+1][1],	mo2->qnpnts[i-1][j+1][2]);
                        glVertex3d(mo2->qrpnts[i-1][j+1][0],	mo2->qrpnts[i-1][j+1][1],	mo2->qrpnts[i-1][j+1][2]);
                        glNormal3d(mo2->qnpnts[i][j+1][0],		mo2->qnpnts[i][j+1][1],		mo2->qnpnts[i][j+1][2]);
                        glVertex3d(mo2->qrpnts[i][j+1][0],		mo2->qrpnts[i][j+1][1],		mo2->qrpnts[i][j+1][2]);
                    }
                    else {
                        glNormal3d(mo2->qnpnts[i-1][j-radial_steps+1][0],	mo2->qnpnts[i-1][j-radial_steps+1][1],	mo2->qnpnts[i-1][j-radial_steps+1][2] );
                        glVertex3d(mo2->qrpnts[i-1][j-radial_steps+1][0],	mo2->qrpnts[i-1][j-radial_steps+1][1],	mo2->qrpnts[i-1][j-radial_steps+1][2] );
                        glNormal3d(mo2->qnpnts[i][j-radial_steps+1][0],		mo2->qnpnts[i][j-radial_steps+1][1],	mo2->qnpnts[i][j-radial_steps+1][2]);
                        glVertex3d(mo2->qrpnts[i][j-radial_steps+1][0],		mo2->qrpnts[i][j-radial_steps+1][1],	mo2->qrpnts[i][j-radial_steps+1][2]);
                    }
                }
            }
            glEnd();
            //<---------------------------
            start =	mo2->markQuatCurve[h]+1;
            end	= mo2->markQuatCurve[h+1];
            if(h==(mo2->markQuatCurveCounter)-1) {
                end	= mo2->numPoints-2;//#@changed from ..numPoints-1;
            }
        }
        
        
    }//$ IF(qshowFace..)
}


void computeQuaternionRings						( void )
{
	int		i			= 0;
	//int		j			= 0;
	double	qvec[3];
	int		limit		= 0;
	double	tangent[3];
	double	crossZT[3];
	double	norm_crossZT= 0.0;
	double	z[3]		= {0.0, 0.0, 1.0};
	double	thetaZT		= 0.0;
	double	theta		= 0.0;
	double	steps		= 50.0;
	double	delta		= (4.0*M_PI)/steps;
	int		cnt			= 0;
	int		counter		= 0;
	double	s			= 255.99/(1.25*mo2->scale);
	int		colorcode	= 0.0;
	quaternion	*q, *f, *b;					// quaternion
	
	//init quaternions
	b =	(quaternion*) my_malloc(sizeof(quaternion));
	f =	(quaternion*) my_malloc(sizeof(quaternion));
	q =	(quaternion*) my_malloc(sizeof(quaternion));

	if(mo2->flag_curve_closed)
		limit = mo2->numPoints;
	else
		limit = mo2->numPoints-2;

    if( mo2->qshowQuatRings || mo2->rshowFace || mo2->rshowEdge || mo2->rshowSurfaceNormals || mo2->rshowSurfacePoints ) {
		for(i=0; i<limit; i++) {

			tangent[0] =	mo2->TangList[i].p3.d[0];
			tangent[1] =	mo2->TangList[i].p3.d[1];
			tangent[2] =	mo2->TangList[i].p3.d[2];

			//compute (z x tangent)
			cross3D( z, tangent, crossZT );
			norm_crossZT = length3D( crossZT );
			crossZT[0] /= norm_crossZT;
			crossZT[1] /= norm_crossZT;
			crossZT[2] /= norm_crossZT;

			//compute arccos(z . tangent)
			thetaZT = acos( dot3D( z, tangent ) );

			//compute b(vhat)
			b->r = cos( thetaZT/2.0 );
			b->i = crossZT[0]*sin( thetaZT/2.0 );
			b->j = crossZT[1]*sin( thetaZT/2.0 );
			b->k = crossZT[2]*sin( thetaZT/2.0 );
			
			theta = 0.0;
			cnt = 0;
			//while( theta <4.0*M_PI ) {
			for(cnt=0;cnt<steps;cnt++) {

				//compute q(theta, vhat)
				q->r = cos( theta/2.0 );
				q->i = tangent[0]*sin( theta/2.0 );
				q->j = tangent[1]*sin( theta/2.0 );
				q->k = tangent[2]*sin( theta/2.0 );
				
				quatQuatProduct( q, b, f );
				getQuatComp( mo2->qwhichQuatComponent, f, qvec );
				
				mo2->rpList[i][cnt]->p3.d[0] = qvec[0];
				mo2->rpList[i][cnt]->p3.d[1] = qvec[1];
				mo2->rpList[i][cnt]->p3.d[2] = qvec[2];

				mo2->rpList[i][cnt]->p4.d[0] = q->r;
				mo2->rpList[i][cnt]->p4.d[1] = q->i;
				mo2->rpList[i][cnt]->p4.d[2] = q->j;
				mo2->rpList[i][cnt]->p4.d[3] = q->k;

				colorcode = (int) ( (f->r+1.0) * s );
				if(colorcode>255)
					colorcode = 255;
				mo2->rpList[i][cnt]->color = colorcode;
				
				theta = theta + delta;
			}
			counter = cnt;

			for(cnt=0; cnt<counter; cnt++) {
				mo2->prevList[cnt]->p3.d[0] = mo2->currList[cnt]->p3.d[0];
				mo2->prevList[cnt]->p3.d[1] = mo2->currList[cnt]->p3.d[1];
				mo2->prevList[cnt]->p3.d[2] = mo2->currList[cnt]->p3.d[2];
			}

		}//# FOR(i..)
		mo2->numRingPoints = counter;
	}
	//#DONE
	my_free(b);
	my_free(f);
	my_free(q);

}
 
//*********************************************//
//
//*********************************************//
void computeQuaternionRingsSurface_normals		( void )
{
	int		i = 0;
	int		j = 0;
	int		limit = 0;
	double	v1[3],v2[3];
	double	n[3];
	double	len		= 0.0;

	if(mo2->flag_curve_closed)
		limit = mo2->numPoints;
	else
		limit = mo2->numPoints-2;

	for(i=0; i<limit; i++) {
		
		for(j=0; j<mo2->numRingPoints; j++) {
			if(j<(mo2->numRingPoints-1)) {
				minus3D(mo2->rpList[i][j+1]->p3.d, mo2->rpList[i][j]->p3.d, v1);
				if(i<(limit-1))
					minus3D(mo2->rpList[i+1][j]->p3.d, mo2->rpList[i][j]->p3.d, v2);
				else
					minus3D(mo2->rpList[0][j]->p3.d, mo2->rpList[i][j]->p3.d, v2);
			}
			else {
				minus3D(mo2->rpList[i][0]->p3.d, mo2->rpList[i][j]->p3.d, v1);
				
				if(i<(limit-1))
					minus3D(mo2->rpList[i+1][j]->p3.d, mo2->rpList[i][j]->p3.d, v2);
				else
					minus3D(mo2->rpList[0][j]->p3.d, mo2->rpList[i][j]->p3.d, v2);
				
			}
			cross3D(v2,v1,n);
			len = length3D(n);
			n[0] /= len;
			n[1] /= len;
			n[2] /= len;
			mo2->rpnList[i][j]->p3.d[0] = n[0];
			mo2->rpnList[i][j]->p3.d[1] = n[1];
			mo2->rpnList[i][j]->p3.d[2] = n[2];	
		}//$ FOR(j=0...)
	}//$ For(i=0...)
}

//*********************************************//
//
//*********************************************//
void drawQuaternionRings						( void )
{
	int		i			= 0;
	int		j			= 0;
	int		limit		= 0;
	Vertex	*nlist;
	int		cnt			= 0;
	int		step		= 0;

	quaternion	*q, *f, *b;					// quaternion
	
	//init quaternions
	b =	(quaternion*) my_malloc(sizeof(quaternion));
	f =	(quaternion*) my_malloc(sizeof(quaternion));
	q =	(quaternion*) my_malloc(sizeof(quaternion));
	nlist = (Vertex*) my_malloc(sizeof(Vertex)*MAX_POINTS);

    
	if(mo2->flag_curve_closed==1)
		limit = mo2->numPoints;
	else
		limit = mo2->numPoints-1;

	//@ RINGS
	if(mo2->qshowQuatRings) {
		glPushAttrib( GL_LIGHTING_BIT | GL_LINE_BIT );
        glDisable(GL_LIGHTING);
		glLineWidth(0.5f);
        if(mo2->numOfRings > 0)
		for(i=0; i<limit; i+=1) {
                            
            if(i%(5-mo2->numOfRings)==0) {
			    if(i==mo2->trackingRing) {
				    glColor3f(0.75f,0.0f,0.0f);
				    glLineWidth(4.0f);
			    }
                else if(i==0) {
                    glLineWidth(2.0f);
           		    glColor3f(1.0f,1.0f,0.0f);
                }
                else if(i==limit-1) {
                    glLineWidth(2.0f);
           		    glColor3f(0.0f,0.85f,0.0f);
                }
			    else {
				    glLineWidth(0.5f);
				    glColor3f(0.1f,0.1f,0.25f);
			    }
			    glBegin(GL_LINE_STRIP);
				    for(cnt=0; cnt <= mo2->numRingPoints; cnt++) {
					    if(cnt<mo2->numRingPoints)
						    glVertex3dv( mo2->rpList[i][cnt]->p3.d );
					    else 
						    glVertex3dv( mo2->rpList[i][0]->p3.d );
				    }
			    glEnd();
            }
		}//$ FOR(i..)
        glPopAttrib();	
	}//$ IF(rshowCurve..)

	
	//@ MESH
	if(mo2->rshowEdge) {
		glPushAttrib( GL_LIGHTING_BIT | GL_LINE_BIT );
        glDisable(GL_LIGHTING);
		glDisable(GL_COLOR_MATERIAL);
		glLineWidth(0.1f);
		glColor3f(0.1f,0.1f,0.3f);
        step = 1;
		for(i=0; i<limit; i+=step) {
			glBegin(GL_LINE_STRIP);
				for(cnt=0; cnt<=mo2->numRingPoints; cnt++) {
					if(cnt<mo2->numRingPoints) 
						glVertex3dv( mo2->rpList[i][cnt]->p3.d );
					else
						glVertex3dv( mo2->rpList[i][0]->p3.d );
				}
			glEnd();
		}//# FOR(i..)

        for(j=0; j<mo2->numRingPoints; j+=step) {
			glBegin(GL_LINE_STRIP);
			for(i=0;i<=limit;i+=1) {
				if(i<limit)
					glVertex3f( mo2->rpList[i][j]->p3.d[0], mo2->rpList[i][j]->p3.d[1], mo2->rpList[i][j]->p3.d[2] );
				else
					glVertex3f( mo2->rpList[0][j]->p3.d[0], mo2->rpList[0][j]->p3.d[1], mo2->rpList[0][j]->p3.d[2] );
			}
			glEnd();
		}
		glPopAttrib();	

    }//$ IF(rshowCurve..)

	
	//@ SURFACE
	if(mo2->rshowFace) {

		float mats[4] = {0.25,0.25,0.25,1.0};
		glPushAttrib(GL_LIGHTING_BIT);
			
			
			glEnable(GL_LIGHTING);
			

			glMaterialfv (GL_FRONT_AND_BACK,GL_SPECULAR,mats);
			glMaterialf  (GL_FRONT_AND_BACK,GL_SHININESS,100.0);

			glEnable(GL_POLYGON_OFFSET_FILL);
			glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);
			glPolygonOffset(1.0f,1.0f);

			if(mo2->rshowColorMode == CODE_SOLID_COLOR) {
				//glEnable(GL_COLOR_MATERIAL);
				//glColor3f(0.65f,0.3f,0.5f);
				glDisable(GL_COLOR_MATERIAL);
				glMaterialfv( GL_FRONT_AND_BACK, GL_AMBIENT, mat_ambient2);
				glMaterialfv( GL_FRONT_AND_BACK, GL_DIFFUSE, mat_diffuse2);
				glMaterialfv( GL_FRONT_AND_BACK, GL_SPECULAR, mat_specular2);
				glMaterialfv( GL_FRONT_AND_BACK, GL_SHININESS, high_shininess2);
				glMaterialfv( GL_FRONT_AND_BACK, GL_EMISSION, no_mat2);
				glColorMaterial(GL_FRONT_AND_BACK, GL_TRUE);
			}
			else if(mo2->rshowColorMode == CODE_TWO_SIDE) {
				glDisable(GL_COLOR_MATERIAL);
				glMaterialfv( GL_FRONT_AND_BACK, GL_AMBIENT, mat_ambient2);
				glMaterialfv( GL_FRONT_AND_BACK, GL_DIFFUSE, mat_diffuse2);
				glMaterialfv( GL_FRONT_AND_BACK, GL_SPECULAR, mat_specular2);
				glMaterialfv( GL_FRONT_AND_BACK, GL_SHININESS, high_shininess2);
				glMaterialfv( GL_FRONT_AND_BACK, GL_EMISSION, no_mat2);
				glColorMaterial(GL_FRONT_AND_BACK, GL_TRUE);

				glMaterialfv( GL_BACK, GL_AMBIENT, mat_ambient4);
				glMaterialfv( GL_BACK, GL_DIFFUSE, mat_diffuse4);
				glMaterialfv( GL_BACK, GL_SPECULAR, mat_specular4);
				glMaterialfv( GL_BACK, GL_SHININESS, high_shininess4);
				glMaterialfv( GL_BACK, GL_EMISSION, no_mat4);
				glColorMaterial(GL_BACK, GL_TRUE);
			}
			else if(mo2->rshowColorMode == CODE_DEPTH) {
				glShadeModel(GL_SMOOTH);
			}

			for(i=0;i<limit;i++) {
				for(cnt=0; cnt<mo2->numRingPoints; cnt++) {
					nlist[cnt].p3.d[0] = mo2->rpnList[i][cnt]->p3.d[0];
					nlist[cnt].p3.d[1] = mo2->rpnList[i][cnt]->p3.d[1];
					nlist[cnt].p3.d[2] = mo2->rpnList[i][cnt]->p3.d[2];
					
					if(i<(limit-1)) {
						glBegin(GL_QUADS);
							glNormal3dv( mo2->rpnList[i][cnt]->p3.d );
							if(mo2->rshowColorMode == CODE_DEPTH)
								glColor4fv(mo2->palette[mo2->rpList[i][cnt]->color]);								
							glVertex3f( mo2->rpList[i][cnt]->p3.d[0],		mo2->rpList[i][cnt]->p3.d[1],		mo2->rpList[i][cnt]->p3.d[2] );

							glNormal3dv( mo2->rpnList[i+1][cnt]->p3.d );
							if(mo2->rshowColorMode == CODE_DEPTH)
								glColor4fv(mo2->palette[mo2->rpList[i][cnt]->color]);
							glVertex3f( mo2->rpList[i+1][cnt]->p3.d[0],		mo2->rpList[i+1][cnt]->p3.d[1],		mo2->rpList[i+1][cnt]->p3.d[2] );
							if(cnt<(mo2->numRingPoints-1)) {		
									glNormal3dv( mo2->rpnList[i+1][cnt+1]->p3.d );
									glVertex3f( mo2->rpList[i+1][cnt+1]->p3.d[0],	mo2->rpList[i+1][cnt+1]->p3.d[1],	mo2->rpList[i+1][cnt+1]->p3.d[2] );
									
									glNormal3dv( mo2->rpnList[i][cnt+1]->p3.d );
									glVertex3f( mo2->rpList[i][cnt+1]->p3.d[0],		mo2->rpList[i][cnt+1]->p3.d[1],		mo2->rpList[i][cnt+1]->p3.d[2] );
							}
							else {	
									glNormal3dv( mo2->rpnList[i+1][0]->p3.d );
									glVertex3f( mo2->rpList[i+1][0]->p3.d[0],	mo2->rpList[i+1][0]->p3.d[1],	mo2->rpList[i+1][0]->p3.d[2] );
									
									glNormal3dv( mo2->rpnList[i][0]->p3.d );
									glVertex3f( mo2->rpList[i][0]->p3.d[0],		mo2->rpList[i][0]->p3.d[1],		mo2->rpList[i][0]->p3.d[2] );
							}
						glEnd();
					}
					else {
						glBegin(GL_QUADS);
							glNormal3dv( mo2->rpnList[i][cnt]->p3.d );
							if(mo2->rshowColorMode == CODE_DEPTH)
								glColor4fv(mo2->palette[mo2->rpList[i][cnt]->color]);
							glVertex3f( mo2->rpList[i][cnt]->p3.d[0],		mo2->rpList[i][cnt]->p3.d[1],		mo2->rpList[i][cnt]->p3.d[2] );

							glNormal3dv( mo2->rpnList[0][cnt]->p3.d );
							if(mo2->rshowColorMode == CODE_DEPTH)
								glColor4fv(mo2->palette[mo2->rpList[i][cnt]->color]);
							glVertex3f( mo2->rpList[0][cnt]->p3.d[0],		mo2->rpList[0][cnt]->p3.d[1],		mo2->rpList[0][cnt]->p3.d[2] );
							if(cnt<(mo2->numRingPoints-1)) {		
									glNormal3dv( mo2->rpnList[0][cnt+1]->p3.d );
									if(mo2->rshowColorMode == CODE_DEPTH)
										glColor4fv(mo2->palette[mo2->rpList[i][cnt]->color]);
									glVertex3f( mo2->rpList[0][cnt+1]->p3.d[0],	mo2->rpList[0][cnt+1]->p3.d[1],	mo2->rpList[0][cnt+1]->p3.d[2] );
									
									glNormal3dv( mo2->rpnList[i][cnt+1]->p3.d );
									if(mo2->rshowColorMode == CODE_DEPTH)
										glColor4fv(mo2->palette[mo2->rpList[i][cnt]->color]);
									glVertex3f( mo2->rpList[i][cnt+1]->p3.d[0],		mo2->rpList[i][cnt+1]->p3.d[1],		mo2->rpList[i][cnt+1]->p3.d[2] );
							}
							else {	
									glNormal3dv( mo2->rpnList[0][0]->p3.d );
									if(mo2->rshowColorMode == CODE_DEPTH)
										glColor4fv(mo2->palette[mo2->rpList[i][cnt]->color]);
									glVertex3f( mo2->rpList[0][0]->p3.d[0],	mo2->rpList[0][0]->p3.d[1],	mo2->rpList[0][0]->p3.d[2] );
									
									glNormal3dv( mo2->rpnList[i][0]->p3.d );
									if(mo2->rshowColorMode == CODE_DEPTH)
										glColor4fv(mo2->palette[mo2->rpList[i][cnt]->color]);
									glVertex3f( mo2->rpList[i][0]->p3.d[0],		mo2->rpList[i][0]->p3.d[1],		mo2->rpList[i][0]->p3.d[2] );
							}
						glEnd();
					}

				}//$ FOR(j=0...)
			}//$ FOR(i=0...)
		glPopAttrib();
	}//$ IF(rshowFace..)

	
	//@ NORMALS
	if(mo2->rshowSurfaceNormals) {
		glPushAttrib( GL_LIGHTING_BIT | GL_LINE_BIT );
        glDisable(GL_LIGHTING);
		glColor3f(0,0.0,0.6);
		glLineWidth(3.0f);
		for(i=0; i<limit; i+=step) {
			for(j=0; j<mo2->numRingPoints; j+=step) {
				glPushMatrix();
					glTranslatef( mo2->rpList[i][j]->p3.d[0], mo2->rpList[i][j]->p3.d[1], mo2->rpList[i][j]->p3.d[2] );
					glScalef(0.1,0.1,0.1);
					glBegin(GL_LINES);
						glVertex3f(0,0,0);
						glVertex3f( mo2->rpnList[i][j]->p3.d[0], mo2->rpnList[i][j]->p3.d[1], mo2->rpnList[i][j]->p3.d[2] );
					glEnd();
				glPopMatrix();
			}
		}
		glPopAttrib();
	}//$ IF(rshowSurfaceNormals..)
	
	//@ SURFACE-POINTS
	if(mo2->rshowSurfacePoints) {
		glPushAttrib( GL_LIGHTING_BIT | GL_LINE_BIT );
        glEnable(GL_LIGHTING);
		glDisable(GL_COLOR_MATERIAL);
		glMaterialfv( GL_FRONT_AND_BACK, GL_AMBIENT, mat_ambient1);
		glMaterialfv( GL_FRONT_AND_BACK, GL_DIFFUSE, mat_diffuse1);
		glMaterialfv( GL_FRONT_AND_BACK, GL_SPECULAR, mat_specular1);
		glMaterialfv( GL_FRONT_AND_BACK, GL_SHININESS, high_shininess1);
		glMaterialfv( GL_FRONT_AND_BACK, GL_EMISSION, no_mat1);
		glColorMaterial(GL_FRONT_AND_BACK, GL_TRUE);

		glLineWidth(0.5f);
		glColor3f(0.95f,0.95f,0.0f);
		glLineWidth(3.0f);
		for(i=0; i<limit; i+=step) {
			for(j=0; j<mo2->numRingPoints; j+=step) {
				glPushMatrix();
					glTranslatef( mo2->rpList[i][j]->p3.d[0], mo2->rpList[i][j]->p3.d[1], mo2->rpList[i][j]->p3.d[2] );
					glutSolidCube(0.02);
				glPopMatrix();

			}
		}
		glPopAttrib();
	}//$ IF(rshowSurfacePoints..)
	
	//#DONE
	my_free(b);
	my_free(f);
	my_free(q);
}


//*********************************************//
//
//*********************************************//
void drawQuaternionRings_forOpenCurve			( void )
{
	int		i			= 0;
	int		j			= 0;
	int		limit		= 0;
	Vertex	*nlist;
	int		cnt			= 0;
	int		step		= 3;

	quaternion	*q, *f, *b;					// quaternion
	
	//init quaternions
	b =	(quaternion*) my_malloc(sizeof(quaternion));
	f =	(quaternion*) my_malloc(sizeof(quaternion));
	q =	(quaternion*) my_malloc(sizeof(quaternion));
	nlist = (Vertex*) my_malloc(sizeof(Vertex)*MAX_POINTS);

	limit = mo2->numPoints-1;


	//@ RINGS
	if(mo2->qshowQuatRings) {
		glPushAttrib( GL_LIGHTING_BIT | GL_LINE_BIT );
        glDisable(GL_LIGHTING);
		glLineWidth(0.5f);
		for(i=0; i<=limit; i+=1) {
                            
			if(i==mo2->trackingRing) {
				glColor3f(0.75f,0.0f,0.0f);
				glLineWidth(4.0f);
			}
			else {
				glLineWidth(0.5f);
				if(i==0) 
					glColor3f(1.0f,1.0f,0.0f);
				else if(i==mo2->numPoints-1) 
					glColor3f(0.0f,1.0f,0.0f);
				else
					glColor3f(0.1f,0.1f,0.25f);
			}

			glBegin(GL_LINE_STRIP);
				for(cnt=0; cnt <= mo2->numRingPoints; cnt++) {
					if(cnt<mo2->numRingPoints)
						glVertex3dv( mo2->rpList[i][cnt]->p3.d );
					else 
						glVertex3dv( mo2->rpList[i][0]->p3.d );
				}
			glEnd();
		}//$ FOR(i..)
		glPopAttrib();	
	}//$ IF(rshowCurve..)

	
	//@ MESH
	if(mo2->rshowEdge) {
		glPushAttrib( GL_LIGHTING_BIT | GL_LINE_BIT );
        glDisable(GL_LIGHTING);
		glDisable(GL_COLOR_MATERIAL);
		glLineWidth(0.1f);
		
		glColor3f(0.1f,0.1f,0.3f);
		for(i=0; i<limit-1; i+=step) {
			glBegin(GL_LINE_STRIP);
				for(cnt=0; cnt<=mo2->numRingPoints; cnt++) {
					if(cnt<mo2->numRingPoints) 
						glVertex3dv( mo2->rpList[i][cnt]->p3.d );
					else
						glVertex3dv( mo2->rpList[i][0]->p3.d );
				}
			glEnd();
		}//# FOR(i..)
		for(j=0; j<mo2->numRingPoints; j+=step) {
			glBegin(GL_LINE_STRIP);
			for(i=0;i<limit-1;i+=1) {
				//if(i<limit)
					glVertex3f( mo2->rpList[i][j]->p3.d[0], mo2->rpList[i][j]->p3.d[1], mo2->rpList[i][j]->p3.d[2] );
				//else
				//	glVertex3f( mo2->rpList[0][j]->p3.d[0], mo2->rpList[0][j]->p3.d[1], mo2->rpList[0][j]->p3.d[2] );
			}
			glEnd();
		}
		glPopAttrib();	
	}//$ IF(rshowCurve..)

	
	//@ SURFACE
	if(mo2->rshowFace) {

		float mats[4] = {0.25,0.25,0.25,1.0};
		glPushAttrib(GL_LIGHTING_BIT);
			
			
			glEnable(GL_LIGHTING);
			

			glMaterialfv (GL_FRONT_AND_BACK,GL_SPECULAR,mats);
			glMaterialf  (GL_FRONT_AND_BACK,GL_SHININESS,100.0);

			glEnable(GL_POLYGON_OFFSET_FILL);
			glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);
			glPolygonOffset(1.0f,1.0f);

			if(mo2->rshowColorMode == CODE_SOLID_COLOR) {
				//glEnable(GL_COLOR_MATERIAL);
				//glColor3f(0.65f,0.3f,0.5f);
				glDisable(GL_COLOR_MATERIAL);
				glMaterialfv( GL_FRONT_AND_BACK, GL_AMBIENT, mat_ambient2);
				glMaterialfv( GL_FRONT_AND_BACK, GL_DIFFUSE, mat_diffuse2);
				glMaterialfv( GL_FRONT_AND_BACK, GL_SPECULAR, mat_specular2);
				glMaterialfv( GL_FRONT_AND_BACK, GL_SHININESS, high_shininess2);
				glMaterialfv( GL_FRONT_AND_BACK, GL_EMISSION, no_mat2);
				glColorMaterial(GL_FRONT_AND_BACK, GL_TRUE);
			}
			else if(mo2->rshowColorMode == CODE_TWO_SIDE) {
				glDisable(GL_COLOR_MATERIAL);
				glMaterialfv( GL_FRONT_AND_BACK, GL_AMBIENT, mat_ambient3);
				glMaterialfv( GL_FRONT_AND_BACK, GL_DIFFUSE, mat_diffuse3);
				glMaterialfv( GL_FRONT_AND_BACK, GL_SPECULAR, mat_specular3);
				glMaterialfv( GL_FRONT_AND_BACK, GL_SHININESS, high_shininess3);
				glMaterialfv( GL_FRONT_AND_BACK, GL_EMISSION, no_mat3);
				glColorMaterial(GL_FRONT_AND_BACK, GL_TRUE);

				glMaterialfv( GL_BACK, GL_AMBIENT, mat_ambient4);
				glMaterialfv( GL_BACK, GL_DIFFUSE, mat_diffuse4);
				glMaterialfv( GL_BACK, GL_SPECULAR, mat_specular4);
				glMaterialfv( GL_BACK, GL_SHININESS, high_shininess4);
				glMaterialfv( GL_BACK, GL_EMISSION, no_mat4);
				glColorMaterial(GL_BACK, GL_TRUE);
			}
			else if(mo2->rshowColorMode == CODE_DEPTH) {
				glShadeModel(GL_SMOOTH);
			}

			for(i=0;i<limit-2;i++) {
				for(cnt=0; cnt<mo2->numRingPoints; cnt++) {
					nlist[cnt].p3.d[0] = mo2->rpnList[i][cnt]->p3.d[0];
					nlist[cnt].p3.d[1] = mo2->rpnList[i][cnt]->p3.d[1];
					nlist[cnt].p3.d[2] = mo2->rpnList[i][cnt]->p3.d[2];
					
					if(i<(limit-2)) {
						glBegin(GL_QUADS);
							glNormal3dv( mo2->rpnList[i][cnt]->p3.d );
							if(mo2->rshowColorMode == CODE_DEPTH)
								glColor4fv(mo2->palette[mo2->rpList[i][cnt]->color]);								
							glVertex3f( mo2->rpList[i][cnt]->p3.d[0],		mo2->rpList[i][cnt]->p3.d[1],		mo2->rpList[i][cnt]->p3.d[2] );

							glNormal3dv( mo2->rpnList[i+1][cnt]->p3.d );
							if(mo2->rshowColorMode == CODE_DEPTH)
								glColor4fv(mo2->palette[mo2->rpList[i][cnt]->color]);
							glVertex3f( mo2->rpList[i+1][cnt]->p3.d[0],		mo2->rpList[i+1][cnt]->p3.d[1],		mo2->rpList[i+1][cnt]->p3.d[2] );
							if(cnt<(mo2->numRingPoints-1)) {		
									glNormal3dv( mo2->rpnList[i+1][cnt+1]->p3.d );
									glVertex3f( mo2->rpList[i+1][cnt+1]->p3.d[0],	mo2->rpList[i+1][cnt+1]->p3.d[1],	mo2->rpList[i+1][cnt+1]->p3.d[2] );
									
									glNormal3dv( mo2->rpnList[i][cnt+1]->p3.d );
									glVertex3f( mo2->rpList[i][cnt+1]->p3.d[0],		mo2->rpList[i][cnt+1]->p3.d[1],		mo2->rpList[i][cnt+1]->p3.d[2] );
							}
							else {	
									glNormal3dv( mo2->rpnList[i+1][0]->p3.d );
									glVertex3f( mo2->rpList[i+1][0]->p3.d[0],	mo2->rpList[i+1][0]->p3.d[1],	mo2->rpList[i+1][0]->p3.d[2] );
									
									glNormal3dv( mo2->rpnList[i][0]->p3.d );
									glVertex3f( mo2->rpList[i][0]->p3.d[0],		mo2->rpList[i][0]->p3.d[1],		mo2->rpList[i][0]->p3.d[2] );
							}
						glEnd();
					}
				}//$ FOR(j=0...)
			}//$ FOR(i=0...)
		glPopAttrib();
	}//$ IF(rshowFace..)
	
	//@ NORMALS
	if(mo2->rshowSurfaceNormals) {
		glPushAttrib( GL_LIGHTING_BIT | GL_LINE_BIT );
        glDisable(GL_LIGHTING);
		glColor3f(0,0.0,0.6);
		glLineWidth(3.0f);
		for(i=0; i<limit; i+=step) {
			for(j=0; j<mo2->numRingPoints; j+=step) {
				glPushMatrix();
					glTranslatef( mo2->rpList[i][j]->p3.d[0], mo2->rpList[i][j]->p3.d[1], mo2->rpList[i][j]->p3.d[2] );
					glScalef(0.1,0.1,0.1);
					glBegin(GL_LINES);
						glVertex3f(0,0,0);
						glVertex3f( mo2->rpnList[i][j]->p3.d[0], mo2->rpnList[i][j]->p3.d[1], mo2->rpnList[i][j]->p3.d[2] );
					glEnd();
				glPopMatrix();
			}
		}
		glPopAttrib();
	}//$ IF(rshowSurfaceNormals..)


	
	//@ SURFACE-POINTS
	if(mo2->rshowSurfacePoints) {
		glPushAttrib( GL_LIGHTING_BIT | GL_LINE_BIT );
        glEnable(GL_LIGHTING);
		glDisable(GL_COLOR_MATERIAL);
		glMaterialfv( GL_FRONT_AND_BACK, GL_AMBIENT, mat_ambient1);
		glMaterialfv( GL_FRONT_AND_BACK, GL_DIFFUSE, mat_diffuse1);
		glMaterialfv( GL_FRONT_AND_BACK, GL_SPECULAR, mat_specular1);
		glMaterialfv( GL_FRONT_AND_BACK, GL_SHININESS, high_shininess1);
		glMaterialfv( GL_FRONT_AND_BACK, GL_EMISSION, no_mat1);
		glColorMaterial(GL_FRONT_AND_BACK, GL_TRUE);

		glLineWidth(0.5f);
		glColor3f(0.95f,0.95f,0.0f);
		glLineWidth(3.0f);
		for(i=0; i<limit; i+=step) {
			for(j=0; j<mo2->numRingPoints; j+=step) {
				glPushMatrix();
					glTranslatef( mo2->rpList[i][j]->p3.d[0], mo2->rpList[i][j]->p3.d[1], mo2->rpList[i][j]->p3.d[2] );
					glutSolidCube(0.02);
				glPopMatrix();

			}
		}
		glPopAttrib();
	}//$ IF(rshowSurfacePoints..)
	
	//#DONE
	my_free(b);
	my_free(f);
	my_free(q);
	my_free(nlist);
}


//### COMPUTATION: quaternion space
void updateQuaternionRings						( void )
{
	int			i			= 0;		// quaternion-ring to be updated
	double		rangle		= 0.0;		// quaternion-ring angle to be updated
	double		theta		= 0.0;
	double		tangent[3];
	double		crossZT[3];
	double		norm_crossZT= 0.0;
	double		thetaZT		= 0.0;
	double		phi			= 0.0;
	double		z[3]		= {0.0, 0.0, 1.0};

	quaternion	*b;
	quaternion	*q;
	quaternion	*r;
	quaternion	*s;

	b =	(quaternion*) my_malloc(sizeof(quaternion));
	q =	(quaternion*) my_malloc(sizeof(quaternion));
	r =	(quaternion*) my_malloc(sizeof(quaternion));
	s =	(quaternion*) my_malloc(sizeof(quaternion));
	i		= mo2->trackingRing;
	rangle	= mo2->rquatRingAngle;

	//process
	if(mo2->trackingRing>=0) {
		//printf("[%d] rangle = %f\n",mo2->trackingRing,rangle);
		switch( mo2->whichFrame ) {
			case FRENETSERRET:
				quatCopy( &(mo2->fsQuatList[i]), s);
				break;
			case PARALLELTRANSPORT:
				quatCopy( &(mo2->ptQuatList[i]), s);
				break;
			case GEODESICREFERENCE:
				quatCopy( &(mo2->grQuatList[i]), s);
				break;
		}
		phi = mo2->rtrackQuatAngle;
		//printf("r= %f, %f, %f, %f\n",r->r, r->i, r->j, r->k );

		tangent[0] =	mo2->TangList[i].p3.d[0];
		tangent[1] =	mo2->TangList[i].p3.d[1];
		tangent[2] =	mo2->TangList[i].p3.d[2];

		//compute (z x tangent)
		cross3D( z, tangent, crossZT );
		norm_crossZT = length3D( crossZT );
		crossZT[0] /= norm_crossZT;
		crossZT[1] /= norm_crossZT;
		crossZT[2] /= norm_crossZT;

		//compute arccos(z . tangent)
		thetaZT = acos( dot3D( z, tangent ) );

		//compute b(vhat)
		b->r = cos( thetaZT/2.0 );
		b->i = crossZT[0]*sin( thetaZT/2.0 );
		b->j = crossZT[1]*sin( thetaZT/2.0 );
		b->k = crossZT[2]*sin( thetaZT/2.0 );

		theta = phi + rangle;
		q->r = cos( theta/2.0 );
		q->i = tangent[0]*sin( theta/2.0 );
		q->j = tangent[1]*sin( theta/2.0 );
		q->k = tangent[2]*sin( theta/2.0 );
		quatQuatProduct( q, b, r );
		quatConjugate( r );

		//getQuatComp( mo2->qwhichQuatComponent, r, qvec );
		//glEnable(GL_LIGHTING);
		//glEnable(GL_COLOR_MATERIAL);
		//glColor3f(.7,.2,6);
		//glPushMatrix();
		//	glTranslatef(qvec[0],qvec[1],qvec[2]);
		//	glutSolidSphere( 0.02,16,16);
		//glPopMatrix();
		switch( mo2->whichFrame ) {
			case FRENETSERRET:
				quatCopy(r, &(mo2->fsQuatList[i]) );
				break;
			case PARALLELTRANSPORT:
				quatCopy(r, &(mo2->ptQuatList[i]) );
				break;
			case GEODESICREFERENCE:
				quatCopy(r, &(mo2->grQuatList[i]) );
				break;
		}
		
	}
	optQuatCurveEnergy();
	my_free(b);
	my_free(q);
	my_free(r);
	my_free(s);
}
//*********************************************//
//
//*********************************************//
void findQuat_Current							( int ind, quaternion *retQuat, double* retAngle ) 
{

	int			i			= ind;
	double		tangent[3];
	double		crossZT[3];
	double		norm_crossZT= 0.0;
	double		thetaZT		= 0.0;
	double		z[3]		= {0.0, 0.0, 1.0};
	double		theta		= 0.0;
	double		delta		= 0.005;
	double		currDP		= 0.0;
	double		prevDP		= 0.0;
	int			cnt			= 0;
	int			index		= 0;
	double		phi			= 0.0;
	quaternion	*b;
	quaternion	*q;
	quaternion	*r;
	quaternion	*s;

	b =	(quaternion*) my_malloc(sizeof(quaternion));
	q =	(quaternion*) my_malloc(sizeof(quaternion));
	r =	(quaternion*) my_malloc(sizeof(quaternion));
	s =	(quaternion*) my_malloc(sizeof(quaternion));
	//i		= mo2->trackingRing;
	i = ind;

	if( mo2->whichFrame == FRENETSERRET )
		quatCopy( &(mo2->fsQuatList[i]), s);
	else if( mo2->whichFrame == PARALLELTRANSPORT )
		quatCopy( &(mo2->ptQuatList[i]), s);
	else if( mo2->whichFrame == GEODESICREFERENCE )
		quatCopy( &(mo2->grQuatList[i]), s);
	else if( mo2->whichFrame == OPTIMIZATION )
		quatCopy( &(mo2->opQuatList[i]), s);

	tangent[0] =	mo2->TangList[i].p3.d[0];
	tangent[1] =	mo2->TangList[i].p3.d[1];
	tangent[2] =	mo2->TangList[i].p3.d[2];

	//printf("#\ttangent = %f, %f, %f\n",tangent[0],tangent[1],tangent[2]);
	//compute (z x tangent)
	cross3D( z, tangent, crossZT );
	norm_crossZT = length3D( crossZT );
	crossZT[0] /= norm_crossZT;
	crossZT[1] /= norm_crossZT;
	crossZT[2] /= norm_crossZT;

	//compute arccos(z . tangent)
	thetaZT = acos( dot3D( z, tangent ) );

	//compute b(vhat)
	b->r = cos( thetaZT/2.0 );
	b->i = crossZT[0]*sin( thetaZT/2.0 );
	b->j = crossZT[1]*sin( thetaZT/2.0 );
	b->k = crossZT[2]*sin( thetaZT/2.0 );
	//printf("#\tb = %f, %f, %f, %f\n",b->r,b->i,b->j,b->k);
	theta = 0.0;
	cnt = 0;
	index = 0;
	prevDP = 1000.0;
	while( theta <=4.0*M_PI+delta ) {
		//compute q(theta, vhat)
		q->r = cos( theta/2.0 );
		q->i = tangent[0]*sin( theta/2.0 );
		q->j = tangent[1]*sin( theta/2.0 );
		q->k = tangent[2]*sin( theta/2.0 );
		
		quatQuatProduct( q, b, r );
		currDP = quatQuatDotProduct( s, r );
		if( (currDP < prevDP) ) {
			index = cnt;
			//printf("[%d] dotProd = %f\n",index,currDP);
			prevDP = currDP;
			phi = theta;
		}
		theta = theta + delta;
		cnt++;
	}
	q->r = cos( phi/2.0 );
	q->i = tangent[0]*sin( phi/2.0 );
	q->j = tangent[1]*sin( phi/2.0 );
	q->k = tangent[2]*sin( phi/2.0 );
	//printf("#\tq = %f, %f, %f, %f\n",q->r,q->i,q->j,q->k);

	quatQuatProduct( q, b, r );
	quatConjugate(r);
	//printf("#\treturn = %f, %f, %f, %f\n",r->r,r->i,r->j,r->k);
	//printf("DP= %f\n",quatQuatDotProduct( r, s ));
	//printf("#\tr = %f, %f, %f, %f\n",r->r,r->i,r->j,r->k);
	quatCopy(r, retQuat);
	*retAngle = phi;
	//printf("l_currQuat = %f\n",phi);
	//printf("l_currQuat  = %f, %f, %f, %f\n",r->r,r->i,r->j,r->k);

	my_free(b);
	my_free(q);
	my_free(r);
	my_free(s);
}
//*********************************************//
//
//*********************************************//
void findQuat_Previous							( int ind, quaternion *prevQuat, double* prevTheta ) 
{

	int			i			= ind;
	double		tangent[3];
	double		crossZT[3];
	double		norm_crossZT= 0.0;
	double		thetaZT		= 0.0;
	double		z[3]		= {0.0, 0.0, 1.0};
	double		l_currTheta = 0.0;
	double		l_prevTheta	= 0.0;
	double		epsilon		= mo2->opt_EPSILON;
	quaternion	*b;
	quaternion	*q;
	quaternion	*l_prevQuat; 
	quaternion	*l_currQuat;

	b			=	(quaternion*) my_malloc(sizeof(quaternion));
	q			=	(quaternion*) my_malloc(sizeof(quaternion));
	l_prevQuat	=	(quaternion*) my_malloc(sizeof(quaternion));
	l_currQuat	=	(quaternion*) my_malloc(sizeof(quaternion));
	i			= mo2->trackingRing;
	l_prevQuat->r = l_prevQuat->i = l_prevQuat->j = l_prevQuat->k = 0.0;

	findQuat_Current( mo2->trackingRing, l_currQuat, &(l_currTheta) );

	tangent[0] =	mo2->TangList[i].p3.d[0];
	tangent[1] =	mo2->TangList[i].p3.d[1];
	tangent[2] =	mo2->TangList[i].p3.d[2];

	//compute (z x tangent)
	cross3D( z, tangent, crossZT );
	norm_crossZT = length3D( crossZT );
	crossZT[0] /= norm_crossZT;
	crossZT[1] /= norm_crossZT;
	crossZT[2] /= norm_crossZT;

	//compute arccos(z . tangent)
	thetaZT = acos( dot3D( z, tangent ) );

	//compute b(vhat)
	b->r = cos( thetaZT/2.0 );
	b->i = crossZT[0]*sin( thetaZT/2.0 );
	b->j = crossZT[1]*sin( thetaZT/2.0 );
	b->k = crossZT[2]*sin( thetaZT/2.0 );

	l_prevTheta = l_currTheta - epsilon;
	q->r =			  cos( l_prevTheta/2.0 );
	q->i = tangent[0]*sin( l_prevTheta/2.0 );
	q->j = tangent[1]*sin( l_prevTheta/2.0 );
	q->k = tangent[2]*sin( l_prevTheta/2.0 );

	quatQuatProduct( q, b, l_prevQuat );
	quatConjugate( l_prevQuat );

	quatCopy( l_prevQuat, prevQuat );
	if( l_prevTheta < 0.0 )
		l_prevTheta = 4.0*M_PI - l_prevTheta;
	if (l_prevTheta > 4.0*M_PI )
		l_prevTheta = l_prevTheta - 4.0*M_PI;

	*prevTheta = l_prevTheta;
	//printf("l_prevTheta = %f\n",l_prevTheta);
	//printf("l_prevQuat  = %f, %f, %f, %f\n",l_prevQuat->r,l_prevQuat->i,l_prevQuat->j,l_prevQuat->k);

	my_free(b);
	my_free(q);
	my_free(l_prevQuat);
	my_free(l_currQuat);
}
//*********************************************//
//
//*********************************************//
void findQuat_Previous2							( int ind, quaternion *prevQuat, double* prevTheta, int SIGN ) 
{

	int			i			= ind;
	double		tangent[3];
	double		crossZT[3];
	double		norm_crossZT= 0.0;
	double		thetaZT		= 0.0;
	double		z[3]		= {0.0, 0.0, 1.0};
	double		l_currTheta = 0.0;
	double		l_prevTheta	= 0.0;
	double		epsilon		= mo2->opt_EPSILON;
	quaternion	*b;
	quaternion	*q;
	quaternion	*l_prevQuat; 
	quaternion	*l_currQuat;

	b			=	(quaternion*) my_malloc(sizeof(quaternion));
	q			=	(quaternion*) my_malloc(sizeof(quaternion));
	l_prevQuat	=	(quaternion*) my_malloc(sizeof(quaternion));
	l_currQuat	=	(quaternion*) my_malloc(sizeof(quaternion));
	i			= mo2->trackingRing;
	l_prevQuat->r = l_prevQuat->i = l_prevQuat->j = l_prevQuat->k = 0.0;

	findQuat_Current( mo2->trackingRing, l_currQuat, &(l_currTheta) );

	tangent[0] =	mo2->TangList[i].p3.d[0];
	tangent[1] =	mo2->TangList[i].p3.d[1];
	tangent[2] =	mo2->TangList[i].p3.d[2];

	//compute (z x tangent)
	cross3D( z, tangent, crossZT );
	norm_crossZT = length3D( crossZT );
	crossZT[0] /= norm_crossZT;
	crossZT[1] /= norm_crossZT;
	crossZT[2] /= norm_crossZT;

	//compute arccos(z . tangent)
	thetaZT = acos( dot3D( z, tangent ) );

	//compute b(vhat)
	b->r = cos( thetaZT/2.0 );
	b->i = crossZT[0]*sin( thetaZT/2.0 );
	b->j = crossZT[1]*sin( thetaZT/2.0 );
	b->k = crossZT[2]*sin( thetaZT/2.0 );

	l_prevTheta = l_currTheta + SIGN*epsilon;
	q->r =			  cos( l_prevTheta/2.0 );
	q->i = tangent[0]*sin( l_prevTheta/2.0 );
	q->j = tangent[1]*sin( l_prevTheta/2.0 );
	q->k = tangent[2]*sin( l_prevTheta/2.0 );

	quatQuatProduct( q, b, l_prevQuat );
	quatConjugate( l_prevQuat );

	quatCopy( l_prevQuat, prevQuat );
	if( l_prevTheta < 0.0 )
		l_prevTheta = 4.0*M_PI - l_prevTheta;
	if (l_prevTheta > 4.0*M_PI )
		l_prevTheta = l_prevTheta - 4.0*M_PI;

	*prevTheta = l_prevTheta;
	//printf("l_prevTheta = %f\n",l_prevTheta);
	//printf("l_prevQuat  = %f, %f, %f, %f\n",l_prevQuat->r,l_prevQuat->i,l_prevQuat->j,l_prevQuat->k);

	my_free(b);
	my_free(q);
	my_free(l_prevQuat);
	my_free(l_currQuat);
}
//*********************************************//
//
//*********************************************//
void findQuat_Update							( int ind, quaternion *retQuat,  double currTheta ) 
{

	int			i			= ind;
	double		tangent[3];
	double		crossZT[3];
	double		norm_crossZT= 0.0;
	double		thetaZT		= 0.0;
	double		z[3]		= {0.0, 0.0, 1.0};
	quaternion	*b;
	quaternion	*q;
	quaternion	*r;

	b			= (quaternion*) my_malloc(sizeof(quaternion));
	q			= (quaternion*) my_malloc(sizeof(quaternion));
	r			= (quaternion*) my_malloc(sizeof(quaternion));
	i			= mo2->trackingRing;

	tangent[0] =	mo2->TangList[i].p3.d[0];
	tangent[1] =	mo2->TangList[i].p3.d[1];
	tangent[2] =	mo2->TangList[i].p3.d[2];

	//compute (z x tangent)
	cross3D( z, tangent, crossZT );
	norm_crossZT = length3D( crossZT );
	crossZT[0] /= norm_crossZT;
	crossZT[1] /= norm_crossZT;
	crossZT[2] /= norm_crossZT;

	//compute arccos(z . tangent)
	thetaZT = acos( dot3D( z, tangent ) );

	//compute b(vhat)
	b->r = cos( thetaZT/2.0 );
	b->i = crossZT[0]*sin( thetaZT/2.0 );
	b->j = crossZT[1]*sin( thetaZT/2.0 );
	b->k = crossZT[2]*sin( thetaZT/2.0 );

	q->r =			  cos( currTheta/2.0 );
	q->i = tangent[0]*sin( currTheta/2.0 );
	q->j = tangent[1]*sin( currTheta/2.0 );
	q->k = tangent[2]*sin( currTheta/2.0 );

	quatQuatProduct( q, b, r );
	quatConjugate( r );

	quatCopy( r, retQuat );

	my_free(b);
	my_free(q);
	my_free(r);
}
//*********************************************//
//
//*********************************************//
void test_FindQuat								( void ) 
{
	int i = 0;
	double		Theta=0.0;
	quaternion	*Quat; 
	double		qvec[3];
	Quat	=	(quaternion*) my_malloc(sizeof(quaternion));	
	Quat->r = Quat->i = Quat->j = Quat->k = 0.0;

	//for(i=0;i<10;i++) {
		findQuat_Previous( i, Quat, &(Theta) );
		getQuatComp( mo2->qwhichQuatComponent, Quat, qvec );
		//printf("Quat  = %f, %f, %f, %f\n",Quat->r,Quat->i,Quat->j,Quat->k);
		glEnable(GL_LIGHTING);
		glEnable(GL_COLOR_MATERIAL);
		glColor3f(.9,.9,.9);
		glPushMatrix();
			glTranslatef(qvec[0],qvec[1],qvec[2]);
			glutSolidSphere( 0.02,16,16);
		glPopMatrix();

	//}
    my_free(Quat);
}
//*********************************************//
//
//*********************************************//
void test_FindQuat								( double ANGLE ) 
{
	int i = 0;
	quaternion	*Quat; 
	double		qvec[3];
	Quat	=	(quaternion*) my_malloc(sizeof(quaternion));	
	Quat->r = Quat->i = Quat->j = Quat->k = 0.0;

	//for(i=0;i<10;i++) {
		findQuat_Update( i, Quat, ANGLE );
		getQuatComp( mo2->qwhichQuatComponent, Quat, qvec );
		//printf("Quat  = %f, %f, %f, %f\n",Quat->r,Quat->i,Quat->j,Quat->k);
		glEnable(GL_LIGHTING);
		glEnable(GL_COLOR_MATERIAL);
		glColor3f(.1,.1,.1);
		glPushMatrix();
			glTranslatef(qvec[0],qvec[1],qvec[2]);
			glutSolidSphere( 0.02,16,16);
		glPopMatrix();

	//}
    my_free(Quat);

}
//### COMPUTATION: quaternion space
void computeTorus								( void )
{
	double u=0.0, v=0.0;
	double r=0.0, a=0.0, b=0.0;
	int    i=0, j=0;
	double dTheta = 2*M_PI/mo2->TORUS_STPS;
	double dPhi   = 2*M_PI/mo2->TORUS_PNTS;

	//normal calculations
	double	v1[3],v2[3];
	double	n[3];
	double	len		= 0.0;

	r = mo2->param_r;
	a = mo2->param_a-0.0075;
	b = mo2->param_b-0.0075;

	//@ compute torus points
	for(i=0; i<mo2->TORUS_PNTS; i++) {
		for(j=0; j<mo2->TORUS_STPS; j++) {
			mo2->tpnts[i][j][0] = (r + a*cos(v))*cos(u);
			mo2->tpnts[i][j][1] = (r + a*cos(v))*sin(u);
			mo2->tpnts[i][j][2] = (b)*sin(v);
			//glPushMatrix();
			//glTranslatef(mo2->tpnts[i][j][0],mo2->tpnts[i][j][1],mo2->tpnts[i][j][2]);
			//glutSolidSphere(0.01,8,8);
			//glPopMatrix();
			v = v+dTheta;
		}
		u = u+dPhi;
	}
	mo2->totalTorusPoints = i;

	//@ compute normals
	for(i=0; i<(mo2->TORUS_PNTS); i++) {
		for(j=0; j<(mo2->TORUS_STPS); j++) {
			if(i<(mo2->TORUS_PNTS-1) ) {
				v1[0] = mo2->tpnts[i+1][j][0] - mo2->tpnts[i][j][0];
				v1[1] = mo2->tpnts[i+1][j][1] - mo2->tpnts[i][j][1];
				v1[2] = mo2->tpnts[i+1][j][2] - mo2->tpnts[i][j][2];
			}
			else {
				v1[0] = mo2->tpnts[0][j][0] - mo2->tpnts[i][j][0];
				v1[1] = mo2->tpnts[0][j][1] - mo2->tpnts[i][j][1];
				v1[2] = mo2->tpnts[0][j][2] - mo2->tpnts[i][j][2];
			}
			if(j<(mo2->TORUS_STPS-1)) {
				v2[0] = mo2->tpnts[i][j+1][0] - mo2->tpnts[i][j][0];
				v2[1] = mo2->tpnts[i][j+1][1] - mo2->tpnts[i][j][1];
				v2[2] = mo2->tpnts[i][j+1][2] - mo2->tpnts[i][j][2];
			}
			else {
				v2[0] = mo2->tpnts[i][0][0] - mo2->tpnts[i][j][0];
				v2[1] = mo2->tpnts[i][0][1] - mo2->tpnts[i][j][1];
				v2[2] = mo2->tpnts[i][0][2] - mo2->tpnts[i][j][2];
			}
			cross3D(v1,v2,n);
			len = length3D(n);
			n[0] /= len;
			n[1] /= len;
			n[2] /= len;
			mo2->upnts[i][j][0] = n[0];
			mo2->upnts[i][j][1] = n[1];
			mo2->upnts[i][j][2] = n[2];
		}//$ FOR(j..)
	}//$ FOR(i..)
}
//*********************************************//
//
//*********************************************//
void computeSphere								( void ) 
{
	double theta=0.0, phi=0.0;
	double r=0.0, a=0.0, b=0.0;
	int    i=0, j=0;
	//double	maxPnts = t_pnts;
	//double	maxStps = t_stps;
	int		maxPnts = mo2->u;
	int		maxStps = mo2->v;
	double dTheta = 2.0*M_PI/mo2->TORUS_STPS;
	double dPhi   = M_PI/mo2->TORUS_PNTS;

	//normal calculations
	double	v1[3],v2[3];
	double	n[3];
	double	len		= 0.0;

	r = mo2->param_r;
	a = mo2->param_a-0.0075;
	b = mo2->param_b-0.0075;
	glPointSize(5.0f);
	//@ compute sphere points
	for(i=0; i<maxPnts; i++) {
		theta = 0.0;
		for(j=0; j<maxStps; j++) {
			mo2->tpnts[i][j][0] = a*( r )*cos(theta)*sin(phi);
			mo2->tpnts[i][j][1] = a*( r )*sin(theta)*sin(phi);
			mo2->tpnts[i][j][2] = b*( r )*cos(phi);
			theta = theta+dTheta;
			//glPushMatrix();
			//glTranslatef(mo2->tpnts[i][j][0],mo2->tpnts[i][j][1],mo2->tpnts[i][j][2]);
			//glutSolidSphere(0.01,8,8);
			//glPopMatrix();
		}
		phi = phi+dPhi;
	}
	mo2->totalTorusPoints = i;

	//@ compute normals
	for(i=0; i<(maxPnts); i++) {
		for(j=0; j<(maxStps); j++) {
			if(i<(mo2->TORUS_PNTS-1) ) {
				v1[0] = mo2->tpnts[i+1][j][0] - mo2->tpnts[i][j][0];
				v1[1] = mo2->tpnts[i+1][j][1] - mo2->tpnts[i][j][1];
				v1[2] = mo2->tpnts[i+1][j][2] - mo2->tpnts[i][j][2];
			}
			else {
				v1[0] = mo2->tpnts[0][j][0] - mo2->tpnts[i][j][0];
				v1[1] = mo2->tpnts[0][j][1] - mo2->tpnts[i][j][1];
				v1[2] = mo2->tpnts[0][j][2] - mo2->tpnts[i][j][2];
			}
			if(j<(mo2->TORUS_STPS-1)) {
				v2[0] = mo2->tpnts[i][j+1][0] - mo2->tpnts[i][j][0];
				v2[1] = mo2->tpnts[i][j+1][1] - mo2->tpnts[i][j][1];
				v2[2] = mo2->tpnts[i][j+1][2] - mo2->tpnts[i][j][2];
			}
			else {
				v2[0] = mo2->tpnts[i][0][0] - mo2->tpnts[i][j][0];
				v2[1] = mo2->tpnts[i][0][1] - mo2->tpnts[i][j][1];
				v2[2] = mo2->tpnts[i][0][2] - mo2->tpnts[i][j][2];
			}
			cross3D(v1,v2,n);
			len = length3D(n);
			n[0] /= len;
			n[1] /= len;
			n[2] /= len;
			mo2->upnts[i][j][0] = n[0];
			mo2->upnts[i][j][1] = n[1];
			mo2->upnts[i][j][2] = n[2];
		}//$ FOR(j..)
	}//$ FOR(i..)	
}

//*********************************************//
//
//*********************************************//
void computeCatenoid							( void ) 
{
	double u=0.0, v=0.0;
	double r=0.0, a=0.0, b=0.0;
	int    i=0, j=0;
	double dTheta = 2*M_PI/mo2->TORUS_STPS;
	double dPhi   = M_PI/mo2->TORUS_PNTS;

	//normal calculations
	double	v1[3],v2[3];
	double	n[3];
	double	len		= 0.0;

	r = mo2->param_r;
	a = mo2->param_a-0.0075;
	b = mo2->param_b-0.0075;

	//@ compute torus points
	u = -M_PI/2.0;
	v = 0.0;
	for(i=0; i<mo2->TORUS_PNTS; i++) {
		for(j=0; j<mo2->TORUS_STPS; j++) {
			mo2->tpnts[i][j][0] = r*a*( cos(v)*cosh(u) );
			mo2->tpnts[i][j][1] = r*a*( sin(v)*cosh(u) );
			mo2->tpnts[i][j][2] = u*b;
			v = v+dTheta;
			//glPushMatrix();
			//glTranslatef(mo2->tpnts[i][j][0],mo2->tpnts[i][j][1],mo2->tpnts[i][j][2]);
			//glutSolidSphere(0.01,8,8);
			//glPopMatrix();
		}
		u = u+dPhi;
	}
	mo2->totalTorusPoints = i;

	//@ compute normals
	for(i=0; i<(mo2->TORUS_PNTS); i++) {
		for(j=0; j<(mo2->TORUS_STPS); j++) {
			if(i<(mo2->TORUS_PNTS-1) ) {
				v1[0] = mo2->tpnts[i+1][j][0] - mo2->tpnts[i][j][0];
				v1[1] = mo2->tpnts[i+1][j][1] - mo2->tpnts[i][j][1];
				v1[2] = mo2->tpnts[i+1][j][2] - mo2->tpnts[i][j][2];
			}
			else {
				v1[0] = mo2->tpnts[0][j][0] - mo2->tpnts[i][j][0];
				v1[1] = mo2->tpnts[0][j][1] - mo2->tpnts[i][j][1];
				v1[2] = mo2->tpnts[0][j][2] - mo2->tpnts[i][j][2];
			}
			if(j<(mo2->TORUS_STPS-1)) {
				v2[0] = mo2->tpnts[i][j+1][0] - mo2->tpnts[i][j][0];
				v2[1] = mo2->tpnts[i][j+1][1] - mo2->tpnts[i][j][1];
				v2[2] = mo2->tpnts[i][j+1][2] - mo2->tpnts[i][j][2];
			}
			else {
				v2[0] = mo2->tpnts[i][0][0] - mo2->tpnts[i][j][0];
				v2[1] = mo2->tpnts[i][0][1] - mo2->tpnts[i][j][1];
				v2[2] = mo2->tpnts[i][0][2] - mo2->tpnts[i][j][2];
			}
			cross3D(v1,v2,n);
			len = length3D(n);
			n[0] /= len;
			n[1] /= len;
			n[2] /= len;
			mo2->upnts[i][j][0] = n[0];
			mo2->upnts[i][j][1] = n[1];
			mo2->upnts[i][j][2] = n[2];
		}//$ FOR(j..)
	}//$ FOR(i..)
}
//*********************************************//
//
//*********************************************//
void computeEnneper								( void ) 
{
	double u=0.0, v=0.0;
	double r=0.0, a=0.0, b=0.0;
	int    i=0, j=0;
	double dTheta = 1.0/mo2->TORUS_STPS;
	double dPhi   = 1.0/mo2->TORUS_PNTS;

	//normal calculations
	double	v1[3],v2[3];
	double	n[3];
	double	len		= 0.0;

	r = mo2->param_r;
	a = mo2->param_a-0.0075;
	b = mo2->param_b-0.0075;

	//@ compute torus points
	u = 0.0;
	v = 0.0;
	for(i=0; i<mo2->TORUS_PNTS; i++) {
		for(j=0; j<mo2->TORUS_STPS; j++) {
			mo2->tpnts[i][j][0] = r*( u - (u*u*u)/3.0 + u*v*v );
			mo2->tpnts[i][j][1] = r*( v - (v*v*v)/3.0 + v*u*u );
			mo2->tpnts[i][j][2] = r*( u*u - v*v );
			v = v+dTheta;
			//glPushMatrix();
			//glTranslatef(mo2->tpnts[i][j][0],mo2->tpnts[i][j][1],mo2->tpnts[i][j][2]);
			//glutSolidSphere(0.01,8,8);
			//glPopMatrix();
		}
		u = u+dPhi;
	}
	mo2->totalTorusPoints = i;

	//@ compute normals
	for(i=0; i<(mo2->TORUS_PNTS); i++) {
		for(j=0; j<(mo2->TORUS_STPS); j++) {
			if(i<(mo2->TORUS_PNTS-1) ) {
				v1[0] = mo2->tpnts[i+1][j][0] - mo2->tpnts[i][j][0];
				v1[1] = mo2->tpnts[i+1][j][1] - mo2->tpnts[i][j][1];
				v1[2] = mo2->tpnts[i+1][j][2] - mo2->tpnts[i][j][2];
			}
			else {
				v1[0] = mo2->tpnts[0][j][0] - mo2->tpnts[i][j][0];
				v1[1] = mo2->tpnts[0][j][1] - mo2->tpnts[i][j][1];
				v1[2] = mo2->tpnts[0][j][2] - mo2->tpnts[i][j][2];
			}
			if(j<(mo2->TORUS_STPS-1)) {
				v2[0] = mo2->tpnts[i][j+1][0] - mo2->tpnts[i][j][0];
				v2[1] = mo2->tpnts[i][j+1][1] - mo2->tpnts[i][j][1];
				v2[2] = mo2->tpnts[i][j+1][2] - mo2->tpnts[i][j][2];
			}
			else {
				v2[0] = mo2->tpnts[i][0][0] - mo2->tpnts[i][j][0];
				v2[1] = mo2->tpnts[i][0][1] - mo2->tpnts[i][j][1];
				v2[2] = mo2->tpnts[i][0][2] - mo2->tpnts[i][j][2];
			}
			cross3D(v1,v2,n);
			len = length3D(n);
			n[0] /= len;
			n[1] /= len;
			n[2] /= len;
			mo2->upnts[i][j][0] = n[0];
			mo2->upnts[i][j][1] = n[1];
			mo2->upnts[i][j][2] = n[2];
		}//$ FOR(j..)
	}//$ FOR(i..)
}
//### RENDERINGS: surfaces
void renderTorus								( void )
{
    int		i		= 0;
	int		j		= 0;
	//double	v1[3],v2[3];
	//double	len		= 0.0;
	//double	scale	= 0.10;
	//double	maxPnts = mo2->TORUS_PNTS;
	//double	maxStps	= mo2->TORUS_STPS;
	double	maxPnts = mo2->u;
	double	maxStps = mo2->v;

	assert(maxPnts<=mo2->TORUS_PNTS);
	assert(maxStps<=mo2->TORUS_STPS);

    ///////////////////////// RING /////////////////////////
    if(mo2->showEdge) {
        glPushAttrib(GL_CURRENT_BIT | GL_LIGHTING_BIT | GL_ENABLE_BIT | GL_LINE_BIT);
        glDisable(GL_LIGHTING);
        glLineWidth(2.5f);
        glColor3f(0.2f,0.2f,0.2f);
        glBegin(GL_LINES);
        for(i=0; i<mo2->TORUS_PNTS; i++) {
            for(j=0; j<mo2->TORUS_STPS; j++) {
                glVertex3d(mo2->tpnts[i][j][0],mo2->tpnts[i][j][1],mo2->tpnts[i][j][2]);
                if(j<(mo2->TORUS_STPS-1))
                    glVertex3d(mo2->tpnts[i][j+1][0],mo2->tpnts[i][j+1][1],mo2->tpnts[i][j+1][2]);
                else
                    glVertex3d(mo2->tpnts[i][0][0],mo2->tpnts[i][0][1],mo2->tpnts[i][0][2]);
            }
        }
        glEnd();
		glPopAttrib();

        ///////////////////////// LONGITUDINAL LINES /////////////////////////
        glPushAttrib(GL_CURRENT_BIT | GL_LIGHTING_BIT | GL_ENABLE_BIT | GL_LINE_BIT);
        glDisable(GL_LIGHTING);
        glLineWidth(2.5f);
		glBegin(GL_LINES);
        glColor3f(0.2f,0.2f,0.2f);
        for(i=1;i<mo2->TORUS_PNTS; i++) {
            for(j=0;j<mo2->TORUS_STPS;j++) {
                glVertex3d(mo2->tpnts[i][j][0],		mo2->tpnts[i][j][1],	mo2->tpnts[i][j][2]); 
                glVertex3d(mo2->tpnts[i-1][j][0],	mo2->tpnts[i-1][j][1],	mo2->tpnts[i-1][j][2]);
                if( i==(mo2->TORUS_PNTS-1) ){
                    glVertex3d(mo2->tpnts[i][j][0],mo2->tpnts[i][j][1],mo2->tpnts[i][j][2]); 
                    glVertex3d(mo2->tpnts[0][j][0],mo2->tpnts[0][j][1],mo2->tpnts[0][j][2]);
                }
            }
        }
        glEnd();
        glPopAttrib();
	}//$ IF(showEdge..)

    ///////////////////////// QUAD FACES /////////////////////////
	if(mo2->showFace) {
		glPushAttrib(GL_CURRENT_BIT | GL_LIGHTING_BIT | GL_ENABLE_BIT | GL_LINE_BIT);
		glDisable(GL_LIGHTING);
		
		glLineWidth(0.1f);
		
		i=0;
		j=0;
		for(i=0; i<(maxPnts); i++) {
			for(j=0; j<(maxStps); j++) {
				glEnable(GL_LIGHTING);
				glEnable(GL_COLOR_MATERIAL);
				glColor3f(mo2->fgcolor[0],mo2->fgcolor[1],mo2->fgcolor[2]);
				glBegin(GL_QUADS);
					//glNormal3f( n[0], n[1], n[2] );
					glNormal3f( mo2->upnts[i][j][0], mo2->upnts[i][j][1], mo2->upnts[i][j][2] );
					if(i<(mo2->TORUS_PNTS-1)) {
						if((j<mo2->TORUS_STPS-1)) {
							glVertex3d(mo2->tpnts[i][j][0],		mo2->tpnts[i][j][1],	mo2->tpnts[i][j][2]); 
							glVertex3d(mo2->tpnts[i+1][j][0],	mo2->tpnts[i+1][j][1],	mo2->tpnts[i+1][j][2]); 
							glVertex3d(mo2->tpnts[i+1][j+1][0],	mo2->tpnts[i+1][j+1][1],mo2->tpnts[i+1][j+1][2]); 
							glVertex3d(mo2->tpnts[i][j+1][0],	mo2->tpnts[i][j+1][1],	mo2->tpnts[i][j+1][2]); 
						}
						else {
							glVertex3d(mo2->tpnts[i][j][0],		mo2->tpnts[i][j][1],	mo2->tpnts[i][j][2]); 
							glVertex3d(mo2->tpnts[i+1][j][0],	mo2->tpnts[i+1][j][1],	mo2->tpnts[i+1][j][2]); 
							glVertex3d(mo2->tpnts[i+1][0][0],	mo2->tpnts[i+1][0][1],	mo2->tpnts[i+1][0][2]); 
							glVertex3d(mo2->tpnts[i][0][0],		mo2->tpnts[i][0][1],	mo2->tpnts[i][0][2]); 
						}
					}
					else {
						if((j<mo2->TORUS_STPS-1)) {
							glVertex3d(mo2->tpnts[i][j][0],		mo2->tpnts[i][j][1],	mo2->tpnts[i][j][2]); 
							glVertex3d(mo2->tpnts[0][j][0],		mo2->tpnts[0][j][1],	mo2->tpnts[0][j][2]); 
							glVertex3d(mo2->tpnts[0][j+1][0],	mo2->tpnts[0][j+1][1],	mo2->tpnts[0][j+1][2]); 
							glVertex3d(mo2->tpnts[i][j+1][0],	mo2->tpnts[i][j+1][1],	mo2->tpnts[i][j+1][2]); 
						}
						else {
							glVertex3d(mo2->tpnts[i][j][0],		mo2->tpnts[i][j][1],	mo2->tpnts[i][j][2]); 
							glVertex3d(mo2->tpnts[0][j][0],		mo2->tpnts[0][j][1],	mo2->tpnts[0][j][2]); 
							glVertex3d(mo2->tpnts[0][0][0],		mo2->tpnts[0][0][1],	mo2->tpnts[0][0][2]); 
							glVertex3d(mo2->tpnts[i][0][0],		mo2->tpnts[i][0][1],	mo2->tpnts[i][0][2]); 
						}
					}
				glEnd();
			}//$ FOR(j..)
		}//$ FOR(i..)
		glPopAttrib();
	}//$ IF(showFace..)

    ///////////////////////// NORMALS /////////////////////////
	if(mo2->showNormals) {
		glPushAttrib(GL_CURRENT_BIT | GL_LIGHTING_BIT | GL_ENABLE_BIT | GL_LINE_BIT);
		glDisable(GL_LIGHTING);
		
		glLineWidth(1.5f);
		
		i=0;
		j=0;
		for(i=0; i<(maxPnts); i++) {
			for(j=0; j<(maxStps); j++) {
				//if(i<(mo2->TORUS_PNTS-1) ) {
				//	v1[0] = mo2->tpnts[i+1][j][0] - mo2->tpnts[i][j][0];
				//	v1[1] = mo2->tpnts[i+1][j][1] - mo2->tpnts[i][j][1];
				//	v1[2] = mo2->tpnts[i+1][j][2] - mo2->tpnts[i][j][2];
				//}
				//else {
				//	v1[0] = mo2->tpnts[0][j][0] - mo2->tpnts[i][j][0];
				//	v1[1] = mo2->tpnts[0][j][1] - mo2->tpnts[i][j][1];
				//	v1[2] = mo2->tpnts[0][j][2] - mo2->tpnts[i][j][2];
				//}
				//if(j<(mo2->TORUS_STPS-1)) {
				//	v2[0] = mo2->tpnts[i][j+1][0] - mo2->tpnts[i][j][0];
				//	v2[1] = mo2->tpnts[i][j+1][1] - mo2->tpnts[i][j][1];
				//	v2[2] = mo2->tpnts[i][j+1][2] - mo2->tpnts[i][j][2];
				//}
				//else {
				//	v2[0] = mo2->tpnts[i][0][0] - mo2->tpnts[i][j][0];
				//	v2[1] = mo2->tpnts[i][0][1] - mo2->tpnts[i][j][1];
				//	v2[2] = mo2->tpnts[i][0][2] - mo2->tpnts[i][j][2];
				//}

				//len = length3D( v1 );
				//v1[0] /= len; 
				//v1[1] /= len; 
				//v1[2] /= len; 

				//len = length3D( v2 );
				//v2[0] /= len; 
				//v2[1] /= len; 
				//v2[2] /= len; 

			//glColor3f(0,0,1);
			//glPushMatrix();
			//	glTranslatef(mo2->tpnts[i][j][0],		mo2->tpnts[i][j][1],	mo2->tpnts[i][j][2]); 
			//	glScalef( scale, scale, scale );
			//	glBegin(GL_LINES);
			//	glVertex3f(0,0,0);
			//	glVertex3f(mo2->upnts[i][j][0],		mo2->upnts[i][j][1],	mo2->upnts[i][j][2]);
			//	glEnd();
			//glPopMatrix();
			//glColor3f(1,0,0);
			//glPushMatrix();
			//	glTranslatef(mo2->tpnts[i][j][0],		mo2->tpnts[i][j][1],	mo2->tpnts[i][j][2]); 
			//	glScalef( scale, scale, scale );
			//	glBegin(GL_LINES);
			//	glVertex3f(0,0,0);
			//	glVertex3f(mo2->grRBasis[i].p3.d[0], mo2->grRBasis[i].p3.d[1], mo2->grRBasis[i].p3.d[2]);
			//	glEnd();
			//glPopMatrix();
			//glColor3f(0,1,0);
			//glPushMatrix();
			//	glTranslatef(mo2->tpnts[i][j][0],		mo2->tpnts[i][j][1],	mo2->tpnts[i][j][2]); 
			//	glScalef( scale, scale, scale );
			//	glBegin(GL_LINES);
			//	glVertex3f(0,0,0);
			//	glVertex3f(mo2->grSBasis[i].p3.d[0], mo2->grSBasis[i].p3.d[1], mo2->grSBasis[i].p3.d[2]);
			//	glEnd();
			//glPopMatrix();
			
			//glColor3f(1,0,0);
				//glPushMatrix();
				//	glTranslatef(mo2->tpnts[i][j][0],		mo2->tpnts[i][j][1],	mo2->tpnts[i][j][2]); 
				//	glScalef( scale, scale, scale );
				//	glBegin(GL_LINES);
				//	glVertex3f(0,0,0);
				//	glVertex3f(mo2->grRBasis[i].p3.d[0], mo2->grRBasis[i].p3.d[1], mo2->grRBasis[i].p3.d[2]);
				//	glEnd();
				//glPopMatrix();

				//glColor3f(0,1,0);
				//glPushMatrix();
				//	glTranslatef(mo2->tpnts[i][j][0],		mo2->tpnts[i][j][1],	mo2->tpnts[i][j][2]); 
				//	glScalef( scale, scale, scale );
				//	glBegin(GL_LINES);
				//	glVertex3f(0,0,0);
				//	glVertex3f(mo2->grSBasis[i].p3.d[0], mo2->grSBasis[i].p3.d[1], mo2->grSBasis[i].p3.d[2]);
				//	glEnd();
				//glPopMatrix();

				//glColor3f(0,0,1);
				//glPushMatrix();
				//	glTranslatef(mo2->tpnts[i][j][0],		mo2->tpnts[i][j][1],	mo2->tpnts[i][j][2]); 
				//	glScalef( scale, scale, scale );
				//	glBegin(GL_LINES);
				//	glVertex3f(0,0,0);
				//	glVertex3f(mo2->upnts[i][j][0],		mo2->upnts[i][j][1],	mo2->upnts[i][j][2]);
				//	glEnd();
				//glPopMatrix();
			}//$ FOR(j..)
		}//$ FOR(i..)
		glPopAttrib();

	}
}
//*********************************************//
//
//*********************************************//
void renderSurfaceObject 						( void )
{
    int		i		= 0;
	int		j		= 0;
	int		limitPnts=0;
	int		limitSpts=0;
	//double	v1[3],v2[3];
	//double	len		= 0.0;
	//double	scale	= 0.10;
	//double	maxPnts = mo2->TORUS_PNTS;
	//double	maxStps	= mo2->TORUS_STPS;
	double	maxPnts = mo2->u;
	double	maxStps = mo2->v;
	limitPnts = maxPnts;
	limitSpts = maxStps;

	assert(maxPnts<=mo2->TORUS_PNTS);
	assert(maxStps<=mo2->TORUS_STPS);
	
    ///////////////////////// RING /////////////////////////
    if(mo2->showEdge) {
        glPushAttrib(GL_CURRENT_BIT | GL_LIGHTING_BIT | GL_ENABLE_BIT | GL_LINE_BIT);
        glDisable(GL_LIGHTING);
        glLineWidth(2.5f);
        glColor3f(0.2f,0.2f,0.2f);
        glBegin(GL_LINES);
        for(i=0; i<limitPnts; i++) {
            for(j=0; j<limitSpts; j++) {
                glVertex3d(mo2->tpnts[i][j][0],mo2->tpnts[i][j][1],mo2->tpnts[i][j][2]);
                if(j<(mo2->TORUS_STPS-1))
                    glVertex3d(mo2->tpnts[i][j+1][0],mo2->tpnts[i][j+1][1],mo2->tpnts[i][j+1][2]);
                else
                    glVertex3d(mo2->tpnts[i][0][0],mo2->tpnts[i][0][1],mo2->tpnts[i][0][2]);
            }
        }
        glEnd();
		glPopAttrib();

        ///////////////////////// LONGITUDINAL LINES /////////////////////////
        glPushAttrib(GL_CURRENT_BIT | GL_LIGHTING_BIT | GL_ENABLE_BIT | GL_LINE_BIT);
        glDisable(GL_LIGHTING);
        glLineWidth(2.5f);
		glBegin(GL_LINES);
        glColor3f(0.2f,0.2f,0.2f);
		if(mo2->surfObjectID==2)
			limitPnts = maxPnts-1;
		else 
			limitPnts = maxPnts;
        for(i=1;i<limitPnts; i++) {
            for(j=0;j<limitSpts;j++) {
                glVertex3d(mo2->tpnts[i][j][0],		mo2->tpnts[i][j][1],	mo2->tpnts[i][j][2]); 
                glVertex3d(mo2->tpnts[i-1][j][0],	mo2->tpnts[i-1][j][1],	mo2->tpnts[i-1][j][2]);
                if( i==(mo2->TORUS_PNTS-1) ){
                    glVertex3d(mo2->tpnts[i][j][0],mo2->tpnts[i][j][1],mo2->tpnts[i][j][2]); 
                    glVertex3d(mo2->tpnts[0][j][0],mo2->tpnts[0][j][1],mo2->tpnts[0][j][2]);
                }
            }
        }
        glEnd();
        glPopAttrib();
	}//$ IF(showEdge..)

    ///////////////////////// QUAD FACES /////////////////////////
	if(mo2->showFace) {
		glPushAttrib(GL_CURRENT_BIT | GL_LIGHTING_BIT | GL_ENABLE_BIT | GL_LINE_BIT);
		glDisable(GL_LIGHTING);
		
		glEnable(GL_POLYGON_OFFSET_FILL);
		glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);
		glPolygonOffset(1.0f,1.0f);
		glLineWidth(0.01f);

		i=0;
		j=0;
		if(mo2->surfObjectID==2)
			limitPnts = maxPnts-1;
		else 
			limitPnts = maxPnts;
		for(i=0; i<(limitPnts); i++) {
			for(j=0; j<(limitSpts); j++) {
				glEnable(GL_LIGHTING);
				glEnable(GL_COLOR_MATERIAL);
				glColor3f(mo2->fgcolor[0],mo2->fgcolor[1],mo2->fgcolor[2]);
				glBegin(GL_QUADS);
					
					if(i<(mo2->TORUS_PNTS-1)) {
						if((j<mo2->TORUS_STPS-1)) {
							glNormal3f( mo2->upnts[i][j][0],    mo2->upnts[i][j][1],    mo2->upnts[i][j][2] );
                            glVertex3d( mo2->tpnts[i][j][0],    mo2->tpnts[i][j][1],	mo2->tpnts[i][j][2] ); 

							glNormal3f( mo2->upnts[i+1][j][0],  mo2->upnts[i+1][j][1],  mo2->upnts[i+1][j][2] );
                            glVertex3d( mo2->tpnts[i+1][j][0],	mo2->tpnts[i+1][j][1],	mo2->tpnts[i+1][j][2] ); 
							
                            glNormal3f( mo2->upnts[i+1][j+1][0],mo2->upnts[i+1][j+1][1],mo2->upnts[i+1][j+1][2] );
                            glVertex3d( mo2->tpnts[i+1][j+1][0],mo2->tpnts[i+1][j+1][1],mo2->tpnts[i+1][j+1][2] ); 
							
                            glNormal3f( mo2->upnts[i][j+1][0],  mo2->upnts[i][j+1][1],  mo2->upnts[i][j+1][2] );
                            glVertex3d( mo2->tpnts[i][j+1][0],	mo2->tpnts[i][j+1][1],	mo2->tpnts[i][j+1][2] ); 
						}
						else {
							glNormal3f( mo2->upnts[i][j][0],    mo2->upnts[i][j][1],    mo2->upnts[i][j][2] );
                            glVertex3d(mo2->tpnts[i][j][0],		mo2->tpnts[i][j][1],	mo2->tpnts[i][j][2]); 
							
                            glNormal3f( mo2->upnts[i+1][j][0],  mo2->upnts[i+1][j][1],  mo2->upnts[i+1][j][2] );
                            glVertex3d(mo2->tpnts[i+1][j][0],	mo2->tpnts[i+1][j][1],	mo2->tpnts[i+1][j][2]); 
							
                            glNormal3f( mo2->upnts[i+1][0][0],  mo2->upnts[i+1][0][1],  mo2->upnts[i+1][0][2] );
                            glVertex3d(mo2->tpnts[i+1][0][0],	mo2->tpnts[i+1][0][1],	mo2->tpnts[i+1][0][2]); 
							
                            glNormal3f(mo2->upnts[i][0][0],		mo2->upnts[i][0][1],	mo2->upnts[i][0][2]); 
                            glVertex3d(mo2->tpnts[i][0][0],		mo2->tpnts[i][0][1],	mo2->tpnts[i][0][2]); 
						}
					}
					else {
						if((j<mo2->TORUS_STPS-1)) {
							glNormal3f( mo2->upnts[i][j][0],    mo2->upnts[i][j][1],    mo2->upnts[i][j][2] );
                            glVertex3d(mo2->tpnts[i][j][0],		mo2->tpnts[i][j][1],	mo2->tpnts[i][j][2]); 
							
                            glNormal3f(mo2->upnts[0][j][0],		mo2->upnts[0][j][1],	mo2->upnts[0][j][2]); 
                            glVertex3d(mo2->tpnts[0][j][0],		mo2->tpnts[0][j][1],	mo2->tpnts[0][j][2]); 
							
                            glNormal3f(mo2->upnts[0][j+1][0],	mo2->upnts[0][j+1][1],	mo2->upnts[0][j+1][2]); 
                            glVertex3d(mo2->tpnts[0][j+1][0],	mo2->tpnts[0][j+1][1],	mo2->tpnts[0][j+1][2]); 
							
                            glNormal3f(mo2->upnts[i][j+1][0],	mo2->upnts[i][j+1][1],	mo2->upnts[i][j+1][2]); 
                            glVertex3d(mo2->tpnts[i][j+1][0],	mo2->tpnts[i][j+1][1],	mo2->tpnts[i][j+1][2]); 
						}
						else {
							glNormal3f( mo2->upnts[i][j][0],    mo2->upnts[i][j][1],    mo2->upnts[i][j][2] );
                            glVertex3d(mo2->tpnts[i][j][0],		mo2->tpnts[i][j][1],	mo2->tpnts[i][j][2]); 
							
                            glNormal3f(mo2->upnts[0][j][0],		mo2->upnts[0][j][1],	mo2->upnts[0][j][2]); 
                            glVertex3d(mo2->tpnts[0][j][0],		mo2->tpnts[0][j][1],	mo2->tpnts[0][j][2]); 
							
                            glNormal3f(mo2->upnts[0][0][0],		mo2->upnts[0][0][1],	mo2->upnts[0][0][2]); 
                            glVertex3d(mo2->tpnts[0][0][0],		mo2->tpnts[0][0][1],	mo2->tpnts[0][0][2]); 
							
                            glNormal3f(mo2->upnts[i][0][0],		mo2->upnts[i][0][1],	mo2->upnts[i][0][2]); 
                            glVertex3d(mo2->tpnts[i][0][0],		mo2->tpnts[i][0][1],	mo2->tpnts[i][0][2]); 
						}
					}
				glEnd();
			}//$ FOR(j..)
		}//$ FOR(i..)
		glPopAttrib();
	}//$ IF(showFace..)

    ///////////////////////// NORMALS /////////////////////////
	if(mo2->showNormals) {
		//glPushAttrib(GL_CURRENT_BIT | GL_LIGHTING_BIT | GL_ENABLE_BIT | GL_LINE_BIT);
		//glDisable(GL_LIGHTING);
		//
		//glLineWidth(1.5f);
		//
		//i=0;
		//j=0;
		//for(i=0; i<(maxPnts); i++) {
		//	for(j=0; j<(maxStps); j++) {

		//	}//$ FOR(j..)
		//}//$ FOR(i..)
		//glPopAttrib();

	}
}
//*********************************************//
//
//*********************************************//
void renderSphereSurface 						( void )
{
    int		i		= 0;
	int		j		= 0;
	//double	v1[3],v2[3];
	//double	len		= 0.0;
	//double	scale	= 0.10;
	//double	maxPnts = mo2->TORUS_PNTS;
	//double	maxStps	= mo2->TORUS_STPS;
	//double	maxPnts = t_pnts;
	//double	maxStps = t_stps;
	double	maxPnts = mo2->u;
	double	maxStps = mo2->v;

	assert(maxPnts<=mo2->TORUS_PNTS);
	assert(maxStps<=mo2->TORUS_STPS);

    ///////////////////////// RING /////////////////////////
    if(mo2->showEdge) {
        glPushAttrib(GL_CURRENT_BIT | GL_LIGHTING_BIT | GL_ENABLE_BIT | GL_LINE_BIT);
        glDisable(GL_LIGHTING);
        glLineWidth(2.5f);
        glColor3f(0.2f,0.2f,0.2f);
        glBegin(GL_LINES);
        for(i=0; i<maxPnts; i++) {
            for(j=0; j<maxStps; j++) {
                glVertex3d(mo2->tpnts[i][j][0],mo2->tpnts[i][j][1],mo2->tpnts[i][j][2]);
                if(j<(mo2->TORUS_STPS-1))
                    glVertex3d(mo2->tpnts[i][j+1][0],mo2->tpnts[i][j+1][1],mo2->tpnts[i][j+1][2]);
                else
                    glVertex3d(mo2->tpnts[i][0][0],mo2->tpnts[i][0][1],mo2->tpnts[i][0][2]);
            }
        }
        glEnd();
		glPopAttrib();

        ///////////////////////// LONGITUDINAL LINES /////////////////////////
        glPushAttrib(GL_CURRENT_BIT | GL_LIGHTING_BIT | GL_ENABLE_BIT | GL_LINE_BIT);
        glDisable(GL_LIGHTING);
        glLineWidth(2.5f);
		glBegin(GL_LINES);
        glColor3f(0.2f,0.2f,0.2f);
        for(i=1;i<maxPnts-1; i++) {
            for(j=0;j<maxStps;j++) {
                glVertex3d(mo2->tpnts[i][j][0],		mo2->tpnts[i][j][1],	mo2->tpnts[i][j][2]); 
                glVertex3d(mo2->tpnts[i-1][j][0],	mo2->tpnts[i-1][j][1],	mo2->tpnts[i-1][j][2]);
                if( i==(mo2->TORUS_PNTS-1) ){
                    glVertex3d(mo2->tpnts[i][j][0],mo2->tpnts[i][j][1],mo2->tpnts[i][j][2]); 
                    glVertex3d(mo2->tpnts[0][j][0],mo2->tpnts[0][j][1],mo2->tpnts[0][j][2]);
                }
            }
        }
        glEnd();
        glPopAttrib();
	}//$ IF(showEdge..)

    ///////////////////////// QUAD FACES /////////////////////////
	if(mo2->showFace) {
		glPushAttrib(GL_CURRENT_BIT | GL_LIGHTING_BIT | GL_ENABLE_BIT | GL_LINE_BIT);
		glDisable(GL_LIGHTING);
		
		glLineWidth(0.1f);
		
		i=0;
		j=0;
		for(i=0; i<(maxPnts); i++) {
			for(j=0; j<(maxStps); j++) {
				glEnable(GL_LIGHTING);
				glEnable(GL_COLOR_MATERIAL);
				glColor3f(mo2->fgcolor[0],mo2->fgcolor[1],mo2->fgcolor[2]);
				glBegin(GL_QUADS);
					//glNormal3f( n[0], n[1], n[2] );
					glNormal3f( mo2->upnts[i][j][0], mo2->upnts[i][j][1], mo2->upnts[i][j][2] );
					if(i<(mo2->TORUS_PNTS-1)) {
						if((j<mo2->TORUS_STPS-1)) {
							glVertex3d(mo2->tpnts[i][j][0],		mo2->tpnts[i][j][1],	mo2->tpnts[i][j][2]); 
							glVertex3d(mo2->tpnts[i+1][j][0],	mo2->tpnts[i+1][j][1],	mo2->tpnts[i+1][j][2]); 
							glVertex3d(mo2->tpnts[i+1][j+1][0],	mo2->tpnts[i+1][j+1][1],mo2->tpnts[i+1][j+1][2]); 
							glVertex3d(mo2->tpnts[i][j+1][0],	mo2->tpnts[i][j+1][1],	mo2->tpnts[i][j+1][2]); 
						}
						else {
							glVertex3d(mo2->tpnts[i][j][0],		mo2->tpnts[i][j][1],	mo2->tpnts[i][j][2]); 
							glVertex3d(mo2->tpnts[i+1][j][0],	mo2->tpnts[i+1][j][1],	mo2->tpnts[i+1][j][2]); 
							glVertex3d(mo2->tpnts[i+1][0][0],	mo2->tpnts[i+1][0][1],	mo2->tpnts[i+1][0][2]); 
							glVertex3d(mo2->tpnts[i][0][0],		mo2->tpnts[i][0][1],	mo2->tpnts[i][0][2]); 
						}
					}
					else {
						if((j<mo2->TORUS_STPS-1)) {
							glVertex3d(mo2->tpnts[i][j][0],		mo2->tpnts[i][j][1],	mo2->tpnts[i][j][2]); 
							glVertex3d(mo2->tpnts[0][j][0],		mo2->tpnts[0][j][1],	mo2->tpnts[0][j][2]); 
							glVertex3d(mo2->tpnts[0][j+1][0],	mo2->tpnts[0][j+1][1],	mo2->tpnts[0][j+1][2]); 
							glVertex3d(mo2->tpnts[i][j+1][0],	mo2->tpnts[i][j+1][1],	mo2->tpnts[i][j+1][2]); 
						}
						else {
							glVertex3d(mo2->tpnts[i][j][0],		mo2->tpnts[i][j][1],	mo2->tpnts[i][j][2]); 
							glVertex3d(mo2->tpnts[0][j][0],		mo2->tpnts[0][j][1],	mo2->tpnts[0][j][2]); 
							glVertex3d(mo2->tpnts[0][0][0],		mo2->tpnts[0][0][1],	mo2->tpnts[0][0][2]); 
							glVertex3d(mo2->tpnts[i][0][0],		mo2->tpnts[i][0][1],	mo2->tpnts[i][0][2]); 
						}
					}
				glEnd();
			}//$ FOR(j..)
		}//$ FOR(i..)
		glPopAttrib();
	}//$ IF(showFace..)

 //   ///////////////////////// NORMALS /////////////////////////
	//if(mo2->showNormals) {
	//	glPushAttrib(GL_CURRENT_BIT | GL_LIGHTING_BIT | GL_ENABLE_BIT | GL_LINE_BIT);
	//	glDisable(GL_LIGHTING);
	//	
	//	glLineWidth(1.5f);
	//	
	//	i=0;
	//	j=0;
	//	for(i=0; i<(maxPnts); i++) {
	//		for(j=0; j<(maxStps); j++) {
	//			//if(i<(mo2->TORUS_PNTS-1) ) {
	//			//	v1[0] = mo2->tpnts[i+1][j][0] - mo2->tpnts[i][j][0];
	//			//	v1[1] = mo2->tpnts[i+1][j][1] - mo2->tpnts[i][j][1];
	//			//	v1[2] = mo2->tpnts[i+1][j][2] - mo2->tpnts[i][j][2];
	//			//}
	//			//else {
	//			//	v1[0] = mo2->tpnts[0][j][0] - mo2->tpnts[i][j][0];
	//			//	v1[1] = mo2->tpnts[0][j][1] - mo2->tpnts[i][j][1];
	//			//	v1[2] = mo2->tpnts[0][j][2] - mo2->tpnts[i][j][2];
	//			//}
	//			//if(j<(mo2->TORUS_STPS-1)) {
	//			//	v2[0] = mo2->tpnts[i][j+1][0] - mo2->tpnts[i][j][0];
	//			//	v2[1] = mo2->tpnts[i][j+1][1] - mo2->tpnts[i][j][1];
	//			//	v2[2] = mo2->tpnts[i][j+1][2] - mo2->tpnts[i][j][2];
	//			//}
	//			//else {
	//			//	v2[0] = mo2->tpnts[i][0][0] - mo2->tpnts[i][j][0];
	//			//	v2[1] = mo2->tpnts[i][0][1] - mo2->tpnts[i][j][1];
	//			//	v2[2] = mo2->tpnts[i][0][2] - mo2->tpnts[i][j][2];
	//			//}

	//			//len = length3D( v1 );
	//			//v1[0] /= len; 
	//			//v1[1] /= len; 
	//			//v1[2] /= len; 

	//			//len = length3D( v2 );
	//			//v2[0] /= len; 
	//			//v2[1] /= len; 
	//			//v2[2] /= len; 

	//		//glColor3f(0,0,1);
	//		//glPushMatrix();
	//		//	glTranslatef(mo2->tpnts[i][j][0],		mo2->tpnts[i][j][1],	mo2->tpnts[i][j][2]); 
	//		//	glScalef( scale, scale, scale );
	//		//	glBegin(GL_LINES);
	//		//	glVertex3f(0,0,0);
	//		//	glVertex3f(mo2->upnts[i][j][0],		mo2->upnts[i][j][1],	mo2->upnts[i][j][2]);
	//		//	glEnd();
	//		//glPopMatrix();
	//		//glColor3f(1,0,0);
	//		//glPushMatrix();
	//		//	glTranslatef(mo2->tpnts[i][j][0],		mo2->tpnts[i][j][1],	mo2->tpnts[i][j][2]); 
	//		//	glScalef( scale, scale, scale );
	//		//	glBegin(GL_LINES);
	//		//	glVertex3f(0,0,0);
	//		//	glVertex3f(mo2->grRBasis[i].p3.d[0], mo2->grRBasis[i].p3.d[1], mo2->grRBasis[i].p3.d[2]);
	//		//	glEnd();
	//		//glPopMatrix();
	//		//glColor3f(0,1,0);
	//		//glPushMatrix();
	//		//	glTranslatef(mo2->tpnts[i][j][0],		mo2->tpnts[i][j][1],	mo2->tpnts[i][j][2]); 
	//		//	glScalef( scale, scale, scale );
	//		//	glBegin(GL_LINES);
	//		//	glVertex3f(0,0,0);
	//		//	glVertex3f(mo2->grSBasis[i].p3.d[0], mo2->grSBasis[i].p3.d[1], mo2->grSBasis[i].p3.d[2]);
	//		//	glEnd();
	//		//glPopMatrix();
	//		
	//		//glColor3f(1,0,0);
	//			//glPushMatrix();
	//			//	glTranslatef(mo2->tpnts[i][j][0],		mo2->tpnts[i][j][1],	mo2->tpnts[i][j][2]); 
	//			//	glScalef( scale, scale, scale );
	//			//	glBegin(GL_LINES);
	//			//	glVertex3f(0,0,0);
	//			//	glVertex3f(mo2->grRBasis[i].p3.d[0], mo2->grRBasis[i].p3.d[1], mo2->grRBasis[i].p3.d[2]);
	//			//	glEnd();
	//			//glPopMatrix();

	//			//glColor3f(0,1,0);
	//			//glPushMatrix();
	//			//	glTranslatef(mo2->tpnts[i][j][0],		mo2->tpnts[i][j][1],	mo2->tpnts[i][j][2]); 
	//			//	glScalef( scale, scale, scale );
	//			//	glBegin(GL_LINES);
	//			//	glVertex3f(0,0,0);
	//			//	glVertex3f(mo2->grSBasis[i].p3.d[0], mo2->grSBasis[i].p3.d[1], mo2->grSBasis[i].p3.d[2]);
	//			//	glEnd();
	//			//glPopMatrix();

	//			//glColor3f(0,0,1);
	//			//glPushMatrix();
	//			//	glTranslatef(mo2->tpnts[i][j][0],		mo2->tpnts[i][j][1],	mo2->tpnts[i][j][2]); 
	//			//	glScalef( scale, scale, scale );
	//			//	glBegin(GL_LINES);
	//			//	glVertex3f(0,0,0);
	//			//	glVertex3f(mo2->upnts[i][j][0],		mo2->upnts[i][j][1],	mo2->upnts[i][j][2]);
	//			//	glEnd();
	//			//glPopMatrix();
	//		}//$ FOR(j..)
	//	}//$ FOR(i..)
	//	glPopAttrib();

	//}
}
//*********************************************//
//
//*********************************************//
void computeGeodesicReferenceFrame_MANIFOLD		( void )
{
	int		i		= 0;				// index into points array
	int		j		= 0;				// index into points array
	double	z[3]	= {0.0, 1.0, 0.0};	// z-hat: reference axis to be tilted into tangent/normals
	double	v[3];						// vector for Tangent
	double	n[3];						// vector for Normal
	double	s[3];						// vector for third	orthonormal	component
	double	cross_ZV[3];				// vector for tangent-tangent cross	product
	double	theta	= 0.0;				// PT initial frame	angle
	double	len		= 0.0;				// vector norm (length)
	quaternion *q;						// quaternion variables
	//double	maxPnts = mo2->TORUS_PNTS;
	//double	maxStps	= mo2->TORUS_STPS;
	//double	maxPnts = t_pnts;
	//double	maxStps = t_stps;
	double	maxPnts = mo2->u;
	double	maxStps = mo2->v;
	double	scale	= 0.1;

	assert(maxPnts<=mo2->TORUS_PNTS);
	assert(maxStps<=mo2->TORUS_STPS);

	//initialize
	q =	(quaternion*) malloc(sizeof(quaternion));

	//@ Compute GEODESIC-REFERENCE FRAME for all points ////////////
	glPushAttrib(GL_LIGHTING_BIT);
	glDisable(GL_LIGHTING);
	glLineWidth(3.0f);
	for(i=0; i<maxPnts; i++) {
		for(j=0; j<maxStps; j++) {
		
			v[0] = mo2->upnts[i][j][0];
			v[1] = mo2->upnts[i][j][1];
			v[2] = mo2->upnts[i][j][2];

			//#NOTE: have to handle the case when v-hat = -(z-hat)
			//obtain axis 'cross_ZV' for rotation of reference-frame 'z-hat'
			cross3D( z, v, cross_ZV );	
			theta = acos( dot3D( z, v ) );

			//normalize cross_ZV
			len = length3D( cross_ZV );
			cross_ZV[0] /= len; 
			cross_ZV[1] /= len; 
			cross_ZV[2] /= len; 

			cross3D( cross_ZV, v, n );
			len = length3D( n );
			n[0] /= len;
			n[1] /= len;
			n[2] /= len;

			mo2->grRBasis[i].p3.d[0] = n[0];
			mo2->grRBasis[i].p3.d[1] = n[1];
			mo2->grRBasis[i].p3.d[2] = n[2];

			cross3D( v, n, s );
			len = length3D( s );
			s[0] /= len;
			s[1] /= len;
			s[2] /= len;

			mo2->grSBasis[i].p3.d[0] = s[0];
			mo2->grSBasis[i].p3.d[1] = s[1];
			mo2->grSBasis[i].p3.d[2] = s[2];

			q->r =			   cos( theta/2.0 );
			q->i = cross_ZV[0]*sin( theta/2.0 );
			q->j = cross_ZV[1]*sin( theta/2.0 );
			q->k = cross_ZV[2]*sin( theta/2.0 );
			quatNormalize(q);

			//asign quaternion
			quatCopy( q, (mo2->surfGRQuatList[i][j]) );

			if(mo2->showFrame) {
				glColor3f(0,0,1);
				glPushMatrix();
					glTranslatef(mo2->tpnts[i][j][0],		mo2->tpnts[i][j][1],	mo2->tpnts[i][j][2]); 
					glScalef( scale, scale, scale );
					glBegin(GL_LINES);
					glVertex3f(0,0,0);
					glVertex3f(v[0], v[1], v[2]);
					glEnd();
				glPopMatrix();
				glColor3f(0,1,0);
				glPushMatrix();
					glTranslatef(mo2->tpnts[i][j][0],		mo2->tpnts[i][j][1],	mo2->tpnts[i][j][2]); 
					glScalef( scale, scale, scale );
					glBegin(GL_LINES);
					glVertex3f(0,0,0);
					glVertex3f(mo2->grRBasis[i].p3.d[0], mo2->grRBasis[i].p3.d[1], mo2->grRBasis[i].p3.d[2]);
					glEnd();
				glPopMatrix();
				glColor3f(1,0,0);
				glPushMatrix();
					glTranslatef(mo2->tpnts[i][j][0],		mo2->tpnts[i][j][1],	mo2->tpnts[i][j][2]); 
					glScalef( scale, scale, scale );
					glBegin(GL_LINES);
					glVertex3f(0,0,0);
					glVertex3f(mo2->grSBasis[i].p3.d[0], mo2->grSBasis[i].p3.d[1], mo2->grSBasis[i].p3.d[2]);
					glEnd();
				glPopMatrix();
			}
		}//$ FOR(j..)
	}//$ FOR(i..)
	glPopAttrib();
	my_free(q);
	//printf("Quaternions for GR frame:\n");
	//for(i=0; i<maxPnts; i++) {
	//	for(j=0; j<maxStps; j++) {
	//		quatPrint(mo2->surfGRQuatList[i][j]);
	//		printf("\n");
	//	}
	//	printf("\n");
	//}
	
}
//*********************************************//
//
//*********************************************//
void renderGaussMap 							( void )
{
    int		i		= 0;
	int		j		= 0;
	//double	maxPnts = t_pnts;
	//double	maxStps = t_stps;
	double	maxPnts = mo2->u;
	double	maxStps = mo2->v;
	double	n[3];
	double	scale	= 0.15;

    ///////////////////////// QUAD FACES /////////////////////////
	if(mo2->qshowFace) {
		glPushAttrib(GL_CURRENT_BIT | GL_LIGHTING_BIT | GL_ENABLE_BIT | GL_LINE_BIT);
		glEnable(GL_LIGHTING);
		glEnable(GL_COLOR_MATERIAL);
		glColor3f(0.55f,0.0f,0.3f);		
		glLineWidth(1.0f);

		for(i=0; i<(maxPnts); i++) {
			for(j=0; j<(maxStps); j++) {

				glBegin(GL_QUADS);
					n[0] = mo2->upnts[i][j][0] - 0.0;
					n[1] = mo2->upnts[i][j][1] - 0.0;
					n[2] = mo2->upnts[i][j][2] - 0.0;
					glNormal3f( n[0], n[1], n[2] );
					if(i<(mo2->TORUS_PNTS-1)) {
						if((j<mo2->TORUS_STPS-1)) {
							glVertex3d(mo2->upnts[i][j][0],		mo2->upnts[i][j][1],	mo2->upnts[i][j][2]); 
							glVertex3d(mo2->upnts[i+1][j][0],	mo2->upnts[i+1][j][1],	mo2->upnts[i+1][j][2]); 
							glVertex3d(mo2->upnts[i+1][j+1][0],	mo2->upnts[i+1][j+1][1],mo2->upnts[i+1][j+1][2]); 
							glVertex3d(mo2->upnts[i][j+1][0],	mo2->upnts[i][j+1][1],	mo2->upnts[i][j+1][2]); 
						}
						else {
							glVertex3d(mo2->upnts[i][j][0],		mo2->upnts[i][j][1],	mo2->upnts[i][j][2]); 
							glVertex3d(mo2->upnts[i+1][j][0],	mo2->upnts[i+1][j][1],	mo2->upnts[i+1][j][2]); 
							glVertex3d(mo2->upnts[i+1][0][0],	mo2->upnts[i+1][0][1],	mo2->upnts[i+1][0][2]); 
							glVertex3d(mo2->upnts[i][0][0],		mo2->upnts[i][0][1],	mo2->upnts[i][0][2]); 
						}
					}
					else {
						if((j<mo2->TORUS_STPS-1)) {
							glVertex3d(mo2->upnts[i][j][0],		mo2->upnts[i][j][1],	mo2->upnts[i][j][2]); 
							glVertex3d(mo2->upnts[0][j][0],		mo2->upnts[0][j][1],	mo2->upnts[0][j][2]); 
							glVertex3d(mo2->upnts[0][j+1][0],	mo2->upnts[0][j+1][1],	mo2->upnts[0][j+1][2]); 
							glVertex3d(mo2->upnts[i][j+1][0],	mo2->upnts[i][j+1][1],	mo2->upnts[i][j+1][2]); 
						}
						else {
							glVertex3d(mo2->upnts[i][j][0],		mo2->upnts[i][j][1],	mo2->upnts[i][j][2]); 
							glVertex3d(mo2->upnts[0][j][0],		mo2->upnts[0][j][1],	mo2->upnts[0][j][2]); 
							glVertex3d(mo2->upnts[0][0][0],		mo2->upnts[0][0][1],	mo2->upnts[0][0][2]); 
							glVertex3d(mo2->upnts[i][0][0],		mo2->upnts[i][0][1],	mo2->upnts[i][0][2]); 
						}
					}
				glEnd();
			}//$ FOR(j..)
		}//$ FOR(i..)
		glPopAttrib();
	}//$ IF(showFace..)

    ///////////////////////// QUAD EDGES /////////////////////////
    if(mo2->qshowEdge) {
        glPushAttrib(GL_CURRENT_BIT | GL_LIGHTING_BIT | GL_ENABLE_BIT | GL_LINE_BIT);
        glDisable(GL_LIGHTING);
        glLineWidth(2.5f);
        glColor3f(0.2f,0.2f,0.2f);
        glBegin(GL_LINES);
        for(i=0; i<maxPnts; i++) {
            for(j=0; j<maxStps; j++) {
                glVertex3d(mo2->upnts[i][j][0],mo2->upnts[i][j][1],mo2->upnts[i][j][2]);
                if(j<(mo2->TORUS_STPS-1))
                    glVertex3d(mo2->upnts[i][j+1][0],mo2->upnts[i][j+1][1],mo2->upnts[i][j+1][2]);
                else
                    glVertex3d(mo2->upnts[i][0][0],mo2->upnts[i][0][1],mo2->upnts[i][0][2]);
            }
        }

        glEnd();
		glPopAttrib();

        ///////////////////////// LONGITUDINAL LINES /////////////////////////
        glPushAttrib(GL_CURRENT_BIT | GL_LIGHTING_BIT | GL_ENABLE_BIT | GL_LINE_BIT);
        glDisable(GL_LIGHTING);
        glLineWidth(2.5f);
		glBegin(GL_LINES);
        glColor3f(0.2f,0.2f,0.2f);
		if(mo2->surfObjectID==1)
			maxPnts = maxPnts-1;
        for(i=1;i<maxPnts; i++) {
            for(j=0;j<maxStps;j++) {
                glVertex3d(mo2->upnts[i][j][0],		mo2->upnts[i][j][1],	mo2->upnts[i][j][2]); 
                glVertex3d(mo2->upnts[i-1][j][0],	mo2->upnts[i-1][j][1],	mo2->upnts[i-1][j][2]);
                if( i==(mo2->TORUS_PNTS-1) ){
                    glVertex3d(mo2->upnts[i][j][0],mo2->upnts[i][j][1],mo2->upnts[i][j][2]); 
                    glVertex3d(mo2->upnts[0][j][0],mo2->upnts[0][j][1],mo2->upnts[0][j][2]);
                }
            }
        }
        glEnd();
        glPopAttrib();
	}//$ IF(showEdge..)

    ///////////////////////// NORMALS /////////////////////////
	if(mo2->qshowNormals) {
	glPushAttrib(GL_CURRENT_BIT | GL_LIGHTING_BIT | GL_ENABLE_BIT | GL_LINE_BIT);
	glDisable(GL_LIGHTING);
	
	glLineWidth(3.0f);
	
	i=0;
	j=0;
	for(i=0; i<maxPnts; i++) {
		for(j=0; j<(maxStps); j++) {
			n[0] = mo2->upnts[i][j][0] - 0.0;
			n[1] = mo2->upnts[i][j][1] - 0.0;
			n[2] = mo2->upnts[i][j][2] - 0.0;
			glNormal3f( n[0], n[1], n[2] );
		
			glColor3f(0,0,1);
			glPushMatrix();
				glTranslatef(mo2->upnts[i][j][0], mo2->upnts[i][j][1], mo2->upnts[i][j][2]); 
				glScalef( scale, scale, scale );
				glBegin(GL_LINES);
				glVertex3f(0,0,0);
				glVertex3f(n[0], n[1], n[2]);
				glEnd();
			glPopMatrix();

		}//$ FOR(j..)
	}//$ FOR(i..)
	glPopAttrib();

}

}

//*********************************************//
//
//*********************************************//
void drawQuaternionCurve_SURFACE				( void ) 
{
	int		i		= 0;
	int		j		= 0;
	int		limitPnts=0;
	int		limitSpts=0;
	//double	maxPnts = t_pnts;
	//double	maxStps = t_stps;
	double	maxPnts = mo2->u;
	double	maxStps = mo2->v;
	double	qvec[3];
	float	qfgcolor[3];

	assert(maxPnts<=mo2->TORUS_PNTS);
	assert(maxStps<=mo2->TORUS_STPS);
	glPushAttrib(GL_ALL_ATTRIB_BITS);
	glEnable(GL_LIGHTING);
	glEnable(GL_COLOR_MATERIAL);
	
    qfgcolor[0] = 0.5f; qfgcolor[1] = 0.0f; qfgcolor[2] = 0.0f;
	glColor3fv( qfgcolor );

	for(i=0; i<maxPnts; i++) {
		for(j=0; j<maxStps; j++) {
			getQuatComp( mo2->qwhichQuatComponent, mo2->surfGRQuatList[i][j], qvec );
			glPushMatrix();
				glTranslatef( qvec[0], qvec[1], qvec[2] );
				glutSolidSphere(0.02,16,16);
			glPopMatrix();
		}
	}

	if(mo2->surfObjectID==1 || mo2->surfObjectID==2) {
		limitPnts = maxPnts-1;
		limitSpts = maxStps+1;
	} 
	else {
		limitPnts = maxPnts;
		limitSpts = maxStps;
	}
	for(i=0; i<limitPnts; i++) {
		glBegin(GL_LINE_STRIP);
		for(j=0; j<limitSpts; j++) {
			if(j<(limitSpts-1)) {
				getQuatComp( mo2->qwhichQuatComponent, mo2->surfGRQuatList[i][j], qvec );
			}
			else
				getQuatComp( mo2->qwhichQuatComponent, mo2->surfGRQuatList[i][0], qvec );

			//glPushMatrix();
				glVertex3f( qvec[0], qvec[1], qvec[2] );
			//	glutSolidSphere(0.02,16,16);
			//glPopMatrix();
		}
		glEnd();
	}
}

//### COMPUTATION: Curve Propperties
void computeProperties							( int index )
{
	int			i = index;
	double		h = 0.0;
	quaternion	*q0, *q1,*q2; 
	quaternion	*q3;
	quaternion	*b;
	quaternion	*q;
	quaternion	*r;
    double s = 0;

	b =	(quaternion*) my_malloc(sizeof(quaternion));
	q =	(quaternion*) my_malloc(sizeof(quaternion));
	r =	(quaternion*) my_malloc(sizeof(quaternion));
	q0=	(quaternion*) my_malloc(sizeof(quaternion));	
	q1=	(quaternion*) my_malloc(sizeof(quaternion));	
	q2=	(quaternion*) my_malloc(sizeof(quaternion));	
	q3=	(quaternion*) my_malloc(sizeof(quaternion));	

	mo2->gmax_curve_qdist	 = -1.0;
	mo2->gmax_curve_qdotprod = -1.0;
	mo2->gmax_curve_qtangacc = -1.0;

	//get previous quaternion
	quatCopy(&(mo2->opQuatList[i-1]),q0);

	//get next quaternion
	quatCopy(&(mo2->opQuatList[i+1]),q2);

	//get current quaternion
	quatCopy(&(mo2->opQuatList[i]),q1);
	
	//@ TURNING ANGLE experienced by 3D-Frames
	mo2->mtaENERGY = 2.0*acos( quatQuatDotProduct(q0,q1) );
	//printf("Energy[%d] = %f\n",i,mo2->mtaENERGY);
	mo2->list_mtaENERGY_angle[i] = mo2->mtaENERGY;
	if(mo2->list_mtaENERGY_angle[i] > mo2->gmax_curve_qdist)
		mo2->gmax_curve_qdist = mo2->list_mtaENERGY_angle[i];
	if(mo2->showColorCoded == CODE_QUAT_TURN_ANGLE) {
		s = 255.99/(2.0*mo2->scale);
		if(mo2->list_mtaENERGY_angle[i]>0.0) {
			mo2->pntList[i].color = (int) ((mo2->list_mtaENERGY_angle[i]*1.0)*s);
		}
		else
			mo2->pntList[i].color = (int) ((-mo2->list_mtaENERGY_angle[i]*1.0)*s);
		if(mo2->pntList[i].color > 255)
			mo2->pntList[i].color = 255;
	}	

    //@ INNER DOT PRODUCT Quat-Quat 
	mo2->mtaENERGY = ( quatQuatDotProduct(q0,q1) );
	//printf("Energy[%d] = %f\n",i,mo2->mtaENERGY);
    mo2->list_mtaENERGY_dist[i] = mo2->mtaENERGY;
	if(mo2->list_mtaENERGY_dist[i] > mo2->gmax_curve_qdotprod)
		mo2->gmax_curve_qdotprod = mo2->list_mtaENERGY_dist[i];

	if(mo2->showColorCoded == CODE_QUAT_DOTPROD) {
		s = 255.99/(2.0*mo2->scale);
		if(mo2->list_mtaENERGY_dist[i]>0.0) {
			mo2->pntList[i].color = (int) ((mo2->list_mtaENERGY_dist[i]*1.0)*s);
		}
		else
			mo2->pntList[i].color = (int) ((-mo2->list_mtaENERGY_dist[i]*1.0)*s);
		if(mo2->pntList[i].color > 255)
			mo2->pntList[i].color = 255;
	}	


	//@ TANGENTIAL ACCELERATION Experienced by Quaternion-Curve
	h = 0.00001;
	calc_num_SecondOrder_diff_3POINT(q0,q1,q2,q3,h);
	calc_TangAcceleration(q3,q1,h,&(mo2->mtaENERGY));
	//printf("Energy[%d] = %f\n",i,mo2->mtaENERGY);
	mo2->list_mtaENERGY_tacc[i] = mo2->mtaENERGY;
	if(mo2->list_mtaENERGY_tacc[i] > mo2->gmax_curve_qtangacc)
		mo2->gmax_curve_qtangacc = mo2->list_mtaENERGY_tacc[i];

	//### DONE
	my_free(b);
	my_free(q);
	my_free(r);
	my_free(q0);
	my_free(q1);
	my_free(q2);
}

//### COMPUTATION: Optimization

void initOPTIM_STRUCT							( void )
{
	//int i = 0;
	//for(i=0; i<MAX_POINTS; i++) {
	//	mo2->opQuatList[i].r = 0.0;
 //       mo2->opQuatList[i].i = 0.0;
 //       mo2->opQuatList[i].j = 0.0;
 //       mo2->opQuatList[i].k = 0.0;
	//}
	//mo2->opt_INDEX = 0;
}

void optCalcPTFrame_TurningAngle				( void )
{
	//@ CALCULATE ENERGY FOR PT FRAMES
	int tmp_whichFrame = mo2->whichFrame;

	mo2->whichFrame = PARALLELTRANSPORT;

	//compute Parallel Transport Quaternion Map
	computeParallelTransportFrame();
	copyQuatToPointList( PARALLELTRANSPORT );
	makeTangent_forQuatMap();
	makeTangentDerivatives_forQuatMap();
	computeParallelTransportFrame_forQuatMap();

	//copy quatlist into opQuatList[]
	copyQuatList_OPT();

	//compute sum-total of Angles
	optQuatCurveEnergy();
	mo2->curveQuatAngle_PT = mo2->curveQuatAngle;

	//mo2->whichFrame = FRENETSERRET;
	//reset opQuatList[] for current frame
	copyQuatList_OPT();

	//compute sum-total of Angles
	optQuatCurveEnergy();

	//reset frame to original
	mo2->whichFrame = tmp_whichFrame;
}

void initOPTIM_TurningAngle						( void )
{
	//@ CALCULATE ENERGY FOR PT FRAMES
	//int tmp_whichFrame = mo2->whichFrame;

	//mo2->whichFrame = PARALLELTRANSPORT;

	////compute Parallel Transport Quaternion Map
	//computeParallelTransportFrame();
	//copyQuatToPointList( PARALLELTRANSPORT );
	//makeTangent_forQuatMap();
	//makeTangentDerivatives_forQuatMap();
	//computeParallelTransportFrame_forQuatMap();

	//copy quatlist into opQuatList[]
	//int i = 0;
	//copyQuatList_OPT();
	//for(i=0; i<mo2->numPoints; i++) {

	//	mo2->list_ENERGY_turnang[i] = 1000.0;
	//}
	////compute sum-total of Angles
	//optQuatCurveEnergy();
	//mo2->curveQuatAngle_PT = mo2->curveQuatAngle;

	////reset frame to original
	//mo2->whichFrame = tmp_whichFrame;
	//
	////reset opQuatList[] for current frame
	////copyQuatList_OPT();

	////set optimization index
	//mo2->opt_INDEX = 0;
	mo2->minTurningAngle = 1000.0;
	//printf("PT Global Frame Angle-sum = %f\n",mo2->curveQuatAngle_PT);
}

void optMinTurningAngle							( int index ) 
{
	int		i = 0;
    int     l_index = index;
    l_index = 0;
	double	qvec[3];
	glPushAttrib(GL_LIGHTING_BIT);
	glEnable(GL_LIGHTING);
	glColor3f(0.0,1.0,0.8);
	for(i=0; i<mo2->numPoints-1; i++) {
		
		glPushMatrix();
			getQuatComp(mo2->qwhichQuatComponent,&(mo2->opQuatList[i]),qvec);
			glTranslatef(qvec[0],qvec[1],qvec[2]);
			glutSolidSphere(0.02,16,16);
		glPopMatrix();
	
	}
	//glColor3f(1.0,0.0,0.8);
	//glPushMatrix();
	//	getQuatComp(mo2->qwhichQuatComponent,&(mo2->currQuat),qvec);
	//	glTranslatef(qvec[0],qvec[1],qvec[2]);
	//	glutSolidSphere(0.03,16,16);
	//glPopMatrix();
	//glColor3f(1.0,0.0,0.5);
	//glPushMatrix();
	//	getQuatComp(mo2->qwhichQuatComponent,&(mo2->prevQuat),qvec);
	//	glTranslatef(qvec[0],qvec[1],qvec[2]);
	//	glutSolidSphere(0.03,16,16);
	//glPopMatrix();
	glPopAttrib();
}

void optCalcTurningAngle2						( int index ) 
{
	//obtain previous and current quaternion, 
	int			i = 0;
	//int			j = 0;
	int			steps = 100;
	double		delta = (1.0)/ ((double)steps);
	//double		dTheta = (4.0*M_PI)/((double)steps);
	double		theta = 0.0;
	double		quatDOT	= 0.0;
	double		quatTHETA=0.0;
	//double		qvec[3];
	double		tangent[3];
	double		crossZT[3];
	double		norm_crossZT= 0.0;
	double		z[3]		= {0.0, 0.0, 1.0};
	double		thetaZT		= 0.0;
	double		cangle, pangle;
	double		F_t0, F_t1,F_t2;
	double		update;
	double		theta1, theta2;
	double		Fdash = 0.0;
	int			cond=0;
	int			tmp_whichFrame = 0;
	double		initDist = 0.0;
	quaternion	*b;
	quaternion	*q;
	quaternion	*r;
	quaternion	*u;
	quaternion	*v;

	b =	(quaternion*) my_malloc(sizeof(quaternion));
	q =	(quaternion*) my_malloc(sizeof(quaternion));
	r =	(quaternion*) my_malloc(sizeof(quaternion));
	u =	(quaternion*) my_malloc(sizeof(quaternion));
	v =	(quaternion*) my_malloc(sizeof(quaternion));

	F_t0 = F_t1 = F_t2 = 0.0;

	//for(i=1; i<mo2->numPoints-1; i++) {
		i = index;

	while(cond<1) {
		//printf("cond = %d\n",cond);
		//printf("rplist[%d].p4  (actual)  = %f, %f, %f, %f\n",i-1,mo2->fsQuatList[i-1].r, mo2->fsQuatList[i-1].i, mo2->fsQuatList[i-1].j, mo2->fsQuatList[i-1].k);
		//printf("mo2->currQuat (computed)= %f, %f, %f, %f\n",s->r, s->i, s->j, s->k);
		tmp_whichFrame = mo2->whichFrame;
		mo2->whichFrame = OPTIMIZATION;
		findQuat_Current(i,u,&cangle);
		quatCopy(u,&(mo2->currQuat));
		findQuat_Current(i-1,v,&pangle);
		quatCopy(v,&(mo2->prevQuat));
		mo2->whichFrame = tmp_whichFrame;
		initDist = quatQuatDotProduct(u,v);
		//printf("[%d]\t initDist = %f\n",i,initDist);

		tangent[0] =	mo2->TangList[i].p3.d[0];
		tangent[1] =	mo2->TangList[i].p3.d[1];
		tangent[2] =	mo2->TangList[i].p3.d[2];

		////compute (z x tangent)
		cross3D( z, tangent, crossZT );
		norm_crossZT = length3D( crossZT );
		crossZT[0] /= norm_crossZT;
		crossZT[1] /= norm_crossZT;
		crossZT[2] /= norm_crossZT;

		//compute arccos(z . tangent)
		thetaZT = acos( dot3D( z, tangent ) );

		//compute b(vhat)
		b->r = cos( thetaZT/2.0 );
		b->i = crossZT[0]*sin( thetaZT/2.0 );
		b->j = crossZT[1]*sin( thetaZT/2.0 );
		b->k = crossZT[2]*sin( thetaZT/2.0 );

		//F_t0
			theta = cangle;
			//printf("\ttheta (t0) = %f\n",theta);
			q->r = cos( theta/2.0 );
			q->i = tangent[0]*sin( theta/2.0 );
			q->j = tangent[1]*sin( theta/2.0 );
			q->k = tangent[2]*sin( theta/2.0 );
			quatQuatProduct( q, b, r );
			quatConjugate(r);
			quatDOT = quatQuatDotProduct( r, v );
			quatTHETA = 2.0*acos( quatDOT );
			F_t0 = quatTHETA;
			//printf("\t F_t0 = %f\n",F_t0);

		//F_t1_1
			theta1 = cangle - delta;
			q->r = cos( theta1/2.0 );
			q->i = tangent[0]*sin( theta1/2.0 );
			q->j = tangent[1]*sin( theta1/2.0 );
			q->k = tangent[2]*sin( theta1/2.0 );
			quatQuatProduct( q, b, r );
			quatConjugate(r);
			quatDOT = quatQuatDotProduct( v, r );
			quatTHETA = 2.0*acos( quatDOT );
			//printf("#\tr (t1)= %f, %f, %f, %f\n",r->r,r->i,r->j,r->k);
			F_t1 = quatTHETA;
			//printf("\tF_t1 = %f\n",F_t1);

		//F_t2
			theta2 = cangle + delta;
			//printf("\ttheta (t2) = %f\n",theta2);
			q->r = cos( theta2/2.0 );
			q->i = tangent[0]*sin( theta2/2.0 );
			q->j = tangent[1]*sin( theta2/2.0 );
			q->k = tangent[2]*sin( theta2/2.0 );
			quatQuatProduct( q, b, r );
			quatConjugate(r);
			quatDOT = quatQuatDotProduct( v, r );
			quatTHETA = 2.0*acos( quatDOT );
			//printf("#\tr (t1)= %f, %f, %f, %f\n",r->r,r->i,r->j,r->k);
			F_t2 = quatTHETA;
			//printf("\tF_t2 = %f\n",F_t2);

		//Fdash_dash = ( F_t1 - 2.0*F_t0 + F_t2 )/(delta*delta);	//! this is 2nd derivative
		Fdash = ( F_t2 - F_t1 )/( 2.0 * delta );		// two-point formula			
		
		//1st derivative using 3 point rule
		//Fdash = ( F_t2 - F_t0 )/( delta );			//from NRecipes
		update = F_t0/Fdash;
	
		theta = cangle - update;
		q->r = cos( theta/2.0 );
		q->i = tangent[0]*sin( theta/2.0 );
		q->j = tangent[1]*sin( theta/2.0 );
		q->k = tangent[2]*sin( theta/2.0 );
		quatQuatProduct( q, b, r );
		quatConjugate(r);
		quatDOT = quatQuatDotProduct( v, r );
		quatTHETA = 2.0*acos( quatDOT );
		
		//printf("[%d]\tmo2->val = %f\n",i,mo2->list_ENERGY_turnang[i]);
		//printf("\t quattheta = %f\n",quatTHETA);

		//theta = cangle + update;
		//q->r = cos( theta/2.0 );
		//q->i = tangent[0]*sin( theta/2.0 );
		//q->j = tangent[1]*sin( theta/2.0 );
		//q->k = tangent[2]*sin( theta/2.0 );
		//quatQuatProduct( q, b, u );
		//quatConjugate(r);
		//quatDOT = quatQuatDotProduct( v, u );
		//quatTHETA = 2.0*acos( quatDOT );
		//printf("\t quattheta = %f\n",quatTHETA);

		//if(quatTHETA < mo2->list_ENERGY_turnang[i]) {
			mo2->list_ENERGY_turnang[i] = quatTHETA;
			//printf("\t# mo2->val = %f\n",mo2->list_ENERGY_turnang[i]);
			quatCopy(r, &(mo2->opQuatList[i]));
			optQuatCurveEnergy();
		//}

		cond += 1;
	}
	//}
    my_free(b);
    my_free(q);
    my_free(r);
    my_free(u);
    my_free(v);
}
void optCalcTurningAngle						( int index ) 
{
	//obtain previous and current quaternion, 
	int			i = 0;
	//int			j = 0;
	int			steps = 10;
	double		delta = (1.0)/ ((double)steps);
	//double		dTheta = (4.0*M_PI)/((double)steps);
	double		theta = 0.0;
	double		quatDOT	= 0.0;
	double		quatTHETA=0.0;
	//double		qvec[3];
	double		tangent[3];
	double		crossZT[3];
	double		norm_crossZT= 0.0;
	double		z[3]		= {0.0, 0.0, 1.0};
	double		thetaZT		= 0.0;
	//double		cangle, pangle;
	double		t0, t1, t2;
	double		F_t0, F_t1,F_t2, Fdash;
	double		update;
	double		currAng;
	//double		Fdash = 0.0;
	//int			cond=0;
	int			tmp_whichFrame = 0;
	//double		initDist = 0.0;
	quaternion	*b;
	quaternion	*q;
	quaternion	*r;
	quaternion	*prev;
	quaternion	*curr;

	b =	(quaternion*) my_malloc(sizeof(quaternion));
	q =	(quaternion*) my_malloc(sizeof(quaternion));
	r =	(quaternion*) my_malloc(sizeof(quaternion));
	prev =	(quaternion*) my_malloc(sizeof(quaternion));
	curr =	(quaternion*) my_malloc(sizeof(quaternion));

	i		= index;
	F_t0	= F_t1 = F_t2 = Fdash = 0.0;
	update	= 0.0;
	currAng	= 0.0;
    
	//printf("### [%d] \n",i);

		//printf("delta = %f\n",delta);
		quatCopy( &(mo2->opQuatList[i-1]), prev );
		
		//printf("\t prev-quat = %f, %f, %f, %f\n",prev->r, prev->i, prev->j, prev->k );
		tmp_whichFrame = mo2->whichFrame;
		mo2->whichFrame = OPTIMIZATION;
		findQuat_Current( i, curr, &currAng );
		mo2->whichFrame = tmp_whichFrame;
		//printf("\t curr-quat = %f, %f, %f, %f\n",curr->r, curr->i, curr->j, curr->k );
		//printf("\t curr-angl = %f\n",currAng);

		t0		= currAng;
		t1		= t0 - delta;
		t2		= t0 + delta;

		//for(i=1; i<mo2->numPoints-1; i++) {
		i = index;

		tangent[0] =	mo2->TangList[i].p3.d[0];
		tangent[1] =	mo2->TangList[i].p3.d[1];
		tangent[2] =	mo2->TangList[i].p3.d[2];

		////compute (z x tangent)
		cross3D( z, tangent, crossZT );
		norm_crossZT = length3D( crossZT );
		crossZT[0] /= norm_crossZT;
		crossZT[1] /= norm_crossZT;
		crossZT[2] /= norm_crossZT;

		//compute arccos(z . tangent)
		thetaZT = acos( dot3D( z, tangent ) );

		//compute b(vhat)
		b->r = cos( thetaZT/2.0 );
		b->i = crossZT[0]*sin( thetaZT/2.0 );
		b->j = crossZT[1]*sin( thetaZT/2.0 );
		b->k = crossZT[2]*sin( thetaZT/2.0 );
		
		//printf("\t ------------->\n");
		//printf("\n");
		//q_t0, F_t0
			theta = t0;
			//printf("\t theta (t0) = %f\n",theta);
			q->r = cos( theta/2.0 );
			q->i = tangent[0]*sin( theta/2.0 );
			q->j = tangent[1]*sin( theta/2.0 );
			q->k = tangent[2]*sin( theta/2.0 );
			quatQuatProduct( q, b, r );
			quatConjugate(r);
			quatDOT = quatQuatDotProduct( r, prev );
			quatTHETA = 2.0*acos( quatDOT );
			F_t0 = quatTHETA;
			//printf("\t F_t0 = %f\n",F_t0);
			//printf("\n");

		//q_t1, F_t1
			theta = t1;
			//printf("\t theta (t1) = %f\n",theta);
			q->r = cos( theta/2.0 );
			q->i = tangent[0]*sin( theta/2.0 );
			q->j = tangent[1]*sin( theta/2.0 );
			q->k = tangent[2]*sin( theta/2.0 );
			quatQuatProduct( q, b, r );
			quatConjugate(r);
			quatDOT = quatQuatDotProduct( r, prev );
			quatTHETA = 2.0*acos( quatDOT );
			//printf("#\tr (t1)= %f, %f, %f, %f\n",r->r,r->i,r->j,r->k);
			F_t1 = quatTHETA;
			//printf("\t F_t1 = %f\n",F_t1);
			//printf("\n");

		//q_t2, F_t2
			theta = t2;
			//printf("\t theta (t2) = %f\n",theta);
			q->r = cos( theta/2.0 );
			q->i = tangent[0]*sin( theta/2.0 );
			q->j = tangent[1]*sin( theta/2.0 );
			q->k = tangent[2]*sin( theta/2.0 );
			quatQuatProduct( q, b, r );
			quatConjugate(r);
			quatDOT = quatQuatDotProduct( r, prev );
			quatTHETA = 2.0*acos( quatDOT );
			//printf("#\tr (t1)= %f, %f, %f, %f\n",r->r,r->i,r->j,r->k);
			F_t2 = quatTHETA;
			//printf("\t F_t2 = %f\n",F_t2);
			//printf("\n");

		//central difference formula
			Fdash = ( F_t2 - F_t1 )/( 2.0 * delta );		// two-point formula
			//printf("\t Fdash = %f\n",Fdash);

			////Fdash = ( F_t2 - F_t0 )/( delta );			//from NRecipes
			update = F_t0/Fdash;
			//printf("\t update= %f\n", update);

			theta = t0 - update;
			if(theta < 0.0) {
				theta = 4.0*M_PI + theta;
			}
			//printf("\t THETA = %f\n",theta);
	
		//theta = cangle - update;
		q->r = cos( theta/2.0 );
		q->i = tangent[0]*sin( theta/2.0 );
		q->j = tangent[1]*sin( theta/2.0 );
		q->k = tangent[2]*sin( theta/2.0 );
		quatQuatProduct( q, b, r );
		quatConjugate(r);
		//quatDOT = quatQuatDotProduct( v, r );
		//quatTHETA = 2.0*acos( quatDOT );
		//
		//mo2->list_ENERGY_turnang[i] = quatTHETA;
		////printf("\t# mo2->val = %f\n",mo2->list_ENERGY_turnang[i]);
		quatCopy(r, &(mo2->opQuatList[i]));
		
		optQuatCurveEnergy();
		if(mo2->opt_INDEX == mo2->numPoints-1){
		if(mo2->curveQuatAngle < mo2->minTurningAngle)
			mo2->minTurningAngle = mo2->curveQuatAngle;
		}
	//}
			//printf("\t <-------------\n");
			//printf("\n");
    my_free(b);
    my_free(q);
    my_free(r);
    my_free(prev);
    my_free(curr);
}

void buildQuaternion_forRings ( int index, double theta, quaternion *quatRet ) 
{
    int         i = 0;
	double		tangent[3];
	double		crossZT[3];
	double		norm_crossZT= 0.0;
	double		z[3]		= {0.0, 0.0, 1.0};
	double		thetaZT		= 0.0;
	quaternion	*b;
	quaternion	*q;
	quaternion	*r;
	b =	(quaternion*) my_malloc(sizeof(quaternion));
	q =	(quaternion*) my_malloc(sizeof(quaternion));
	r =	(quaternion*) my_malloc(sizeof(quaternion));
    
    i = index;

    tangent[0] =	mo2->TangList[i].p3.d[0];
    tangent[1] =	mo2->TangList[i].p3.d[1];
    tangent[2] =	mo2->TangList[i].p3.d[2];

    ////compute (z x tangent)
    cross3D( z, tangent, crossZT );
    norm_crossZT = length3D( crossZT );
    crossZT[0] /= norm_crossZT;
    crossZT[1] /= norm_crossZT;
    crossZT[2] /= norm_crossZT;

    //compute arccos(z . tangent)
    thetaZT = acos( dot3D( z, tangent ) );

    //compute b(vhat)
    b->r = cos( thetaZT/2.0 );
    b->i = crossZT[0]*sin( thetaZT/2.0 );
    b->j = crossZT[1]*sin( thetaZT/2.0 );
    b->k = crossZT[2]*sin( thetaZT/2.0 );

    q->r = cos( theta/2.0 );
    q->i = tangent[0]*sin( theta/2.0 );
    q->j = tangent[1]*sin( theta/2.0 );
    q->k = tangent[2]*sin( theta/2.0 );
    quatQuatProduct( q, b, r );
    quatConjugate(r);

    quatCopy(r, quatRet);

    my_free(b);
    my_free(q);
    my_free(r);

}

void optPT_step0 ( void ) 
{
    //set last quat same as 0th quat

    int         i           = mo2->numPoints-1;
	double		currAng     = 0.0;
    double      theta       = 0.0;
	//quaternion	*prev;
	quaternion	*curr;
	int         l_whichFrame= mo2->whichFrame;

	//prev    =	(quaternion*) my_malloc(sizeof(quaternion));
	curr    =	(quaternion*) my_malloc(sizeof(quaternion));

    ////@ Update LAST to FIRST quaternion
    //findQuat_Current( 0, prev, &prevAng );
    //buildQuaternion_forRings(i, prevAng, curr);
    //quatCopy( curr, &(mo2->opQuatList[i]) );

    //@ Update LAST to rquatRing angle:
    mo2->whichFrame = PARALLELTRANSPORT;    //@ if this is OPTIMIZATION then it accumulates angles; 
                                            //- has to be the current one opQuatList is emulating.

    findQuat_Current( i, curr, &currAng );
    theta = currAng - mo2->rquatRingAngle;
    if(theta>4.0*M_PI)
        theta = theta - 4.0*M_PI;
    if(theta<0.0)
        theta = theta + 4.0*M_PI;

    buildQuaternion_forRings( i, theta, curr );
    quatCopy( curr, &(mo2->opQuatList[i]) );

    mo2->whichFrame = l_whichFrame;

    //my_free(prev);
    my_free(curr);

}
void optPT_step4 ( int index ) 
{
    int         i           = index;
    //double      theta0      = 0.0;
    double      theta       = 0.0;
	int			steps       = 100;
	double		delta       = (1.0)/ ((double)steps);
	double		quatDOT	    = 0.0;
	double		quatTHETA   = 0.0;
	int         l_whichFrame= mo2->whichFrame;
	double		t0, t1, t2;
	double		F_t0, F_t1,F_t2, Fdash;
	double		update;
	double		currAng,prevAng;
    quaternion	*p;
	quaternion	*q;
	quaternion	*prev;
	quaternion	*curr;

	//init variables
    p       =	(quaternion*) my_malloc(sizeof(quaternion));
	q       =	(quaternion*) my_malloc(sizeof(quaternion));
	prev    =	(quaternion*) my_malloc(sizeof(quaternion));
	curr    =	(quaternion*) my_malloc(sizeof(quaternion));
	F_t0	= F_t1 = F_t2 = Fdash = 0.0;
	update	= 0.0;
	currAng	= 0.0;
    mo2->whichFrame = OPTIMIZATION;

    printf("# i = %d\n",i);

    //find previous quaternion
    if(i==mo2->numPoints-1)
        findQuat_Current( 0, prev, &prevAng );
    else
        findQuat_Current( i+1, prev, &prevAng );

    printf("\t prevang = %f\n",prevAng);

    //find current quaternion
    findQuat_Current( i, curr, &currAng );
    
    //init angles
	t0		= currAng;
	t1		= t0 - delta;
	t2		= t0 + delta;

    //printf("\t PT [%d] = %f, %f, %f, %f\n",i+1,mo2->ptQuatList[i+1].r, mo2->ptQuatList[i+1].i, mo2->ptQuatList[i+1].j, mo2->ptQuatList[i+1].k); 
    //printf("\t prev    = %f, %f, %f, %f\n",prev->r, prev->i, prev->j, prev->k); 
    //
    //printf("\t PT [%d] = %f, %f, %f, %f\n",i,mo2->ptQuatList[i].r, mo2->ptQuatList[i].i, mo2->ptQuatList[i].j, mo2->ptQuatList[i].k); 

    theta = t0;
    //printf("\t theta (t0) = %f\n",theta);
    buildQuaternion_forRings( i, theta, curr );
	quatDOT = quatQuatDotProduct( curr, prev );
	quatTHETA = 2.0*acos( quatDOT );
	//printf("\t curr (t0)= %f, %f, %f, %f angle = %f\n",curr->r,curr->i,curr->j,curr->k, quatTHETA);
	F_t0 = quatTHETA;    

    theta = t1;
    //printf("\t theta (t1) = %f\n",theta);
    buildQuaternion_forRings( i, theta, curr );
	quatDOT = quatQuatDotProduct( curr, prev );
	quatTHETA = 2.0*acos( quatDOT );
	//printf("\t curr (t1)= %f, %f, %f, %f angle = %f\n",curr->r,curr->i,curr->j,curr->k, quatTHETA);
	F_t1 = quatTHETA;    

    theta = t2;
    //printf("\t theta (t2) = %f\n",theta);
    buildQuaternion_forRings( i, theta, curr );
	quatDOT = quatQuatDotProduct( curr, prev );
	quatTHETA = 2.0*acos( quatDOT );
	//printf("\t curr (t2)= %f, %f, %f, %f angle = %f\n",curr->r,curr->i,curr->j,curr->k, quatTHETA);
	F_t2 = quatTHETA;    

	Fdash = ( F_t2 - F_t1 )/( 2.0 * delta );		// two-point formula
    //printf("\t Fdash = %f\n",Fdash);

    update = F_t0/Fdash;
    update = update / (4.0*M_PI);
	//printf("\t update = %f\n", update);

	theta = t0 - update;

    printf("\t new theta = %f\n",theta);
    buildQuaternion_forRings( i, theta, curr );
    quatCopy( curr, &(mo2->opQuatList[i]) );

    //### DONE
    mo2->whichFrame = l_whichFrame;
    
    my_free(p);
    my_free(q);
    my_free(prev);
    my_free(curr);
}

void optPT_step3 ( int index ) 
{
    int         i           = index;
    //double      theta0      = 0.0;
    double      theta       = 0.0;
	int			steps       = 100;
	double		delta       = (1.0)/ ((double)steps);
	//double		quatDOT	    = 0.0;
	//double		quatTHETA   = 0.0;
    double      tmpdist     = 10000.0;
	int         l_whichFrame= mo2->whichFrame;
	double		t0, t1, t2;
	double		F_t0, F_t1,F_t2, Fdash;
	double		update;
	double		currAng,prevAng;
    quaternion	*p;
	quaternion	*q;
	quaternion	*prev;
	quaternion	*curr;

	//init variables
    p       =	(quaternion*) my_malloc(sizeof(quaternion));
	q       =	(quaternion*) my_malloc(sizeof(quaternion));
	prev    =	(quaternion*) my_malloc(sizeof(quaternion));
	curr    =	(quaternion*) my_malloc(sizeof(quaternion));
	F_t0	= F_t1 = F_t2 = Fdash = 0.0;
	update	= 0.0;
	currAng	= 0.0;
    mo2->whichFrame = OPTIMIZATION;

    //printf("# i = %d\n",i);

    //find previous quaternion
    if(i==mo2->numPoints-1)
        findQuat_Current( 0, prev, &prevAng );
    else
        findQuat_Current( i+1, prev, &prevAng );

    //printf("\t prevang = %f\n",prevAng);

    //find current quaternion
    findQuat_Current( i, curr, &currAng );
    
    //init angles
	t0		= currAng;
	t1		= t0 - delta;
	t2		= t0 + delta;

    //printf("\t PT [%d] = %f, %f, %f, %f\n",i+1,mo2->ptQuatList[i+1].r, mo2->ptQuatList[i+1].i, mo2->ptQuatList[i+1].j, mo2->ptQuatList[i+1].k); 
    //printf("\t prev    = %f, %f, %f, %f\n",prev->r, prev->i, prev->j, prev->k); 
    //
    //printf("\t PT [%d] = %f, %f, %f, %f\n",i,mo2->ptQuatList[i].r, mo2->ptQuatList[i].i, mo2->ptQuatList[i].j, mo2->ptQuatList[i].k); 

   // theta = t0;
 //   //printf("\t theta (t0) = %f\n",theta);
 //   buildQuaternion_forRings( i, theta, curr );
	//quatDOT = quatQuatDotProduct( curr, prev );
	//quatTHETA = 2.0*acos( quatDOT );
	////printf("\t curr (t0)= %f, %f, %f, %f angle = %f\n",curr->r,curr->i,curr->j,curr->k, quatTHETA);
	//F_t0 = quatTHETA;    

 //   theta = t1;
 //   //printf("\t theta (t1) = %f\n",theta);
 //   buildQuaternion_forRings( i, theta, curr );
	//quatDOT = quatQuatDotProduct( curr, prev );
	//quatTHETA = 2.0*acos( quatDOT );
	////printf("\t curr (t1)= %f, %f, %f, %f angle = %f\n",curr->r,curr->i,curr->j,curr->k, quatTHETA);
	//F_t1 = quatTHETA;    

 //   theta = t2;
 //   //printf("\t theta (t2) = %f\n",theta);
 //   buildQuaternion_forRings( i, theta, curr );
	//quatDOT = quatQuatDotProduct( curr, prev );
	//quatTHETA = 2.0*acos( quatDOT );
	////printf("\t curr (t2)= %f, %f, %f, %f angle = %f\n",curr->r,curr->i,curr->j,curr->k, quatTHETA);
	//F_t2 = quatTHETA;    

	//Fdash = ( F_t2 - F_t1 )/( 2.0 * delta );		// two-point formula
 //   //printf("\t Fdash = %f\n",Fdash);

 //   update = F_t0/Fdash;
 //   update = update / (4.0*M_PI);
	////printf("\t update = %f\n", update);
    tmpdist = 1000.0;
    theta = 0.0;
    delta = mo2->opt_EPSILON;
    while(theta<=4.0*M_PI) {

        buildQuaternion_forRings( i, theta, curr );
        quatCopy( curr, &(mo2->opQuatList[i]) );
        //optPT_QuatCurveEnergy();
        mo2->pt_curve_dist = 2.0*acos(quatQuatDotProduct(curr, &(mo2->opQuatList[i+1]))) + 2.0*acos(quatQuatDotProduct(curr, &(mo2->opQuatList[i-1])));
        //quatCopy( curr, &(mo2->opQuatList[i]) );
        //mo2->form->notifyOGLWidgets();
        if(mo2->pt_curve_dist < tmpdist) {
            currAng = theta;
            tmpdist = mo2->pt_curve_dist;
        }
        theta = theta + delta;
    }
    //currAng = t0+delta;
    //buildQuaternion_forRings(i, currAng, curr);
    //quatCopy(curr,&(mo2->opQuatList[i]));
    ////optPT_QuatCurveEnergy();
    //mo2->pt_curve_dist = 2.0*acos(quatQuatDotProduct(curr, &(mo2->opQuatList[i+1]))) + 2.0*acos(quatQuatDotProduct(curr, &(mo2->opQuatList[i-1])));

    //F_t1 = mo2->pt_curve_dist;

    //currAng = t0-delta;
    //buildQuaternion_forRings(i, currAng, curr);
    //quatCopy(curr,&(mo2->opQuatList[i]));
    //mo2->pt_curve_dist = 2.0*acos(quatQuatDotProduct(curr, &(mo2->opQuatList[i+1]))) + 2.0*acos(quatQuatDotProduct(curr, &(mo2->opQuatList[i-1])));

    //F_t2 = mo2->pt_curve_dist;

    //if(F_t1 < F_t2)
    //    theta = t0 + delta;
    //else
        //theta = t0 - delta;

	//optPT_QuatCurveEnergy();
    //printf("#%d\tquat energy = %f\n",i,mo2->curveQuatDistance);
    //printf("\t new theta = %f\n",theta);
    theta = currAng;
    buildQuaternion_forRings( i, theta, curr );
    quatCopy( curr, &(mo2->opQuatList[i]) );
    //quatCopy( curr, &(mo2->ptQuatList[i]) );
    //mo2->computeObject_forOpt();

    //### DONE
    mo2->whichFrame = l_whichFrame;
    
    my_free(p);
    my_free(q);
    my_free(prev);
    my_free(curr);

}

void optPT_step2 ( int index ) 
{
    int         i           = index;
    int         j           = 0;
    //double      theta0      = 0.0;
    double      theta       = 0.0;
	int			steps       = 100;
	double		delta       = (1.0)/ ((double)steps);
	//double		quatDOT	    = 0.0;
	//double		quatTHETA   = 0.0;
    double      tmpdist     = 10000.0;
	int         l_whichFrame= mo2->whichFrame;
	double		t0;
	double		F_t0, F_t1,F_t2, Fdash;
	double		update;
	double		currAng;
    quaternion	*p;
	quaternion	*q;
	quaternion	*prev;
	quaternion	*curr;

    //double      t1[3];
    //double      cross_T1T2[3];
    //double      len = 0.0;

	//init variables
    p       =	(quaternion*) my_malloc(sizeof(quaternion));
	q       =	(quaternion*) my_malloc(sizeof(quaternion));
	prev    =	(quaternion*) my_malloc(sizeof(quaternion));
	curr    =	(quaternion*) my_malloc(sizeof(quaternion));
	F_t0	= F_t1 = F_t2 = Fdash = 0.0;
	update	= 0.0;
	currAng	= 0.0;
    mo2->whichFrame = PARALLELTRANSPORT;

    //printf("# i = %d\n",i);
    findQuat_Current( i, curr, &currAng );

    theta = currAng + mo2->rquatRingAngle;
    if(theta>4.0*M_PI)
        theta = theta - 4.0*M_PI;
    //printf("\t > curr val = %f\n",theta);

    buildQuaternion_forRings( i, theta, curr );
    quatCopy( curr, &(mo2->opQuatList[i]) );

    //printf(" item = %f\n", ( mo2->numPoints-2 )*( (mo2->rquatRingAngle + (M_PI/16.0))/(2.0*M_PI)));
    mo2->pt_update_ring_index = ((int)( mo2->numPoints-2 )*(mo2->rquatRingAngle/(2.0*M_PI)));
    if(mo2->pt_update_ring_index > mo2->numPoints-2)
        mo2->pt_update_ring_index = mo2->numPoints-2;

    delta = 0.05;
    int rindex = 0; // real index
    //for(j=1;j<mo2->numPoints-2;j++) {
    for(j=1; j<mo2->pt_update_ring_index; j++) {
        tmpdist = 1000.0;
        t0 = 0.0;
        rindex = i-j;
        //### UPDATE METHOD
        
        //# Naive method
        while(t0<4.0*M_PI) {
            buildQuaternion_forRings( i-j, t0, curr );
            mo2->pt_curve_dist = 2.0*( acos(quatQuatDotProduct(curr, &(mo2->opQuatList[i])) ));// + acos(quatQuatDotProduct(curr, &(mo2->opQuatList[i-2]))) );
            if(mo2->pt_curve_dist < tmpdist) {
                theta = t0;
                tmpdist = mo2->pt_curve_dist;
                buildQuaternion_forRings( i-j, theta, prev );
                quatCopy(prev, &(mo2->opQuatList[i-j]));
            }
            t0 = t0 + delta;
        }

        //# Better method
        //findQuat_Current( rindex, curr, &currAng );
        //theta = currAng + delta;
        //buildQuaternion_forRings( i-j, theta, prev );
        //quatCopy(prev, &(mo2->opQuatList[i-j]));

        //t1[0] =	mo2->TangList[rindex].p3.d[0];
        //t1[1] =	mo2->TangList[rindex].p3.d[1];
        //t1[2] =	mo2->TangList[rindex].p3.d[2];

        //t2[0] =	mo2->TangList[rindex+1].p3.d[0];
        //t2[1] =	mo2->TangList[rindex+1].p3.d[1];
        //t2[2] =	mo2->TangList[rindex+1].p3.d[2];

        ////cross3D(t1,t2, cross_T1T2);
        ////len	= length3D(cross_T1T2);

        ////cross_T1T2[0] /= len;
        ////cross_T1T2[1] /= len;
        ////cross_T1T2[2] /= len;

        //theta =	acos(dot3D(t1,t2));
        ////printf("\t update-theta = %f\n",theta);
        //theta = currAng + theta;
        //buildQuaternion_forRings( i-j, theta, prev );
        //quatCopy(prev, &(mo2->opQuatList[i-j]));

    }
    optPT_QuatCurveEnergy();

    double diff = quatQuatDotProduct(&(mo2->opQuatList[0]), &(mo2->opQuatList[mo2->numPoints-1]));
    //printf("\t Diff bn 1st and nth frame = %f\n",diff );
    mo2->opt_ERROR = diff;
    if(diff > 0.996)
        mo2->pt_flag_terminate = 1;
    //### DONE
    mo2->whichFrame = l_whichFrame;
    my_free(p);
    my_free(q);
    my_free(prev);
    my_free(curr);

    //mo2->pt_update_ring_index = ((int)( mo2->numPoints-2 )*(mo2->rquatRingAngle/(2.0*M_PI)));
    //if(mo2->pt_update_ring_index > mo2->numPoints-2)
    //    mo2->pt_update_ring_index = mo2->numPoints-2;

    ////delta = 0.000001;
    ////int     rindex = 0; // real index

    ////for(j=1; j<mo2->pt_update_ring_index; j++) {

    ////    rindex = i-j;


    ////    //# Naive method
    ////    theta = theta + delta;
    ////    buildQuaternion_forRings(rindex, theta, &(mo2->opQuatList[rindex]) );
    ////}
    ////optPT_QuatCurveEnergy();

    ////double diff = quatQuatDotProduct(&(mo2->opQuatList[0]), &(mo2->opQuatList[mo2->numPoints-1]));
    //////printf("\t Diff bn 1st and nth frame = %f\n",diff );
    ////mo2->opt_ERROR = diff;
    ////if(diff > 0.996)
    ////    mo2->pt_flag_terminate = 1;
}

void optPT_step1 ( int index ) 
{
    int         i           = index;
    //int         j           = 0;
    //int         rindex      = 0;
    //int         ind         = 0;
    double      theta       = 0.0;
	int			steps       = 100;
	double		delta       = (1.0)/ ((double)steps);
	//double		quatDOT	    = 0.0;
	//double		quatTHETA   = 0.0;
    double      tmpdist     = 0.0;
    //double      linkdist    = 1000.0;
	int         l_whichFrame= mo2->whichFrame;
	//double		TMAX;
    double      t0, t1, t2;
	double		F_t0, F_t1,F_t2, Fdash;
	double		update;
	double		currAng;
    quaternion	*p;
	quaternion	*q;
	quaternion	*prev;
	quaternion	*curr;

    //double      t1[3];
    //double      cross_T1T2[3];
    //double      len = 0.0;

	//init variables
    p       =	(quaternion*) my_malloc(sizeof(quaternion));
	q       =	(quaternion*) my_malloc(sizeof(quaternion));
	prev    =	(quaternion*) my_malloc(sizeof(quaternion));
	curr    =	(quaternion*) my_malloc(sizeof(quaternion));
	F_t0	= F_t1 = F_t2 = Fdash = 0.0;
	update	= 0.0;
	currAng	= 0.0;
    
    mo2->pt_update_ring_index = ((int)( mo2->numPoints-2 )*(mo2->rquatRingAngle/(2.0*M_PI)));
    if(mo2->pt_update_ring_index > mo2->numPoints-2)
        mo2->pt_update_ring_index = mo2->numPoints-2;

    delta = 0.05;
    int cmdprnt = 0;
    int sign = 1;
    int cond = 0;

    for(i=mo2->numPoints-2; i>=1; i--) {
        //ind = i;
        //for( j = 0; j<mo2->numRingPoints; j++) {
        //    tmpdist = 0.0;
        //    q->r = mo2->rpList[ind][j]->p4.d[0];
        //    q->i = mo2->rpList[ind][j]->p4.d[1];
        //    q->j = mo2->rpList[ind][j]->p4.d[2];
        //    q->k = mo2->rpList[ind][j]->p4.d[3];
        //    tmpdist = quatQuatDotProduct( q, &(mo2->opQuatList[j+1]) );
        //    tmpdist = tmpdist + quatQuatDotProduct( q, &(mo2->opQuatList[j-1]) );
        //    if(tmpdist<linkdist) {
        //        rindex = j;
        //        linkdist = tmpdist;
        //    }
        //}
        ////printf("\t rindex = %d\n",rindex);
        //mo2->opQuatList[i].r = mo2->rpList[i][rindex]->p4.d[0];
        //mo2->opQuatList[i].i = mo2->rpList[i][rindex]->p4.d[1];
        //mo2->opQuatList[i].j = mo2->rpList[i][rindex]->p4.d[2];
        //mo2->opQuatList[i].k = mo2->rpList[i][rindex]->p4.d[3];
        //findQuat_Current(i, p, &currAng);
        //printf("[%d]\t currAng = %f\n",i,currAng);
        //printf("\t compang = %f\n",mo2->opQuatList[i].theta);
        printf("#i = %d\n",i);
        currAng = mo2->opQuatList[i].theta;
        
        //F_t0 =  2.0*acos( quatQuatDotProduct( &(mo2->opQuatList[i]), &(mo2->opQuatList[i+1])) ) ;//+ 2.0*acos( quatQuatDotProduct( &(mo2->opQuatList[i]), &(mo2->opQuatList[i-1])) );
        //if(cmdprnt) printf("\t F_t0 = %f\n", F_t0);
        //
        //t1 = currAng + delta;
        //buildQuaternion_forRings( i, t1, q );
        //F_t1 =  2.0*acos( quatQuatDotProduct(q, &(mo2->opQuatList[i+1])) ) ;//+ 2.0*acos( quatQuatDotProduct(q, &(mo2->opQuatList[i-1])) );
        //if(cmdprnt) printf("\t F_t1 = %f\n", F_t1);

        //t2 = currAng - delta;
        //buildQuaternion_forRings( i, t2, q );
        //F_t2 =  2.0*acos( quatQuatDotProduct(q, &(mo2->opQuatList[i+1])) ) ;//+2.0*acos( quatQuatDotProduct(q, &(mo2->opQuatList[i-1])) );
        //if(cmdprnt) printf("\t F_t2 = %f\n", F_t2);

        ////@ Fdash = [ f(x+h) - f(x-h) ] / 2h
        //Fdash = ( F_t1 - F_t2 ) / ( 2.0*delta );
        //if(cmdprnt) printf("\t Fdash = %f\n",Fdash);

        ////@ update = f(x) / f'(x) 
        //update = F_t0 / Fdash;
        //if(cmdprnt) printf("\t update = %f\n",update);

        //theta = currAng - update;

        ////if(F_t2 < F_t1)
        ////    theta = t2;
        ////else
        ////    theta = t1;

        //
        //if(theta>4.0*M_PI)
        //    theta = theta/(4.0*M_PI);
        F_t0 =  2.0*acos( quatQuatDotProduct( &(mo2->opQuatList[i]), &(mo2->opQuatList[i+1])) ) ;
        //if(cmdprnt) printf("\t F_t0 = %f\n",F_t0);
        t1 = currAng+delta;
        buildQuaternion_forRings(i,t1,q);
        F_t1 = 2.0*acos( quatQuatDotProduct( q, &(mo2->opQuatList[i+1])) ) ;

        t2 = currAng-delta;
        buildQuaternion_forRings(i,t2,q);
        F_t2 = 2.0*acos( quatQuatDotProduct( q, &(mo2->opQuatList[i+1])) ) ;

        if(F_t1 < F_t2)
            sign = +1;
        else
            sign = -1;

        tmpdist = 1000.0;
        t0 = currAng;
        //printf("\t t0 = %f\n",t0);
        cond = 0;
        t0 = 0.0;
        while(t0<= 4.0*M_PI) {
            buildQuaternion_forRings(i, t0, q);
            //printf("\t SIGN = %d\n",sign);
            F_t1 =  2.0*acos( quatQuatDotProduct( q, &(mo2->opQuatList[i+1])) ) ;
            if(F_t1 < tmpdist) {
                if(!cond)
                    cond = 1;
                theta = t0;
                tmpdist = F_t1;
                //printf("\t tmpdist = %f, theta = %f\n",tmpdist,theta);
            }
            if(sign == -1) 
                if(cond && F_t1>tmpdist) {
                    //t0 = t0 + delta;             
                    break;
                }
            if(sign == 1)
                if(cond && F_t1<tmpdist) {
                    break;
                }
                
            t0 = t0 + sign * delta;
        }

        if(cmdprnt) printf("\t F_t0 = %f, currAng   = %f\n",F_t0,currAng);
        if(cmdprnt) printf("\t F_t1 = %f, new theta = %f\n",tmpdist,theta);
        buildQuaternion_forRings(i, theta, q);
        tmpdist =  2.0*acos( quatQuatDotProduct(q, &(mo2->opQuatList[i+1])) ) ;//+ 2.0*acos( quatQuatDotProduct(q, &(mo2->opQuatList[i-1])) );
        if(cmdprnt) printf("\t DIST = %f\n",tmpdist);

        buildQuaternion_forRings(i, theta, &(mo2->opQuatList[i]) );
        mo2->opQuatList[i].theta = theta;
    }
    if( (quatQuatDotProduct( &(mo2->opQuatList[mo2->numPoints-1]), &(mo2->opQuatList[0]) )) >= 0.995)
        mo2->flag_opt_continue = 0;
    //### DONE
    mo2->whichFrame = l_whichFrame;
    my_free(p);
    my_free(q);
    my_free(prev);
    my_free(curr);
}


void optPT_step ( int index ) 
{
    int         i           =index;
    int         limit       = mo2->numPoints-2;
    double      theta       = 0.0;
    double      minTheta    = 0.0;

    double      MINDIST     = 1000.0;
    double      tmpdist     = 0.0;
    double      delta       = 0.0;
    quaternion	*p;
	quaternion	*q;

	//init variables
    p       =	(quaternion*) my_malloc(sizeof(quaternion));
	q       =	(quaternion*) my_malloc(sizeof(quaternion));

    mo2->pt_update_ring_index = ((int)( mo2->numPoints-2 )*(mo2->rquatRingAngle/(2.0*M_PI)));
    if(mo2->pt_update_ring_index > mo2->numPoints-2)
        mo2->pt_update_ring_index = mo2->numPoints-2;

    delta = 0.01;
    //i = index;
    //for(i=limit; i>=(limit-mo2->opt_INDEX-1); i--) {
    for(i = (limit-mo2->opt_INDEX); i<=limit; i++) {
        while(theta <=4.0*M_PI) {

            buildQuaternion_forRings( i, theta, q );
            tmpdist = quatQuatDotProduct( q, &(mo2->opQuatList[i+1]) );

            if(tmpdist < MINDIST) {
                //printf("\t dist = %f\n",tmpdist);
                MINDIST = tmpdist;
                minTheta= theta;
                //quatCopy(q,p);
                //quatCopy(q, &(mo2->opQuatList[i]));
            }
            theta = theta + delta;
        }

        //printf("\t MINDIST = %f\n",MINDIST);
        buildQuaternion_forRings( i, minTheta, p );
        q->r = -p->r;
        q->i = -p->i;
        q->j = -p->j;
        q->k = -p->k;

        tmpdist = quatQuatDotProduct( q, &(mo2->opQuatList[i+1]) );
        //printf("\t FINAL = %f\n",tmpdist);
        quatCopy(q, &(mo2->opQuatList[i]));
        //quatCopy(q, &(mo2->ptQuatList[i]));
        //mo2->computeObject();
    }
    double diff = quatQuatDotProduct( &(mo2->opQuatList[limit]), &(mo2->opQuatList[limit+1]) );
    printf("\t diff = %f\n",diff);
    if(diff > 0.99)
        mo2->flag_opt_continue = 0;
    //optPT_QuatCurveEnergy();

}
/////////////////////////////////////////////////////////////////
void optPT_QuatCurveEnergy 						( void ) 
{
	int			i		= 0;
	int			limit	= 0;
	double		quatDOT	= 0.0;
	double		quatTHETA=0.0;
	double		quat_dist=0.0;
	double		quat_angle=0.0;

	quaternion	*p;
	quaternion	*q;

	p =	(quaternion*) my_malloc(sizeof(quaternion));
	q =	(quaternion*) my_malloc(sizeof(quaternion));

	if(mo2->flag_curve_closed)
		limit = mo2->numPoints-1;
	else
		limit = mo2->numPoints-2;

    quatCopy( &(mo2->opQuatList[0]), p );
	for(i=1;i<limit;i++) {
		quatCopy( &(mo2->opQuatList[i]), q );
		quatDOT = quatQuatDotProduct( p, q );
		quatTHETA = 2.0*acos( quatDOT );
		quatCopy( q, p );
		quat_dist = quat_dist  + quatDOT;
		quat_angle= quat_angle + quatTHETA;
	}
	mo2->curveQuatAngle = quat_angle;
    mo2->curveQuatDistance = quat_dist;
    if(quat_dist<mo2->minTurningAngle)
        mo2->minTurningAngle = quat_dist;
    //printf("\t curve dist = %f\n",quat_dist);
    
    my_free(p);
    my_free(q);

}
void optQuatCurveEnergy 						( void ) 
{
	int			i		= 0;
	int			limit	= 0;
	double		quatDOT	= 0.0;
	double		quatTHETA=0.0;
	double		quat_dist=0.0;
	double		quat_angle=0.0;

	quaternion	*p;
	quaternion	*q;

	p =	(quaternion*) my_malloc(sizeof(quaternion));
	q =	(quaternion*) my_malloc(sizeof(quaternion));

	if(mo2->flag_curve_closed)
		limit = mo2->numPoints-1;
	else
		limit = mo2->numPoints-2;

	//if( mo2->whichFrame == FRENETSERRET )
	//	quatCopy( &(mo2->fsQuatList[0]), p );
	//else if( mo2->whichFrame == PARALLELTRANSPORT )
	//	quatCopy( &(mo2->ptQuatList[0]), p );
	//else if( mo2->whichFrame == GEODESICREFERENCE )
	//	quatCopy( &(mo2->grQuatList[0]), p );
	quatCopy( &(mo2->opQuatList[0]), p );
	for(i=1;i<limit;i++) {
		//if( mo2->whichFrame == FRENETSERRET )
		//	quatCopy( &(mo2->fsQuatList[i]), q );
		//else if( mo2->whichFrame == PARALLELTRANSPORT )
		//	quatCopy( &(mo2->ptQuatList[i]), q );
		//else if( mo2->whichFrame == GEODESICREFERENCE )
		//	quatCopy( &(mo2->grQuatList[i]), q );
		quatCopy( &(mo2->opQuatList[i]), q );
		quatDOT = quatQuatDotProduct( p, q );
		quatTHETA = 2.0*acos( quatDOT );
		quatCopy( q, p );
		quat_dist = quat_dist  + quatDOT;
		quat_angle= quat_angle + quatTHETA;
	}
	mo2->curveQuatDistance	= quat_dist;
	mo2->curveQuatAngle		= quat_angle;
	//printf("Quatdist = %f\n",quatDIST);
	//printf("Quatangle= %f\n",quatTHETA);
    
    my_free(p);
    my_free(q);

}
void optQuatCurveEnergy_CNT						( int MAX ) 
{
	int			i		= 0;
	int			limit	= MAX;
	double		quatDOT	= 0.0;
	double		quatTHETA=0.0;
	double		quat_dist=0.0;
	double		quat_angle=0.0;

	quaternion	*p;
	quaternion	*q;

	p =	(quaternion*) my_malloc(sizeof(quaternion));
	q =	(quaternion*) my_malloc(sizeof(quaternion));

	//if(mo2->flag_curve_closed)
	//	limit = mo2->numPoints-1;
	//else
	//	limit = mo2->numPoints-2;

	//if( mo2->whichFrame == FRENETSERRET )
	//	quatCopy( &(mo2->fsQuatList[0]), p );
	//else if( mo2->whichFrame == PARALLELTRANSPORT )
	//	quatCopy( &(mo2->ptQuatList[0]), p );
	//else if( mo2->whichFrame == GEODESICREFERENCE )
	//	quatCopy( &(mo2->grQuatList[0]), p );
	quatCopy( &(mo2->opQuatList[0]), p );

	for(i=1;i<=limit;i++) {
		//if( mo2->whichFrame == FRENETSERRET )
		//	quatCopy( &(mo2->fsQuatList[i]), q );
		//else if( mo2->whichFrame == PARALLELTRANSPORT )
		//	quatCopy( &(mo2->ptQuatList[i]), q );
		//else if( mo2->whichFrame == GEODESICREFERENCE )
		//	quatCopy( &(mo2->grQuatList[i]), q );
		quatCopy( &(mo2->opQuatList[i]), q );
		quatDOT = quatQuatDotProduct( p, q );
		quatTHETA = 2.0*acos( quatDOT );
		quatCopy( q, p );
		quat_dist = quat_dist  + quatDOT;
		quat_angle= quat_angle + quatTHETA;
		//printf("[%d] sigma(angle) = %f\n",i,quat_angle);
	}
	mo2->curveQuatDistance_CNT	= quat_dist;
	mo2->curveQuatAngle_CNT		= quat_angle;
	//printf("[0-%d] quat angle= %f\n",limit,mo2->curveQuatAngle_CNT);

    
    my_free(p);
    my_free(q);
}

void optNRmethod								( void )
{
	//int i = 0;
	//for(i=1; i<mo2->numPoints-1; i++) {
		optNRstep(mo2->opt_INDEX);
	//}
}

void optNRstep									( int index )
{
	int			i			= index;
	double		FP			= 0;
	//int			cnt			= 0;
	//double		currAngle	= 0.0;
	double		tangent[3];
	double		crossZT[3];
	double		norm_crossZT= 0.0;
	double		thetaZT		= 0.0;
	double		z[3]		= {0.0, 0.0, 1.0};
	double		theta		= 0.0;
	double		phi			= 0.0;
	double		delta		= 0.0001;
	double		currDP		= 0.0;
	double		prevDP		= 0.0;

	//double		Eold		= 0.0;
	//double		Enew		= 0.0;
	//double		dE			= 0.0;
	//double		dTheta		= 0.0;
	//double		newAng		= 0.0;

	quaternion	*q0, *q1,*q2; 
	quaternion	*b;
	quaternion	*q;
	quaternion	*r;

	b =	(quaternion*) my_malloc(sizeof(quaternion));
	q =	(quaternion*) my_malloc(sizeof(quaternion));
	r =	(quaternion*) my_malloc(sizeof(quaternion));
	q0=	(quaternion*) my_malloc(sizeof(quaternion));	
	q1=	(quaternion*) my_malloc(sizeof(quaternion));	
	q2=	(quaternion*) my_malloc(sizeof(quaternion));	

	//@ get previous quaternion
	quatCopy(&(mo2->opQuatList[i-1]),q0);

	//@ get current tangent
	tangent[0] =	mo2->TangList[i].p3.d[0];
	tangent[1] =	mo2->TangList[i].p3.d[1];
	tangent[2] =	mo2->TangList[i].p3.d[2];
	//printf("#\ttangent = %f, %f, %f\n",tangent[0],tangent[1],tangent[2]);
	
	//@ compute (z x tangent)
	cross3D( z, tangent, crossZT );
	norm_crossZT = length3D( crossZT );
	crossZT[0] /= norm_crossZT;
	crossZT[1] /= norm_crossZT;
	crossZT[2] /= norm_crossZT;

	//@ compute arccos(z . tangent)
	thetaZT = acos( dot3D( z, tangent ) );

	//@ compute b(vhat)
	b->r = cos( thetaZT/2.0 );
	b->i = crossZT[0]*sin( thetaZT/2.0 );
	b->j = crossZT[1]*sin( thetaZT/2.0 );
	b->k = crossZT[2]*sin( thetaZT/2.0 );
	//printf("#\tb = %f, %f, %f, %f\n",b->r,b->i,b->j,b->k);

	theta = 0.0;
	delta = mo2->opt_EPSILON;
	//cnt = 0;
	//index = 0;
	prevDP = 1000.0;
	
	//@ find angle for minimum distance
	while( theta <=4.0*M_PI+delta ) {
		q->r = cos( theta/2.0 );
		q->i = tangent[0]*sin( theta/2.0 );
		q->j = tangent[1]*sin( theta/2.0 );
		q->k = tangent[2]*sin( theta/2.0 );
		
		quatQuatProduct( q, b, r );
		currDP = quatQuatDotProduct( q0, r );
		if( (currDP < prevDP) ) {
			//index = cnt;
			//printf("[%d] dotProd = %f, theta = %f\n",index,currDP,theta);
			prevDP = currDP;
			phi = theta;
		}
		theta = theta + delta;
		//cnt++;
	}

	//@ compute q(theta, vhat)
	q->r = cos( phi/2.0 );
	q->i = tangent[0]*sin( phi/2.0 );
	q->j = tangent[1]*sin( phi/2.0 );
	q->k = tangent[2]*sin( phi/2.0 );
	//printf("#\tq = %f, %f, %f, %f\n",q->r,q->i,q->j,q->k);

	//@ compute new quaternion to update
	quatQuatProduct( q, b, q2 );
	quatConjugate(q2);
	//printf("#\tNEW = %f, %f, %f, %f\n",q2->r,q2->i,q2->j,q2->k);
	//printf("qnorm(q2) = %f\n",qNorm(q2));
	//printf("minAngle = %f\n",phi);
	currDP = quatQuatDotProduct(q0,q2);
	//printf("minDist  = %f\n",currDP);
	quatCopy(q2,&(mo2->opQuatList[i]));

	optQuatCurveEnergy( );
	FP = mo2->curveQuatAngle;
	//printf("GLOBAL = %f\n",FP);
	mo2->opt_ERROR = FP - mo2->curveQuatAngle_PT;
	//printf("ERROR = %f\n",mo2->opt_ERROR);
	
	if(mo2->opt_ERROR < 0.0001) {
		mo2->flag_opt_continue = 0;
		mo2->trackingRing = -1;
	}
	
	optQuatCurveEnergy();
	//### DONE ###
	my_free(b);
	my_free(q);
	my_free(r);
	my_free(q0);
	my_free(q1);
	my_free(q2);
}
void optNR_naive_method							( void )
{
	//int i = 0;
	//printf("mo2->whichFrame =\n");
	//switch(mo2->whichFrame) {
	//	case FRENETSERRET:
	//		printf("Frenet");
	//		break;
	//	case PARALLELTRANSPORT:
	//		printf("PT");
	//		break;
	//	case GEODESICREFERENCE:
	//		printf("GR");
	//		break;
	//}
	//printf("\n");
	//initOPTIM_TurningAngle();

	//for(i=1; i<mo2->numPoints-1; i++) {
	//	optNR_naive_step(i);
	//}
	//printf("opt_Index = %d\n",mo2->opt_INDEX);
	optNR_naive_step(mo2->opt_INDEX);

}
void optNR_naive_step							( int index )
{
	int			i			= index;
	double		FP			= 0;
	//int			cnt			= 0;
	//double		currAngle	= 0.0;
	double		tangent[3];
	double		crossZT[3];
	double		norm_crossZT= 0.0;
	double		thetaZT		= 0.0;
	double		z[3]		= {0.0, 0.0, 1.0};
	double		theta		= 0.0;
	double		phi			= 0.0;
	double		delta		= 0.0001;
	double		currDP		= 0.0;
	double		prevDP		= 0.0;

	//double		Eold		= 0.0;
	//double		Enew		= 0.0;
	//double		dE			= 0.0;
	//double		dTheta		= 0.0;
	//double		newAng		= 0.0;

	quaternion	*q0, *q1,*q2; 
	quaternion	*b;
	quaternion	*q;
	quaternion	*r;

	b =	(quaternion*) my_malloc(sizeof(quaternion));
	q =	(quaternion*) my_malloc(sizeof(quaternion));
	r =	(quaternion*) my_malloc(sizeof(quaternion));
	q0=	(quaternion*) my_malloc(sizeof(quaternion));	
	q1=	(quaternion*) my_malloc(sizeof(quaternion));	
	q2=	(quaternion*) my_malloc(sizeof(quaternion));	

	//@ get previous quaternion
	quatCopy(&(mo2->opQuatList[i-1]),q0);

	//@ get current tangent
	tangent[0] =	mo2->TangList[i].p3.d[0];
	tangent[1] =	mo2->TangList[i].p3.d[1];
	tangent[2] =	mo2->TangList[i].p3.d[2];
	//printf("#\ttangent = %f, %f, %f\n",tangent[0],tangent[1],tangent[2]);
	
	//@ compute (z x tangent)
	cross3D( z, tangent, crossZT );
	norm_crossZT = length3D( crossZT );
	crossZT[0] /= norm_crossZT;
	crossZT[1] /= norm_crossZT;
	crossZT[2] /= norm_crossZT;

	//@ compute arccos(z . tangent)
	thetaZT = acos( dot3D( z, tangent ) );

	//@ compute b(vhat)
	b->r = cos( thetaZT/2.0 );
	b->i = crossZT[0]*sin( thetaZT/2.0 );
	b->j = crossZT[1]*sin( thetaZT/2.0 );
	b->k = crossZT[2]*sin( thetaZT/2.0 );
	//printf("#\tb = %f, %f, %f, %f\n",b->r,b->i,b->j,b->k);

	theta = 0.0;
	delta = mo2->opt_EPSILON;
	//cnt = 0;
	//index = 0;
	prevDP = 1000.0;
	
	//@ find angle for minimum distance
	while( theta <=4.0*M_PI+delta ) {
		q->r = cos( theta/2.0 );
		q->i = tangent[0]*sin( theta/2.0 );
		q->j = tangent[1]*sin( theta/2.0 );
		q->k = tangent[2]*sin( theta/2.0 );
		
		quatQuatProduct( q, b, r );
		currDP = quatQuatDotProduct( q0, r );
		if( (currDP < prevDP) ) {
			//index = cnt;
			//printf("[%d] dotProd = %f, theta = %f\n",index,currDP,theta);
			prevDP = currDP;
			phi = theta;
		}
		theta = theta + delta;
		//cnt++;
	}

	//@ compute q(theta, vhat)
	q->r = cos( phi/2.0 );
	q->i = tangent[0]*sin( phi/2.0 );
	q->j = tangent[1]*sin( phi/2.0 );
	q->k = tangent[2]*sin( phi/2.0 );
	//printf("#\tq = %f, %f, %f, %f\n",q->r,q->i,q->j,q->k);

	//@ compute new quaternion to update
	quatQuatProduct( q, b, q2 );
	quatConjugate(q2);
	//printf("#\tNEW = %f, %f, %f, %f\n",q2->r,q2->i,q2->j,q2->k);
	//printf("qnorm(q2) = %f\n",qNorm(q2));
	//printf("minAngle = %f\n",phi);
	currDP = quatQuatDotProduct(q0,q2);
	//printf("minDist  = %f\n",currDP);
	quatCopy(q2,&(mo2->opQuatList[i]));

	optQuatCurveEnergy( );
	FP = mo2->curveQuatAngle;
	//printf("GLOBAL = %f\n",FP);
	mo2->opt_ERROR = FP-mo2->curveQuatAngle_PT;
	//printf("ERROR = %f\n",mo2->opt_ERROR);
	
	if(mo2->opt_ERROR < 0.0001) {
		mo2->flag_opt_continue = 0;
		mo2->trackingRing = -1;
	}
	
	optQuatCurveEnergy();
	//### DONE ###
	my_free(b);
	my_free(q);
	my_free(r);
	my_free(q0);
	my_free(q1);
	my_free(q2);
}
void optNAIVEmethod								( void )
{
    //printf("opt_Index = %d\n",mo2->opt_INDEX);
	optNAIVEstep( mo2->opt_INDEX );
}

void optNAIVEstep								( int index )
{
	int			i		= index;
	double		currVal = 0.0;
	double		prevVal = 0.0;

	double		nextPHI = 0.0;
	double		fp0		= 0.0;
	double		fp1		= 0.0;
	double		FP		= 0.0;
	quaternion	*Quat; 

	Quat	=	(quaternion*) my_malloc(sizeof(quaternion));	
	Quat->r = Quat->i = Quat->j = Quat->k = 0.0;

	optQuatCurveEnergy_CNT( i );
	FP = mo2->curveQuatAngle_CNT;
	//printf("[0-%d] FP = %f\n",i,FP);

	optQuatCurveEnergy_CNT( i-1 );
	fp0=mo2->curveQuatAngle_CNT;
	//printf("[0-%d] fp0 = %f\n",(i-1),fp0);

	findQuat_Current( i, Quat, &(currVal) );
	findQuat_Previous2( i, Quat, &(prevVal), +1 );
	findQuat_Update( i, Quat, prevVal );
	nextPHI = 2.0*acos( quatQuatDotProduct( Quat, &(mo2->opQuatList[i-1]) ) );
	fp1 = fp0 + nextPHI;

	if(fp1 < FP) {
		quatCopy( Quat, &(mo2->opQuatList[i]) );
		optQuatCurveEnergy_CNT( i );
		FP = mo2->curveQuatAngle_CNT;
	}
	else {
		findQuat_Previous2( i, Quat, &(prevVal), -1 );
		findQuat_Update( i, Quat, prevVal );
		nextPHI = 2.0*acos( quatQuatDotProduct( Quat, &(mo2->opQuatList[i-1]) ) );
		fp1 = fp0 + nextPHI;
		if(fp1 < FP) {
			quatCopy( Quat, &(mo2->opQuatList[i]) );
			optQuatCurveEnergy_CNT( i );
			FP = mo2->curveQuatAngle_CNT;
		}
	}
	//printf("[NEW*] QUAT ANGLE = %f\n",FP);
	optQuatCurveEnergy( );
	FP = mo2->curveQuatAngle;
	//printf("GLOBAL = %f\n",FP);
	mo2->opt_ERROR = FP-mo2->curveQuatAngle_PT;
	//printf("ERROR = %f\n",mo2->opt_ERROR);
	if(mo2->opt_ERROR < 0.0001)
		mo2->flag_opt_continue = 0;
	
    my_free(Quat);
}
void optSECANTmethod							( void )
{
	int		i		= mo2->trackingRing;
	double	currVal = 0.0;
	double	prevVal = 0.0;
	double	nextVal = 0.0;
	int		index	= mo2->trackingRing;

	quaternion	*Quat; 
	Quat	=	(quaternion*) my_malloc(sizeof(quaternion));	
	Quat->r = Quat->i = Quat->j = Quat->k = 0.0;

	findQuat_Current( index, Quat, &(currVal) );
	findQuat_Previous( index, Quat, &(prevVal) );

	optSECANTstep( currVal, prevVal, &(nextVal) );
	//printf("CURR = %f\n",currVal );
	//printf("NEXT = %f\n",nextVal );
	mo2->updatedTHETA = nextVal;
	findQuat_Update( i, Quat, mo2->updatedTHETA );
	//printf("Quat[%d] = %f, %f, %f, %f\n",i,Quat->r, Quat->i, Quat->j, Quat->k);
	quatCopy( Quat, &(mo2->opQuatList[i]) );
	optQuatCurveEnergy();
	printf("EN  = %f\n",mo2->curveQuatAngle);
    my_free(Quat);
}
void optSECANTstep								( double currVal, double prevVal, double* nextVal )
{
	double		p0		= prevVal;
	double		p1		= currVal;
	double		p2		= 0.0;
	double		fp0		= 0.0;
	double		fp1		= 0.0;
	int			index	= mo2->trackingRing;
	double		Theta	=0.0;
	quaternion	*Quat; 
	Quat	=	(quaternion*) my_malloc(sizeof(quaternion));	
	Quat->r = Quat->i = Quat->j = Quat->k = 0.0;


	optQuatCurveEnergy_CNT( index-1 );
	fp0 = mo2->curveQuatAngle;
	findQuat_Previous( index, Quat, &(Theta) );
	optQuatCurveEnergy_CNT( index );
	fp1 = mo2->curveQuatAngle;
	printf("fp1 = %f\nfp0 = %f\n",fp1,fp0);

	p2 = p1 - ( fp1*(p1-p0) )/( fp1 - fp0 );
	
	*nextVal = p2;
    my_free(Quat);
}
void optMTAsetup								( void )
{
	//@ setup
	//set keyframes
	mo2->list_keyframes[0] = 0;
	mo2->list_keyframes[1] = 3;
	mo2->list_keyframes[2] = 5;
	mo2->num_keyframes = 3;
	
	//initialize keyframes
	quatCopy(&(mo2->opQuatList[0]),&(mo2->kfQuatList[0]));
	quatCopy(&(mo2->opQuatList[3]),&(mo2->kfQuatList[1]));
	quatCopy(&(mo2->opQuatList[5]),&(mo2->kfQuatList[2]));
	mo2->curve_energy	= 1000.0;
	mo2->limit_pmin		= 2;
	mo2->limit_pmax		= mo2->numPoints-4;
}
void optMTAmethod								( void )
{
	if( (mo2->opt_INDEX >= mo2->limit_pmin) && (mo2->opt_INDEX <= mo2->limit_pmax) ) {
	//	printf("INDEX = %d\n",mo2->opt_INDEX);
		mo2->index_q0 = mo2->opt_INDEX-1;
		mo2->index_q1 = mo2->opt_INDEX;
		mo2->index_q2 = mo2->opt_INDEX+1;
		optMTAstep(mo2->opt_INDEX);
	}
	if(mo2->opt_INDEX==mo2->numPoints-2)
		mo2->flag_opt_continue = 0;

}
void optMTAstep									( int index )
{
	int i = index;
	//int	pmin = 0;
	//int pmax = 0;
	//double ENERGY = 0.0;
	//double newENERGY = 0.0;
	//double		tangent[3];
	//double		crossZT[3];
	//double		norm_crossZT= 0.0;
	//double		thetaZT		= 0.0;
	//double		z[3]		= {0.0, 0.0, 1.0};
	//double		theta		= 0.0;
	//double		phi			= 0.0;
	//double		delta		= 0.0001;
	//double		currDP		= 0.0;
	//double		prevDP		= 0.0;
	//bool		condition	= TRUE;
	double		h			= 0.0;
	quaternion	*q0, *q1,*q2; 
	quaternion	*q3;
	quaternion	*b;
	quaternion	*q;
	quaternion	*r;

	b =	(quaternion*) my_malloc(sizeof(quaternion));
	q =	(quaternion*) my_malloc(sizeof(quaternion));
	r =	(quaternion*) my_malloc(sizeof(quaternion));
	q0=	(quaternion*) my_malloc(sizeof(quaternion));	
	q1=	(quaternion*) my_malloc(sizeof(quaternion));	
	q2=	(quaternion*) my_malloc(sizeof(quaternion));	
	q3=	(quaternion*) my_malloc(sizeof(quaternion));	

	mo2->gmax_curve_qdist	 = -1.0;
	mo2->gmax_curve_qdotprod = -1.0;
	mo2->gmax_curve_qtangacc = -1.0;
	//printf("*** index = %d\n",i);

	//get previous quaternion
	quatCopy(&(mo2->opQuatList[i-1]),q0);

	//get next quaternion
	quatCopy(&(mo2->opQuatList[i+1]),q2);

	//get current quaternion
	quatCopy(&(mo2->opQuatList[i]),q1);
	
	//@ TURNING ANGLE experienced by 3D-Frames
	mo2->mtaENERGY = 2.0*acos( quatQuatDotProduct(q0,q1) );
	//printf("Energy[%d] = %f\n",i,mo2->mtaENERGY);
	mo2->list_mtaENERGY_angle[i] = mo2->mtaENERGY;
	if(mo2->list_mtaENERGY_angle[i] > mo2->gmax_curve_qdist)
		mo2->gmax_curve_qdist = mo2->list_mtaENERGY_angle[i];
	
	//@ INNER DOT PRODUCT Quat-Quat 
	mo2->mtaENERGY = ( quatQuatDotProduct(q0,q1) );
	//printf("Energy[%d] = %f\n",i,mo2->mtaENERGY);
    mo2->list_mtaENERGY_dist[i] = mo2->mtaENERGY;
	if(mo2->list_mtaENERGY_dist[i] > mo2->gmax_curve_qdotprod)
		mo2->gmax_curve_qdotprod = mo2->list_mtaENERGY_dist[i];

	//@ TANGENTIAL ACCELERATION Experienced by Quaternion-Curve
	h = 0.00001;
	calc_num_SecondOrder_diff_3POINT(q0,q1,q2,q3,h);
	calc_TangAcceleration(q3,q1,h,&(mo2->mtaENERGY));
	//printf("Energy[%d] = %f\n",i,mo2->mtaENERGY);
	mo2->list_mtaENERGY_tacc[i] = mo2->mtaENERGY;
	if(mo2->list_mtaENERGY_tacc[i] > mo2->gmax_curve_qtangacc)
		mo2->gmax_curve_qtangacc = mo2->list_mtaENERGY_tacc[i];

	//construct current
		//@ get current tangent
		////tangent[0] =	mo2->TangList[i].p3.d[0];
		////tangent[1] =	mo2->TangList[i].p3.d[1];
		////tangent[2] =	mo2->TangList[i].p3.d[2];
		//////printf("#\ttangent = %f, %f, %f\n",tangent[0],tangent[1],tangent[2]);
		////
		//////@ compute (z x tangent)
		////cross3D( z, tangent, crossZT );
		////norm_crossZT = length3D( crossZT );
		////crossZT[0] /= norm_crossZT;
		////crossZT[1] /= norm_crossZT;
		////crossZT[2] /= norm_crossZT;

		//////@ compute arccos(z . tangent)
		////thetaZT = acos( dot3D( z, tangent ) );

		//////@ compute b(vhat)
		////b->r = cos( thetaZT/2.0 );
		////b->i = crossZT[0]*sin( thetaZT/2.0 );
		////b->j = crossZT[1]*sin( thetaZT/2.0 );
		////b->k = crossZT[2]*sin( thetaZT/2.0 );

		////delta = mo2->opt_EPSILON;
		////phi = 0.0;
		////while(condition) {
		////	//construct quat with latest value of PHI
		////	q->r = cos( phi/2.0 );
		////	q->i = tangent[0]*sin( phi/2.0 );
		////	q->j = tangent[1]*sin( phi/2.0 );
		////	q->k = tangent[2]*sin( phi/2.0 );
		////	quatQuatProduct( q, b, q2 );

		////	//compute energy of this section (LOCAL)
		////		//use tangential acceleration to calculate energy

		////	//compare energy with previous ENERGY
		////		//if newENERGY smaller than ENERGY then set ENERGY = newENERGY
		////		//esle continue looking

		////	//increment PHI by delta
		////}

		//hopefully finds the quaternion that minimizes the tangential acceleration
		//at index=i; what happens when subsequent quaternion (i+1) is changed?
	//### DONE ###
	my_free(b);
	my_free(q);
	my_free(r);
	my_free(q0);
	my_free(q1);
	my_free(q2);

}
void calc_num_SecondOrder_diff_3POINT			( const quaternion *q0, const quaternion *q1, const quaternion *q2, quaternion *q3, double h ) 
{
	// q'' = q2 - 2q1 + q0
	//		 -------------
	//			  h^2

	q3->r = (q2->r - 2.0*q1->r + q0->r)/(h*h);
	q3->i = (q2->i - 2.0*q1->i + q0->i)/(h*h);
	q3->j = (q2->j - 2.0*q1->j + q0->j)/(h*h);
	q3->k = (q2->k - 2.0*q1->k + q0->k)/(h*h);
	quatNormalize(q3);
	//printf("#\tq'' = %f, %f, %f, %f\n",q3->r,q3->i,q3->j,q3->k);
}

void calc_TangAcceleration						( quaternion *qDD, quaternion *q, double h, double *retValue )
{
	//tang acc. at this point in quaternion space is:
	// 
	//	E(q) = h * | qDD \ q |^2
	//
	//	qDD \ q = qDD - a(q)
	//
	//	a = (qDD.q)/(q.q)

	double	alpha = 0.0;
	double  norm  = 0.0;
	quaternion *alphaQ;
	quaternion *qret;
	alphaQ	= (quaternion*) my_malloc(sizeof(quaternion));
	qret	= (quaternion*) my_malloc(sizeof(quaternion));

	//	a = (qDD.q)/(q.q)
	alpha = ( quatQuatDotProduct(qDD,q) ) / ( quatQuatDotProduct(q,q) );
	//printf("\talpha [a.b/b.b] = %f\n",alpha);
	alphaQ->r = alpha * q->r;
	alphaQ->i = alpha * q->i;
	alphaQ->j = alpha * q->j;
	alphaQ->k = alpha * q->k;
	//printf("#\talphaQ = %f, %f, %f, %f\n",alphaQ->r,alphaQ->i,alphaQ->j,alphaQ->k);

	//	qDD \ q = qDD - a(q)
	q->r = qDD->r - alphaQ->r;
	q->i = qDD->i - alphaQ->i;
	q->j = qDD->j - alphaQ->j;
	q->k = qDD->k - alphaQ->k;
	//printf("#\tq [qDD-a(q)] = %f, %f, %f, %f\n",q->r,q->i,q->j,q->k);

	norm = qNorm(q);
	//printf("\tnorm = %f\n",norm);

	//	E(q) = h * | qDD \ q |^2
	*retValue = h*norm*norm;
	//printf("\tretValue [h*norm^2] = %f\n",*retValue);

	//### DONE ###
    my_free(alphaQ);
	my_free(qret);
}

//*********************************************//
//
//*********************************************//
void TEST										( void ) 
{
#ifndef TESTING
#define TESTING
#endif
	double		m[16];
	quaternion	*q, *p, *r;					// quaternion
	double a = M_PI/4.0;
	double s = sin(a);
	double c = cos(a);

	//init quaternions
	p =	(quaternion*) my_malloc(sizeof(quaternion));
	q =	(quaternion*) my_malloc(sizeof(quaternion));
	r =	(quaternion*) my_malloc(sizeof(quaternion));

	printf(">>> TEST CASES\n\n");

	//TESTING quatQuatProd;
	p->r = 1.0; p->i = p->j = p->k = 0.0;

	printf("IDENTITY Quaternion: \n");
	printf("I = %f, %f, %f, %f\n",p->r, p->i, p->j, p->k );
	printf("testing... quatToRotationMatrix( I ) = \n");
	quatToRotationMatrix( p, m );
	printf("\n");

	printf("testing... ExtractQuaternionFromMat( m ) = \n");
	m[0] = 1; m[4] = 0;	m[8] = 0; m[12] = 0;
	m[1] = 0; m[5] = 1; m[9] = 0; m[13] = 0;
	m[2] = 0; m[6] = 0; m[10]= 1; m[14] = 0;
	m[3] = 0; m[7] = 0; m[11]= 0; m[15] = 1;
	ExtractQuaternionFromMat( m, q );
	printf("[I] q= %f, %f, %f, %f\n",q->r, q->i, q->j, q->k );
	printf("\n");

	printf("testing... quatQuatProduct( I*i ) = i:\n");
	p->r = 1.0; p->i = p->j = p->k = 0.0;
	q->i = 1.0; q->r = q->j = q->k = 0.0;
	printf("I = %f, %f, %f, %f\n",p->r, p->i, p->j, p->k );
	printf("i = %f, %f, %f, %f\n",q->r, q->i, q->j, q->k );
	quatQuatProduct( p, q, r );
	printf(" I*i = %f, %f, %f, %f\n",r->r, r->i, r->j, r->k );
	printf("\n");

	printf("testing... quatQuatProduct( i*j ) = k:\n");
	p->i = 1.0; p->r = p->j = p->k = 0.0;
	q->j = 1.0; q->r = q->i = q->k = 0.0;
	printf("i = %f, %f, %f, %f\n",p->r, p->i, p->j, p->k );
	printf("j = %f, %f, %f, %f\n",q->r, q->i, q->j, q->k );
	quatQuatProduct( p, q, r );
	printf("i*j = k = %f, %f, %f, %f\n",r->r, r->i, r->j, r->k );
	printf("\n");

	printf("Obtaining back rotation matrix for k:\n");
	quatToRotationMatrix( r, m );
	printf("\n");

	printf("testing... ExtractQuaternionFromMat( k ) = \n");
	m[0] =-1; m[4] = 0;	m[8] = 0; m[12] = 0;
	m[1] = 0; m[5] =-1; m[9] = 0; m[13] = 0;
	m[2] = 0; m[6] = 0; m[10]= 1; m[14] = 0;
	m[3] = 0; m[7] = 0; m[11]= 0; m[15] = 1;
	ExtractQuaternionFromMat( m, q );
	printf("    k= %f, %f, %f, %f\n",q->r, q->i, q->j, q->k );
	printf("\n");

	printf("Changing order for mult: j*i\n");
	p->i = 1.0; p->r = p->j = p->k = 0.0;
	q->j = 1.0; q->r = q->i = q->k = 0.0;
	printf("i = %f, %f, %f, %f\n",p->r, p->i, p->j, p->k );
	printf("j = %f, %f, %f, %f\n",q->r, q->i, q->j, q->k );
	quatQuatProduct( q, p, r );
	printf(" j*i = %f, %f, %f, %f\n",r->r, r->i, r->j, r->k );
	printf("\n");

	printf("Obtaining back rotation matrix for -k:\n");
	quatToRotationMatrix( r, m );
	printf("\n");

	printf("testing... ExtractQuaternionFromMat( -k ) = \n");
	m[0] =-1; m[4] = 0;	m[8] = 0; m[12] = 0;
	m[1] = 0; m[5] =-1; m[9] = 0; m[13] = 0;
	m[2] = 0; m[6] = 0; m[10]= 1; m[14] = 0;
	m[3] = 0; m[7] = 0; m[11]= 0; m[15] = 1;
	ExtractQuaternionFromMat( m, q );
	printf("    k= %f, %f, %f, %f\n",q->r, q->i, q->j, q->k );
	printf("\n");

	printf("Obtaining back rotation matrix for -I:\n");
	p->r = -1.0; p->i = p->j = p->k = 0.0;
	quatToRotationMatrix( p, m );
	printf("\n");

	a = M_PI/4.0;
	s = sin(a);
	c = cos(a);
	printf("testing... Rotation by 45-deg along 0,0,1 = \n");
	m[0] = c; m[4] =-s;	m[8] = 0; m[12] = 0;
	m[1] = s; m[5] = c; m[9] = 0; m[13] = 0;
	m[2] = 0; m[6] = 0; m[10]= 1; m[14] = 0;
	m[3] = 0; m[7] = 0; m[11]= 0; m[15] = 1;
	ExtractQuaternionFromMat( m, q );
	printf("   q= %f, %f, %f, %f\n",q->r, q->i, q->j, q->k );
	printf("\n");

	a = M_PI/2.0;
	s = sin(a);
	c = cos(a);
	printf("testing... Rotation by 90-deg along 0,0,1 = \n");
	m[0] = c; m[4] =-s;	m[8] = 0; m[12] = 0;
	m[1] = s; m[5] = c; m[9] = 0; m[13] = 0;
	m[2] = 0; m[6] = 0; m[10]= 1; m[14] = 0;
	m[3] = 0; m[7] = 0; m[11]= 0; m[15] = 1;
	ExtractQuaternionFromMat( m, q );
	printf("   q= %f, %f, %f, %f\n",q->r, q->i, q->j, q->k );
	printf("\n");

	a = M_PI;
	s = sin(a);
	c = cos(a);
	printf("testing... Rotation by 180-deg along 0,0,1 = \n");
	m[0] = c; m[4] =-s;	m[8] = 0; m[12] = 0;
	m[1] = s; m[5] = c; m[9] = 0; m[13] = 0;
	m[2] = 0; m[6] = 0; m[10]= 1; m[14] = 0;
	m[3] = 0; m[7] = 0; m[11]= 0; m[15] = 1;
	ExtractQuaternionFromMat( m, q );
	printf("   q= %f, %f, %f, %f\n",q->r, q->i, q->j, q->k );
	printf("\n");

	a = (M_PI/180.0)*33.0;
	s = sin(a);
	c = cos(a);
	printf("testing... Rotation by 33-deg along 0,0,1 = \n");
	m[0] = c; m[4] =-s;	m[8] = 0; m[12] = 0;
	m[1] = s; m[5] = c; m[9] = 0; m[13] = 0;
	m[2] = 0; m[6] = 0; m[10]= 1; m[14] = 0;
	m[3] = 0; m[7] = 0; m[11]= 0; m[15] = 1;
	ExtractQuaternionFromMat( m, q );
	printf("   q= %f, %f, %f, %f\n",q->r, q->i, q->j, q->k );
	printf("manual calc of q: = q(cos(theta/2.0), vhat*sin(theta/2.0))\n");
	q->r = cos(a/2.0);
	q->i = q->j = 0.0;
	q->k = 1.0*sin(a/2.0);
	printf("   q= %f, %f, %f, %f\n",q->r, q->i, q->j, q->k );
	printf("\n");
#undef TESTING

}



//### Load MESHVIEW FILES
void	enqueue			( double value, queue *q )
{
    node *tmp;

    if( (tmp = (node *) my_malloc(sizeof(node)) ) == NULL) {
        
        printf("Warning : malloc failed.\n");
    }

    tmp->v = value;
    tmp->next = NULL;

    if(q->count == 0) {
        q->begin = tmp;  
        q->end = tmp;
    } else {
        (q->end)->next = tmp;
        q->end = tmp;
    }
    q->count++;
}

double	dequeue			( queue *q ) 
{
    node *tmp;
    double v;

    if(q->count == 0) {
        printf("Queue is empty.\n"); /* shouldn't get here */
        exit(0);
    }
    v = (q->begin)->v;
    tmp = q->begin;
    q->begin = (q->begin)->next;
    free(tmp);
    q->count--;
    return v;
}

void	qfree			( queue *q ) 
{
    node *tmp;

    while(q->count > 0) {
        tmp = q->begin;
        q->begin = (q->begin)->next;
        free(tmp);
        q->count--;
    }
}

void	free_obj		( Obj *obj ) 
{
    //int i;

    if(obj != NULL) {

        //if(obj->vlist != NULL) {
        //    for(i = 0; i < obj->nvertices; i++) {
        //        my_free(obj->vlist[i].faces);
        //    }
        //}
        //if(obj->flist != NULL) {
        //    for(i = 0; i < obj->nfaces; i++)
        //        my_free(obj->flist[i].vertices);
        //}
        my_free(obj->vlist);
        //my_free(obj->flist);
        
        //for(i = 0; i < obj->nframes; i++)
        //    free_obj(obj->frames[i]);
        //my_free(obj->frames-1);
        my_free(obj);
    }
}



/********************************************************************************
* Date   : Sat Oct 29 12:04:49 EST 2005
* Method : void    globalVarInit   ( void )
* Args   : none
* Return : none
* Purpose: initialize data structures for lines
********************************************************************************/
void    globalVarInit   ( void )
{
    mo2->olist          = (Obj **) malloc(sizeof(Obj *) * MAXOBJ);  //init: object list for single line-object
    mo2->lineList       = (Obj ***) my_malloc(sizeof(Obj **)) ;     //init: top-level object list for multiple line-object
    mo2->lineList[0]    = (Obj **) my_malloc(sizeof(Obj *)*MAXOBJ); //init: 2nd-level object list for multiple line-object
}

/********************************************************************************
* Date   : Sat Oct 29 12:04:49 EST 2005
* Method : void load_data ( const char *filename )
* Args   : filname to be loaded, reference of object to be initialized and returned
* Return : none
* Purpose: call appropriate routines to load MESHVIEW file for LINE(S)
* Notes  : Check if multiple/single line(s) to be loaded; empty out structs 
*        : accordingly
********************************************************************************/
void    load_data		( const char *filename )
{
    Obj     *objs[MAXOBJ];  //object list    
    int     i   = 0;        //index
    int     num = 0;        //number of objects initialized
    char    *extn;          //input-file extension
    int     objNUM = 0;

    
    
    extn = strrchr(filename, '.'); 
    extn++; /* find beginning of the extention */

    //@ read only file header to determine number of lines
    if(strcmp(extn, "lines") == 0) { 
        Obj ***testobj;
        testobj       = (Obj***)my_malloc(sizeof(Obj));
        testobj[0]    = (Obj**)my_malloc(sizeof(Obj));
        testobj[0][0] = (Obj*)my_malloc(sizeof(Obj));
        num = loadLINES(filename, testobj, FALSE);
        my_free(testobj[0][0]);
        my_free(testobj[0]);
        my_free(testobj);
    }

    //# free list for multiple lines (2nd time onwards)
    if(mo2->flag_multi_lines && !mo2->flag_multi_lines_first_time ) {
        for(i=0;i<MAXOBJ;i++)                   
            my_free(mo2->tmpobj[i]);
        my_free(mo2->tmpobj);                                   //init: Data-struct
        mo2->flag_multi_lines_first_time = 0;                   //init: Flag for multiple lines first-time check
        mo2->nlines = 0;                                        //init: Number of lines to be read from file
    }

    //# initialize list for multiple lines
    if(strcmp(extn, "lines") == 0) { 
        mo2->tmpobj = (Obj**) my_malloc(sizeof(Obj));
        for(i=0; i<mo2->nlines; i++) 
            mo2->tmpobj[i] = (Obj*) my_malloc(sizeof(Obj));
        mo2->mobjs = 0;                                         //init: Number of objects for multiple lines
    }
    //# initialize objects for single lines
    else if(strcmp(extn, "line") == 0) { 
        mo2->nobjs = 0;                                         //init: Number of objects for single line file
    }

    //# Load file for multiple lines
    if(strcmp(extn, "lines") == 0) {
        if((num = loadLINES(filename, &(mo2->tmpobj), TRUE)) > 0) {   //call: Attemp loading file with multiple lines
            //@ continue
            //printf("Successfully read file...\n");
        }

        for(i = 0; i < num && mo2->mobjs < MAXOBJ; i++) {       //admin: Assign number of objects to global-var
            mo2->lineList[mo2->mobjs] = mo2->tmpobj;            //       list of multiple-lines; 
            mo2->mobjs++;                                       //       Increment counter         
        }
        mo2->flag_multi_lines = TRUE;                           //init:  Set flag for multiple-lines    
    }
    //# Load file for single line
    else if(strcmp(extn,"line") == 0) {
        if((num = loadLINE(filename, objs)) > 0) {              //call: Attemp loading file with single line
            //@ continue
            //printf("Successfully read file...\n");
        }
        for(i = 0; i < num && mo2->nobjs < MAXOBJ; i++) {       //admin: Assign number of objects to global-list
            mo2->olist[mo2->nobjs] = objs[i];                   //       
            mo2->nobjs++;                                       //       Increment counter
        }
        mo2->flag_multi_lines = FALSE;
        if(i < num) {
            printf("You exceeded MAX number of objects.\n"
	            "%d  new objects discarded.\n", num - i);
            for(;i < num; i++)
                free_obj(objs[i]);
        }
    }


    //### DONE 
}

/********************************************************************************
* Date   : Sat Oct 29 12:04:49 EST 2005
* Method : int		loadLINE		( const char *filename, Obj **objret )
* Args   : filname to be loaded, reference of object to be initialized and returned
* Return : none
* Purpose: load a SINGLE LINE from MESHVIEW format file
********************************************************************************/
int		loadLINE		( const char *filename, Obj **objret )
{
    FILE *in;
    char str[1000], *str_ptr, *end_str;
    int vcolor = 0;
    int dim = 3;
    int gformat;
    int cformat = NOCOLOR;
    int nvertices, nfaces;
    int size_flag = 0;
    float size = 1.0;
    Vertex *vlist;
    Face *flist;    /* faces are used to hold lines */
    Color4 cfront;
    Obj *obj;

    int i, j, tmp_value;
    double value;
    queue q;

    q.count = 0;

    if((in = fopen(filename, "r")) == NULL) {
        printf("Couldn't open file %s.\n", filename);
        return 0;
    }

        /* parse the header */
    while(fgets(str, 1000, in) != NULL && str[0] == '#');
    str_ptr = str;
  
    if(*str_ptr == 'S') {
        size_flag = ON;
        str_ptr++;
    }

    if(*str_ptr == 'C') {
        vcolor = 1;
        str_ptr++;
    }

    if(*str_ptr == '3') {
        str_ptr++;
    } else if(*str_ptr == '4') {
        dim = 4;
        str_ptr++;
    }

    if(!strncmp(str_ptr, "LINE", 4)) {
        str_ptr += 4;
        gformat = LINE;
    } else {
/*     printf("File: %s Invalid LINE format.\n", filename); */
        return 0;
    }
  
    if(!strncmp(str_ptr, "_LOOP", 5)) {
        str_ptr += 5;
        gformat = LINE_LOOP;    
    } else if(!strncmp(str_ptr, "_STRIP", 6)) {
        str_ptr += 6;
        gformat = LINE_STRIP;
    }

    if(*str_ptr == 'C' || *str_ptr == 'c') {
        str_ptr++;
        cformat = UNIFORM1;
    }

        /* put all values into a queue */
    while(fgets(str, 1000, in) != NULL) {
        str_ptr = str;
        for(;;) {
            value = strtod(str_ptr, &end_str);
            if(str_ptr == end_str) break;
            str_ptr = end_str;
            enqueue(value, &q);
        }
    }
    fclose(in);

        /* check if there is enough values for object color and #points */
    if(q.count<(cformat == UNIFORM1 ? 4 : 0)+(size_flag == ON ? 1 : 0)+2){
        printf("File %s. Not enough values.\n", filename);
        qfree(&q);
        return 0;
    }
  
        /* read in object color(s) */
    if(cformat == UNIFORM1) {
        cfront.f[0] = dequeue(&q);
        cfront.f[1] = dequeue(&q);
        cfront.f[2] = dequeue(&q);
        cfront.f[3] = dequeue(&q);
    }

    if(size_flag == ON) size = dequeue(&q);
    if(size <= 0) size = 1.0;

    nvertices = (int)dequeue(&q);
    nfaces = (int)dequeue(&q);

        /* check if there's enough values for vertices in a queue */
    if(q.count<nvertices*(dim  + (vcolor == 1 ? 4 : 0))+(gformat == LINE ? nfaces * 2 : 0)) {
        printf("File: %s. Not enough values.\n", filename);
        qfree(&q);
        return 0;
    }

        /* allocate and zero out memory */
    obj = (Obj *)malloc(sizeof(Obj));
    vlist = (Vertex*)malloc(sizeof(Vertex) * nvertices);
    flist = (Face*) malloc(sizeof(Face) * nfaces);

    if(obj == NULL || vlist == NULL) {
        printf("File: %s. Malloc failed.\n", filename);
        qfree(&q);
        my_free(obj);
        my_free(vlist);
        my_free(flist);
        return 0;
    }
#ifdef LINUX
    bzero(obj, sizeof(Obj));
    bzero(vlist,sizeof(Vertex) * nvertices);
#endif
        /* Initialize object */
    obj->gformat = gformat;
    obj->cformat = cformat;
    obj->cfront = cfront;
    obj->texture_flag = OFF;
    obj->vcolor = vcolor;
    obj->fcolor = OFF;
    obj->nvertices = nvertices;
    obj->nfaces = nfaces;
    obj->size_flag = size_flag;
    obj->size = size;
    obj->vlist = vlist;
    obj->flist = flist;
    obj->scale = 1.0;

        /* read vertices */
    for(i = 0; i < nvertices; i++) {
        vlist[i].index = i;
            /* all inputs are treated as 3D */
        vlist[i].p3.d[0] = dequeue(&q);
        vlist[i].p3.d[1] = dequeue(&q);
        vlist[i].p3.d[2] = dequeue(&q);
        //if(dim == 3)
        //    vlist[i].p4.d[3] = 0.0;
        //else
        //    vlist[i].p4.d[3] = dequeue(&q);
        if(vcolor == 1) {
            vlist[i].cfront.f[0] = dequeue(&q);
            vlist[i].cfront.f[1] = dequeue(&q);
            vlist[i].cfront.f[2] = dequeue(&q);
            vlist[i].cfront.f[3] = dequeue(&q);
        }
    }

    if(gformat != LINE_LOOP && gformat != LINE_STRIP) {

            /* read lines */
        for(i = 0; i < nfaces; i++) {
      
            flist[i].index = i;
            flist[i].nvert = 2;
      
                /* allocate and zero out memory for vertex array  */
            if((flist[i].vertices = (Vertex **)malloc(sizeof(Vertex*) * 2)) == NULL) {
                printf("File %s. Malloc failed.\n", filename);
                qfree(&q);
                free_obj(obj);
                return 0;
            }
#ifdef LINUX
            bzero(flist[i].vertices, sizeof(Vertex*) * 2);
#endif
                /* read vertex indices for the current face */
            for(j = 0; j < 2; j++) {
                tmp_value = (int)dequeue(&q);
                if(tmp_value < 0 || tmp_value >=  nvertices){
                    printf("File: %s. Vertex index is out of range.\n", filename);
                    qfree(&q);
                    free_obj(obj);
                    return 0;
                }
                flist[i].vertices[j] = &vlist[tmp_value];
                //printf("vertex[%d][%d] = %f,%f,%f\n",i,j,flist[i].vertices[j]->p3.d[0], flist[i].vertices[j]->p3.d[1], flist[i].vertices[j]->p3.d[2]);
            }
        }
    }

    if(q.count > 0) {
        printf("File %s %d numbers left unread.\n", filename, q.count);
        qfree(&q);
    }

    *objret = obj;
  
#if DEBUG_LOAD
    print_obj(obj, filename);
#endif
  
    printf("File %s has been loaded\n", filename);
    return 1;
}
/********************************************************************************
* Date   : Sat Oct 29 12:04:49 EST 2005
* Method : int		loadLINES		( const char *filename, Obj **objret )
* Args   : filname to be loaded, reference of object to be initialized and returned
* Return : none
* Purpose: load MULTIPLE LINES from MESHVIEW format file
********************************************************************************/
int		loadLINES		( const char *filename, Obj ***objret, int flag )
{
    FILE *in;
    char str[1000], *str_ptr, *end_str;
    int vcolor = 0;
    int dim = 3;
    int gformat;
    int cformat = NOCOLOR;
    int nvertices, nfaces;
    int nlines; //@sid
    int size_flag = 0;
    float size = 1.0;
    Vertex *vlist;
    Face *flist;    /* faces are used to hold lines */
    Color4 cfront;
    Obj **obj;


    int i, j;
    double value;
    queue q;

    q.count = 0;

    if((in = fopen(filename, "r")) == NULL) {
        printf("Couldn't open file %s.\n", filename);
        return 0;
    }

    /* parse the header */
    while(fgets(str, 1000, in) != NULL && str[0] == '#');
    str_ptr = str;
  
    if(*str_ptr == 'S') {
        size_flag = ON;
        str_ptr++;
    }

    if(*str_ptr == 'C') {
        vcolor = 1;
        str_ptr++;
    }

    if(*str_ptr == '3') {
        str_ptr++;
    } else if(*str_ptr == '4') {
        dim = 4;
        str_ptr++;
    }

    if(!strncmp(str_ptr, "LINE", 4)) {
        str_ptr += 4;
        gformat = LINE;
    } else {
        /* printf("File: %s Invalid LINE format.\n", filename); */
        return 0;
    }
  
    if(!strncmp(str_ptr, "_LOOP", 5)) {
        str_ptr += 5;
        gformat = LINE_LOOP;    
    } else if(!strncmp(str_ptr, "_STRIP", 6)) {
        str_ptr += 6;
        gformat = LINE_STRIP;
    } 
    else if(!strncmp(str_ptr, "_LIST", 5)) {
        str_ptr += 5;
        gformat = LINE_LIST;
    }

    if(*str_ptr == 'C' || *str_ptr == 'c') {
        str_ptr++;
        cformat = UNIFORM1;
    }

    /* put all values into a queue */
    while(fgets(str, 1000, in) != NULL) {
        str_ptr = str;
        for(;;) {
            value = strtod(str_ptr, &end_str);
            if(str_ptr == end_str) break;
            str_ptr = end_str;
            enqueue(value, &q);
        }
    }
    fclose(in);

    /* check if there is enough values for object color and #points */
    if(q.count<(cformat == UNIFORM1 ? 4 : 0)+(size_flag == ON ? 1 : 0)+2){
        printf("File %s. Not enough values.\n", filename);
        qfree(&q);
        return 0;
    }
  
    /* read in object color(s) */
    if(cformat == UNIFORM1) {
        cfront.f[0] = dequeue(&q);
        cfront.f[1] = dequeue(&q);
        cfront.f[2] = dequeue(&q);
        cfront.f[3] = dequeue(&q);
    }

    if(size_flag == ON) size = dequeue(&q);
    if(size <= 0) size = 1.0;

    nlines = (int) dequeue(&q);
    mo2->nlines = nlines;
    printf("\t num lines to be loaded = %d\n",nlines);
    if(!flag)
        return 0;
    /* allocate and zero out memory */
    obj = (Obj **)malloc(sizeof(Obj) * nlines); //@sid

    /* loop through number of lines */
    for(i=0; i<nlines; i++) {

        nvertices = (int)dequeue(&q);

        obj[i] = (Obj *)malloc(sizeof(Obj));
        vlist = (Vertex*)malloc(sizeof(Vertex) * nvertices);
        //flist = (Face*) malloc(sizeof(Face) * nfaces); /* don't need this */

        if(obj == NULL || vlist == NULL) {
            printf("File: %s. Malloc failed.\n", filename);
            qfree(&q);
            my_free(obj[i]);
            my_free(vlist);
            my_free(flist);
            return 0;
        }
#ifdef LINUX
        bzero(obj, sizeof(Obj)); 
        bzero(vlist,sizeof(Vertex) * nvertices);
#endif
            /* Initialize object */
        obj[i]->gformat = gformat;
        obj[i]->cformat = cformat;
        obj[i]->cfront = cfront;
        obj[i]->texture_flag = OFF;
        obj[i]->vcolor = vcolor;
        obj[i]->fcolor = OFF;
        obj[i]->nvertices = nvertices;
        obj[i]->nfaces = nfaces;
        obj[i]->nlines = nlines;
        obj[i]->size_flag = size_flag;
        obj[i]->size = size;
        obj[i]->vlist = vlist;
        obj[i]->flist = flist;
        obj[i]->scale = 1.0;

            /* read vertices */
        for(j = 0; j < nvertices; j++) {

            vlist[j].index = i;

            /* all inputs are treated as 3D */
            vlist[j].p3.d[0] = dequeue(&q);
            vlist[j].p3.d[1] = dequeue(&q);
            vlist[j].p3.d[2] = dequeue(&q);

            if(vcolor == 1) {
                vlist[j].cfront.f[0] = dequeue(&q);
                vlist[j].cfront.f[1] = dequeue(&q);
                vlist[j].cfront.f[2] = dequeue(&q);
                vlist[j].cfront.f[3] = dequeue(&q);
            }
        }
    }
    *objret = obj;
  
#if DEBUG_LOAD
    print_obj(obj, filename);
#endif
  
    return 1;
}
/********************************************************************************
* Date   : Sat Oct 29 12:04:49 EST 2005
* Method : void    draw_line       ( Obj *obj, int flag )
* Args   : object reference, color flag
* Return : none
* Purpose: render a SINGLE LINE that has been read from MESHVIEW format file
********************************************************************************/
void    draw_line       ( Obj *obj, int flag )
{
    int i, j;
    Vertex *vertex;

    if(obj->size_flag == ON)
        glLineWidth(obj->size);
    else
        glLineWidth(2.0);

    glDisable(GL_LIGHTING);
    
    if(obj->gformat == LINE) {
        glBegin(GL_LINES);
            for(i=0; i<obj->nfaces; i++) {
	
                for(j=0; j<obj->flist[i].nvert; j++) {
	                vertex = obj->flist[i].vertices[j];
	                switch(flag) {
	                    case VERT_FRONT:
	                        glColor4fv(vertex->cfront.f);
	                        break;
	                    //case DEPTH_C:
	                    //    glColor4fv(palette[vertex->depth]);
	                    //    break;
	                }
	                glVertex3dv(vertex->p3.d);
                }
            }
        glEnd();

    } else {

        if(obj->gformat == LINE_STRIP) 
            glBegin(GL_LINE_STRIP);
        else
            glBegin(GL_LINE_LOOP);
    
        for(i=0; i<obj->nvertices; i++) {
              
            vertex = &(obj->vlist[i]);
            switch(flag) {
	            case VERT_FRONT:
	                glColor4fv(vertex->cfront.f);
	                break;
	            //case DEPTH_C:
	            //    glColor4fv(palette[vertex->depth]);
	            //    break;
	        }
            glVertex3dv(vertex->p3.d);
        }
        glEnd();
    }

    glEnable(GL_LIGHTING);
}

/********************************************************************************
* Date   : Sat Oct 29 12:04:49 EST 2005
* Method : void    draw_line       ( Obj *obj, int flag )
* Args   : object reference, color flag
* Return : none
* Purpose: render MULTIPLE LINES that has been read from MESHVIEW format file
********************************************************************************/
void    draw_lines		( Obj *obj, int flag )
{
    //printf("\t > obj.vertices = %d\n",obj->nvertices);
    int i, j;
    Vertex *vertex;

    if(obj->size_flag == ON)
        glLineWidth(obj->size);
    else
        glLineWidth(2.0);

    glDisable(GL_LIGHTING);
    
    if(obj->gformat == LINE) {
        //printf("#\t nverts = %d, nfaces = %d\n",obj->nvertices, obj->nfaces);
        glBegin(GL_LINES);
            for(i=0; i<obj->nfaces; i++) {
	
                for(j=0; j<obj->flist[i].nvert; j++) {
	                vertex = obj->flist[i].vertices[j];
	                switch(flag) {
	                    case VERT_FRONT:
	                        glColor4fv(vertex->cfront.f);
	                        break;
	                    //case DEPTH_C:
	                    //    glColor4fv(palette[vertex->depth]);
	                    //    break;
	                }
	                glVertex3dv(vertex->p3.d);
                }
            }
        glEnd();

    } else {

        if(obj->gformat == LINE_STRIP) 
            glBegin(GL_LINE_STRIP);
        else if(obj->gformat == LINE_LIST)
            glBegin(GL_LINE_STRIP);
        else
            glBegin(GL_LINE_LOOP);
    
        for(i=0; i<obj->nvertices-3; i++) {
              
            vertex = &(obj->vlist[i]);
            switch(flag) {
	            case VERT_FRONT:
	                glColor4fv(vertex->cfront.f);
	                break;
	            //case DEPTH_C:
	            //    glColor4fv(palette[vertex->depth]);
	            //    break;
	        }
            glVertex3dv(vertex->p3.d);
        }
        glEnd();
    }
    
    glEnable(GL_LIGHTING);
}

