#include "objectKnot.h"
#include "QuaternionSet.h"

//global declarations
MathObject		*mo2;
char			obj_qstr[128];
static GLUquadricObj *qobjCYL, *qobjSPH;
void			reflect_QuatList();

//Rings structs
int MAX_RING_PNTS = 300;
Vertex  **prevList;
Vertex	**firstList;
Vertex	**currList;


//# Initializing Routines and Destructor
void stext(char* string) {
  char* p;
  for(p=string;*p;p++) {
    glutBitmapCharacter((void*)mo2->qfont,*p);
  }
}

void initKnotFirst ( MathObject	*mathobjRef	)
{
	int	i =	0;
	mo2	= mathobjRef;
	
	mo2->objectID =	OBJKNOT;
	
	mo2->pntList = (Vertex*) my_malloc(sizeof(Vertex)*MAX_POINTS);
	for(i=0; i<MAX_POINTS; i++)	{
		mo2->pntList->p3.d[0] =	0.0;
		mo2->pntList->p3.d[1] =	0.0;
		mo2->pntList->p3.d[2] =	0.0;
	}
	
	mo2->increment		= 0.05;
	mo2->numPoints		= 0;

	mo2->showAxes		= 1;
	mo2->showCurve		= 1;
	mo2->showFrame		= 0;
	mo2->showFace		= 1;
	mo2->showEdge		= 0;
	mo2->showLabels		= 1;
	mo2->showMesh		= 0;
	mo2->showNormals	= 0;
	mo2->showQuatMap	= 1;
	mo2->showViewWYZ	= 1;
	mo2->TORUS_PNTS		= 50;
	mo2->TORUS_STPS		= 30;
	mo2->showTSpace		= 0;
	mo2->whichFrame	  =	PARALLELTRANSPORT;
//	  	mo2->whichFrame		= FRENETSERRET;
	mo2->whichQuatFrame	= PARALLELTRANSPORT;
	mo2->whoCommandsMainWin = MAIN_WINDOW;

	// quaternion map
	mo2->qfont			= GLUT_BITMAP_HELVETICA_12;
	mo2->kflip			= 1;
	mo2->qshowAxes		= 1;
	mo2->qshowFrame		= 0;
	mo2->qshowFace		= 1;
	mo2->qshowEdge		= 0;
	mo2->qshowNormals	= 0;
	mo2->qwhichQuatComponent = SHOWXYZ;
	mo2->qshowQuatRings	= 0;
	mo2->qshowCloudPoints=0;
	mo2->qshowSphere	= 1;

	// quatenrion ring
	mo2->numOfRings		= 4;	
	mo2->rshowAxes		= 1;
	mo2->rshowCurve		= 0;
	mo2->rshowEdge		= 0;
	mo2->rshowFace		= 1;
	mo2->rshowFrame		= 0;

	mo2->m			= 2;
	mo2->n			= 2;
	mo2->r			= 0.8;
	mo2->a			= 0.4;
	mo2->b			= 0.5;
	mo2->lock_a_b	= 0;
	mo2->startTheta	= 0.0;
	mo2->leave_gap	= 1;
	mo2->go_around_twice = 0;
	mo2->flag_end_rings_match = 0;

	mo2->fgcolor[0]	= 0.1; 
	mo2->fgcolor[1]	= 0.37;
	mo2->fgcolor[2]	= 0.22;
	
	mo2->qset		= new QuaternionSet();

	initKnot();
	
	qobjCYL		= gluNewQuadric();
	qobjSPH		= gluNewQuadric();

	//Ring structs
	firstList = (Vertex**) my_malloc(sizeof(Vertex)*MAX_RING_PNTS);
	for(i=0;i<MAX_RING_PNTS;i++)
		firstList[i] = (Vertex*) my_malloc(sizeof(Vertex));

	prevList = (Vertex**) my_malloc(sizeof(Vertex)*500);
	for(i=0;i<MAX_RING_PNTS;i++)
		prevList[i] = (Vertex*) my_malloc(sizeof(Vertex));

	currList = (Vertex**) my_malloc(sizeof(Vertex)*MAX_RING_PNTS);
	for(i=0;i<MAX_RING_PNTS;i++)
		currList[i] = (Vertex*) my_malloc(sizeof(Vertex));

}

void initKnot (	void )
{
	int	i=0, j=0, k=0;
	
	mo2->tpnts		= (double***) my_malloc(sizeof(double)*MAX_POINTS);
	for(i=0;i<mo2->TORUS_PNTS;i++)
		mo2->tpnts[i] =	(double**) my_malloc(sizeof(double)*mo2->TORUS_STPS);
	
	for(i=0;i<mo2->TORUS_PNTS;i++)
		for(j=0;j<mo2->TORUS_STPS;j++)
			mo2->tpnts[i][j] = (double*) my_malloc(sizeof(double)*3);

	for(i=0;i<mo2->TORUS_PNTS;i++)
		for(j=0;j<mo2->TORUS_STPS;j++)
			for(k=0;k<3;k++)
				mo2->tpnts[i][j][k]=0.0;

}

void cleanupKnot ( void	)
{
	int	i =	0, j = 0;
	for(i=0;i<mo2->TORUS_PNTS;i++)
		for(j=0;j<mo2->TORUS_STPS;j++)
			my_free(mo2->tpnts[i][j]);
	
	for(i=0;i<mo2->TORUS_PNTS;i++)
		my_free(mo2->tpnts[i]);
	
	my_free(mo2->tpnts);

	//Rings structs
	for(i=0;i<MAX_RING_PNTS;i++)
		my_free(firstList[i]);
	my_free(firstList);
	for(i=0;i<MAX_RING_PNTS;i++)
		my_free(prevList[i]);
	my_free(prevList);
	for(i=0;i<MAX_RING_PNTS;i++)
		my_free(currList[i]);
	my_free(currList);
}


//# Rendering Routines
void displayKnot ( void	)
{

	glPushMatrix();
	glRotatef(-90,0,1,0);
	glRotatef(-90,1,0,0);
	drawKnot();
	glPopMatrix();
	if(mo2->showAxes)
		mo2->renderAxes();
	
}

void drawKnot (	void )
{
	int	i=0;
	int limit = 0;
	double tubeThickness = 0.075;
	if(mo2->go_around_twice)
		limit = mo2->numPoints/2;
	else
		limit = mo2->numPoints;

	constructKnot();
	makeTangent();
	makeTangentDerivatives();

	//knotted curve
	glPushAttrib(GL_ALL_ATTRIB_BITS);
	glColor3f(1.0,0.0,0.0);
	glLineWidth(2.0);
	if(mo2->showCurve) {
		glDisable(GL_LIGHTING);
		glDisable(GL_COLOR_MATERIAL);
		glBegin(GL_LINE_STRIP);
		for(i=0;i<=limit;i++) {
			if(i<mo2->numPoints)
				glVertex3f(mo2->pntList[i].p3.d[0],	mo2->pntList[i].p3.d[1], mo2->pntList[i].p3.d[2]);
			else if(i==mo2->numPoints && !mo2->leave_gap)
				glVertex3f(mo2->pntList[0].p3.d[0],	mo2->pntList[0].p3.d[1], mo2->pntList[0].p3.d[2]);
		}
		glEnd();
		//glEnable(GL_LIGHTING);
		//glEnable(GL_COLOR_MATERIAL);
		//for(i=0;i<limit;i++) {
		//	if(i==0 || i==(mo2->numPoints-1))
		//		glColor3f(0.8f,0.0f,0.0f);
		//	else
		//		glColor3f(0.8f,0.7f,0.0f);
		//	glPushMatrix();
		//	glTranslatef(mo2->pntList[i].p3.d[0],mo2->pntList[i].p3.d[1],mo2->pntList[i].p3.d[2]);
		//	//glutSolidSphere(0.02,16,16);
		//	glPopMatrix();
		//}
	}
	glPopAttrib();

	//function to compute tube-frameing
	mo2->fgcolor[0]	= 0.7f;
	mo2->fgcolor[1]	= 0.5f;
	mo2->fgcolor[2]	= 0.2f;

	if(	mo2->whichFrame	== PARALLELTRANSPORT ) {
		
		computeParallelTransportFrame();
		renderFramedCurve(tubeThickness);		 
	}
	else if( mo2->whichFrame ==	FRENETSERRET ) {

		computeFrenetFrame();
		renderFramedCurve(tubeThickness);		
		computeFrenetQuaternionMap6();
	}
	else if( mo2->whichFrame ==	NOFRAME	) {
		renderThickenedCurve(tubeThickness);
	}


}

void drawFrames	( void ) 
{
	int		i	  =	0;
	double	scale =	0.25;
	int		limit =	0;
	int		start =	0;
	if(mo2->go_around_twice)
		limit = mo2->numPoints/2;
	else
		limit = mo2->numPoints;

	if(mo2->numPoints>2) {
		glPushAttrib(GL_ALL_ATTRIB_BITS);
		for( i=start; i<(limit); i++) {
			glPushMatrix();
				glDisable(GL_LIGHTING);
				glTranslatef(mo2->pntList[i].p3.d[0], mo2->pntList[i].p3.d[1], mo2->pntList[i].p3.d[2]);
				glScalef(scale,scale,scale);
				glLineWidth(1.5);
				glBegin(GL_LINES);
				glColor3f(0.0, 0.0, 1.0);
				glVertex3d(0.0,0.0,0.0);
				glVertex3d(mo2->TangList[i].p3.d[0],mo2->TangList[i].p3.d[1],mo2->TangList[i].p3.d[2]);
				glEnd();
			glPopMatrix();
			glPushMatrix();
				glDisable(GL_LIGHTING);
				glTranslatef(mo2->pntList[i].p3.d[0], mo2->pntList[i].p3.d[1], mo2->pntList[i].p3.d[2]);
				glScalef(scale,scale,scale);
				glLineWidth(1.5);
				glBegin(GL_LINES);
				glColor3f(0.0, 1.0, 0.0);
				glVertex3d(0.0,0.0,0.0);
				glVertex3d(mo2->basisRList[i].p3.d[0],mo2->basisRList[i].p3.d[1],mo2->basisRList[i].p3.d[2]);
				glEnd();
			glPopMatrix();
			glPushMatrix();
				glDisable(GL_LIGHTING);
				glTranslatef(mo2->pntList[i].p3.d[0],mo2-> pntList[i].p3.d[1], mo2->pntList[i].p3.d[2]);
				glScalef(scale,scale,scale);
				glLineWidth(1.5);
				glBegin(GL_LINES);
				glColor3f(1.0, 0.0, 0.0);
				glVertex3d(0.0,0.0,0.0);
				glVertex3d(mo2->basisSList[i].p3.d[0],mo2->basisSList[i].p3.d[1],mo2->basisSList[i].p3.d[2]);
				glEnd();
			glPopMatrix();
		}	
		glPopAttrib();
	}
}

void drawTangDTangSpace ( void ) {
	int		i	  =	0;
	int		limit =	mo2->numPoints;
	double	v[3];
	if(mo2->numPoints>2)
	for( i=0; i<(limit); i++) {
		v[0] = mo2->TangList[i].p3.d[0];
		v[1] = mo2->TangList[i].p3.d[1];
		v[2] = mo2->TangList[i].p3.d[2];

		glPushMatrix();
			glTranslatef(mo2->pntList[i].p3.d[0], mo2->pntList[i].p3.d[1], mo2->pntList[i].p3.d[2]);
			glScalef(0.5,0.5,0.5);
			glColor3f(1.0f,0.0f,1.0f);
			glBegin(GL_LINES);
				glVertex3f(0,0,0);
				glVertex3f(v[0],v[1],v[2]);
			glEnd();
			glPushMatrix();
				glTranslatef(v[0],v[1],v[2]);
				gluOrtho2D(-0.1,0.1,-0.1,0.1);
				glDisable(GL_LIGHTING);
				glColor3f(0.2f,0.2f,0.2f);
				sprintf(obj_qstr,"%d",i);
				glRasterPos2f(0.0,0.0);
				stext(obj_qstr);
			glPopMatrix();
		glPopMatrix();

		v[0] = mo2->dTangList[i].p3.d[0];
		v[1] = mo2->dTangList[i].p3.d[1];
		v[2] = mo2->dTangList[i].p3.d[2];

		glPushMatrix();
			glTranslatef(mo2->pntList[i].p3.d[0],mo2->pntList[i].p3.d[1],mo2->pntList[i].p3.d[2]);
			glScalef(0.45,0.45,0.45);
			glColor3f(1.0f,1.0f,0.0f);
			glBegin(GL_LINES);
				glVertex3f(0,0,0);
				glVertex3f(v[0],v[1],v[2]);
			glEnd();
			glPushMatrix();
				glTranslatef(v[0],v[1],v[2]);
				gluOrtho2D(-0.1,0.1,-0.1,0.1);
				glDisable(GL_LIGHTING);
				glColor3f(0.2f,0.2f,0.2f);
				sprintf(obj_qstr,"%d",i);
				glRasterPos2f(0.0,0.0);
				stext(obj_qstr);
			glPopMatrix();
		glPopMatrix();
	}
}
void renderFramedCurve ( double	radius ) 
{
	int		i		= 0;
	int		j		= 0;
	//int	fcnt	= 0;
	int		p		= 0;
	int		start	= 0;
	//int	end		= 0;
	int		renderPoints = 0;
	double	theta		 = 0.0;
	double	dTheta		 = 0.0;
	double	n_radius	 = 2.0*radius;
	int		radialSteps	 = mo2->radialSteps;
	

	/////////////////////////////////////////////////
	// initialize variables
	/////////////////////////////////////////////////
	dTheta = 2.0*M_PI/mo2->radialSteps;
	
	if(mo2->go_around_twice)
		renderPoints = mo2->numPoints/2;
	else
		renderPoints = mo2->numPoints;

	renderPoints = mo2->numPoints-1;

	/////////////////////////////////////////////////
	// compute frames
	/////////////////////////////////////////////////
	for	( i=start; i<renderPoints; i++ ) {

			theta= 0.0;
			//printf("%d > %f,%f,%f\n",i,mo2->pntList[i].p3.d[0],mo2->pntList[i].p3.d[1],mo2->pntList[i].p3.d[2]);
			for	( j=0; j<radialSteps; j++ )	{
				
				mo2->rpnts[i][j][0]	= mo2->pntList[i].p3.d[0] +	radius*mo2->basisRList[i].p3.d[0]*cos(theta) + radius*mo2->basisSList[i].p3.d[0]*sin(theta);
				mo2->rpnts[i][j][1]	= mo2->pntList[i].p3.d[1] +	radius*mo2->basisRList[i].p3.d[1]*cos(theta) + radius*mo2->basisSList[i].p3.d[1]*sin(theta);
				mo2->rpnts[i][j][2]	= mo2->pntList[i].p3.d[2] +	radius*mo2->basisRList[i].p3.d[2]*cos(theta) + radius*mo2->basisSList[i].p3.d[2]*sin(theta);
				
				mo2->npnts[i][j][0]	= n_radius*mo2->basisRList[i].p3.d[0]*cos(theta)+n_radius*mo2->basisSList[i].p3.d[0]*sin(theta);
				mo2->npnts[i][j][1]	= n_radius*mo2->basisRList[i].p3.d[1]*cos(theta)+n_radius*mo2->basisSList[i].p3.d[1]*sin(theta);
				mo2->npnts[i][j][2]	= n_radius*mo2->basisRList[i].p3.d[2]*cos(theta)+n_radius*mo2->basisSList[i].p3.d[2]*sin(theta);
				
				theta =	theta +	dTheta;
			}
	}
		
	if(mo2->showMesh) {
			glPushAttrib(GL_CURRENT_BIT	| GL_LIGHTING_BIT |	GL_ENABLE_BIT |	GL_LINE_BIT);
			glDisable(GL_LIGHTING);
			
			glLineWidth(2.0);
			
			///////////////////////// RING /////////////////////////
			for(i=start; i<(renderPoints); i++) {
				
				glBegin(GL_LINE_STRIP);
				glColor3f(0.0,0.0,0.0);
				for(j=0; j<=radialSteps; j++) {
					if(j<(radialSteps))
						glVertex3d(mo2->rpnts[i][j][0],mo2->rpnts[i][j][1],mo2->rpnts[i][j][2]);
				else
					glVertex3d(mo2->rpnts[i][0][0],mo2->rpnts[i][0][1],mo2->rpnts[i][0][2]);
				}
				glEnd();
			}
			
			///////////////////////// LONGITUDINAL LINES /////////////////////////
			glBegin(GL_LINES);
			glColor3f(0.0,0.0,0.0);
			for(i=start+1;i<(renderPoints); i++) {
				if(i<(renderPoints)) {
					for(j=0;j<radialSteps;j++) {
						glVertex3d(mo2->rpnts[i][j][0],	mo2->rpnts[i][j][1], mo2->rpnts[i][j][2]); 
						glVertex3d(mo2->rpnts[i-1][j][0],mo2->rpnts[i-1][j][1],mo2->rpnts[i-1][j][2]);
						if(j<(radialSteps-1)){
							glVertex3d(mo2->rpnts[i-1][j+1][0],mo2->rpnts[i-1][j+1][1],mo2->rpnts[i-1][j+1][2]);
							glVertex3d(mo2->rpnts[i][j+1][0],mo2->rpnts[i][j+1][1],mo2->rpnts[i][j+1][2]);
						}
						else {
							glVertex3d(mo2->rpnts[i-1][j-radialSteps+1][0],	mo2->rpnts[i-1][j-radialSteps+1][1],mo2->rpnts[i-1][j-radialSteps+1][2]	);
							glVertex3d(mo2->rpnts[i][j-radialSteps+1][0],	mo2->rpnts[i][j-radialSteps+1][1],	mo2->rpnts[i][j-radialSteps+1][2]);
						}
					}
				}
			}
			
		/////////////////////////////////////////////////
		// draw	last section
		/////////////////////////////////////////////////
			if(!mo2->leave_gap)	{
				i=0;
				p=renderPoints-1;
				for(j=0;j<radialSteps;j++) {
					glVertex3d(mo2->rpnts[i][j][0],	mo2->rpnts[i][j][1], mo2->rpnts[i][j][2]); 
					glVertex3d(mo2->rpnts[p][j][0],	mo2->rpnts[p][j][1], mo2->rpnts[p][j][2]);
					if(j<(radialSteps-1)){
						glVertex3d(mo2->rpnts[p][j+1][0], mo2->rpnts[p][j+1][1], mo2->rpnts[p][j+1][2]);
						glVertex3d(mo2->rpnts[i][j+1][0], mo2->rpnts[i][j+1][1], mo2->rpnts[i][j+1][2]);
					}
					else {
						glVertex3d(mo2->rpnts[p][j-radialSteps+1][0], mo2->rpnts[p][j-radialSteps+1][1], mo2->rpnts[p][j-radialSteps+1][2] );
						glVertex3d(mo2->rpnts[i][j-radialSteps+1][0], mo2->rpnts[i][j-radialSteps+1][1], mo2->rpnts[i][j-radialSteps+1][2]);
					}
				}
			}
			glEnd();
		}	

	if(mo2->showFrame) {
		drawFrames();	
	}
	
	if(mo2->showTSpace) {
		drawTangDTangSpace();
	}
	
	if(mo2->showNormals) {
			///////////////////////// NORMALS /////////////////////////
		glPushAttrib(GL_CURRENT_BIT	| GL_LIGHTING_BIT);
		glDisable(GL_LIGHTING);
		glLineWidth(0.5);
		for(i=start; i<renderPoints; i++) {
			//printf("%d > %f,%f,%f\n",i,mo2->fsTList[i].p3.d[0],mo2->fsTList[i].p3.d[1],mo2->fsTList[i].p3.d[2]);

			for(j=0; j<radialSteps;	j++) {
				glPushMatrix();
				glTranslatef(mo2->rpnts[i][j][0], mo2->rpnts[i][j][1], mo2->rpnts[i][j][2]);
				glBegin(GL_LINES);
				glColor3f(0.95f,0.95f,0.0f);
				glVertex3d(0.0,0.0,0.0);
				glVertex3d(mo2->npnts[i][j][0],	mo2->npnts[i][j][1], mo2->npnts[i][j][2]);
				glEnd();
				glPopMatrix();
			}
			
		}
		glPopAttrib();
	}
		
	if(mo2->showFace) {
		///////////////////////// QUAD FACES /////////////////////////
			glPushAttrib(GL_CURRENT_BIT	| GL_ENABLE_BIT	| GL_LIGHTING_BIT);
			
			glEnable(GL_COLOR_MATERIAL);
			
			glEnable(GL_NORMALIZE);
			glEnable(GL_LIGHTING);
			glDisable(GL_CULL_FACE);
			
			
			/*if(showTexture) {
				  glEnable(GL_TEXTURE_2D);
				  glBindTexture(GL_TEXTURE_2D, texName[0]);
				  glDisable(GL_COLOR_MATERIAL);
				  glColor3f(0.0,0.0,0.0);
				  
				  }*/
			
				//if(!showQuatMap) {
			if(1) {
				glBegin(GL_QUADS);
				for(i=start+1;i<(renderPoints); i++) {
					if(i==(start+1))
						glColor3f(0.9,0.1,0.2);
					else
						glColor3f(mo2->fgcolor[0],mo2->fgcolor[1],mo2->fgcolor[2]);
					for(j=0;j<radialSteps;j++) {
						
						//glTexCoord2f(1.0,	1.0);
						//glTexCoord2f(t, t);
						glNormal3d(mo2->npnts[i][j][0],	mo2->npnts[i][j][1], mo2->npnts[i][j][2]);
						glVertex3d(mo2->rpnts[i][j][0],	mo2->rpnts[i][j][1], mo2->rpnts[i][j][2]);
						
						//glTexCoord2f(1.0,	0.0);
						//glTexCoord2f(t, 0.0);
						glNormal3d(mo2->npnts[i-1][j][0], mo2->npnts[i-1][j][1], mo2->npnts[i-1][j][2]);
						glVertex3d(mo2->rpnts[i-1][j][0], mo2->rpnts[i-1][j][1], mo2->rpnts[i-1][j][2]);
						if(j<(radialSteps-1)){
							//glTexCoord2f(0.0,	0.0);
							glNormal3d(mo2->npnts[i-1][j+1][0],	mo2->npnts[i-1][j+1][1], mo2->npnts[i-1][j+1][2]);
							glVertex3d(mo2->rpnts[i-1][j+1][0],	mo2->rpnts[i-1][j+1][1], mo2->rpnts[i-1][j+1][2]);
							
							//glTexCoord2f(0.0,	1.0);
							//glTexCoord2f(0.0,	t);
							glNormal3d(mo2->npnts[i][j+1][0],mo2->npnts[i][j+1][1],mo2->npnts[i][j+1][2]);
							glVertex3d(mo2->rpnts[i][j+1][0],mo2->rpnts[i][j+1][1],mo2->rpnts[i][j+1][2]);
						}
						else {
							//glTexCoord2f(0.0,	0.0);
							glNormal3d(mo2->npnts[i-1][j-radialSteps+1][0],	mo2->npnts[i-1][j-radialSteps+1][1], mo2->npnts[i-1][j-radialSteps+1][2] );
							glVertex3d(mo2->rpnts[i-1][j-radialSteps+1][0],	mo2->rpnts[i-1][j-radialSteps+1][1], mo2->rpnts[i-1][j-radialSteps+1][2] );
							
							//glTexCoord2f(0.0,	1.0);
							//glTexCoord2f(0.0,	t);
							glNormal3d(mo2->npnts[i][j-radialSteps+1][0], mo2->npnts[i][j-radialSteps+1][1], mo2->npnts[i][j-radialSteps+1][2]);
							glVertex3d(mo2->rpnts[i][j-radialSteps+1][0], mo2->rpnts[i][j-radialSteps+1][1], mo2->rpnts[i][j-radialSteps+1][2]);
						}
					}
				}
				
				/////////////////////////////////////////////////
				// draw	last section
				/////////////////////////////////////////////////
				if(!mo2->leave_gap ) {
					i=0;
					p=renderPoints-1;
					for(j=0;j<radialSteps;j++) {
						
						//glTexCoord2f(1.0,	1.0);
						//glTexCoord2f(t, t);
						glNormal3d(mo2->npnts[i][j][0],	mo2->npnts[i][j][1], mo2->npnts[i][j][2]);
						glVertex3d(mo2->rpnts[i][j][0],	mo2->rpnts[i][j][1], mo2->rpnts[i][j][2]);
						
						//glTexCoord2f(1.0,	0.0);
						//glTexCoord2f(t, 0.0);
						glNormal3d(mo2->npnts[p][j][0],	mo2->npnts[p][j][1], mo2->npnts[p][j][2]);
						glVertex3d(mo2->rpnts[p][j][0],	mo2->rpnts[p][j][1], mo2->rpnts[p][j][2]);
						if(j<(radialSteps-1)){
							//glTexCoord2f(0.0,	0.0);
							glNormal3d(mo2->npnts[p][j+1][0], mo2->npnts[p][j+1][1], mo2->npnts[p][j+1][2]);
							glVertex3d(mo2->rpnts[p][j+1][0], mo2->rpnts[p][j+1][1], mo2->rpnts[p][j+1][2]);
							
							//glTexCoord2f(0.0,	1.0);
							//glTexCoord2f(0.0,	t);
							glNormal3d(mo2->npnts[i][j+1][0],mo2->npnts[i][j+1][1],mo2->npnts[i][j+1][2]);
							glVertex3d(mo2->rpnts[i][j+1][0],mo2->rpnts[i][j+1][1],mo2->rpnts[i][j+1][2]);
						}
						else {
							//glTexCoord2f(0.0,	0.0);
							glNormal3d(mo2->npnts[p][j-radialSteps+1][0], mo2->npnts[p][j-radialSteps+1][1], mo2->npnts[p][j-radialSteps+1][2] );
							glVertex3d(mo2->rpnts[p][j-radialSteps+1][0], mo2->rpnts[p][j-radialSteps+1][1], mo2->rpnts[p][j-radialSteps+1][2] );
							
							//glTexCoord2f(0.0,	1.0);
							//glTexCoord2f(0.0,	t);
							glNormal3d(mo2->npnts[i][j-radialSteps+1][0], mo2->npnts[i][j-radialSteps+1][1], mo2->npnts[i][j-radialSteps+1][2]);
							glVertex3d(mo2->rpnts[i][j-radialSteps+1][0], mo2->rpnts[i][j-radialSteps+1][1], mo2->rpnts[i][j-radialSteps+1][2]);
						}
					}//end for()
				}//end if()
				glEnd();
			}//	end	if()
			
			glDisable(GL_TEXTURE_2D);
			glPopAttrib();			  
	}//end if()
}


void renderFramedQuaternionCurve ( double radius ) 
{
	int		h			 = 0;
	int		i			 = 0;
	int		j			 = 0;
//	int		fcnt		 = 0;
	int		end			 = 0;
	int		start		 = 0;
	int		renderPoints = 0;
	double	theta		 = 0.0;
	double	dTheta		 = 0.0;
	double	n_radius	 = 2.0*radius;
//	double	tempColor[3];
	int		radialSteps	 = mo2->radialSteps;
	double	ccolor[3];
	ccolor[0] =	ccolor[1] =	ccolor[2] =	0.0;
	
	/////////////////////////////////////////////////
	// initialize variables
	/////////////////////////////////////////////////
	dTheta = 2.0*M_PI/radialSteps;
	
	renderPoints = mo2->numPoints;

	if(mo2->showQuatMap)
			start += 1;
		
	/////////////////////////////////////////////////
	// compute frames
	/////////////////////////////////////////////////
	for	( i=start; i<renderPoints; i++ ) {

		theta= 0.0;
		for	( j=0; j<radialSteps; j++ )	{
			
			mo2->rpnts[i][j][0]	= mo2->pntList[i].p3.d[0] +	radius*mo2->basisRList[i].p3.d[0]*cos(theta) + radius*mo2->basisSList[i].p3.d[0]*sin(theta);
			mo2->rpnts[i][j][1]	= mo2->pntList[i].p3.d[1] +	radius*mo2->basisRList[i].p3.d[1]*cos(theta) + radius*mo2->basisSList[i].p3.d[1]*sin(theta);
			mo2->rpnts[i][j][2]	= mo2->pntList[i].p3.d[2] +	radius*mo2->basisRList[i].p3.d[2]*cos(theta) + radius*mo2->basisSList[i].p3.d[2]*sin(theta);
			
			mo2->npnts[i][j][0]	= n_radius*mo2->basisRList[i].p3.d[0]*cos(theta)+n_radius*mo2->basisSList[i].p3.d[0]*sin(theta);
			mo2->npnts[i][j][1]	= n_radius*mo2->basisRList[i].p3.d[1]*cos(theta)+n_radius*mo2->basisSList[i].p3.d[1]*sin(theta);
			mo2->npnts[i][j][2]	= n_radius*mo2->basisRList[i].p3.d[2]*cos(theta)+n_radius*mo2->basisSList[i].p3.d[2]*sin(theta);
			
			theta =	theta +	dTheta;
		}
	}
	
	///////////////////////// MESH /////////////////////////
	if(mo2->qshowEdge) {
		glPushAttrib(GL_CURRENT_BIT	| GL_LIGHTING_BIT |	GL_ENABLE_BIT |	GL_LINE_BIT);
		glDisable(GL_LIGHTING);
		
		glLineWidth(1.0);
 
		///////////////////////// RING /////////////////////////
		if(mo2->showQuatMap	&& mo2->whichFrame==PARALLELTRANSPORT) {
			start=1;
			for(i=start; i<(renderPoints-2); i++) {
				
				glBegin(GL_LINE_STRIP);
				glColor3f(0.4f,0.4f,0.4f);
				for(j=0; j<=radialSteps; j++) {
					if(j<(radialSteps))
						glVertex3d(mo2->rpnts[i][j][0],mo2->rpnts[i][j][1],mo2->rpnts[i][j][2]);
					else
						glVertex3d(mo2->rpnts[i][0][0],mo2->rpnts[i][0][1],mo2->rpnts[i][0][2]);
				}
				glEnd();
			}
		}
		else if(mo2->showQuatMap &&	mo2->whichFrame==FRENETSERRET) {
			glLineWidth(1.0f);
			glColor3f(0.4f,0.4f,0.4f);
			start =	1;
			if(mo2->markFrameCounter ==	0)
				end	= mo2->numPoints-2;
			else
				end	= mo2->markFrame[0]-2;
			for(h=0; h<=mo2->markFrameCounter; h++)	{
				//--------------------------->
				glBegin(GL_LINE_STRIP);
				glColor3f(0.4f,0.4f,0.4f);
				for(i=start; i<end;	i++) {
					for(j=0; j<=radialSteps; j++) {
						if(j<(radialSteps))
							glVertex3d(mo2->rpnts[i][j][0],mo2->rpnts[i][j][1],mo2->rpnts[i][j][2]);
						else
							glVertex3d(mo2->rpnts[i][0][0],mo2->rpnts[i][0][1],mo2->rpnts[i][0][2]);
					}
				}
				glEnd();
				//<---------------------------
				start =	mo2->markFrame[h]+1;
				end	= mo2->markFrame[h+1];
				//fprintf(stderr,"Start= %f, end= %f\n",start,end);
				if(h==(mo2->markFrameCounter)-1) {
					end	= mo2->numPoints-1;
					//glPushMatrix();
					//glTranslatef(mo2->pntList[i].p3.d[0],mo2->pntList[i].p3.d[1],mo2->pntList[i].p3.d[2]);
					//glutSolidSphere(0.09,16,16);
					//glPopMatrix();
				}
			}
		}

		///////////////////////// LONGITUDINAL LINES /////////////////////////
		if(mo2->showQuatMap	&& mo2->whichFrame==PARALLELTRANSPORT) {
			glLineWidth(1.0f);
			glBegin(GL_LINES);
			glColor3f(0.4f,0.4f,0.4f);
			start = 1;
			if(mo2->markFrameCounter ==	0)
				end	= mo2->numPoints-2;
			else
				end	= mo2->markFrame[0]-2;
			for(i=start+1;i<(renderPoints-1); i++) {
				if(	i<(renderPoints-2) ) {
					for(j=0;j<radialSteps;j++) {
						glVertex3d(mo2->rpnts[i][j][0],	mo2->rpnts[i][j][1], mo2->rpnts[i][j][2]); 
						glVertex3d(mo2->rpnts[i-1][j][0],mo2->rpnts[i-1][j][1],mo2->rpnts[i-1][j][2]);
						if(j<(radialSteps-1)){
							glVertex3d(mo2->rpnts[i-1][j+1][0],mo2->rpnts[i-1][j+1][1],mo2->rpnts[i-1][j+1][2]);
							glVertex3d(mo2->rpnts[i][j+1][0],mo2->rpnts[i][j+1][1],mo2->rpnts[i][j+1][2]);
						}
						else {
							glVertex3d(mo2->rpnts[i-1][j-radialSteps+1][0],mo2->rpnts[i-1][j-radialSteps+1][1],mo2->rpnts[i-1][j-radialSteps+1][2] );
							glVertex3d(mo2->rpnts[i][j-radialSteps+1][0],mo2->rpnts[i][j-radialSteps+1][1],mo2->rpnts[i][j-radialSteps+1][2]);
						}
					}
				}
			}
			glEnd();
		}
		else if(mo2->showQuatMap &&	mo2->whichFrame==FRENETSERRET) {
			glLineWidth(1.0f);
			start =	2;
			if(mo2->markFrameCounter ==	0)
				end	= mo2->numPoints-1;
			else
				end	= mo2->markFrame[0];
			glBegin(GL_LINES);
			glColor3f(0.4f,0.4f,0.4f);
			for(h=0; h <= mo2->markFrameCounter; h++) {
				for(i=start; i<end;	i++) {
					for(j=0;j<radialSteps;j++) {
						glVertex3d(mo2->rpnts[i][j][0],	mo2->rpnts[i][j][1], mo2->rpnts[i][j][2]); 
						glVertex3d(mo2->rpnts[i-1][j][0],mo2->rpnts[i-1][j][1],mo2->rpnts[i-1][j][2]);
						if(j<(radialSteps-1)){
							glVertex3d(mo2->rpnts[i-1][j+1][0],mo2->rpnts[i-1][j+1][1],mo2->rpnts[i-1][j+1][2]);
							glVertex3d(mo2->rpnts[i][j+1][0],mo2->rpnts[i][j+1][1],mo2->rpnts[i][j+1][2]);
						}
						else {
							glVertex3d(mo2->rpnts[i-1][j-radialSteps+1][0],mo2->rpnts[i-1][j-radialSteps+1][1],mo2->rpnts[i-1][j-radialSteps+1][2] );
							glVertex3d(mo2->rpnts[i][j-radialSteps+1][0],mo2->rpnts[i][j-radialSteps+1][1],mo2->rpnts[i][j-radialSteps+1][2]);
						}
					}
				}
				start =	mo2->markFrame[h]+2;
				end	= mo2->markFrame[h+1];
				if(h ==	(mo2->markFrameCounter)-1) {
					end	= mo2->numPoints-1;
				}
			}
			glEnd();
		}
		glPopAttrib();
	}	
	/////////////////////////////////////////////////////////////
		
	///////////////////////// NORMALS /////////////////////////
	if(mo2->qshowNormals) {
		glPushAttrib(GL_CURRENT_BIT	| GL_LIGHTING_BIT |	GL_ENABLE_BIT |	GL_LINE_BIT	);
		glDisable(GL_LIGHTING);
		glLineWidth(0.5);
		start =	1;
		if(mo2->showQuatMap	&& mo2->whichFrame==PARALLELTRANSPORT) {
			for(i=start; i<renderPoints; i++) {
				for(j=0; j<radialSteps;	j++) {
					glPushMatrix();
					glTranslatef(mo2->rpnts[i][j][0], mo2->rpnts[i][j][1], mo2->rpnts[i][j][2]);
					glBegin(GL_LINES);
					glColor3f(0.95f,0.95f,0.0f);
					glVertex3d(0.0,0.0,0.0);
					glVertex3d(mo2->npnts[i][j][0],	mo2->npnts[i][j][1], mo2->npnts[i][j][2]);
					glEnd();
					glPopMatrix();
				}
				
			}
		}
		else if(mo2->showQuatMap &&	mo2->whichFrame==FRENETSERRET) {
			start = 0;
			for(i=start; i<renderPoints; i++) {

				for(j=0; j<radialSteps;	j++) {
					glPushMatrix();
					glTranslatef(mo2->rpnts[i][j][0], mo2->rpnts[i][j][1], mo2->rpnts[i][j][2]);
					glBegin(GL_LINES);
					glColor3f(0.95f,0.95f,0.0f);
					glVertex3d(0.0,0.0,0.0);
					glVertex3d(mo2->npnts[i][j][0],	mo2->npnts[i][j][1], mo2->npnts[i][j][2]);
					glEnd();
					glPopMatrix();
				}
			}
		}
		glPopAttrib();
	}
		
	///////////////////////// QUAD FACES /////////////////////////
	if(mo2->qshowFace) {
		glPushAttrib(GL_CURRENT_BIT	| GL_LIGHTING_BIT |	GL_ENABLE_BIT |	GL_LINE_BIT);
		
		glEnable(GL_COLOR_MATERIAL);
		glColor3f(mo2->fgcolor[0],mo2->fgcolor[1],mo2->fgcolor[2]);
		
		glEnable(GL_NORMALIZE);
		glEnable(GL_LIGHTING);
		glDisable(GL_CULL_FACE);
		
		if(mo2->showQuatMap	&& mo2->whichFrame==PARALLELTRANSPORT) {
		
			start =	1;
			glBegin(GL_QUADS);
			for(i=start+1;i<(renderPoints-2); i++) {
				for(j=0;j<radialSteps;j++) {
					glNormal3d(mo2->npnts[i][j][0],	mo2->npnts[i][j][1], mo2->npnts[i][j][2]);
					glVertex3d(mo2->rpnts[i][j][0],	mo2->rpnts[i][j][1], mo2->rpnts[i][j][2]);
					glNormal3d(mo2->npnts[i-1][j][0], mo2->npnts[i-1][j][1], mo2->npnts[i-1][j][2]);
					glVertex3d(mo2->rpnts[i-1][j][0], mo2->rpnts[i-1][j][1], mo2->rpnts[i-1][j][2]);
					if(j<(radialSteps-1)){
						glNormal3d(mo2->npnts[i-1][j+1][0],	mo2->npnts[i-1][j+1][1], mo2->npnts[i-1][j+1][2]);
						glVertex3d(mo2->rpnts[i-1][j+1][0],	mo2->rpnts[i-1][j+1][1], mo2->rpnts[i-1][j+1][2]);
						glNormal3d(mo2->npnts[i][j+1][0],mo2->npnts[i][j+1][1],mo2->npnts[i][j+1][2]);
						glVertex3d(mo2->rpnts[i][j+1][0],mo2->rpnts[i][j+1][1],mo2->rpnts[i][j+1][2]);
					}
					else {
						glNormal3d(mo2->npnts[i-1][j-radialSteps+1][0],	mo2->npnts[i-1][j-radialSteps+1][1], mo2->npnts[i-1][j-radialSteps+1][2] );
						glVertex3d(mo2->rpnts[i-1][j-radialSteps+1][0],	mo2->rpnts[i-1][j-radialSteps+1][1], mo2->rpnts[i-1][j-radialSteps+1][2] );
						glNormal3d(mo2->npnts[i][j-radialSteps+1][0], mo2->npnts[i][j-radialSteps+1][1], mo2->npnts[i][j-radialSteps+1][2]);
						glVertex3d(mo2->rpnts[i][j-radialSteps+1][0], mo2->rpnts[i][j-radialSteps+1][1], mo2->rpnts[i][j-radialSteps+1][2]);
					}
				}
			}
			glEnd();
		
		}//	end	if()
		else if(mo2->showQuatMap &&	mo2->whichFrame	== FRENETSERRET) {

			start =	1;
			if(mo2->markFrameCounter ==	0)
				end	= mo2->numPoints-1;
			else
				end	= mo2->markFrame[0];
			for(h=0; h<=mo2->markFrameCounter; h++)	{
				//fprintf(stderr,"FACES: start = %d, end = %d\n",start,end);				
				//--------------------------->
				int	colnum = h;
				if(h==mo2->markFrameCounter)
					colnum = 0;
				colnum = 0;
				getSegmentColor(colnum,	ccolor);
				glColor3f(ccolor[0],ccolor[1],ccolor[2]);

				glBegin(GL_QUADS);
				for(i =	start+1; i < end; i++) {

					for(j=0;j<radialSteps;j++) {
						glNormal3d(mo2->npnts[i][j][0],	mo2->npnts[i][j][1], mo2->npnts[i][j][2]);
						glVertex3d(mo2->rpnts[i][j][0],	mo2->rpnts[i][j][1], mo2->rpnts[i][j][2]);
						glNormal3d(mo2->npnts[i-1][j][0], mo2->npnts[i-1][j][1], mo2->npnts[i-1][j][2]);
						glVertex3d(mo2->rpnts[i-1][j][0], mo2->rpnts[i-1][j][1], mo2->rpnts[i-1][j][2]);
						if(j<(radialSteps-1)){
							glNormal3d(mo2->npnts[i-1][j+1][0],	mo2->npnts[i-1][j+1][1], mo2->npnts[i-1][j+1][2]);
							glVertex3d(mo2->rpnts[i-1][j+1][0],	mo2->rpnts[i-1][j+1][1], mo2->rpnts[i-1][j+1][2]);
							glNormal3d(mo2->npnts[i][j+1][0],mo2->npnts[i][j+1][1],mo2->npnts[i][j+1][2]);
							glVertex3d(mo2->rpnts[i][j+1][0],mo2->rpnts[i][j+1][1],mo2->rpnts[i][j+1][2]);
						}
						else {
							glNormal3d(mo2->npnts[i-1][j-radialSteps+1][0],	mo2->npnts[i-1][j-radialSteps+1][1], mo2->npnts[i-1][j-radialSteps+1][2] );
							glVertex3d(mo2->rpnts[i-1][j-radialSteps+1][0],	mo2->rpnts[i-1][j-radialSteps+1][1], mo2->rpnts[i-1][j-radialSteps+1][2] );
							glNormal3d(mo2->npnts[i][j-radialSteps+1][0], mo2->npnts[i][j-radialSteps+1][1], mo2->npnts[i][j-radialSteps+1][2]);
							glVertex3d(mo2->rpnts[i][j-radialSteps+1][0], mo2->rpnts[i][j-radialSteps+1][1], mo2->rpnts[i][j-radialSteps+1][2]);
						}
					}
				}
				glEnd();
				//<---------------------------
				start =	mo2->markFrame[h]+1;
				end	= mo2->markFrame[h+1];
				if(h==(mo2->markFrameCounter)-1) {
					end	= mo2->numPoints-1;
				}
			}

		}
		glDisable(GL_TEXTURE_2D);
		glPopAttrib();			  
	}//end if()
}

void renderFramedQuaternionCurve2 ( double radius ) 
{
	int		h			 = 0;
	int		i			 = 0;
	int		j			 = 0;
	int		end			 = 0;
	int		start		 = 0;
	int		renderPoints = 0;
	double	theta		 = 0.0;
	double	dTheta		 = 0.0;

	double	n_radius	 = 2.0*radius;
	int		radialSteps	 = mo2->radialSteps;
	float	mcolor[3];	mcolor[0] = mcolor[1] = mcolor[2] = 0.4f;
	float	mwidth		= 0.0f;
	double	ccolor[3];	ccolor[0] =	ccolor[1] =	ccolor[2] =	0.0;
	double	qvec[3];	qvec[0] = qvec[1] = qvec[2] = 0.0;   	
	quaternion *q,*r;
	q = (quaternion*) my_malloc(sizeof(quaternion));
	r = (quaternion*) my_malloc(sizeof(quaternion));

	/////////////////////////////////////////////////
	// initialize variables
	/////////////////////////////////////////////////
	dTheta = 2.0*M_PI/radialSteps;
	renderPoints = mo2->numPoints;

	if(mo2->showQuatMap)
			start += 1;
	
	if(mo2->qshowFace) {
		mwidth = 2.0f;
		mcolor[0] = mcolor[1] = mcolor[2] = 0.0f;
	}
	else {
		mwidth = 1.25f;
		mcolor[0] = mcolor[1] = mcolor[2] = 0.25f;
	}
		
	/////////////////////////////////////////////////
	// compute frames
	/////////////////////////////////////////////////
	for	( i=start; i<renderPoints; i++ ) {

		theta= 0.0;
		for	( j=0; j<radialSteps; j++ )	{
			
			mo2->rpnts[i][j][0]	= mo2->pntList[i].p3.d[0] +	radius*mo2->basisRList[i].p3.d[0]*cos(theta) + radius*mo2->basisSList[i].p3.d[0]*sin(theta);
			mo2->rpnts[i][j][1]	= mo2->pntList[i].p3.d[1] +	radius*mo2->basisRList[i].p3.d[1]*cos(theta) + radius*mo2->basisSList[i].p3.d[1]*sin(theta);
			mo2->rpnts[i][j][2]	= mo2->pntList[i].p3.d[2] +	radius*mo2->basisRList[i].p3.d[2]*cos(theta) + radius*mo2->basisSList[i].p3.d[2]*sin(theta);
			
			mo2->npnts[i][j][0]	= n_radius*mo2->basisRList[i].p3.d[0]*cos(theta)+n_radius*mo2->basisSList[i].p3.d[0]*sin(theta);
			mo2->npnts[i][j][1]	= n_radius*mo2->basisRList[i].p3.d[1]*cos(theta)+n_radius*mo2->basisSList[i].p3.d[1]*sin(theta);
			mo2->npnts[i][j][2]	= n_radius*mo2->basisRList[i].p3.d[2]*cos(theta)+n_radius*mo2->basisSList[i].p3.d[2]*sin(theta);
			
			theta =	theta +	dTheta;
		}
	}
	
	///////////////////////// MESH /////////////////////////
	if(mo2->qshowEdge) {
		glPushAttrib(GL_CURRENT_BIT	| GL_LIGHTING_BIT |	GL_ENABLE_BIT |	GL_LINE_BIT);
		glDisable(GL_LIGHTING);
		
		glLineWidth(2.0);
 
		///////////////////////// RING /////////////////////////
		if(mo2->showQuatMap	&& mo2->whichFrame==PARALLELTRANSPORT) {
			start=1;
			for(i=start; i<(renderPoints-2); i++) {
				
				glColor3fv(mcolor);
				glLineWidth(mwidth);
				glBegin(GL_LINE_STRIP);				
				for(j=0; j<=radialSteps; j++) {
					if(j<(radialSteps))
						glVertex3d(mo2->rpnts[i][j][0],mo2->rpnts[i][j][1],mo2->rpnts[i][j][2]);
					else
						glVertex3d(mo2->rpnts[i][0][0],mo2->rpnts[i][0][1],mo2->rpnts[i][0][2]);
				}
				glEnd();
			}
		}
		else if(mo2->showQuatMap &&	mo2->whichFrame==FRENETSERRET) {
			glColor3fv(mcolor);
			glLineWidth(mwidth);
			start =	1;
			if(mo2->markQuatCurveCounter==	0)
				end	= mo2->numPoints-2;
			else
				end	= mo2->markQuatCurve[0];
			//printf("Quat counter = %d\n",mo2->markQuatCurveCounter);

			for(h=0; h<=mo2->markQuatCurveCounter; h++)	{
				//--------------------------->
				glBegin(GL_LINE_STRIP);
				for(i=start; i<end;	i++) {
					for(j=0; j<=radialSteps; j++) {
						if(j<(radialSteps))
							glVertex3d(mo2->rpnts[i][j][0],mo2->rpnts[i][j][1],mo2->rpnts[i][j][2]);
						else
							glVertex3d(mo2->rpnts[i][0][0],mo2->rpnts[i][0][1],mo2->rpnts[i][0][2]);
					}
				}
				glEnd();
				//<---------------------------
				start =	mo2->markQuatCurve[h]+1;
				end	= mo2->markQuatCurve[h+1];
				if(h==(mo2->markQuatCurveCounter)-1) {
					end	= mo2->numPoints-1;
				}
			}
		}

		///////////////////////// LONGITUDINAL LINES /////////////////////////
		if(mo2->showQuatMap	&& mo2->whichFrame==PARALLELTRANSPORT) {
			glColor3fv(mcolor);
			glLineWidth(mwidth);
			glBegin(GL_LINES);
			start = 1;
			if(mo2->markFrameCounter ==	0)
				end	= mo2->numPoints-2;
			else
				end	= mo2->markFrame[0]-2;

			for(i=start+1;i<(renderPoints-1); i++) {
				if(	i<(renderPoints-2) ) {
					for(j=0;j<radialSteps;j++) {
						glVertex3d(mo2->rpnts[i][j][0],	mo2->rpnts[i][j][1], mo2->rpnts[i][j][2]); 
						glVertex3d(mo2->rpnts[i-1][j][0],mo2->rpnts[i-1][j][1],mo2->rpnts[i-1][j][2]);
						if(j<(radialSteps-1)){
							glVertex3d(mo2->rpnts[i-1][j+1][0],mo2->rpnts[i-1][j+1][1],mo2->rpnts[i-1][j+1][2]);
							glVertex3d(mo2->rpnts[i][j+1][0],mo2->rpnts[i][j+1][1],mo2->rpnts[i][j+1][2]);
						}
						else {
							glVertex3d(mo2->rpnts[i-1][j-radialSteps+1][0],mo2->rpnts[i-1][j-radialSteps+1][1],mo2->rpnts[i-1][j-radialSteps+1][2] );
							glVertex3d(mo2->rpnts[i][j-radialSteps+1][0],mo2->rpnts[i][j-radialSteps+1][1],mo2->rpnts[i][j-radialSteps+1][2]);
						}
					}
				}
			}
			glEnd();

		}
		else if(mo2->showQuatMap &&	mo2->whichFrame==FRENETSERRET) {
			glColor3fv(mcolor);
			glLineWidth(mwidth);
			start =	1;
			if(mo2->markQuatCurveCounter==	0)
				end	= mo2->numPoints-2;
			else
				end	= mo2->markQuatCurve[0];
			//printf("Quat counter = %d\n",mo2->markQuatCurveCounter);

			for(h=0; h<=mo2->markQuatCurveCounter; h++)	{
				//--------------------------->
				int	colnum = h;
				if(h==mo2->markFrameCounter)
					colnum = 0;
				colnum = 0;
				getSegmentColor(colnum,	ccolor);
				glBegin(GL_LINES);
				if(h==0)
					start = 1;
				for(i=start+1; i<end;	i++) {
					for(j=0;j<radialSteps;j++) {
						glVertex3d(mo2->rpnts[i][j][0],	mo2->rpnts[i][j][1], mo2->rpnts[i][j][2]); 
						glVertex3d(mo2->rpnts[i-1][j][0],mo2->rpnts[i-1][j][1],mo2->rpnts[i-1][j][2]);
						if(j<(radialSteps-1)){
							glVertex3d(mo2->rpnts[i-1][j+1][0],mo2->rpnts[i-1][j+1][1],mo2->rpnts[i-1][j+1][2]);
							glVertex3d(mo2->rpnts[i][j+1][0],mo2->rpnts[i][j+1][1],mo2->rpnts[i][j+1][2]);
						}
						else {
							glVertex3d(mo2->rpnts[i-1][j-radialSteps+1][0],mo2->rpnts[i-1][j-radialSteps+1][1],mo2->rpnts[i-1][j-radialSteps+1][2] );
							glVertex3d(mo2->rpnts[i][j-radialSteps+1][0],mo2->rpnts[i][j-radialSteps+1][1],mo2->rpnts[i][j-radialSteps+1][2]);
						}
					}
				}
				glEnd();
				//<---------------------------
				start =	mo2->markQuatCurve[h]+1;
				end	= mo2->markQuatCurve[h+1];
				if(h==(mo2->markQuatCurveCounter)-1) {
					end	= mo2->numPoints-1;
				}
			}
		}
		glPopAttrib();
	}	
	/////////////////////////////////////////////////////////////
		
	///////////////////////// NORMALS /////////////////////////
	if(mo2->qshowNormals) {
		glPushAttrib(GL_CURRENT_BIT	| GL_LIGHTING_BIT |	GL_ENABLE_BIT |	GL_LINE_BIT	);
		glDisable(GL_LIGHTING);
		glLineWidth(0.5);
		start =	1;
		if(mo2->showQuatMap	&& mo2->whichFrame==PARALLELTRANSPORT) {
			for(i=start; i<renderPoints; i++) {
				for(j=0; j<radialSteps;	j++) {
					glPushMatrix();
					glTranslatef(mo2->rpnts[i][j][0], mo2->rpnts[i][j][1], mo2->rpnts[i][j][2]);
					glBegin(GL_LINES);
					glColor3f(0.95f,0.95f,0.0f);
					glVertex3d(0.0,0.0,0.0);
					glVertex3d(mo2->npnts[i][j][0],	mo2->npnts[i][j][1], mo2->npnts[i][j][2]);
					glEnd();
					glPopMatrix();
				}
				
			}
		}
		else if(mo2->showQuatMap &&	mo2->whichFrame==FRENETSERRET) {
			start = 0;
			for(i=start; i<renderPoints; i++) {

				for(j=0; j<radialSteps;	j++) {
					glPushMatrix();
					glTranslatef(mo2->rpnts[i][j][0], mo2->rpnts[i][j][1], mo2->rpnts[i][j][2]);
					glBegin(GL_LINES);
					glColor3f(0.95f,0.95f,0.0f);
					glVertex3d(0.0,0.0,0.0);
					glVertex3d(mo2->npnts[i][j][0],	mo2->npnts[i][j][1], mo2->npnts[i][j][2]);
					glEnd();
					glPopMatrix();
				}
			}
		}
		glPopAttrib();
	}
		
	///////////////////////// QUAD FACES /////////////////////////
	if(mo2->qshowFace) {
		glPushAttrib(GL_CURRENT_BIT	| GL_LIGHTING_BIT |	GL_ENABLE_BIT |	GL_LINE_BIT);
		
		glEnable(GL_COLOR_MATERIAL);
		glColor3f(mo2->fgcolor[0],mo2->fgcolor[1],mo2->fgcolor[2]);
		
		glEnable(GL_NORMALIZE);
		glEnable(GL_LIGHTING);
		glDisable(GL_CULL_FACE);
		
		if(mo2->showQuatMap	&& mo2->whichFrame==PARALLELTRANSPORT) {
		
			start =	1;
			glBegin(GL_QUADS);
			for(i=start+1;i<(renderPoints-2); i++) {
				for(j=0;j<radialSteps;j++) {
					glNormal3d(mo2->npnts[i][j][0],	mo2->npnts[i][j][1], mo2->npnts[i][j][2]);
					glVertex3d(mo2->rpnts[i][j][0],	mo2->rpnts[i][j][1], mo2->rpnts[i][j][2]);
					glNormal3d(mo2->npnts[i-1][j][0], mo2->npnts[i-1][j][1], mo2->npnts[i-1][j][2]);
					glVertex3d(mo2->rpnts[i-1][j][0], mo2->rpnts[i-1][j][1], mo2->rpnts[i-1][j][2]);
					if(j<(radialSteps-1)){
						glNormal3d(mo2->npnts[i-1][j+1][0],	mo2->npnts[i-1][j+1][1], mo2->npnts[i-1][j+1][2]);
						glVertex3d(mo2->rpnts[i-1][j+1][0],	mo2->rpnts[i-1][j+1][1], mo2->rpnts[i-1][j+1][2]);
						glNormal3d(mo2->npnts[i][j+1][0],mo2->npnts[i][j+1][1],mo2->npnts[i][j+1][2]);
						glVertex3d(mo2->rpnts[i][j+1][0],mo2->rpnts[i][j+1][1],mo2->rpnts[i][j+1][2]);
					}
					else {
						glNormal3d(mo2->npnts[i-1][j-radialSteps+1][0],	mo2->npnts[i-1][j-radialSteps+1][1], mo2->npnts[i-1][j-radialSteps+1][2] );
						glVertex3d(mo2->rpnts[i-1][j-radialSteps+1][0],	mo2->rpnts[i-1][j-radialSteps+1][1], mo2->rpnts[i-1][j-radialSteps+1][2] );
						glNormal3d(mo2->npnts[i][j-radialSteps+1][0], mo2->npnts[i][j-radialSteps+1][1], mo2->npnts[i][j-radialSteps+1][2]);
						glVertex3d(mo2->rpnts[i][j-radialSteps+1][0], mo2->rpnts[i][j-radialSteps+1][1], mo2->rpnts[i][j-radialSteps+1][2]);
					}
				}
			}
			glEnd();
		
		}//	end	if()
		else if(mo2->showQuatMap &&	mo2->whichFrame	== FRENETSERRET) {

			start =	1;
			if(mo2->markQuatCurveCounter==	0)
				end	= mo2->numPoints-2;
			else
				end	= mo2->markQuatCurve[0];

			for(h=0; h<=mo2->markQuatCurveCounter; h++)	{
				//--------------------------->
				getSegmentColor(0,	ccolor);
				glColor3f(ccolor[0],ccolor[1],ccolor[2]);

				glBegin(GL_QUADS);
				if(h==0)
					start = 1;
				for(i =	start+1; i<end; i++) {

					for(j=0;j<radialSteps;j++) {
						glNormal3d(mo2->npnts[i][j][0],	mo2->npnts[i][j][1], mo2->npnts[i][j][2]);
						glVertex3d(mo2->rpnts[i][j][0],	mo2->rpnts[i][j][1], mo2->rpnts[i][j][2]);
						glNormal3d(mo2->npnts[i-1][j][0], mo2->npnts[i-1][j][1], mo2->npnts[i-1][j][2]);
						glVertex3d(mo2->rpnts[i-1][j][0], mo2->rpnts[i-1][j][1], mo2->rpnts[i-1][j][2]);
						if(j<(radialSteps-1)){
							glNormal3d(mo2->npnts[i-1][j+1][0],	mo2->npnts[i-1][j+1][1], mo2->npnts[i-1][j+1][2]);
							glVertex3d(mo2->rpnts[i-1][j+1][0],	mo2->rpnts[i-1][j+1][1], mo2->rpnts[i-1][j+1][2]);
							glNormal3d(mo2->npnts[i][j+1][0],mo2->npnts[i][j+1][1],mo2->npnts[i][j+1][2]);
							glVertex3d(mo2->rpnts[i][j+1][0],mo2->rpnts[i][j+1][1],mo2->rpnts[i][j+1][2]);
						}
						else {
							glNormal3d(mo2->npnts[i-1][j-radialSteps+1][0],	mo2->npnts[i-1][j-radialSteps+1][1], mo2->npnts[i-1][j-radialSteps+1][2] );
							glVertex3d(mo2->rpnts[i-1][j-radialSteps+1][0],	mo2->rpnts[i-1][j-radialSteps+1][1], mo2->rpnts[i-1][j-radialSteps+1][2] );
							glNormal3d(mo2->npnts[i][j-radialSteps+1][0], mo2->npnts[i][j-radialSteps+1][1], mo2->npnts[i][j-radialSteps+1][2]);
							glVertex3d(mo2->rpnts[i][j-radialSteps+1][0], mo2->rpnts[i][j-radialSteps+1][1], mo2->rpnts[i][j-radialSteps+1][2]);
						}
					}
				}
				glEnd();
				//<---------------------------
				start =	mo2->markQuatCurve[h]+1;
				end	= mo2->markQuatCurve[h+1];
				if(h==(mo2->markQuatCurveCounter)-1) {
					end	= mo2->numPoints-1;
				}
			}

			//start =	1;
			//if(mo2->markQuatCurveCounter2==	0)
			//	end	= mo2->numPoints-2;
			//else
			//	end	= mo2->markQuatCurve2[0];
			//printf("Quat counter = %d\n",mo2->markQuatCurveCounter2);

			//for(h=0; h<=mo2->markQuatCurveCounter2; h++)	{
			//	//--------------------------->
			//	int	colnum = h;
			//	if(h==mo2->markQuatCurveCounter2)
			//		colnum = 0;

			//	getSegmentColor(colnum,	ccolor);
			//	printf("colnum= %d,ccolor = %f, %f, %f\n",h,ccolor[0],ccolor[1],ccolor[2]);
			//	glColor3f(ccolor[0],ccolor[1],ccolor[2]);

			//	glBegin(GL_QUADS);
			//	if(h==0)
			//		start = 1;
			//	for(i =	start+1; i<end; i++) {

			//		for(j=0;j<radialSteps;j++) {
			//			glNormal3d(mo2->npnts[i][j][0],	mo2->npnts[i][j][1], mo2->npnts[i][j][2]);
			//			glVertex3d(mo2->rpnts[i][j][0],	mo2->rpnts[i][j][1], mo2->rpnts[i][j][2]);
			//			glNormal3d(mo2->npnts[i-1][j][0], mo2->npnts[i-1][j][1], mo2->npnts[i-1][j][2]);
			//			glVertex3d(mo2->rpnts[i-1][j][0], mo2->rpnts[i-1][j][1], mo2->rpnts[i-1][j][2]);
			//			if(j<(radialSteps-1)){
			//				glNormal3d(mo2->npnts[i-1][j+1][0],	mo2->npnts[i-1][j+1][1], mo2->npnts[i-1][j+1][2]);
			//				glVertex3d(mo2->rpnts[i-1][j+1][0],	mo2->rpnts[i-1][j+1][1], mo2->rpnts[i-1][j+1][2]);
			//				glNormal3d(mo2->npnts[i][j+1][0],mo2->npnts[i][j+1][1],mo2->npnts[i][j+1][2]);
			//				glVertex3d(mo2->rpnts[i][j+1][0],mo2->rpnts[i][j+1][1],mo2->rpnts[i][j+1][2]);
			//			}
			//			else {
			//				glNormal3d(mo2->npnts[i-1][j-radialSteps+1][0],	mo2->npnts[i-1][j-radialSteps+1][1], mo2->npnts[i-1][j-radialSteps+1][2] );
			//				glVertex3d(mo2->rpnts[i-1][j-radialSteps+1][0],	mo2->rpnts[i-1][j-radialSteps+1][1], mo2->rpnts[i-1][j-radialSteps+1][2] );
			//				glNormal3d(mo2->npnts[i][j-radialSteps+1][0], mo2->npnts[i][j-radialSteps+1][1], mo2->npnts[i][j-radialSteps+1][2]);
			//				glVertex3d(mo2->rpnts[i][j-radialSteps+1][0], mo2->rpnts[i][j-radialSteps+1][1], mo2->rpnts[i][j-radialSteps+1][2]);
			//			}
			//		}
			//	}
			//	glEnd();
			//	//<---------------------------
			//	start =	mo2->markQuatCurve2[h]+1;
			//	end	= mo2->markQuatCurve2[h+1];
			//	if(h==(mo2->markQuatCurveCounter2)-1) {
			//		end	= mo2->numPoints-1;
			//	}
			//}

		}
		glDisable(GL_TEXTURE_2D);
		glPopAttrib();			  
	}//end if()

	if(mo2->qshowCloudPoints && mo2->whichFrame==FRENETSERRET) {
		glPushAttrib(GL_ALL_ATTRIB_BITS);
			//glEnable(GL_LIGHTING);
			//glEnable(GL_COLOR_MATERIAL);
			////# Original points in RED
			//glColor3f(1.0f,0.0f,0.0f);
			//for(i=0;i<mo2->numPoints;i++) {
			//	q->r = -(mo2->qpntFliped[i]->r);
			//	q->i = -(mo2->qpntFliped[i]->i);
			//	q->j = -(mo2->qpntFliped[i]->j);
			//	q->k = -(mo2->qpntFliped[i]->k);

			//	getQuatComp(mo2->qwhichQuatComponent,q,qvec);
			//	glPushMatrix();
			//		glTranslatef(qvec[0],qvec[1],qvec[2]);
			//		glutSolidSphere(0.03,16,16);
			//	glPopMatrix();
			//}

			////# Flipped points in GREEN
			//glColor3f(1.0f,1.0f,0.0f);
			//for(i=0;i<mo2->numPoints;i++) {
			//	getQuatComp(mo2->qwhichQuatComponent,mo2->qpntFliped[i],qvec);
			//	glPushMatrix();
			//		glTranslatef(qvec[0],qvec[1],qvec[2]);
			//		glutSolidSphere(0.03,16,16);
			//	glPopMatrix();
			//}

			////# Original points in RED
			//glColor3f(0.8f,0.0f,0.0f);
			//start =	0;
			//if(mo2->markQuatCurveCounter==	0)
			//	end	= mo2->numPoints-2;
			//else
			//	end	= mo2->markQuatCurve[0];

			//for(h=0; h<=mo2->markQuatCurveCounter; h++)	{
			//	//--------------------------->
			//	if(h==0)
			//		start = 1;
			//	for(i =	start; i<end; i++) {
			//		getQuatComp(mo2->qwhichQuatComponent,mo2->qpntOrig[i],qvec);
			//		glPushMatrix();
			//			glTranslatef(qvec[0],qvec[1],qvec[2]);
			//			glutSolidSphere(0.03,16,16);
			//		glPopMatrix();
			//	}
			//	//<---------------------------
			//	start =	mo2->markQuatCurve[h]+1;
			//	end	= mo2->markQuatCurve[h+1];
			//	if(h==(mo2->markQuatCurveCounter)-1) {
			//		end	= mo2->numPoints-1;
			//	}
			//}

			////# Corrected Curve Points in GREEN
			//start =	0;
			//if(mo2->markQuatCurveCounter==	0)
			//	end	= mo2->numPoints-2;
			//else
			//	end	= mo2->markQuatCurve[0];

			//for(h=0; h<=mo2->markQuatCurveCounter; h++)	{
			//	//--------------------------->
			//	if(h==0)
			//		start = 1;
			//	for(i =	start; i<end; i++) {
			//		
			//		getQuatComp(mo2->qwhichQuatComponent,mo2->qpntFliped[i],qvec);
			//		glPushMatrix();
			//			glColor3f(0.0f,0.7f,0.0f);
			//			glTranslatef(qvec[0],qvec[1],qvec[2]);
			//			glutSolidSphere(0.03,16,16);
			//			//glPushMatrix();
			//			//	glTranslatef(delta,delta,delta);
			//			//	gluOrtho2D(-0.1,0.1,-0.1,0.1);
			//			//	glDisable(GL_LIGHTING);
			//			//	glColor3f(0.9f,0.9f,0.9f);
			//			//	sprintf(obj_qstr,"%d",i);
			//			//	glRasterPos2f(0.0,0.0);
			//			//	stext(obj_qstr);
			//			//	glEnable(GL_LIGHTING);
			//			//glPopMatrix();
			//		glPopMatrix();
			//	}
			//	//<---------------------------
			//	start =	mo2->markQuatCurve[h]+1;
			//	end	= mo2->markQuatCurve[h+1];
			//	if(h==(mo2->markQuatCurveCounter)-1) {
			//		end	= mo2->numPoints-1;
			//	}
			//}

			////# Points that will be flipped in ORANGE
			//glColor3f(0.7f,0.5f,0.0f);
			//for(h=0; h<mo2->qflipcounter;) {
			//	for(i=mo2->qflipindex[h];i<=mo2->qflipindex[h+1];i++) {
			//		getQuatComp(mo2->qwhichQuatComponent,mo2->qpntFliped[i],qvec);
			//		glPushMatrix();
			//			glTranslatef(qvec[0],qvec[1],qvec[2]);
			//			glutSolidSphere(0.03,16,16);
			//		glPopMatrix();
			//	}
			//	h = h+2;
			//}
		glPopAttrib();
	}
	if(mo2->qshowSphere) {
		glPushAttrib(GL_ALL_ATTRIB_BITS);
			glEnable(GL_BLEND);
			glDisable(GL_LIGHTING);
			glColor4f(0.7f,0.7f,0.7f,0.1f);
			glutWireSphere(1.0f,32,32);
		glPopAttrib();
	}
	if(mo2->qshowAxes) {
		double len = 0.5;
		double rad = 0.025;
	    
		GLUquadricObj *qobj;
		qobj = gluNewQuadric();

		glPushAttrib(GL_CURRENT_BIT | GL_LIGHTING_BIT);
			glDisable(GL_LIGHTING);
			glPushMatrix();
				glTranslatef(0.0,0.0,0.0);
				glColor3f(0.0, 0.0, 1.0);
				glRotatef(-90,1,0,0);
				gluQuadricDrawStyle(qobj, GLU_FILL);
				gluCylinder(qobj, rad, rad, len, 16, 16);
			glPopMatrix();
			glPushMatrix();
				glTranslatef(0.0,0.0,0.0);
				glColor3f(0.0, 1.0, 0.0);
				glRotatef(90,0,1,0);
				gluQuadricDrawStyle(qobj, GLU_FILL);
				gluCylinder(qobj, rad, rad, len, 16, 16);
			glPopMatrix();
			glPushMatrix();
				glTranslatef(0.0,0.0,0.0);
				glColor3f(1.0, 0.0, 0.0);
				glRotatef(0,0,0,1);
				gluQuadricDrawStyle(qobj, GLU_FILL);
				gluCylinder(qobj, rad, rad, len, 16, 16);
			glPopMatrix();
		glPopAttrib();
	}
}

void renderFramedQuaternionCurve3 ( double radius, double *color, int flag_use_extern_color ) 
{
	int		h			 = 0;
	int		i			 = 0;
	int		j			 = 0;
	int		end			 = 0;
	int		start		 = 0;
	int		renderPoints = 0;
	double	theta		 = 0.0;
	double	dTheta		 = 0.0;
	double	n_radius	 = 2.0*radius;
	int		radialSteps	 = mo2->radialSteps;
	float	mcolor[3];	mcolor[0] = mcolor[1] = mcolor[2] = 0.4f;
	float	mwidth		= 0.0f;
	double	ccolor[3];	ccolor[0] =	ccolor[1] =	ccolor[2] =	0.0;
	double	qvec[3];	qvec[0] = qvec[1] = qvec[2] = 0.0;   	
	quaternion *q,*r;
	q = (quaternion*) my_malloc(sizeof(quaternion));
	r = (quaternion*) my_malloc(sizeof(quaternion));

	/////////////////////////////////////////////////
	// initialize variables
	/////////////////////////////////////////////////
	dTheta = 2.0*M_PI/radialSteps;
	renderPoints = mo2->numPoints;

	if(mo2->showQuatMap)
			start += 1;
	
	if(mo2->qshowFace) {
		mwidth = 2.0f;
		mcolor[0] = mcolor[1] = mcolor[2] = 0.0f;
	}
	else {
		mwidth = 1.25f;
		mcolor[0] = mcolor[1] = mcolor[2] = 0.25f;
	}
		
	/////////////////////////////////////////////////
	// compute frames
	/////////////////////////////////////////////////
	for	( i=start; i<renderPoints; i++ ) {

		theta= 0.0;
		for	( j=0; j<radialSteps; j++ )	{
			
			mo2->rpnts[i][j][0]	= mo2->pntList[i].p3.d[0] +	radius*mo2->basisRList[i].p3.d[0]*cos(theta) + radius*mo2->basisSList[i].p3.d[0]*sin(theta);
			mo2->rpnts[i][j][1]	= mo2->pntList[i].p3.d[1] +	radius*mo2->basisRList[i].p3.d[1]*cos(theta) + radius*mo2->basisSList[i].p3.d[1]*sin(theta);
			mo2->rpnts[i][j][2]	= mo2->pntList[i].p3.d[2] +	radius*mo2->basisRList[i].p3.d[2]*cos(theta) + radius*mo2->basisSList[i].p3.d[2]*sin(theta);
			
			mo2->npnts[i][j][0]	= n_radius*mo2->basisRList[i].p3.d[0]*cos(theta)+n_radius*mo2->basisSList[i].p3.d[0]*sin(theta);
			mo2->npnts[i][j][1]	= n_radius*mo2->basisRList[i].p3.d[1]*cos(theta)+n_radius*mo2->basisSList[i].p3.d[1]*sin(theta);
			mo2->npnts[i][j][2]	= n_radius*mo2->basisRList[i].p3.d[2]*cos(theta)+n_radius*mo2->basisSList[i].p3.d[2]*sin(theta);
			
			theta =	theta +	dTheta;
		}
	}
	
	///////////////////////// MESH /////////////////////////
	if(mo2->qshowEdge) {
		glPushAttrib(GL_CURRENT_BIT	| GL_LIGHTING_BIT |	GL_ENABLE_BIT |	GL_LINE_BIT);
		glDisable(GL_LIGHTING);
		
		glLineWidth(2.0);
 
		///////////////////////// RING /////////////////////////
		if(mo2->showQuatMap	&& mo2->whichFrame==PARALLELTRANSPORT) {
			start=1;
			for(i=start; i<(renderPoints-2); i++) {
				
				glColor3fv(mcolor);
				glLineWidth(mwidth);
				glBegin(GL_LINE_STRIP);				
				for(j=0; j<=radialSteps; j++) {
					if(j<(radialSteps))
						glVertex3d(mo2->rpnts[i][j][0],mo2->rpnts[i][j][1],mo2->rpnts[i][j][2]);
					else
						glVertex3d(mo2->rpnts[i][0][0],mo2->rpnts[i][0][1],mo2->rpnts[i][0][2]);
				}
				glEnd();
			}
		}
		else if(mo2->showQuatMap &&	mo2->whichFrame==FRENETSERRET) {
			glColor3fv(mcolor);
			glLineWidth(mwidth);
			start =	1;
			if(mo2->markQuatCurveCounter==	0)
				end	= mo2->numPoints-2;
			else
				end	= mo2->markQuatCurve[0];
			//printf("Quat counter = %d\n",mo2->markQuatCurveCounter);

			for(h=0; h<=mo2->markQuatCurveCounter; h++)	{
				//--------------------------->
				glBegin(GL_LINE_STRIP);
				for(i=start; i<end;	i++) {
					for(j=0; j<=radialSteps; j++) {
						if(j<(radialSteps))
							glVertex3d(mo2->rpnts[i][j][0],mo2->rpnts[i][j][1],mo2->rpnts[i][j][2]);
						else
							glVertex3d(mo2->rpnts[i][0][0],mo2->rpnts[i][0][1],mo2->rpnts[i][0][2]);
					}
				}
				glEnd();
				//<---------------------------
				start =	mo2->markQuatCurve[h]+1;
				end	= mo2->markQuatCurve[h+1];
				if(h==(mo2->markQuatCurveCounter)-1) {
					end	= mo2->numPoints-1;
				}
			}
		}

		///////////////////////// LONGITUDINAL LINES /////////////////////////
		if(mo2->showQuatMap	&& mo2->whichFrame==PARALLELTRANSPORT) {
			glColor3fv(mcolor);
			glLineWidth(mwidth);
			glBegin(GL_LINES);
			start = 1;
			if(mo2->markFrameCounter ==	0)
				end	= mo2->numPoints-2;
			else
				end	= mo2->markFrame[0]-2;

			for(i=start+1;i<(renderPoints-1); i++) {
				if(	i<(renderPoints-2) ) {
					for(j=0;j<radialSteps;j++) {
						glVertex3d(mo2->rpnts[i][j][0],	mo2->rpnts[i][j][1], mo2->rpnts[i][j][2]); 
						glVertex3d(mo2->rpnts[i-1][j][0],mo2->rpnts[i-1][j][1],mo2->rpnts[i-1][j][2]);
						if(j<(radialSteps-1)){
							glVertex3d(mo2->rpnts[i-1][j+1][0],mo2->rpnts[i-1][j+1][1],mo2->rpnts[i-1][j+1][2]);
							glVertex3d(mo2->rpnts[i][j+1][0],mo2->rpnts[i][j+1][1],mo2->rpnts[i][j+1][2]);
						}
						else {
							glVertex3d(mo2->rpnts[i-1][j-radialSteps+1][0],mo2->rpnts[i-1][j-radialSteps+1][1],mo2->rpnts[i-1][j-radialSteps+1][2] );
							glVertex3d(mo2->rpnts[i][j-radialSteps+1][0],mo2->rpnts[i][j-radialSteps+1][1],mo2->rpnts[i][j-radialSteps+1][2]);
						}
					}
				}
			}
			glEnd();

		}
		else if(mo2->showQuatMap &&	mo2->whichFrame==FRENETSERRET) {
			glColor3fv(mcolor);
			glLineWidth(mwidth);
			start =	1;
			if(mo2->markQuatCurveCounter==	0)
				end	= mo2->numPoints-2;
			else
				end	= mo2->markQuatCurve[0];
			//printf("Quat counter = %d\n",mo2->markQuatCurveCounter);

			for(h=0; h<=mo2->markQuatCurveCounter; h++)	{
				//--------------------------->
				int	colnum = h;
				if(h==mo2->markFrameCounter)
					colnum = 0;
				colnum = 0;
				getSegmentColor(colnum,	ccolor);
				glBegin(GL_LINES);
				if(h==0)
					start = 1;
				for(i=start+1; i<end;	i++) {
					for(j=0;j<radialSteps;j++) {
						glVertex3d(mo2->rpnts[i][j][0],	mo2->rpnts[i][j][1], mo2->rpnts[i][j][2]); 
						glVertex3d(mo2->rpnts[i-1][j][0],mo2->rpnts[i-1][j][1],mo2->rpnts[i-1][j][2]);
						if(j<(radialSteps-1)){
							glVertex3d(mo2->rpnts[i-1][j+1][0],mo2->rpnts[i-1][j+1][1],mo2->rpnts[i-1][j+1][2]);
							glVertex3d(mo2->rpnts[i][j+1][0],mo2->rpnts[i][j+1][1],mo2->rpnts[i][j+1][2]);
						}
						else {
							glVertex3d(mo2->rpnts[i-1][j-radialSteps+1][0],mo2->rpnts[i-1][j-radialSteps+1][1],mo2->rpnts[i-1][j-radialSteps+1][2] );
							glVertex3d(mo2->rpnts[i][j-radialSteps+1][0],mo2->rpnts[i][j-radialSteps+1][1],mo2->rpnts[i][j-radialSteps+1][2]);
						}
					}
				}
				glEnd();
				//<---------------------------
				start =	mo2->markQuatCurve[h]+1;
				end	= mo2->markQuatCurve[h+1];
				if(h==(mo2->markQuatCurveCounter)-1) {
					end	= mo2->numPoints-1;
				}
			}
		}
		glPopAttrib();
	}	
	/////////////////////////////////////////////////////////////
		
	///////////////////////// NORMALS /////////////////////////
	if(mo2->qshowNormals) {
		glPushAttrib(GL_CURRENT_BIT	| GL_LIGHTING_BIT |	GL_ENABLE_BIT |	GL_LINE_BIT	);
		glDisable(GL_LIGHTING);
		glLineWidth(0.5);
		start =	1;
		if(mo2->showQuatMap	&& mo2->whichFrame==PARALLELTRANSPORT) {
			for(i=start; i<renderPoints; i++) {
				for(j=0; j<radialSteps;	j++) {
					glPushMatrix();
					glTranslatef(mo2->rpnts[i][j][0], mo2->rpnts[i][j][1], mo2->rpnts[i][j][2]);
					glBegin(GL_LINES);
					glColor3f(0.95f,0.95f,0.0f);
					glVertex3d(0.0,0.0,0.0);
					glVertex3d(mo2->npnts[i][j][0],	mo2->npnts[i][j][1], mo2->npnts[i][j][2]);
					glEnd();
					glPopMatrix();
				}
				
			}
		}
		else if(mo2->showQuatMap &&	mo2->whichFrame==FRENETSERRET) {
			start = 0;
			for(i=start; i<renderPoints; i++) {

				for(j=0; j<radialSteps;	j++) {
					glPushMatrix();
					glTranslatef(mo2->rpnts[i][j][0], mo2->rpnts[i][j][1], mo2->rpnts[i][j][2]);
					glBegin(GL_LINES);
					glColor3f(0.95f,0.95f,0.0f);
					glVertex3d(0.0,0.0,0.0);
					glVertex3d(mo2->npnts[i][j][0],	mo2->npnts[i][j][1], mo2->npnts[i][j][2]);
					glEnd();
					glPopMatrix();
				}
			}
		}
		glPopAttrib();
	}
		
	///////////////////////// QUAD FACES /////////////////////////
	if(mo2->qshowFace) {
		glPushAttrib(GL_CURRENT_BIT	| GL_LIGHTING_BIT |	GL_ENABLE_BIT |	GL_LINE_BIT);
		
		glEnable(GL_COLOR_MATERIAL);
		
		glEnable(GL_NORMALIZE);
		glEnable(GL_LIGHTING);
		glDisable(GL_CULL_FACE);
		
		if(mo2->showQuatMap	&& mo2->whichFrame==PARALLELTRANSPORT) {
		
			if(flag_use_extern_color)
				glColor3f(color[0],color[1],color[2]);
			else
				glColor3f(mo2->fgcolor[0],mo2->fgcolor[1],mo2->fgcolor[2]);
			start =	1;
			glBegin(GL_QUADS);
			for(i=start+1;i<(renderPoints-2); i++) {
				for(j=0;j<radialSteps;j++) {
					glNormal3d(mo2->npnts[i][j][0],	mo2->npnts[i][j][1], mo2->npnts[i][j][2]);
					glVertex3d(mo2->rpnts[i][j][0],	mo2->rpnts[i][j][1], mo2->rpnts[i][j][2]);
					glNormal3d(mo2->npnts[i-1][j][0], mo2->npnts[i-1][j][1], mo2->npnts[i-1][j][2]);
					glVertex3d(mo2->rpnts[i-1][j][0], mo2->rpnts[i-1][j][1], mo2->rpnts[i-1][j][2]);
					if(j<(radialSteps-1)){
						glNormal3d(mo2->npnts[i-1][j+1][0],	mo2->npnts[i-1][j+1][1], mo2->npnts[i-1][j+1][2]);
						glVertex3d(mo2->rpnts[i-1][j+1][0],	mo2->rpnts[i-1][j+1][1], mo2->rpnts[i-1][j+1][2]);
						glNormal3d(mo2->npnts[i][j+1][0],mo2->npnts[i][j+1][1],mo2->npnts[i][j+1][2]);
						glVertex3d(mo2->rpnts[i][j+1][0],mo2->rpnts[i][j+1][1],mo2->rpnts[i][j+1][2]);
					}
					else {
						glNormal3d(mo2->npnts[i-1][j-radialSteps+1][0],	mo2->npnts[i-1][j-radialSteps+1][1], mo2->npnts[i-1][j-radialSteps+1][2] );
						glVertex3d(mo2->rpnts[i-1][j-radialSteps+1][0],	mo2->rpnts[i-1][j-radialSteps+1][1], mo2->rpnts[i-1][j-radialSteps+1][2] );
						glNormal3d(mo2->npnts[i][j-radialSteps+1][0], mo2->npnts[i][j-radialSteps+1][1], mo2->npnts[i][j-radialSteps+1][2]);
						glVertex3d(mo2->rpnts[i][j-radialSteps+1][0], mo2->rpnts[i][j-radialSteps+1][1], mo2->rpnts[i][j-radialSteps+1][2]);
					}
				}
			}
			glEnd();
		
		}//	end	if()
		else if(mo2->showQuatMap &&	mo2->whichFrame	== FRENETSERRET) {

			if(flag_use_extern_color)
				glColor3f(color[0],color[1],color[2]);
			else
				glColor3f(mo2->fgcolor[0],mo2->fgcolor[1],mo2->fgcolor[2]);

			start =	1;
			if(mo2->markQuatCurveCounter==	0)
				end	= mo2->numPoints-2;
			else
				end	= mo2->markQuatCurve[0];

			for(h=0; h<=mo2->markQuatCurveCounter; h++)	{
				//--------------------------->
				//getSegmentColor(0,	ccolor);
				//glColor3f(ccolor[0],ccolor[1],ccolor[2]);

				glBegin(GL_QUADS);
				if(h==0)
					start = 1;
				for(i =	start+1; i<end; i++) {

					for(j=0;j<radialSteps;j++) {
						glNormal3d(mo2->npnts[i][j][0],	mo2->npnts[i][j][1], mo2->npnts[i][j][2]);
						glVertex3d(mo2->rpnts[i][j][0],	mo2->rpnts[i][j][1], mo2->rpnts[i][j][2]);
						glNormal3d(mo2->npnts[i-1][j][0], mo2->npnts[i-1][j][1], mo2->npnts[i-1][j][2]);
						glVertex3d(mo2->rpnts[i-1][j][0], mo2->rpnts[i-1][j][1], mo2->rpnts[i-1][j][2]);
						if(j<(radialSteps-1)){
							glNormal3d(mo2->npnts[i-1][j+1][0],	mo2->npnts[i-1][j+1][1], mo2->npnts[i-1][j+1][2]);
							glVertex3d(mo2->rpnts[i-1][j+1][0],	mo2->rpnts[i-1][j+1][1], mo2->rpnts[i-1][j+1][2]);
							glNormal3d(mo2->npnts[i][j+1][0],mo2->npnts[i][j+1][1],mo2->npnts[i][j+1][2]);
							glVertex3d(mo2->rpnts[i][j+1][0],mo2->rpnts[i][j+1][1],mo2->rpnts[i][j+1][2]);
						}
						else {
							glNormal3d(mo2->npnts[i-1][j-radialSteps+1][0],	mo2->npnts[i-1][j-radialSteps+1][1], mo2->npnts[i-1][j-radialSteps+1][2] );
							glVertex3d(mo2->rpnts[i-1][j-radialSteps+1][0],	mo2->rpnts[i-1][j-radialSteps+1][1], mo2->rpnts[i-1][j-radialSteps+1][2] );
							glNormal3d(mo2->npnts[i][j-radialSteps+1][0], mo2->npnts[i][j-radialSteps+1][1], mo2->npnts[i][j-radialSteps+1][2]);
							glVertex3d(mo2->rpnts[i][j-radialSteps+1][0], mo2->rpnts[i][j-radialSteps+1][1], mo2->rpnts[i][j-radialSteps+1][2]);
						}
					}
				}
				glEnd();
				//<---------------------------
				start =	mo2->markQuatCurve[h]+1;
				end	= mo2->markQuatCurve[h+1];
				if(h==(mo2->markQuatCurveCounter)-1) {
					end	= mo2->numPoints-1;
				}
			}

			//start =	1;
			//if(mo2->markQuatCurveCounter2==	0)
			//	end	= mo2->numPoints-2;
			//else
			//	end	= mo2->markQuatCurve2[0];
			//printf("Quat counter = %d\n",mo2->markQuatCurveCounter2);

			//for(h=0; h<=mo2->markQuatCurveCounter2; h++)	{
			//	//--------------------------->
			//	int	colnum = h;
			//	if(h==mo2->markQuatCurveCounter2)
			//		colnum = 0;

			//	getSegmentColor(colnum,	ccolor);
			//	printf("colnum= %d,ccolor = %f, %f, %f\n",h,ccolor[0],ccolor[1],ccolor[2]);
			//	glColor3f(ccolor[0],ccolor[1],ccolor[2]);

			//	glBegin(GL_QUADS);
			//	if(h==0)
			//		start = 1;
			//	for(i =	start+1; i<end; i++) {

			//		for(j=0;j<radialSteps;j++) {
			//			glNormal3d(mo2->npnts[i][j][0],	mo2->npnts[i][j][1], mo2->npnts[i][j][2]);
			//			glVertex3d(mo2->rpnts[i][j][0],	mo2->rpnts[i][j][1], mo2->rpnts[i][j][2]);
			//			glNormal3d(mo2->npnts[i-1][j][0], mo2->npnts[i-1][j][1], mo2->npnts[i-1][j][2]);
			//			glVertex3d(mo2->rpnts[i-1][j][0], mo2->rpnts[i-1][j][1], mo2->rpnts[i-1][j][2]);
			//			if(j<(radialSteps-1)){
			//				glNormal3d(mo2->npnts[i-1][j+1][0],	mo2->npnts[i-1][j+1][1], mo2->npnts[i-1][j+1][2]);
			//				glVertex3d(mo2->rpnts[i-1][j+1][0],	mo2->rpnts[i-1][j+1][1], mo2->rpnts[i-1][j+1][2]);
			//				glNormal3d(mo2->npnts[i][j+1][0],mo2->npnts[i][j+1][1],mo2->npnts[i][j+1][2]);
			//				glVertex3d(mo2->rpnts[i][j+1][0],mo2->rpnts[i][j+1][1],mo2->rpnts[i][j+1][2]);
			//			}
			//			else {
			//				glNormal3d(mo2->npnts[i-1][j-radialSteps+1][0],	mo2->npnts[i-1][j-radialSteps+1][1], mo2->npnts[i-1][j-radialSteps+1][2] );
			//				glVertex3d(mo2->rpnts[i-1][j-radialSteps+1][0],	mo2->rpnts[i-1][j-radialSteps+1][1], mo2->rpnts[i-1][j-radialSteps+1][2] );
			//				glNormal3d(mo2->npnts[i][j-radialSteps+1][0], mo2->npnts[i][j-radialSteps+1][1], mo2->npnts[i][j-radialSteps+1][2]);
			//				glVertex3d(mo2->rpnts[i][j-radialSteps+1][0], mo2->rpnts[i][j-radialSteps+1][1], mo2->rpnts[i][j-radialSteps+1][2]);
			//			}
			//		}
			//	}
			//	glEnd();
			//	//<---------------------------
			//	start =	mo2->markQuatCurve2[h]+1;
			//	end	= mo2->markQuatCurve2[h+1];
			//	if(h==(mo2->markQuatCurveCounter2)-1) {
			//		end	= mo2->numPoints-1;
			//	}
			//}

		}
		glDisable(GL_TEXTURE_2D);
		glPopAttrib();			  
	}//end if()

	if(mo2->qshowCloudPoints && mo2->whichFrame==FRENETSERRET) {
		glPushAttrib(GL_ALL_ATTRIB_BITS);
			//glEnable(GL_LIGHTING);
			//glEnable(GL_COLOR_MATERIAL);
			////# Original points in RED
			//glColor3f(1.0f,0.0f,0.0f);
			//for(i=0;i<mo2->numPoints;i++) {
			//	q->r = -(mo2->qpntFliped[i]->r);
			//	q->i = -(mo2->qpntFliped[i]->i);
			//	q->j = -(mo2->qpntFliped[i]->j);
			//	q->k = -(mo2->qpntFliped[i]->k);

			//	getQuatComp(mo2->qwhichQuatComponent,q,qvec);
			//	glPushMatrix();
			//		glTranslatef(qvec[0],qvec[1],qvec[2]);
			//		glutSolidSphere(0.03,16,16);
			//	glPopMatrix();
			//}

			////# Flipped points in GREEN
			//glColor3f(1.0f,1.0f,0.0f);
			//for(i=0;i<mo2->numPoints;i++) {
			//	getQuatComp(mo2->qwhichQuatComponent,mo2->qpntFliped[i],qvec);
			//	glPushMatrix();
			//		glTranslatef(qvec[0],qvec[1],qvec[2]);
			//		glutSolidSphere(0.03,16,16);
			//	glPopMatrix();
			//}

			////# Original points in RED
			//glColor3f(0.8f,0.0f,0.0f);
			//start =	0;
			//if(mo2->markQuatCurveCounter==	0)
			//	end	= mo2->numPoints-2;
			//else
			//	end	= mo2->markQuatCurve[0];

			//for(h=0; h<=mo2->markQuatCurveCounter; h++)	{
			//	//--------------------------->
			//	if(h==0)
			//		start = 1;
			//	for(i =	start; i<end; i++) {
			//		getQuatComp(mo2->qwhichQuatComponent,mo2->qpntOrig[i],qvec);
			//		glPushMatrix();
			//			glTranslatef(qvec[0],qvec[1],qvec[2]);
			//			glutSolidSphere(0.03,16,16);
			//		glPopMatrix();
			//	}
			//	//<---------------------------
			//	start =	mo2->markQuatCurve[h]+1;
			//	end	= mo2->markQuatCurve[h+1];
			//	if(h==(mo2->markQuatCurveCounter)-1) {
			//		end	= mo2->numPoints-1;
			//	}
			//}

			////# Corrected Curve Points in GREEN
			//start =	0;
			//if(mo2->markQuatCurveCounter==	0)
			//	end	= mo2->numPoints-2;
			//else
			//	end	= mo2->markQuatCurve[0];

			//for(h=0; h<=mo2->markQuatCurveCounter; h++)	{
			//	//--------------------------->
			//	if(h==0)
			//		start = 1;
			//	for(i =	start; i<end; i++) {
			//		
			//		getQuatComp(mo2->qwhichQuatComponent,mo2->qpntFliped[i],qvec);
			//		glPushMatrix();
			//			glColor3f(0.0f,0.7f,0.0f);
			//			glTranslatef(qvec[0],qvec[1],qvec[2]);
			//			glutSolidSphere(0.03,16,16);
			//			//glPushMatrix();
			//			//	glTranslatef(delta,delta,delta);
			//			//	gluOrtho2D(-0.1,0.1,-0.1,0.1);
			//			//	glDisable(GL_LIGHTING);
			//			//	glColor3f(0.9f,0.9f,0.9f);
			//			//	sprintf(obj_qstr,"%d",i);
			//			//	glRasterPos2f(0.0,0.0);
			//			//	stext(obj_qstr);
			//			//	glEnable(GL_LIGHTING);
			//			//glPopMatrix();
			//		glPopMatrix();
			//	}
			//	//<---------------------------
			//	start =	mo2->markQuatCurve[h]+1;
			//	end	= mo2->markQuatCurve[h+1];
			//	if(h==(mo2->markQuatCurveCounter)-1) {
			//		end	= mo2->numPoints-1;
			//	}
			//}

			////# Points that will be flipped in ORANGE
			//glColor3f(0.7f,0.5f,0.0f);
			//for(h=0; h<mo2->qflipcounter;) {
			//	for(i=mo2->qflipindex[h];i<=mo2->qflipindex[h+1];i++) {
			//		getQuatComp(mo2->qwhichQuatComponent,mo2->qpntFliped[i],qvec);
			//		glPushMatrix();
			//			glTranslatef(qvec[0],qvec[1],qvec[2]);
			//			glutSolidSphere(0.03,16,16);
			//		glPopMatrix();
			//	}
			//	h = h+2;
			//}
		glPopAttrib();
	}
	if(mo2->qshowSphere) {
		glPushAttrib(GL_ALL_ATTRIB_BITS);
			glEnable(GL_BLEND);
			glDisable(GL_LIGHTING);
			glColor4f(0.7f,0.7f,0.7f,0.1f);
			glutWireSphere(1.0f,32,32);
		glPopAttrib();
	}
	if(mo2->qshowAxes) {
		double len = 0.5;
		double rad = 0.025;
	    
		GLUquadricObj *qobj;
		qobj = gluNewQuadric();

		glPushAttrib(GL_CURRENT_BIT | GL_LIGHTING_BIT);
			glDisable(GL_LIGHTING);
			glPushMatrix();
				glTranslatef(0.0,0.0,0.0);
				glColor3f(0.0, 0.0, 1.0);
				glRotatef(-90,1,0,0);
				gluQuadricDrawStyle(qobj, GLU_FILL);
				gluCylinder(qobj, rad, rad, len, 16, 16);
			glPopMatrix();
			glPushMatrix();
				glTranslatef(0.0,0.0,0.0);
				glColor3f(0.0, 1.0, 0.0);
				glRotatef(90,0,1,0);
				gluQuadricDrawStyle(qobj, GLU_FILL);
				gluCylinder(qobj, rad, rad, len, 16, 16);
			glPopMatrix();
			glPushMatrix();
				glTranslatef(0.0,0.0,0.0);
				glColor3f(1.0, 0.0, 0.0);
				glRotatef(0,0,0,1);
				gluQuadricDrawStyle(qobj, GLU_FILL);
				gluCylinder(qobj, rad, rad, len, 16, 16);
			glPopMatrix();
		glPopAttrib();
	}
	if(mo2->qshowFrame)
		drawFrames();
}

void renderThickenedCurve( double radius ) {

	double p1[3];					//point	1
	double p2[3];					//point	2
	double v[3];					//vector p2-p1
	double r[3];					//vector
	double theta1	= 0.0;
	double theta2	= 0.0;
	double length	= 0.0;
	int	   renderPoints	= mo2->numPoints;
	
	if(	mo2->leave_gap)
		renderPoints -=	2;
	
	for(int	i=0;i<(renderPoints);i++)	{
		
		p1[0] =	mo2->pntList[i].p3.d[0];
		p1[1] =	mo2->pntList[i].p3.d[1];
		p1[2] =	mo2->pntList[i].p3.d[2];
		
		if(i==(renderPoints-1)) {
			p2[0] =	mo2->pntList[0].p3.d[0];
			p2[1] =	mo2->pntList[0].p3.d[1];
			p2[2] =	mo2->pntList[0].p3.d[2];
		}
		else {
			p2[0] =	mo2->pntList[i+1].p3.d[0];
			p2[1] =	mo2->pntList[i+1].p3.d[1];
			p2[2] =	mo2->pntList[i+1].p3.d[2];
		}
		minus3D(p2,p1,v);
		
		r[0] = -v[1]*v[2];
		r[1] =	v[0]*v[2];
		r[2] =	0.0;
		
		length = sqrt( v[0]*v[0] + v[1]*v[1] + v[2]*v[2] );
		double zero	= 1.0e-3;
		
		if (fabs(v[2]) < zero) {
				// rotation	angle in x-y plane
			theta1 = (180/M_PI)*acos( v[0] / length	);	
			if ( v[0] <= 0.0 ) theta1 =	-theta1;
		}
		else {
				// rotation	angle
			theta2 = (180/M_PI)*acos( v[2] / length	);	
			if ( v[2] <= 0.0 ) theta2 =	-theta2;
		}
		
		glPushAttrib (GL_ALL_ATTRIB_BITS);
		
		glDisable(GL_CULL_FACE);
		glEnable(GL_LIGHTING);
		glEnable(GL_COLOR_MATERIAL);
		glColor3f(mo2->fgcolor[0],mo2->fgcolor[1],mo2->fgcolor[2]);
		
		glPushMatrix();
		glTranslatef(p1[0],	p1[1], p1[2] );	
		gluSphere (qobjSPH,	radius-0.001, 16, 16);
		glPopMatrix();
		if(i==(renderPoints-2))	{
			glPushMatrix();
			glTranslatef(p2[0],	p2[1], p2[2] );	
			gluQuadricTexture(qobjSPH, GL_TRUE);
			gluSphere (qobjSPH,	radius-0.0001, 16, 16);
			glPopMatrix();
		}
		glPushMatrix();

		// translate to	point 1
		glTranslatef(p1[0],	p1[1], p1[2] );			
		if(fabs(v[2])<=zero) {
			glRotatef(90.0,	 0.0, 1.0, 0.0);
			if((v[0]<0.0 &&	v[1]>0.0) || (v[0]>0.0 && v[1]<0.0))
				glRotatef(theta1, 1.0, 0.0,	0.0);
			else if	((v[0]<0.0 && v[1]<0.0)	|| (v[0]>0.0 &&	v[1]>0.0))
				glRotatef(theta1, -1.0,	0.0, 0.0);
				}
		else {
			// Rotate about	rotation vector
			glRotatef(theta2, r[0],	r[1], r[2]);	
		}
		gluQuadricTexture(qobjCYL, GL_TRUE);
		gluCylinder	(qobjCYL, radius, radius, length+0.01, 16, 16);
		glPopMatrix();
		glPopAttrib();
		
	}
}

void renderThickenedCurve2( double radius ) {

	int		i		= 0;
	int		h		= 0;
	double	theta1	= 0.0;
	double	theta2	= 0.0;
	int		start	= 0;
	int		end		= 0;
	double	length	= 0.0;
	//int		renderPoints	= mo2->numPoints;
	double	ccolor[3];
	double	p1[3];					//point	1
	double	p2[3];					//point	2
	double	v[3];					//vector p2-p1
	double	r[3];					//vector
	
	start =	1;
	if(mo2->markQuatCurveCounter==	0)
		end	= mo2->numPoints-2;
	else
		end	= mo2->markQuatCurve[0]-1;

	for(h=0; h<=mo2->markQuatCurveCounter; h++)	{
		//--------------------------->
		int	colnum = h;
		if(h==mo2->markFrameCounter)
			colnum = 0;

		getSegmentColor(colnum,	ccolor);
		glColor3f(ccolor[0],ccolor[1],ccolor[2]);

		if(h==0) {
			start = 1;
			end = end-1;
		}
		for(i =	start; i<end; i++) {

			p1[0] =	mo2->pntList[i].p3.d[0];
			p1[1] =	mo2->pntList[i].p3.d[1];
			p1[2] =	mo2->pntList[i].p3.d[2];
			
			p2[0] =	mo2->pntList[i+1].p3.d[0];
			p2[1] =	mo2->pntList[i+1].p3.d[1];
			p2[2] =	mo2->pntList[i+1].p3.d[2];

			minus3D(p2,p1,v);
			
			r[0] = -v[1]*v[2];
			r[1] =	v[0]*v[2];
			r[2] =	0.0;

			length = sqrt( v[0]*v[0] + v[1]*v[1] + v[2]*v[2] );
			double zero	= 1.0e-3;
			
			if (fabs(v[2]) < zero) {
					// rotation	angle in x-y plane
				theta1 = (180/M_PI)*acos( v[0] / length	);	
				if ( v[0] <= 0.0 ) theta1 =	-theta1;
			}
			else {
					// rotation	angle
				theta2 = (180/M_PI)*acos( v[2] / length	);	
				if ( v[2] <= 0.0 ) theta2 =	-theta2;
			}
			
			glPushAttrib (GL_ALL_ATTRIB_BITS);
			
			glDisable(GL_CULL_FACE);
			glEnable(GL_LIGHTING);
			glEnable(GL_COLOR_MATERIAL);
			glColor3f(ccolor[0],ccolor[1],ccolor[2]);
			
			glPushMatrix();
			glTranslatef(p1[0],	p1[1], p1[2] );	
			gluSphere (qobjSPH,	radius-0.001, 16, 16);
			glPopMatrix();
				glPushMatrix();
				glTranslatef(p2[0],	p2[1], p2[2] );	
				gluQuadricTexture(qobjSPH, GL_TRUE);
				gluSphere (qobjSPH,	radius-0.0001, 16, 16);
				glPopMatrix();
			glPushMatrix();

			// translate to	point 1
			glTranslatef(p1[0],	p1[1], p1[2] );			
			if(fabs(v[2])<=zero) {
				glRotatef(90.0,	 0.0, 1.0, 0.0);
				if((v[0]<0.0 &&	v[1]>0.0) || (v[0]>0.0 && v[1]<0.0))
					glRotatef(theta1, 1.0, 0.0,	0.0);
				else if	((v[0]<0.0 && v[1]<0.0)	|| (v[0]>0.0 &&	v[1]>0.0))
					glRotatef(theta1, -1.0,	0.0, 0.0);
					}
			else {
				// Rotate about	rotation vector
				glRotatef(theta2, r[0],	r[1], r[2]);	
			}
			gluQuadricTexture(qobjCYL, GL_TRUE);
			gluCylinder	(qobjCYL, radius, radius, length+0.01, 16, 16);
			glPopMatrix();
			glPopAttrib();
		}

		start =	mo2->markQuatCurve[h]+1;
		end	= mo2->markQuatCurve[h+1];
		if(h==(mo2->markQuatCurveCounter)-1) {
			end	= mo2->numPoints-2;
		}
	}
	//for(int	i=0;i<(renderPoints);i++)	{
	//	
	//	p1[0] =	mo2->pntList[i].p3.d[0];
	//	p1[1] =	mo2->pntList[i].p3.d[1];
	//	p1[2] =	mo2->pntList[i].p3.d[2];
	//	
	//	if(i==(renderPoints-1)) {
	//		p2[0] =	mo2->pntList[0].p3.d[0];
	//		p2[1] =	mo2->pntList[0].p3.d[1];
	//		p2[2] =	mo2->pntList[0].p3.d[2];
	//	}
	//	else {
	//		p2[0] =	mo2->pntList[i+1].p3.d[0];
	//		p2[1] =	mo2->pntList[i+1].p3.d[1];
	//		p2[2] =	mo2->pntList[i+1].p3.d[2];
	//	}
	//	minus3D(p2,p1,v);
	//	
	//	r[0] = -v[1]*v[2];
	//	r[1] =	v[0]*v[2];
	//	r[2] =	0.0;
	//	
	//	length = sqrt( v[0]*v[0] + v[1]*v[1] + v[2]*v[2] );
	//	double zero	= 1.0e-3;
	//	
	//	if (fabs(v[2]) < zero) {
	//			// rotation	angle in x-y plane
	//		theta1 = (180/M_PI)*acos( v[0] / length	);	
	//		if ( v[0] <= 0.0 ) theta1 =	-theta1;
	//	}
	//	else {
	//			// rotation	angle
	//		theta2 = (180/M_PI)*acos( v[2] / length	);	
	//		if ( v[2] <= 0.0 ) theta2 =	-theta2;
	//	}
	//	
	//	glPushAttrib (GL_ALL_ATTRIB_BITS);
	//	
	//	glDisable(GL_CULL_FACE);
	//	glEnable(GL_LIGHTING);
	//	glEnable(GL_COLOR_MATERIAL);
	//	glColor3f(mo2->fgcolor[0],mo2->fgcolor[1],mo2->fgcolor[2]);
	//	
	//	glPushMatrix();
	//	glTranslatef(p1[0],	p1[1], p1[2] );	
	//	gluSphere (qobjSPH,	radius-0.001, 16, 16);
	//	glPopMatrix();
	//	if(i==(renderPoints-2))	{
	//		glPushMatrix();
	//		glTranslatef(p2[0],	p2[1], p2[2] );	
	//		gluQuadricTexture(qobjSPH, GL_TRUE);
	//		gluSphere (qobjSPH,	radius-0.0001, 16, 16);
	//		glPopMatrix();
	//	}
	//	glPushMatrix();

	//	// translate to	point 1
	//	glTranslatef(p1[0],	p1[1], p1[2] );			
	//	if(fabs(v[2])<=zero) {
	//		glRotatef(90.0,	 0.0, 1.0, 0.0);
	//		if((v[0]<0.0 &&	v[1]>0.0) || (v[0]>0.0 && v[1]<0.0))
	//			glRotatef(theta1, 1.0, 0.0,	0.0);
	//		else if	((v[0]<0.0 && v[1]<0.0)	|| (v[0]>0.0 &&	v[1]>0.0))
	//			glRotatef(theta1, -1.0,	0.0, 0.0);
	//			}
	//	else {
	//		// Rotate about	rotation vector
	//		glRotatef(theta2, r[0],	r[1], r[2]);	
	//	}
	//	gluQuadricTexture(qobjCYL, GL_TRUE);
	//	gluCylinder	(qobjCYL, radius, radius, length+0.01, 16, 16);
	//	glPopMatrix();
	//	glPopAttrib();
	//	
	//}
}

void drawQuaternionCurve ( double *color ) 
{
	int	i			=	0;
	int	start		=	1;
	double radius	= 0.0;
	double qvec[3];

	glPushAttrib(GL_LIGHTING_BIT | GL_LINE_BIT | GL_LIGHTING_BIT);
		glColor3f(0.9,0.8,0.0);
		glLineWidth(2.0);
		glDisable(GL_LIGHTING);

		//# Draw curve
		glBegin(GL_LINE_STRIP);
		for( i=start; i<(mo2->numPoints-2);	i++	) {
			getQuatComp(mo2->qwhichQuatComponent,&(mo2->quatList[i]),qvec);
			glVertex3f(qvec[0],qvec[1],qvec[2]);
		}
		glEnd();
		//copy_QuatList();
		//renderThickenedCurve2(0.02);
		glEnable(GL_LIGHTING);
		glEnable(GL_COLOR_MATERIAL);
		glColor3f(color[0],color[1],color[2]);

		for( i=start; i<(mo2->numPoints-2) ;i++	) {
			//if(i%10==1){
			if(i==start) {
				glColor3f(1.0f,1.0f,0.0f);
				//radius = 0.04;
			}
			else if(i==(mo2->numPoints-3)) {
				glColor3f(0.0f,1.0f,1.0f);
			}
			else {
				glColor3f(color[0],color[1],color[2]);
				//radius = 0.02;
			}
			//}
			//else {
				radius = 0.02;
			//}
			getQuatComp(mo2->qwhichQuatComponent,&(mo2->quatList[i]),qvec);
			glPushMatrix();
				glTranslatef(qvec[0],qvec[1],qvec[2]);
				glutSolidSphere(radius,16,16);
				//double delta = 0.02;
				//glPushMatrix();
				//	glTranslatef(delta,delta,delta);
				//	gluOrtho2D(-0.1,0.1,-0.1,0.1);
				//	glDisable(GL_LIGHTING);
				//	glColor3f(0.9f,0.9f,0.9f);
				//	sprintf(obj_qstr,"%d",i);
				//	glRasterPos2f(0.0,0.0);
				//	stext(obj_qstr);
				//glPopMatrix();
			glPopMatrix();
		}
	glPopAttrib();
	if(mo2->showLabels) {
		glPushAttrib(GL_ALL_ATTRIB_BITS);
			glDisable(GL_LIGHTING);
			glPushMatrix();
				glColor3f(1.0f,1.0f,0.0f);
				getQuatComp(mo2->qwhichQuatComponent,&(mo2->quatList[start]),qvec);
				glTranslatef(qvec[0], qvec[1], qvec[2]); 
				glTranslatef(0.05,0.05,0.05);
				glRasterPos2f(0.0,0.0);
				sprintf(obj_qstr,"FIRST Quaternion");
				stext(obj_qstr);
			glPopMatrix();
			glPushMatrix();
				glColor3f(0.0f,1.0f,1.0f);
				getQuatComp(mo2->qwhichQuatComponent,&(mo2->quatList[mo2->numPoints-2]),qvec);
				glTranslatef(qvec[0], qvec[1], qvec[2]); 
				glTranslatef(0.05,0.05,0.05);
				glRasterPos2f(0.0,0.0);
				sprintf(obj_qstr,"FIRST Quaternion");
				stext(obj_qstr);
			glPopMatrix();
		glPopAttrib();
	}

}


void drawQuaternionCurve_forRings ( double *color ) 
{
	int	i			=	0;
	int	start		=	0;
	double radius	= 0.0;
	double qvec[3];

	glPushAttrib(GL_LIGHTING_BIT | GL_LINE_BIT | GL_LIGHTING_BIT);
		glColor3f(0.9,0.8,0.0);
		glLineWidth(2.0);
		glDisable(GL_LIGHTING);

		//# Draw curve
		glBegin(GL_LINE_STRIP);
		for( i=start; i<(mo2->numPoints);	i++	) {
			getQuatComp(mo2->qwhichQuatComponent,&(mo2->quatList[i]),qvec);
			glVertex3f(qvec[0],qvec[1],qvec[2]);
		}
		glEnd();
		//copy_QuatList();
		//renderThickenedCurve2(0.02);
		glEnable(GL_LIGHTING);
		glEnable(GL_COLOR_MATERIAL);
		glColor3f(color[0],color[1],color[2]);

		for( i=start; i<(mo2->numPoints) ;i++	) {
			//if(i%10==1){
			if(i==0) {
				glColor3f(0.6f,0.6f,0.0f);
				radius = 0.03;
			}
			else if(i==(mo2->numPoints-1)) {
				glColor3f(0.0f,0.6f,0.6f);
				radius = 0.03;
			}
			else {
				glColor3f(color[0],color[1],color[2]);
				radius = 0.02;
			}

			getQuatComp(mo2->qwhichQuatComponent,&(mo2->quatList[i]),qvec);
			glPushMatrix();
				glTranslatef(qvec[0],qvec[1],qvec[2]);
				glutSolidSphere(radius,16,16);
				//double delta = 0.02;
				//glPushMatrix();
				//	glTranslatef(delta,delta,delta);
				//	gluOrtho2D(-0.1,0.1,-0.1,0.1);
				//	glDisable(GL_LIGHTING);
				//	glColor3f(0.9f,0.9f,0.9f);
				//	sprintf(obj_qstr,"%d",i);
				//	glRasterPos2f(0.0,0.0);
				//	stext(obj_qstr);
				//glPopMatrix();
			glPopMatrix();
		}
	glPopAttrib();
}


void drawQuaternionCurve ( int start, int end, double* color ) 
{
	int	i =	0;
	double qvec[3];

	glPushAttrib(GL_LIGHTING_BIT | GL_LINE_BIT | GL_LIGHTING_BIT);
	glColor3f(color[0],color[1],color[2]);
	glLineWidth(2.0);
	glBegin(GL_LINE_STRIP);
	for( i=start; i<(end-1); i++ ) {
		switch(mo2->qwhichQuatComponent) {
			case SHOWXYZ:
				glVertex3f(mo2->quatList[i].i, mo2->quatList[i].j, mo2->quatList[i].k);
				break;
			case SHOWWYZ:
				glVertex3f(mo2->quatList[i].r, mo2->quatList[i].j, mo2->quatList[i].k);
				break;
			case SHOWWXZ:
				glVertex3f(mo2->quatList[i].r, mo2->quatList[i].i, mo2->quatList[i].k);
				break;
			case SHOWWXY:
				glVertex3f(mo2->quatList[i].r, mo2->quatList[i].i, mo2->quatList[i].j);
				break;
			default:
				break;
		}
	}
	glEnd();
	glEnable(GL_COLOR_MATERIAL);
	glEnable(GL_LIGHTING);
	for( i=start; i<end-1; i++ ) {
		if(i==start	|| i==(end))
			glColor3f(0.95f,0.95f,0.6f);
		else
			glColor3f(color[0],color[1],color[2]);
		glPushMatrix();
			getQuatComp(mo2->qwhichQuatComponent,&(mo2->quatList[i]),qvec);
			glTranslatef(qvec[0],qvec[1],qvec[2]);
			glutSolidSphere(0.01,16,16);
		glPopMatrix();
	}
	glPopAttrib();
}

void displayKnotQuaternionMap (	void )
{
	int i=0;
	double color[3];
	int	useColor = 1;
	quaternion *r;
	r = (quaternion*) my_malloc(sizeof(quaternion));

	glPushMatrix();
	glRotatef(-90,0,1,0);
	glRotatef(-90,1,0,0);
		color[0] =	0.4; color[1]	= 0.7; color[2] = 0.6;
		glRotatef(0.0,0.0f,0.0f,1.0f);
		computeCurveQuaternionMap();
		
		if(mo2->qshowCurve)
			drawQuaternionCurve(color);

		copy_QuatList();
		
		makeTangent();
		makeTangentDerivatives();

		if(mo2->whichQuatFrame == PARALLELTRANSPORT) {
			computeParallelTransportFrame();
		}
		else if(mo2->whichQuatFrame	== FRENETSERRET){
			computeFrenetFrame();
		}
		renderFramedQuaternionCurve3(0.075, color, useColor);

		if(mo2->qshowCloudPoints) {

			reflect_QuatList();
			makeTangent();
			makeTangentDerivatives();

			color[0] =	0.7; color[1]	= 0.5; color[2] = 0.4;
			if(mo2->whichQuatFrame == PARALLELTRANSPORT) {
				computeParallelTransportFrame();
			}
			else if(mo2->whichQuatFrame	== FRENETSERRET){
				computeFrenetFrame();
			}

			renderFramedQuaternionCurve3(0.075, color, useColor);
			for(i=0;i<mo2->numPoints;i++) {
					//printf("[%d] q =  %f, %f ,%f \n",i,mo2->quatList[i].r,mo2->quatList[i].i,mo2->quatList[i].j);
					quatCopy(&(mo2->quatList[i]),r);
					mo2->quatList[i].r = -r->r;
					mo2->quatList[i].i = -r->i;
					mo2->quatList[i].j = -r->j;
					mo2->quatList[i].k = -r->k;
			}
			if(mo2->showCurve)
				drawQuaternionCurve(color);
		}
		//if(mo2->qshowQuatRings ) {
		//	mo2->qset->whichQuatComponent = mo2->qwhichQuatComponent;
		//	mo2->qset->setTangSet(mo2->TangList,mo2->numPoints);
		//	//printf("numtset = %d\n",mo2->qset->numtset);
		//	mo2->qset->computeQuaternionSet(1);
		//	mo2->qset->generateQuaternionRings();
		//	mo2->qset->displayQuaternions();
		//}
		//if(mo2->qshowQuatRings ) {
		//	constructKnot();
		//	makeTangent();
		//	glPushAttrib(GL_ALL_ATTRIB_BITS);
		//	glDisable(GL_LIGHTING);
		//	glLineWidth(1.5f);

		//	for(i=0;i<mo2->numPoints;i++) {
		//		theta = 0.0;
		//		glBegin(GL_LINE_STRIP);
		//			t[0] = mo2->TangList[i].p3.d[0];
		//			t[1] = mo2->TangList[i].p3.d[1];
		//			t[2] = mo2->TangList[i].p3.d[2];
		//			while(theta<=(4.0*M_PI)) {
		//				makeFamilyOfQuaternions(theta, t, qf[i]);
		//				getQuatComp(mo2->qwhichQuatComponent, qf[i], qvec);
		//				glVertex3f(qvec[0], qvec[1], qvec[2]);
		//				theta = theta + 0.5;
		//			}
		//		glEnd();			
		//	}
		//	glPopAttrib();
		//}
		//drawRings();

	glPopMatrix();
}



void displayQuaternionRings ( void )
{
	//int i=0;
	double color[3];
	//int	useColor = 1;
	quaternion *r;
	r = (quaternion*) my_malloc(sizeof(quaternion));

	glPushMatrix();
	glRotatef(-90,0,1,0);
	glRotatef(-90,1,0,0);
		color[0] =	0.4; color[1]	= 0.7; color[2] = 0.6;
		computeCurveQuaternionMap();
		
		if(mo2->qshowCurve)
			drawQuaternionCurve_forRings(color);
	drawQuaternionRingSpace();

		//copy_QuatList();
		//
		//makeTangent();
		//makeTangentDerivatives();

		//if(mo2->whichQuatFrame == PARALLELTRANSPORT) {
		//	computeParallelTransportFrame();
		//}
		//else if(mo2->whichQuatFrame	== FRENETSERRET){
		//	computeFrenetFrame();
		//}
		//renderFramedQuaternionCurve3(0.075, color, useColor);

		//if(mo2->qshowCloudPoints) {

		//	reflect_QuatList();
		//	makeTangent();
		//	makeTangentDerivatives();

		//	color[0] =	0.7; color[1]	= 0.5; color[2] = 0.4;
		//	if(mo2->whichQuatFrame == PARALLELTRANSPORT) {
		//		computeParallelTransportFrame();
		//	}
		//	else if(mo2->whichQuatFrame	== FRENETSERRET){
		//		computeFrenetFrame();
		//	}

		//	renderFramedQuaternionCurve3(0.075, color, useColor);
		//	for(i=0;i<mo2->numPoints;i++) {
		//			quatCopy(&(mo2->quatList[i]),r);
		//			mo2->quatList[i].r = -r->r;
		//			mo2->quatList[i].i = -r->i;
		//			mo2->quatList[i].j = -r->j;
		//			mo2->quatList[i].k = -r->k;
		//	}
		//	if(mo2->showCurve)
		//		drawQuaternionCurve(color);
		//}
	glPopMatrix();

	if(mo2->rshowAxes)
		mo2->renderAxes();
	//drawRings();
}
//# Utilities
void getQuatComp ( int whichQuatComponent, quaternion* q, double* qvec) {
	switch(whichQuatComponent) {

		case SHOWXYZ:
			qvec[0] = q->i;
			qvec[1] = q->j;
			qvec[2] = q->k;
			break;

		case SHOWWYZ:
			qvec[0] = q->r;
			qvec[1] = q->j;
			qvec[2] = q->k;
			break;
		
		case SHOWWXY:
			qvec[0] = q->r;
			qvec[1] = q->i;
			qvec[2] = q->j;
			break;

		case SHOWWXZ:
			qvec[0] = q->r;
			qvec[1] = q->i;
			qvec[2] = q->k;
			break;

		default:
			break;
		}	
}

void copy_QuatList ( void )	
{
	int	i =	0;
	for(i=0;i<mo2->numPoints;i++) {
		//printf("[%d] q =  %f, %f ,%f \n",i,mo2->quatList[i].r,mo2->quatList[i].i,mo2->quatList[i].j);
		switch(mo2->qwhichQuatComponent) {
			case SHOWXYZ:
				mo2->pntList[i].p3.d[0]	= mo2->quatList[i].i;
				mo2->pntList[i].p3.d[1]	= mo2->quatList[i].j;
				mo2->pntList[i].p3.d[2]	= mo2->quatList[i].k;
				break;
			case SHOWWYZ:
				mo2->pntList[i].p3.d[0]	= mo2->quatList[i].r;
				mo2->pntList[i].p3.d[1]	= mo2->quatList[i].j;
				mo2->pntList[i].p3.d[2]	= mo2->quatList[i].k;
				break;
			case SHOWWXZ:
				mo2->pntList[i].p3.d[0]	= mo2->quatList[i].r;
				mo2->pntList[i].p3.d[1]	= mo2->quatList[i].i;
				mo2->pntList[i].p3.d[2]	= mo2->quatList[i].k;
				break;
			case SHOWWXY:
				mo2->pntList[i].p3.d[0]	= mo2->quatList[i].r;
				mo2->pntList[i].p3.d[1]	= mo2->quatList[i].i;
				mo2->pntList[i].p3.d[2]	= mo2->quatList[i].j;
				break;
			default:
				break;
		}

		//if(mo2->showViewXYZ==1) {
		//	  mo2->pntList[i].p3.d[0] =	mo2->quatList[i].i;
		//	  mo2->pntList[i].p3.d[1] =	mo2->quatList[i].j;
		//	  mo2->pntList[i].p3.d[2] =	mo2->quatList[i].k;
		//}
		//else if(mo2->showViewWYZ==1) {
		//	  mo2->pntList[i].p3.d[0] =	mo2->quatList[i].r;
		//	  mo2->pntList[i].p3.d[1] =	mo2->quatList[i].j;
		//	  mo2->pntList[i].p3.d[2] =	mo2->quatList[i].k;
		//}
		//else if(mo2->showViewWXZ==1) {
		//	  mo2->pntList[i].p3.d[0] =	mo2->quatList[i].r;
		//	  mo2->pntList[i].p3.d[1] =	mo2->quatList[i].i;
		//	  mo2->pntList[i].p3.d[2] =	mo2->quatList[i].k;
		//}
		//else if(mo2->showViewWXY==1) {
		//	  mo2->pntList[i].p3.d[0] =	mo2->quatList[i].r;
		//	  mo2->pntList[i].p3.d[1] =	mo2->quatList[i].i;
		//	  mo2->pntList[i].p3.d[2] =	mo2->quatList[i].j;
		//}
	}
}


void reflect_QuatList ( void )	
{
	int	i =	0;

	for(i=0;i<mo2->numPoints;i++) {
		//printf("[%d] q =  %f, %f ,%f \n",i,mo2->quatList[i].r,mo2->quatList[i].i,mo2->quatList[i].j);
		/*quatCopy(&(mo2->quatList[i]),r);
		mo2->quatList[i].r = -r->r;
		mo2->quatList[i].i = -r->i;
		mo2->quatList[i].j = -r->j;
		mo2->quatList[i].k = -r->k;*/
		switch(mo2->qwhichQuatComponent) {
			case SHOWXYZ:
				mo2->pntList[i].p3.d[0]	= -mo2->quatList[i].i;
				mo2->pntList[i].p3.d[1]	= -mo2->quatList[i].j;
				mo2->pntList[i].p3.d[2]	= -mo2->quatList[i].k;
				break;
			case SHOWWYZ:
				mo2->pntList[i].p3.d[0]	= -mo2->quatList[i].r;
				mo2->pntList[i].p3.d[1]	= -mo2->quatList[i].j;
				mo2->pntList[i].p3.d[2]	= -mo2->quatList[i].k;
				break;
			case SHOWWXZ:
				mo2->pntList[i].p3.d[0]	= -mo2->quatList[i].r;
				mo2->pntList[i].p3.d[1]	= -mo2->quatList[i].i;
				mo2->pntList[i].p3.d[2]	= -mo2->quatList[i].k;
				break;
			case SHOWWXY:
				mo2->pntList[i].p3.d[0]	= -mo2->quatList[i].r;
				mo2->pntList[i].p3.d[1]	= -mo2->quatList[i].i;
				mo2->pntList[i].p3.d[2]	= -mo2->quatList[i].j;
				break;
			default:
				break;
		}
	}
}


void getSegmentColor(int i,	double *ccolor)	{

	switch(i) {
		case 0:
			ccolor[0] =	0.2; ccolor[1]	= 0.55; ccolor[2] = 0.35;
			break;
		case 1:
			ccolor[0] =	0.75; ccolor[1]	= ccolor[2]	= 0.2;
			break;
		case 2:
			ccolor[1] =	0.75; ccolor[0]	= ccolor[2]	= 0.2;
			break;
		case 3:
			ccolor[2] =	0.75; ccolor[1]	= ccolor[0]	= 0.2;
			break;
		case 4:
			ccolor[0] =	0.0; ccolor[1] = ccolor[2] = 0.75;
			break;
		case 5:
			ccolor[2] =	0.0; ccolor[1] = ccolor[0] = 0.75;
			break;
		case 6:
			ccolor[1] =	0.0; ccolor[0] = ccolor[2] = 0.75;
			break;
		case 7:
			ccolor[0] =	ccolor[1] =	ccolor[2] =	0.2;
			break;
		case 8:
			ccolor[0] =	ccolor[1] =	ccolor[2] =	0.4;
			break;
		case 9:
			ccolor[0] =	ccolor[1] =	ccolor[2] =	0.6;
			break;
		case 10:
			ccolor[0] =	ccolor[1] =	ccolor[2] =	1.0;
			break;
		case 11:
			ccolor[2] =	0.0; ccolor[1] = ccolor[0] = 1.0;
			break;
		case 12:
			ccolor[2] =	0.2; ccolor[1] = 0.5; ccolor[0]	= 1.0; 
			break;

		default:
			ccolor[0] =	ccolor[1] =	ccolor[2] =	1.0;
			break;
	}
}
/********************************************************************************
* Date   : Tue Aug 16 10:15:40 EST 2005
* Method : void constructKnot ( void )
* Args   : none
* Return : none
* Purpose: compute a (m,n)-Torus Knot
********************************************************************************/
void constructKnot ( void )
{
	double	t	= 0.0;
	int		i	= 0;
	int		mod	= 0;
	double	div	= 1.0;
	double	inc	= 0.0;
	int		m =	mo2->m;
	int		n =	mo2->n;
	double	a =	mo2->a;
	double	b =	mo2->b;
	double	r =	mo2->r;
	double	maxAngle = 0.0;
	
	/* // Equation of Torus	Knot
	///////////////////////////////////
	torusknot[a_,b_,c_][p_,q_][t_] :=
	{(a	+ b*Cos[q*t])*Cos[p*t],
	(a + b*Cos[q*t])*Sin[p*t],
	c*Sin[q*t]}
	//////////////////////////////////
	*/

	// WARNING:
	// date: 01-17-2005
	// for : solving the double	covering problem
	// :( needs	the	prime numbers to make sure the cases are accounted for.	But
	// this	handles	only a few cases. A	prime-number generater would make it generic.
	if(m==n)
		div=(double)m;

	if(m>n)	{
		mod	= m%n;
		if(mod==0) {
			div=(double)n;
		}
		else {
			if(m%2==0 && n%2==0	&& mod%2==0){
				if((m%mod)%2==0)
				{
					if(m%4==0 && n%4==0)
						div=(double)4;
					else
						div=(double)2;
				}
				else 
					div=(double)mod;
			}

			if(m%3==0 && n%3==0	&& mod%3==0){
				if((m%mod)%3==0)
				{
					if(m%6==0 && n%6==0)
						div=(double)6;
					else
						div=(double)3;
				}			
				else
					div=(double)mod;
			}
			
			if(m%5==0 && n%5==0	&& mod%5==0){
				if((n%mod)%5==0)
					div=(double)5;
				else
					div=(double)mod;
			}

			if(m%7==0 && n%7==0	&& mod%7==0){
				if((n%mod)%7==0)
					div=(double)7;
				else
					div=(double)mod;
			}
		}
		
	}
	
	else {
		mod	= n%m;
		if(mod==0) {
			div=(double)m;
		}
		else {
			if(m%2==0 && n%2==0	&& mod%2==0){
				if((m%mod)%2==0)
				{
					if(m%4==0 && n%4==0)
						div=(double)4;
					else
						div=(double)2;
				}
				else
					div=(double)mod;
			}
			if(m%3==0 && n%3==0	&& mod%3==0){
				if((m%mod)%3==0)
				{
					if(m%6==0 && n%6==0)
						div=(double)6;
					else
						div=(double)3;
				}			
				else
					div=(double)mod;
			}
			if(m%5==0 && n%5==0	&& mod%5==0){
				if((m%mod)%5==0)
					div=(double)5;
				else
					div=(double)mod;
			}
			if(m%7==0 && n%7==0	&& mod%7==0){
				if((m%mod)%7==0)
					div=(double)7;
				else
					div=(double)mod;
			}
		}
	}
	//@@

	inc	= Max(0.04,mo2->increment/Max(m,n));
	mo2->numPoints = 0;
	if(div>0) {
		//while(t<=2*M_PI/div) {
		if(!mo2->go_around_twice)
			maxAngle = 2.0;
		else
			maxAngle = 4.0;
		while(t<=maxAngle*M_PI/div) {
			
			mo2->pntList[i].p3.d[0]	= (r + a*cos(n*t))*cos(m*t);
			mo2->pntList[i].p3.d[1]	= (r + a*cos(n*t))*sin(m*t);
			mo2->pntList[i].p3.d[2]	= b*sin( n*t );
			//fprintf(stderr,"pntList[%d] =	%f,	%f,	%f\n",i, mo2->pntList[i].p3.d[0],mo2->pntList[i].p3.d[1],mo2->pntList[i].p3.d[2]);
			i++;
			t =	t+inc;
		}
		//NOTE: original> 
		mo2->numPoints = i-1;
		//mo2->numPoints = 30;
	}
}

//# Compute Knot Tangents and First-Derivative of Tangents
void makeTangent ( void	) 
{
	double p1[3],p2[3];				//points on	spline	  
	double len;
	
	double v[3];
	int	   i = 0;

	//////////////// BUILD Tangents	/////////////////////////////////
	for	( i=0; i<(mo2->numPoints); i++ ) {

		p1[0] =	mo2->pntList[i].p3.d[0];
		p1[1] =	mo2->pntList[i].p3.d[1];
		p1[2] =	mo2->pntList[i].p3.d[2];

		if(i==(mo2->numPoints-1)) {
			p2[0] =	mo2->pntList[0].p3.d[0];
			p2[1] =	mo2->pntList[0].p3.d[1];
			p2[2] =	mo2->pntList[0].p3.d[2];
		}
		else {
			p2[0] =	mo2->pntList[i+1].p3.d[0];
			p2[1] =	mo2->pntList[i+1].p3.d[1];
			p2[2] =	mo2->pntList[i+1].p3.d[2];
		}
		//printf("[%d] Point2 = %f, %f ,%f\n",i,p2[0],p2[1],p2[2]);

		v[0] = p2[0]-p1[0];
		v[1] = p2[1]-p1[1];
		v[2] = p2[2]-p1[2];
	
		len	= length3D(v);
	
		v[0] = v[0]/len;
		v[1] = v[1]/len;
		v[2] = v[2]/len;

		mo2->TangList[i].p3.d[0] = v[0];
		mo2->TangList[i].p3.d[1] = v[1];
		mo2->TangList[i].p3.d[2] = v[2];
		//printf("[%d] Tang = %f, %f ,%f\n",i,v[0],v[1],v[2]);
	}
}

void makeTangentDerivatives	( void ) 
{
	int	   i = 0;
	double t = 0;
	double len	 = 0.0;
	
	double T0[3];
	double T1[3];
	double TT[3];

	for(i=0; i<mo2->numPoints; i++) {
		t=0;
		T0[0] =	mo2->TangList[i].p3.d[0];;
		T0[1] =	mo2->TangList[i].p3.d[1];;
		T0[2] =	mo2->TangList[i].p3.d[2];;
		
		if(i==(mo2->numPoints-1)) {
			T1[0] =	mo2->TangList[0].p3.d[0];;
			T1[1] =	mo2->TangList[0].p3.d[1];;
			T1[2] =	mo2->TangList[0].p3.d[2];;
		}
		else {
			T1[0] =	mo2->TangList[i+1].p3.d[0];;
			T1[1] =	mo2->TangList[i+1].p3.d[1];;
			T1[2] =	mo2->TangList[i+1].p3.d[2];;
		}
		minus3D(T1,T0,TT);

		len	= length3D(TT);
		
		TT[0] =	TT[0]/len;
		TT[1] =	TT[1]/len;
		TT[2] =	TT[2]/len;

		mo2->dTangList[i].p3.d[0] =	TT[0];
		mo2->dTangList[i].p3.d[1] =	TT[1];
		mo2->dTangList[i].p3.d[2] =	TT[2];
		//printf("[%d] dTang = %f, %f ,%f\n",i,TT[0],TT[1],TT[2]);
	}
}

//# Compute Frames
void computeParallelTransportFrame	( void )
{
	int		i=0;			// index into points array
	double	t[3];			// arbitrary vector
	double	v[3];			// vector for Tangent
	double	s[3];			// vector for third	orthonormal	component
	double	v0[3];			// vector for initial normal vector
	
	double	len;			// vector lenght
	double	theta =	0.0;			// PT initial frame	angle
	double	t1[3];			// vector for Tangent
	double	t2[3];			// vector for Tangent
	double	cross_T1T2[3];			// vector for tangent-tangent cross	product
	int		numPointsToCompute = 0;	// local variable for number of	points for frame calculation
	numPointsToCompute = mo2->numPoints;

	//rings
	//double	mat[16];
	//double	qtheta= 0.0;
	//double	qvec[3];
	//double	delta = 0.2;
	//double	dot_vt = 0.0;
	//double	phi= 0.0;
	//quaternion *p,*q,*u, *qf, *b;
	//b = (quaternion*) my_malloc(sizeof(quaternion));
	//p =	(quaternion*) malloc(sizeof(quaternion));
	//q =	(quaternion*) malloc(sizeof(quaternion));
	//u =	(quaternion*) malloc(sizeof(quaternion));
	//qf = (quaternion*) malloc(sizeof(quaternion));

	assert((mo2->numPoints)>0);
	if(numPointsToCompute >	2) {

		// make	initial	Normal Vector
		v[0] = mo2->TangList[0].p3.d[0];
		v[1] = mo2->TangList[0].p3.d[1];
		v[2] = mo2->TangList[0].p3.d[2];
		
		// select some direction different from	v
		t[0] = 0.0;
		t[1] = 0.0;
		t[2] = 1.0;
		
		cross3D(v,t,v0);
		len=length3D(v0);
		v0[0] /= len;
		v0[1] /= len;
		v0[2] /= len;
		
		// rotate about	v
		rotate3D_3(v0, v, (mo2->startTheta)/2.0);
		len=length3D(v0);
		v0[0] /= len;
		v0[1] /= len;
		v0[2] /= len;

			// copy	v0 into	basisRList
		mo2->basisRList[0].p3.d[0] = v0[0];
		mo2->basisRList[0].p3.d[1] = v0[1];
		mo2->basisRList[0].p3.d[2] = v0[2];
		
		cross3D(v0,v,s);
		
			// copy	s into basisSList
		mo2->basisSList[0].p3.d[0] = s[0];
		mo2->basisSList[0].p3.d[1] = s[1];
		mo2->basisSList[0].p3.d[2] = s[2];
		for(i=0;i<(numPointsToCompute-1);i++) {
			
			t1[0] =	mo2->TangList[i].p3.d[0];
			t1[1] =	mo2->TangList[i].p3.d[1];
			t1[2] =	mo2->TangList[i].p3.d[2];

			t2[0] =	mo2->TangList[i+1].p3.d[0];
			t2[1] =	mo2->TangList[i+1].p3.d[1];
			t2[2] =	mo2->TangList[i+1].p3.d[2];

			cross3D(t1,t2, cross_T1T2);
			len	= length3D(cross_T1T2);

			if(len == 0) {
					//copy current basisR into next	basisRList
				mo2->basisRList[i+1].p3.d[0] = mo2->basisRList[i].p3.d[0];
				mo2->basisRList[i+1].p3.d[1] = mo2->basisRList[i].p3.d[1];
				mo2->basisRList[i+1].p3.d[2] = mo2->basisRList[i].p3.d[2];
				
					//copy basisRList into v0
				v0[0] =	mo2->basisRList[i+1].p3.d[0];
				v0[1] =	mo2->basisRList[i+1].p3.d[1];
				v0[2] =	mo2->basisRList[i+1].p3.d[2];
			}
			else {
				cross_T1T2[0] /= len;
				cross_T1T2[1] /= len;
				cross_T1T2[2] /= len;
				theta =	acos(dot3D(t1,t2));
				rotate3D_3(v0, cross_T1T2, theta);
				
				//copy v0 into basisRList
				mo2->basisRList[i+1].p3.d[0] = v0[0];
				mo2->basisRList[i+1].p3.d[1] = v0[1];
				mo2->basisRList[i+1].p3.d[2] = v0[2];
			}
			
			cross3D(v0,t2,s);
			len	= length3D(s);
			s[0] /=	len;
			s[1] /=	len;
			s[2] /=	len;

				//printf("s[%d]	= %f, %f, %f\n",i,s[0],s[1],s[2]);
			
				// copy	s into basisSList
			mo2->basisSList[i+1].p3.d[0] = s[0];
			mo2->basisSList[i+1].p3.d[1] = s[1];
			mo2->basisSList[i+1].p3.d[2] = s[2];
		}
	}
}

void computeFrenetFrame	( void )
{
	int	   i	 = 0;		// index into points array

	double norm_crossT_TT;	// lenght of various vectors
	double cross_t_tt[3];	// dot and cross product results
	double cross_BV[3];		// vector for cross	product	of Binormal	and	Tangent
	double t[3];			// vector for Tangent
	double tt[3];			// vector for first	derivative of Tangent
	int	   numPointsToCompute =	0;	// local variable for number of	points for frame calculation
	
	numPointsToCompute =  mo2->numPoints;

	assert(numPointsToCompute);
	
	if(numPointsToCompute >	2) {

		for(i=0;i<(numPointsToCompute);i++) {
			
			mo2->fsTList[i].p3.d[0]	= mo2->TangList[i].p3.d[0];
			mo2->fsTList[i].p3.d[1]	= mo2->TangList[i].p3.d[1];
			mo2->fsTList[i].p3.d[2]	= mo2->TangList[i].p3.d[2];

			t[0] = mo2->TangList[i].p3.d[0];
			t[1] = mo2->TangList[i].p3.d[1];
			t[2] = mo2->TangList[i].p3.d[2];
			
			tt[0] =	mo2->dTangList[i].p3.d[0];
			tt[1] =	mo2->dTangList[i].p3.d[1];
			tt[2] =	mo2->dTangList[i].p3.d[2];
			
			//obtain cross product of tangent and derivative of	tangent
			cross3D(t, tt, cross_t_tt);
			
			//obtain length	of vector cross_vt
			norm_crossT_TT = length3D(cross_t_tt);
			
			//normalize	vector cross_vt
			cross_t_tt[0] =	cross_t_tt[0]/norm_crossT_TT;
			cross_t_tt[1] =	cross_t_tt[1]/norm_crossT_TT;
			cross_t_tt[2] =	cross_t_tt[2]/norm_crossT_TT;
			
		   //save cross_vt into	vector list
			mo2->basisRList[i].p3.d[0] = cross_t_tt[0];
			mo2->basisRList[i].p3.d[1] = cross_t_tt[1];
			mo2->basisRList[i].p3.d[2] = cross_t_tt[2];

			//obtain cross product of Binormal and Tangent
			//cross3D(frenetBList[i].d,	frenetTList[i].d, cross_BV);
			cross3D(cross_t_tt,t, cross_BV);
			
			//save cross_BV	into vector	list
			mo2->basisSList[i].p3.d[0] = cross_BV[0];
			mo2->basisSList[i].p3.d[1] = cross_BV[1];
			mo2->basisSList[i].p3.d[2] = cross_BV[2];
		}
	}
}


//# Compute Frame Quaternion-Maps
void computeCurveQuaternionMap ( void )	
{
	mo2->markFrameCounter =	0;
	constructKnot();
	makeTangent();
	makeTangentDerivatives();

	if(mo2->whichFrame == PARALLELTRANSPORT) {
		computeParallelTransportQuaternionMap2();
	}
	if(mo2->whichFrame == FRENETSERRET)	{
		computeFrenetQuaternionMap8();
	}
}

void computeParallelTransportQuaternionMap ( void )	
{
	int		i = 0;
	int		j = 0;
	double	t1[3], t2[3], t[3];
	double	v[3], s[3];
	double	v0[3];
	
	double len,	dot_vt;
	double cross_vv0[3];
	double cross_T1T2[3];
	double theta = 0.0;
	double phi	 = 0.0;

	//# Quat rings
	double	qtheta = 0.0;
	double	qvec[3];
	double	scalar=0.0;
	double	angle = 0.0;
	double  vecpart = 0.0;
	double	qvecpart[3];
	Vertex  **prevList;
	Vertex	**firstList;
	Vertex	**currList;
	int		counter = 0;
	//double	normvec[3];
	double	delta = 0.2;

	firstList = (Vertex**) my_malloc(sizeof(Vertex)*500);
	for(i=0;i<500;i++)
		firstList[i] = (Vertex*) my_malloc(sizeof(Vertex));

	prevList = (Vertex**) my_malloc(sizeof(Vertex)*500);
	for(i=0;i<500;i++)
		prevList[i] = (Vertex*) my_malloc(sizeof(Vertex));

	currList = (Vertex**) my_malloc(sizeof(Vertex)*500);
	for(i=0;i<500;i++)
		currList[i] = (Vertex*) my_malloc(sizeof(Vertex));

	quaternion *p,*q,*u, *qf;
	p =	(quaternion*) malloc(sizeof(quaternion));
	q =	(quaternion*) malloc(sizeof(quaternion));
	u =	(quaternion*) malloc(sizeof(quaternion));
	qf = (quaternion*) malloc(sizeof(quaternion));

	if(mo2->numPoints >	2) {

		//----------------------------------------------->
		//@	take initial tangent vector	and	an arbitrary
		//	vector 't',	compute	cross-product and obtain
		//	the	normalized vector normal to	first tangent
		v[0] = mo2->TangList[0].p3.d[0];
		v[1] = mo2->TangList[0].p3.d[1];
		v[2] = mo2->TangList[0].p3.d[2];
		

		if(mo2->qshowQuatRings ) {
			glPushAttrib(GL_ALL_ATTRIB_BITS);
			glDisable(GL_LIGHTING);
			glColor3f(0.0f,0.0f,0.0f);
			glLineWidth(0.5f);
			qtheta = 0.0;
			j=0;

			//glBegin(GL_LINE_STRIP);
			while(qtheta<(4.0*M_PI)) {
				makeFamilyOfQuaternions(qtheta, v, qf);
				getQuatComp(mo2->qwhichQuatComponent, qf, qvecpart);
				prevList[j]->p3.d[0] = qvecpart[0];
				prevList[j]->p3.d[1] = qvecpart[1];
				prevList[j]->p3.d[2] = qvecpart[2];

				firstList[j]->p3.d[0] = qvecpart[0];
				firstList[j]->p3.d[1] = qvecpart[1];
				firstList[j]->p3.d[2] = qvecpart[2];
				//glVertex3f(qvecpart[0], qvecpart[1], qvecpart[2]);
				qtheta = qtheta + delta;
				j++;
			}
			//glEnd();
			counter = j;

			//glColor3f(0.5f,0.0f,0.0f);
			//for(j=0; j<counter; j++) {
			//	glPushMatrix();
			//		glTranslatef(prevList[j]->p3.d[0],prevList[j]->p3.d[1],prevList[j]->p3.d[2]);
			//		glutSolidSphere(0.01,16,16);
			//	glPopMatrix();
			//}
			glPopAttrib();
		}

		t[0] = 0.0;
		t[1] = 0.0;
		t[2] = 1.0;

		dot_vt = dot3D(v,t);
		phi	   = acos(dot_vt);
		
		cross3D(v,t,v0);
		len=length3D(v0);
		v0[0] /= len;
		v0[1] /= len;
		v0[2] /= len;
		// <----------------------------------------------

		//----------------------------------------------->
		//@	build the first	quaternion based on	first 
		//	tangent	vector and starting	angle
		mo2->quatList[0].r = cos(mo2->startTheta/2.0);
		mo2->quatList[0].i = v0[0]*sin(mo2->startTheta/2.0);
		mo2->quatList[0].j = v0[1]*sin(mo2->startTheta/2.0);
		mo2->quatList[0].k = v0[2]*sin(mo2->startTheta/2.0);
		// <----------------------------------------------

		mo2->basisQRList[0].p3.d[0]	= v0[0];
		mo2->basisQRList[0].p3.d[1]	= v0[1];
		mo2->basisQRList[0].p3.d[2]	= v0[2];
		
		cross3D(v0,v,cross_vv0);
		mo2->basisQSList[0].p3.d[0]	= cross_vv0[0];
		mo2->basisQSList[0].p3.d[1]	= cross_vv0[1];
		mo2->basisQSList[0].p3.d[2]	= cross_vv0[2];
		
		/////////////////////////////////////////////////////
		// build the Quaternions corresponding to Frames
		/////////////////////////////////////////////////////
		for(i=0;i<(mo2->numPoints);i++)	{
			
			t1[0] =	mo2->TangList[i].p3.d[0];
			t1[1] =	mo2->TangList[i].p3.d[1];
			t1[2] =	mo2->TangList[i].p3.d[2];

			t2[0] =	mo2->TangList[i+1].p3.d[0];
			t2[1] =	mo2->TangList[i+1].p3.d[1];
			t2[2] =	mo2->TangList[i+1].p3.d[2];

			cross3D(t1,t2, cross_T1T2);
			len	= length3D(cross_T1T2);

			p->r = mo2->quatList[i].r;
			p->i = mo2->quatList[i].i;
			p->j = mo2->quatList[i].j;
			p->k = mo2->quatList[i].k;
			
			if(len == 0) {
				mo2->basisQRList[i+1].p3.d[0] =	mo2->basisQRList[i].p3.d[0];
				mo2->basisQRList[i+1].p3.d[1] =	mo2->basisQRList[i].p3.d[1];
				mo2->basisQRList[i+1].p3.d[2] =	mo2->basisQRList[i].p3.d[2];
			}
			else {

				cross_T1T2[0] /= len;
				cross_T1T2[1] /= len;
				cross_T1T2[2] /= len;
				theta =	acos(dot3D(t1,t2));

				q->r = cos(theta/2.0);
				q->i = cross_T1T2[0]*sin(theta/2.0);
				q->j = cross_T1T2[1]*sin(theta/2.0);
				q->k = cross_T1T2[2]*sin(theta/2.0);
				
				quatQuatProduct	(q,	p, u);
				quatNormalize(u);

				rotate3D_3(v0, cross_T1T2, theta);
				
				mo2->basisQRList[i+1].p3.d[0] =	v0[0];
				mo2->basisQRList[i+1].p3.d[1] =	v0[1];
				mo2->basisQRList[i+1].p3.d[2] =	v0[2];
				
				mo2->quatList[i+1].r = u->r;
				mo2->quatList[i+1].i = u->i;
				mo2->quatList[i+1].j = u->j;
				mo2->quatList[i+1].k = u->k;

			}
			if(mo2->qshowQuatRings && i>=1) {
				glPushAttrib(GL_ALL_ATTRIB_BITS);
				glEnable(GL_LIGHTING);
				glColor3f(0.0f,0.0f,0.0f);
				glLineWidth(0.5f);
				//printf("norm(v) = %f\n",length3D(v));
				qtheta = 0.0;
				j = 0;
				//printf("[0] v = %f, %f,%f\n",v[0],v[1],v[2]);
				//glBegin(GL_LINE_STRIP);
				while(qtheta<(4.0*M_PI)) {
					makeFamilyOfQuaternions(qtheta, t1, qf);
					getQuatComp(mo2->qwhichQuatComponent, qf, qvecpart);
					currList[j]->p3.d[0] = qvecpart[0];
					currList[j]->p3.d[1] = qvecpart[1];
					currList[j]->p3.d[2] = qvecpart[2];
					//printf("[%f] qf = %f, %f, %f, %f\n",qtheta,qf->r, qf->i, qf->j, qf->k);
					//glPushMatrix();
						//glVertex3f(qvecpart[0], qvecpart[1], qvecpart[2]);
						//glutSolidSphere(0.03,16,16);
					//glPopMatrix();
					qtheta = qtheta + delta;
					j++;
				}
				//glEnd();
				counter = j;
			
				//if(i==1){
				//	glColor3f(0.5f,0.0f,0.0f);
				//	for(j=0; j<counter; j++) {

				//		glPushMatrix();
				//			glTranslatef(currList[j]->p3.d[0],currList[j]->p3.d[1],currList[j]->p3.d[2]);
				//			if(j==(counter-1) || j==0)
				//				glutSolidSphere(0.0075,16,16);
				//		glPopMatrix();
				//	}
				//}
				if(mo2->qshowQuatRingsSurface) {
					glEnable(GL_COLOR_MATERIAL);
					glColor3f(0.8f, 0.7f, 0.0f);
					qtheta = 0.0;
					j=0;
					//while(qtheta <=(4.0*M_PI)) {
					for(j=0;j<counter;j++) {
						glBegin(GL_QUADS);
							//if(qtheta<(4.0*M_PI-delta)) {
							if(j<(counter-1)) {
								glNormal3f(prevList[j]->p3.d[0],	prevList[j]->p3.d[1],	prevList[j]->p3.d[2]);
								glVertex3f(prevList[j]->p3.d[0],	prevList[j]->p3.d[1],	prevList[j]->p3.d[2]); 
								glVertex3f(prevList[j+1]->p3.d[0],	prevList[j+1]->p3.d[1],	prevList[j+1]->p3.d[2]); 
								glVertex3f(currList[j+1]->p3.d[0],	currList[j+1]->p3.d[1],	currList[j+1]->p3.d[2]); 
								glVertex3f(currList[j]->p3.d[0],	currList[j]->p3.d[1],	currList[j]->p3.d[2]); 
							}
							else {
								glNormal3f(prevList[0]->p3.d[0],	prevList[0]->p3.d[1],	prevList[0]->p3.d[2]);
								glVertex3f(prevList[0]->p3.d[0],	prevList[0]->p3.d[1],	prevList[0]->p3.d[2]); 
								glVertex3f(prevList[j]->p3.d[0],	prevList[j]->p3.d[1],	prevList[j]->p3.d[2]); 
								glVertex3f(currList[j]->p3.d[0],	currList[j]->p3.d[1],	currList[j]->p3.d[2]); 
								glVertex3f(currList[0]->p3.d[0],	currList[0]->p3.d[1],	currList[0]->p3.d[2]); 
							}
						glEnd();
						//qtheta = qtheta + delta;
						//j++;
					}
					if(i==(mo2->numPoints-1)) {
						qtheta = 0.0;
						j=0;
						//while(qtheta <=(4.0*M_PI-delta)) {
						for(j=0;j<counter;j++) {
							glBegin(GL_QUADS);
							if(j<(counter-1)) {
								glNormal3f(firstList[j]->p3.d[0],	firstList[j]->p3.d[1],	firstList[j]->p3.d[2]);
								glVertex3f(firstList[j]->p3.d[0],	firstList[j]->p3.d[1],	firstList[j]->p3.d[2]); 
								glVertex3f(firstList[j+1]->p3.d[0],	firstList[j+1]->p3.d[1],firstList[j+1]->p3.d[2]); 
								glVertex3f(currList[j+1]->p3.d[0],	currList[j+1]->p3.d[1],	currList[j+1]->p3.d[2]); 
								glVertex3f(currList[j]->p3.d[0],	currList[j]->p3.d[1],	currList[j]->p3.d[2]); 
							}
							else {
								glNormal3f(prevList[0]->p3.d[0],	prevList[0]->p3.d[1],	prevList[0]->p3.d[2]);
								glVertex3f(prevList[0]->p3.d[0],	prevList[0]->p3.d[1],	prevList[0]->p3.d[2]); 
								glVertex3f(prevList[j]->p3.d[0],	prevList[j]->p3.d[1],	prevList[j]->p3.d[2]); 
								glVertex3f(currList[j]->p3.d[0],	currList[j]->p3.d[1],	currList[j]->p3.d[2]); 
								glVertex3f(currList[0]->p3.d[0],	currList[0]->p3.d[1],	currList[0]->p3.d[2]); 
							}
							glEnd();
							//qtheta = qtheta + delta;
							//j++;
						}

					}

				}
				if(mo2->qshowQuatRingsSurfaceMesh && i>=1) {
					glColor3f(0.2f, 0.2f, 0.2f);
					qtheta = 0.0;
					j=0;
					//while(qtheta <(4.0*M_PI-0.25)) {
					for(j=0; j<counter; j++) {
						glBegin(GL_LINE_STRIP);
							glVertex3f(prevList[j]->p3.d[0],prevList[j]->p3.d[1],prevList[j]->p3.d[2]); 
							glVertex3f(prevList[j+1]->p3.d[0],prevList[j+1]->p3.d[1],prevList[j+1]->p3.d[2]); 
							glVertex3f(currList[j+1]->p3.d[0],currList[j+1]->p3.d[1],currList[j+1]->p3.d[2]); 
							glVertex3f(currList[j]->p3.d[0],currList[j]->p3.d[1],currList[j]->p3.d[2]); 
						glEnd();
						//qtheta = qtheta + 0.25;
						//j++;
					}
				}

				if(i>=1) {
					qtheta=0.0;
					j=0;
					while(qtheta<(4.0*M_PI)) {
						prevList[j]->p3.d[0] = currList[j]->p3.d[0];
						prevList[j]->p3.d[1] = currList[j]->p3.d[1];
						prevList[j]->p3.d[2] = currList[j]->p3.d[2];
						qtheta = qtheta + delta;
						j++;
					}	
				}
				glPopAttrib();
			}

		}
		// copy	vector as Third	component of 
		// Orthonormal basis
		len	= length3D(s);
		s[0] /=	len;
		s[1] /=	len;
		s[2] /=	len;

		mo2->basisQSList[i+1].p3.d[0] =	s[0];
		mo2->basisQSList[i+1].p3.d[1] =	s[1];
		mo2->basisQSList[i+1].p3.d[2] =	s[2];
		
		//---------------------->
		if(mo2->qshowQuatRings & 0) {
			//## QUATERNION RINGS
			glPushAttrib(GL_ALL_ATTRIB_BITS | GL_LIGHTING_BIT | GL_LINE_BIT);
			glDisable(GL_LIGHTING);
			glLineWidth(2.25f);
			for(i=0;i<mo2->numPoints;i++) {
				scalar = mo2->quatList[i].r;
				angle  = 2.0*acos(scalar);
				vecpart= sin(angle/2.0);
				qvecpart[0] = mo2->quatList[i].i/vecpart;
				qvecpart[1] = mo2->quatList[i].j/vecpart;
				qvecpart[2] = mo2->quatList[i].k/vecpart;
				if(i%11==0) {
					qtheta=0.0;
					if(i==11)
						glColor3f(0.0f,0.9f,0.4f);
					else
						glColor3f(0.3f,0.3f,0.3f);
					glBegin(GL_LINE_STRIP);
					while(qtheta<=16.0) {
						makeFamilyOfQuaternions(qtheta,qvecpart,qf);
						getQuatComp(mo2->qwhichQuatComponent,qf,qvec);
						glVertex3f(qvec[0],qvec[1],qvec[2]);
						qtheta = qtheta + 0.05;
					}	
					glEnd();
				}
			}
			glPushMatrix();
				glLineWidth(1.25f);
				glColor3f(0.4f,0.4f,0.4f);
				glTranslatef(0.0f,0.0f,0.0f);
				glutWireSphere(1.0f,16,16);
			glPopMatrix();
			glPopAttrib();
		}
		//<----------------------
	}
	//drawRings();
	for(i=0;i<500;i++)
		my_free(firstList[i]);
	my_free(firstList);
	for(i=0;i<500;i++)
		my_free(prevList[i]);
	my_free(prevList);
	for(i=0;i<500;i++)
		my_free(currList[i]);
	my_free(currList);

}

void computeParallelTransportQuaternionMap2 ( void )	
{
	int		i = 0;
	//int		j = 0;
	double	t1[3], t2[3], t[3];
	double	v[3];
	//double	s[3];
	double	v0[3];
	
	double len,	dot_vt;
	//double cross_vv0[3];
	double cross_T1T2[3];
	double theta = 0.0;
	double phi	 = 0.0;

	//# Quat rings
	//double	z[3] = {0,0,1};
	//double	qtheta = 0.0;
	//double	qvec[3];
	//double	axis[3];
	//double	crossZV[3];
	//double	crossZV_len;
	//double	anglezv = 0.0;
	//double	rangle = 0.0;
	//double	angle = 0.0;
	//double  vecpart = 0.0;
	//double	qvecpart[3];
	//Vertex	**qringList;
	//Vertex  **prevList;
	//Vertex	**firstList;
	//Vertex	**currList;
	//int		counter = 0;
	//double		vbase[3];
	//double delta = 0.0;
	//double temp[3];

	quaternion *p,*q,*u, *qf, *b, *base;
	b = (quaternion*) my_malloc(sizeof(quaternion));
	p =	(quaternion*) malloc(sizeof(quaternion));
	q =	(quaternion*) malloc(sizeof(quaternion));
	u =	(quaternion*) malloc(sizeof(quaternion));
	qf = (quaternion*) malloc(sizeof(quaternion));
	base =	(quaternion*) malloc(sizeof(quaternion));

	constructKnot();
	makeTangent();

	if(mo2->numPoints >	2) {

		//----------------------------------------------->
		//@	take initial tangent vector	and	an arbitrary
		//	vector 't',	compute	cross-product and obtain
		//	the	normalized vector normal to	first tangent
		v[0] = mo2->TangList[0].p3.d[0];
		v[1] = mo2->TangList[0].p3.d[1];
		v[2] = mo2->TangList[0].p3.d[2];
	
		t[0] = 0.0;
		t[1] = 0.0;
		t[2] = 1.0;

		dot_vt = dot3D(t,v);
		phi	   = acos(dot_vt);
		
		cross3D(v,t,v0);
		len=length3D(v0);
		v0[0] /= len;
		v0[1] /= len;
		v0[2] /= len;
		
		// <----------------------------------------------

		//----------------------------------------------->
		//@	build the first	quaternion based on	first 
		//	tangent	vector and starting	angle
		//mo2->quatList[0].r = cos(mo2->startTheta/2.0);
		//mo2->quatList[0].i = v[0]*sin(mo2->startTheta/2.0);
		//mo2->quatList[0].j = v[1]*sin(mo2->startTheta/2.0);
		//mo2->quatList[0].k = v[2]*sin(mo2->startTheta/2.0);

		p->r = 0.0;
		p->i = 0.0;
		p->j = 1.0;
		p->k = 0.0;

		q->r = cos(mo2->startTheta/2.0);
		q->i = v[0]*sin(mo2->startTheta/2.0);
		q->j = v[1]*sin(mo2->startTheta/2.0);
		q->k = v[2]*sin(mo2->startTheta/2.0);

		quatQuatProduct(q, p, u);
		quatNormalize(u);
		mo2->quatList[0].r = u->r;
		mo2->quatList[0].i = u->i;
		mo2->quatList[0].j = u->j;
		mo2->quatList[0].k = u->k;

		// <----------------------------------------------

		/////////////////////////////////////////////////////
		// build the Quaternions corresponding to Frames
		/////////////////////////////////////////////////////
		for(i=1;i<(mo2->numPoints);i++)	{
			
			// #1
			t1[0] =	mo2->TangList[i-1].p3.d[0];
			t1[1] =	mo2->TangList[i-1].p3.d[1];
			t1[2] =	mo2->TangList[i-1].p3.d[2];

			t2[0] =	mo2->TangList[i].p3.d[0];
			t2[1] =	mo2->TangList[i].p3.d[1];
			t2[2] =	mo2->TangList[i].p3.d[2];

			cross3D(t1,t2, cross_T1T2);
			len	= length3D(cross_T1T2);

			// #2. Previous frame
			p->r = mo2->quatList[i-1].r;
			p->i = mo2->quatList[i-1].i;
			p->j = mo2->quatList[i-1].j;
			p->k = mo2->quatList[i-1].k;
			
			if(len == 0) {
				//DO NOTHING!
				quatCopy(p,&(mo2->quatList[i]));
			}
			else {
				
				// #3
				cross_T1T2[0] /= len;
				cross_T1T2[1] /= len;
				cross_T1T2[2] /= len;
				theta =	acos(dot3D(t1,t2));

				q->r = cos(theta/2.0);
				q->i = cross_T1T2[0]*sin(theta/2.0);
				q->j = cross_T1T2[1]*sin(theta/2.0);
				q->k = cross_T1T2[2]*sin(theta/2.0);
				quatQuatProduct ( q, p, u);

				quatNormalize(u);
				
				mo2->quatList[i].r = u->r;
				mo2->quatList[i].i = u->i;
				mo2->quatList[i].j = u->j;
				mo2->quatList[i].k = u->k;

			}
		}//# end FOR

		//glPushAttrib(GL_ALL_ATTRIB_BITS);
		//glDisable(GL_LIGHTING);
		//glColor3f(0,0,0);

		//// Initialize
		//t[0] = 0.0; 
		//t[1] = 0.0;
		//t[2] = 1.0;
		//qvec[0] = qvec[1] = qvec[2] = 0.0;

		//v[0] = mo2->TangList[0].p3.d[0];
		//v[1] = mo2->TangList[0].p3.d[1];
		//v[2] = mo2->TangList[0].p3.d[2];

		//dot_vt = dot3D(t,v);
		//phi	   = acos(dot_vt);
		//
		//cross3D(v,t,v0);
		//len=length3D(v0);
		//v0[0] /= len;
		//v0[1] /= len;
		//v0[2] /= len;
		//
		//p->r = 0.0; 
		//p->i = 0.0; 
		//p->j = 0.0; 
		//p->k = 1.0;
		//
		//q->r =		cos(mo2->startTheta/2.0); 
		//q->i = v[0]*sin(mo2->startTheta/2.0); 
		//q->j = v[1]*sin(mo2->startTheta/2.0); 
		//q->k = v[2]*sin(mo2->startTheta/2.0);

		//quatQuatProduct(q, p, base);
		////getQuatComp(mo2->qwhichQuatComponent,base,qvec);
		////glPushMatrix();
		////glTranslatef(qvec[0],qvec[1],qvec[2]);
		////glutSolidSphere(0.03,8,8);
		////glPopMatrix();

		//int counter = 0;
		//delta = 0.2;

		//for(i=0; i<mo2->numPoints; i++) {
		//	qtheta = 0.0;
		//	counter= 0;
		//	v[0] = mo2->TangList[i].p3.d[0];
		//	v[1] = mo2->TangList[i].p3.d[1];
		//	v[2] = mo2->TangList[i].p3.d[2];

		//	if(i==0)
		//		glColor3f(1.0f,1.0f,0.0f);
		//	else
		//		glColor3f(0.0f,0.0f,0.0f);

		//	glBegin(GL_LINE_STRIP);
		//	while ( qtheta <=4.0*M_PI+delta) {

		//		q->r =		cos(qtheta/2.0); 
		//		q->i = v[0]*sin(qtheta/2.0); 
		//		q->j = v[1]*sin(qtheta/2.0); 
		//		q->k = v[2]*sin(qtheta/2.0);
		//		quatQuatProduct(q, p, u);
		//		getQuatComp(mo2->qwhichQuatComponent,u,qvec);
		//		//makeFamilyOfQuaternions(qtheta, v, qf);
		//		//getQuatComp(mo2->qwhichQuatComponent,qf,qvec);

		//		if(qtheta==0) {
		//			temp[0] = qvec[0]; temp[1] = qvec[1]; temp[2] = qvec[2];
		//		}
		//		if(qtheta<=(4.0*M_PI))
		//			glVertex3f(qvec[0],qvec[1],qvec[2]);
		//		else
		//			glVertex3f(temp[0],temp[1],temp[2]);
		//			//currList[i]->p3.d[0] = qvec[0]; currList[i]->p3.d[1] = qvec[1]; currList[i]->p3.d[2] = qvec[2];
		//		qtheta += delta;
		//		counter++;
		//	}//#end WHILE
		//	glEnd();

		//	glPushAttrib(GL_ALL_ATTRIB_BITS);
		//	glPushMatrix();
		//	glColor3f(0.0f,0.8f,0.0f);
		//	glTranslatef(temp[0],temp[1],temp[2]);
		//	glutSolidSphere(0.01,8,8);
		//	glPopMatrix();
		//	glPopAttrib();
		//	
		//	//# displays beads on the rings (DEBUG ONLY)
		//	//glEnable(GL_LIGHTING);
		//	//qtheta = 0.0;
		//	//while ( qtheta <=4.0*M_PI+0.2) {
		//	//	//makeFamilyOfQuaternions(qtheta,v,qf);
		//	//	q->r =		cos(qtheta/2.0); 
		//	//	q->i = v[0]*sin(qtheta/2.0); 
		//	//	q->j = v[1]*sin(qtheta/2.0); 
		//	//	q->k = v[2]*sin(qtheta/2.0);
		//	//	quatQuatProduct(q, p, u);

		//	//	getQuatComp(mo2->qwhichQuatComponent,u,qvec);

		//	//	glPushMatrix();
		//	//	glTranslatef(qvec[0],qvec[1],qvec[2]);
		//	//	glutSolidSphere(0.01,8,8);
		//	//	glPopMatrix();
		//	//	qtheta += 0.2;
		//	//}//#end WHILE

		//	//update base ref frame
		//	t1[0] =	mo2->TangList[i].p3.d[0];	t1[1] =	mo2->TangList[i].p3.d[1];	t1[2] =	mo2->TangList[i].p3.d[2];
		//	t2[0] =	mo2->TangList[i+1].p3.d[0]; t2[1] =	mo2->TangList[i+1].p3.d[1]; t2[2] =	mo2->TangList[i+1].p3.d[2];
		//	cross3D(t1,t2, cross_T1T2);
		//	len	= length3D(cross_T1T2);
		//	if(len!=0) {
		//		cross_T1T2[0] /= len;
		//		cross_T1T2[1] /= len;
		//		cross_T1T2[2] /= len;
		//		theta =	acos(dot3D(t1,t2));
		//		
		//		base->r = p->r;
		//		base->i = p->i;
		//		base->j = p->j;
		//		base->k = p->k;
		//		q->r = cos(theta/2.0);
		//		q->i = cross_T1T2[0]*sin(theta/2.0);
		//		q->j = cross_T1T2[1]*sin(theta/2.0);
		//		q->k = cross_T1T2[2]*sin(theta/2.0);
		//		quatQuatProduct ( q, base, p);
		//		quatNormalize(p);
		//	}

		//}//#end FOR

		glPopAttrib();
	}
}

void computeFrenetQuaternionMap	( void )
{
	int		i	  =	0;			// index into points array
//	int		limit =	0;			// max number of frames	to be calculated
	int		start =	0;			// start index for the for-loop

	double	t[3];
	double	tt[3];
	double	len	  =	0.0;

	double	prevBinormal[3];	// vector for previous Binormal	vector
	double	currBinormal[3];	// vector for current Binormal vector
	double	currNormal[3];
	double	dotPrevCurrBinormal;// scalar, dot-product of previous and current Binormal
	
	double	mat[16];			// rotation	matrix
	quaternion *p, *q, *r;		//
	double	dot_PQ;				// quaternion inner-product

	//family>
	double	quatAngle =	0.0;
//	double	v[3];
//	double	z[3];
//	double	crossZV[3];
//	double	dotZV=0.0;
	quaternion *fq,*fb,*ff;
//	double	theta =	0.0;
//	double	dTheta = 4*M_PI/20.0;	

	fb = (quaternion*) my_malloc(sizeof(quaternion));
	fq = (quaternion*) my_malloc(sizeof(quaternion));
	ff = (quaternion*) my_malloc(sizeof(quaternion));
	//<family
	//----------------------------------------------->
	//@	INIT section
	p =	(quaternion*) my_malloc(sizeof(quaternion));
	q =	(quaternion*) my_malloc(sizeof(quaternion));
	r =	(quaternion*) my_malloc(sizeof(quaternion));

	p->r = 0;
	p->i = 0;
	p->j = 0;
	p->k = 0;
	
	dotPrevCurrBinormal	= 0.0;
	// <----------------------------------------------


	if(mo2->numPoints >	2) {

		printf("[%d]\n",i);
		/////////////////////////////////////////////////////
		// make	Initial	Binormal Vector
		/////////////////////////////////////////////////////
		// obtain tangent
		t[0] = mo2->TangList[i].p3.d[0];
		t[1] = mo2->TangList[i].p3.d[1];
		t[2] = mo2->TangList[i].p3.d[2];
		/*t[0] = mo2->fsTList[i].p3.d[0];
		t[1] = mo2->fsTList[i].p3.d[1];
		t[2] = mo2->fsTList[i].p3.d[2];*/
		//printf("t	= %f, %f, %f\n",t[0],t[1],t[2]);
		// obtain first-derivative of tangent
		tt[0] =	mo2->dTangList[i].p3.d[0];
		tt[1] =	mo2->dTangList[i].p3.d[1];
		tt[2] =	mo2->dTangList[i].p3.d[2];

		//----------------------------------------------->
		//@	obtain cross-product of	Tangent	and	First
		//	Derivative of tangent (Binormal	vector)
		cross3D(t, tt, currBinormal);
		len	= length3D(currBinormal);
		currBinormal[0]	= currBinormal[0]/len;
		currBinormal[1]	= currBinormal[1]/len;
		currBinormal[2]	= currBinormal[2]/len;
		/*currBinormal[0] =	mo2->basisRList[i].p3.d[0];
		currBinormal[1]	= mo2->basisRList[i].p3.d[1];
		currBinormal[2]	= mo2->basisRList[i].p3.d[2];*/
		//printf("b	= %f, %f, %f\n",currBinormal[0],currBinormal[1],currBinormal[2]);
		// <----------------------------------------------

		//----------------------------------------------->
		//@	obtain the third orthonormal component
		//	from cross-product of Tangent and Binormal
		cross3D(currBinormal,t,	currNormal);
		len	= length3D(currNormal);
		currNormal[0] =	currNormal[0]/len;
		currNormal[1] =	currNormal[1]/len;
		currNormal[2] =	currNormal[2]/len;
		/*currNormal[0]	= mo2->basisSList[i].p3.d[0];
		currNormal[1] =	mo2->basisSList[i].p3.d[1];
		currNormal[2] =	mo2->basisSList[i].p3.d[2];*/
		//printf("n	= %f, %f, %f\n",currNormal[0],currNormal[1],currNormal[2]);
		// <----------------------------------------------

		//----------------------------------------------->
		//@	Frenet-Serret frame	is given by	the	set	of 
		//	coordinate axes:
		//			_  _  _	
		//		  [	T  N  B	]
		//
		// <----------------------------------------------

		//----------------------------------------------->
		//@	build the 4x4 Rotation matris with the
		//	columns	as the orthonormal basis constructed
		//	above using: Tangent, Binormal and Third component
		i=0;
			/*
			mat[0] = t[0];
			mat[1] = t[1];
			mat[2] = t[2];

			mat[4] = currBinormal[0];
			mat[5] = currBinormal[1];
			mat[6] = currBinormal[2];

			mat[8] = currNormal[0];
			mat[9] = currNormal[1];
			mat[10]= currNormal[2];		  
			*/

		mat[0] = t[0];
		mat[1] = t[1];
		mat[2] = t[2];

		mat[4] = currNormal[0];
		mat[5] = currNormal[1];
		mat[6] = currNormal[2];

		mat[8] = currBinormal[0];
		mat[9] = currBinormal[1];
		mat[10]= currBinormal[2];		

		mat[3]	= mat[7] = mat[11] = mat[12] = mat[13] = mat[14] = 0.0;
		mat[15]	= 1.0;
		// <----------------------------------------------

		//----------------------------------------------->
		//@	obtain the quaternion from rotation	matrix
		//	formed by the orthonormal basis	of the frame
		//	that have three	basis-vectors at a point p:
		//	Tangent, Binormal and Third	component
		ExtractQuaternionFromMat(mat,p);
		quatNormalize(p);	
		// <----------------------------------------------
		
		mo2->quatList[i].r = p->r;
		mo2->quatList[i].i = p->i;
		mo2->quatList[i].j = p->j;
		mo2->quatList[i].k = p->k;				  

		quatAngle =	2*acos(p->r);
		mo2->markQuatAngle[i] =	quatAngle;

		start =	1;

		prevBinormal[0]	= currBinormal[0];
		prevBinormal[1]	= currBinormal[1];
		prevBinormal[2]	= currBinormal[2];

		//printf("******** \n",i);
		/////////////////////////////////////////////////////
		// build the Quaternions corresponding to Frames
		/////////////////////////////////////////////////////
		for(i=start;i<(mo2->numPoints-2);i++) {

			printf("[%d]\n",i);
			// obtain tangent
			t[0] = mo2->TangList[i].p3.d[0];
			t[1] = mo2->TangList[i].p3.d[1];
			t[2] = mo2->TangList[i].p3.d[2];
			/*t[0] = mo2->fsTList[i].p3.d[0];
			t[1] = mo2->fsTList[i].p3.d[1];
			t[2] = mo2->fsTList[i].p3.d[2];*/
			//printf("t	= %f, %f, %f\n",t[0],t[1],t[2]);
			
			// obtain first-derivative of tangent
			tt[0] =	mo2->dTangList[i].p3.d[0];
			tt[1] =	mo2->dTangList[i].p3.d[1];
			tt[2] =	mo2->dTangList[i].p3.d[2];

			//----------------------------------------------->
			//@	obtain cross-product of	Tangent	and	First
			//	Derivative of tangent (Binormal	vector)
			cross3D(t, tt, currBinormal);
			len	= length3D(currBinormal);
			currBinormal[0]	= currBinormal[0]/len;
			currBinormal[1]	= currBinormal[1]/len;
			currBinormal[2]	= currBinormal[2]/len;
			/*currBinormal[0] =	mo2->basisRList[i].p3.d[0];
			currBinormal[1]	= mo2->basisRList[i].p3.d[1];
			currBinormal[2]	= mo2->basisRList[i].p3.d[2];*/
			//printf("b	= %f, %f, %f\n",currBinormal[0],currBinormal[1],currBinormal[2]);
			// <----------------------------------------------

			//----------------------------------------------->
			//@	obtain the third orthonormal component
			//	from cross-product of Tangent and Binormal
			cross3D(currBinormal,t,	currNormal);
			len	= length3D(currNormal);
			currNormal[0] =	currNormal[0]/len;
			currNormal[1] =	currNormal[1]/len;
			currNormal[2] =	currNormal[2]/len;
			/*currNormal[0]	= mo2->basisSList[i].p3.d[0];
			currNormal[1] =	mo2->basisSList[i].p3.d[1];
			currNormal[2] =	mo2->basisSList[i].p3.d[2];*/
			//printf("n	= %f, %f, %f\n",currNormal[0],currNormal[1],currNormal[2]);
			// <----------------------------------------------
			
			//----------------------------------------------->
			//@	check the dot products for orthonormality
			printf("dotProd(t,B) = %f\n",dot3D(t,currBinormal));
			printf("dotProd(t,S) = %f\n",dot3D(t,currNormal));
			printf("dotProd(S,B) = %f\n",dot3D(currNormal,currBinormal));
			// <----------------------------------------------

			//----------------------------------------------->
			//@	not	doing this right now but this is 
			//	one	of the steps. Not sure about this.
			//dotPrevCurrBinormal =	dot3D(prevBinormal,	currBinormal);
			// <----------------------------------------------

			// copy	currBinormal into prevBinormal
			prevBinormal[0]	= currBinormal[0];
			prevBinormal[1]	= currBinormal[1];
			prevBinormal[2]	= currBinormal[2];

			//----------------------------------------------->
			//------------ Frenet to Quaternion	------------->
			
			//@	check on Quaternion	4D inner-product
			/*
			mat[0] = t[0];
			mat[1] = t[1];
			mat[2] = t[2];

			mat[4] = currBinormal[0];
			mat[5] = currBinormal[1];
			mat[6] = currBinormal[2];

			mat[8] = currNormal[0];
			mat[9] = currNormal[1];
			mat[10]= currNormal[2];		  
			*/
			mat[0] = t[0];
			mat[1] = t[1];
			mat[2] = t[2];

			mat[4] = currNormal[0];
			mat[5] = currNormal[1];
			mat[6] = currNormal[2];

			mat[8] = currBinormal[0];
			mat[9] = currBinormal[1];
			mat[10]= currBinormal[2];		

			mat[3]	= mat[7] = mat[11] = mat[12] = mat[13] = mat[14] = 0;
			mat[15]	= 1.0;

			/*
			printf("[%d] \n",i);
			
			printf("t =	%f,	%f,	%f\n",t[0],t[1],t[2]);
			printf("n =	%f,	%f,	%f\n",currNormal[0],currNormal[1],currNormal[2]);
			printf("b =	%f,	%f,	%f\n",currBinormal[0],currBinormal[1],currBinormal[2]);
			*/
			ExtractQuaternionFromMat(mat,q);
			quatNormalize(q);

			//	check if dot product of	neighboring	quaternions
			//	is near	negative one
			dot_PQ = quatQuatDotProduct(p,q);
			//printf("[%d] dot_PQ =	%f\n",i,dot_PQ);
			if(dot_PQ <= 0.0) {

				// flip	one	of the quaternions
				r->r = (q->r);
				r->i = -(q->i);
				r->j = -(q->j);
				r->k = -(q->k);
				quatCopy(r,q);
				
				//printf("	   ... dotProd after flipping =	%f\n",quatQuatDotProduct(p,q));
			}
			if(	(dot_PQ>0.0) &&	(dot_PQ<0.75)) {
				mo2->markFrame[mo2->markFrameCounter]=i;
				mo2->markFrameCounter++;
				//printf("[%d] ...NEAR ZERO	marking	this (+)\n",i);
			}
			
			quatAngle =	2*acos(q->r);
			mo2->markQuatAngle[i] =	quatAngle;
			// <----------------------------------------------
			// <----------------------------------------------

			// save	out	resulting quaternion
			mo2->quatList[i].r = q->r;
			mo2->quatList[i].i = q->i;
			mo2->quatList[i].j = q->j;
			mo2->quatList[i].k = q->k;				  

			// save	current	quaternion as 'previous-quaternion'
			// for next	iteration
			quatCopy(q,p);
			//printf("norm = %f\n",qNorm(q));

			
		}
		mo2->markFrame[mo2->markFrameCounter] =	mo2->numPoints-2;
		mo2->markFrameCounter++;

		/*
		printf("***\n");
		for(i=0;i<markFrameCounter;i++)	{
			printf("%d ",markFrame[i]);
		}
		printf("\n");
		*/
	}
	my_free(fb);
	my_free(fq);
	my_free(ff);
	my_free(p);
	my_free(q);
	my_free(r);
}


void computeFrenetQuaternionMap2 ( void	)
{
	int		i		= 0;			// index into points array
	int		j		= 0;
//	int		limit	= 0;			// max number of frames	to be calculated
	int		start	= 0;			// start index for the for-loop
	int		end		= 0;
	double	t[3];
	double	tt[3];
	double	len		= 0.0;

	double	prevBinormal[3];	// vector for previous Binormal	vector
	double	currBinormal[3];	// vector for current Binormal vector
	double	currNormal[3];
	double	dotPrevCurrBinormal;// scalar, dot-product of previous and current Binormal
	
	double	mat[16];			// rotation	matrix
	quaternion *p, *q, *r;		//
//	double	dot_PQ;				// quaternion inner-product

	mo2->markFrameCounter =	0;

	//family>
	double	quatAngle =	0.0;
	double	v[3];
//	double	z[3];
//	double	crossZV[3];
//	double	dotZV=0.0;
//	double	theta =	0.0;
//	double	dTheta = 4*M_PI/20.0;	

	//<family
	//----------------------------------------------->
	//@	INIT section
	p =	(quaternion*) my_malloc(sizeof(quaternion));
	q =	(quaternion*) my_malloc(sizeof(quaternion));
	r =	(quaternion*) my_malloc(sizeof(quaternion));

	p->r = 0;
	p->i = 0;
	p->j = 0;
	p->k = 0;
	
	dotPrevCurrBinormal	= 0.0;
	// <----------------------------------------------


	if(mo2->numPoints >	2) {
		//printf("[%d]\n",i);
		/////////////////////////////////////////////////////
		// make	Initial	Binormal Vector
		/////////////////////////////////////////////////////
		// obtain tangent
		t[0] = mo2->TangList[i].p3.d[0];
		t[1] = mo2->TangList[i].p3.d[1];
		t[2] = mo2->TangList[i].p3.d[2];
		//printf("t	= %f, %f, %f\n",t[0],t[1],t[2]);

		// obtain first-derivative of tangent
		tt[0] =	mo2->dTangList[i].p3.d[0];
		tt[1] =	mo2->dTangList[i].p3.d[1];
		tt[2] =	mo2->dTangList[i].p3.d[2];

		//----------------------------------------------->
		//@	obtain cross-product of	Tangent	and	First
		//	Derivative of tangent (Binormal	vector)
		cross3D(t, tt, currBinormal);
		len	= length3D(currBinormal);
		currBinormal[0]	= currBinormal[0]/len;
		currBinormal[1]	= currBinormal[1]/len;
		currBinormal[2]	= currBinormal[2]/len;
		//printf("b	= %f, %f, %f\n",currBinormal[0],currBinormal[1],currBinormal[2]);
		// <----------------------------------------------

		//----------------------------------------------->
		//@	obtain the third orthonormal component
		//	from cross-product of Tangent and Binormal
		cross3D(currBinormal,t,	currNormal);
		len	= length3D(currNormal);
		currNormal[0] =	currNormal[0]/len;
		currNormal[1] =	currNormal[1]/len;
		currNormal[2] =	currNormal[2]/len;
		//printf("n	= %f, %f, %f\n",currNormal[0],currNormal[1],currNormal[2]);
		// <----------------------------------------------

		//----------------------------------------------->
		//@	Frenet-Serret frame	is given by	the	set	of 
		//	coordinate axes:
		//			_  _  _	
		//		  [	T  N  B	]
		//
		// <----------------------------------------------

		//----------------------------------------------->
		//@	build the 4x4 Rotation matris with the
		//	columns	as the orthonormal basis constructed
		//	above using: Tangent, Binormal and Third component
		i=0;
			/*
			mat[0] = t[0];
			mat[1] = t[1];
			mat[2] = t[2];

			mat[4] = currBinormal[0];
			mat[5] = currBinormal[1];
			mat[6] = currBinormal[2];

			mat[8] = currNormal[0];
			mat[9] = currNormal[1];
			mat[10]= currNormal[2];		  
			*/

		mat[0] = t[0];
		mat[1] = t[1];
		mat[2] = t[2];

		mat[4] = currNormal[0];
		mat[5] = currNormal[1];
		mat[6] = currNormal[2];

		mat[8] = currBinormal[0];
		mat[9] = currBinormal[1];
		mat[10]= currBinormal[2];		

		mat[3]	= mat[7] = mat[11] = mat[12] = mat[13] = mat[14] = 0.0;
		mat[15]	= 1.0;
		// <----------------------------------------------

		//----------------------------------------------->
		//@	obtain the quaternion from rotation	matrix
		//	formed by the orthonormal basis	of the frame
		//	that have three	basis-vectors at a point p:
		//	Tangent, Binormal and Third	component
		ExtractQuaternionFromMat(mat,p);
		quatNormalize(p);	
		// <----------------------------------------------
		
		mo2->quatList[i].r = p->r;
		mo2->quatList[i].i = p->i;
		mo2->quatList[i].j = p->j;
		mo2->quatList[i].k = p->k;				  

		quatAngle =	2*acos(p->r);
		mo2->markQuatAngle[i] =	quatAngle;

		start =	1;

		prevBinormal[0]	= currBinormal[0];
		prevBinormal[1]	= currBinormal[1];
		prevBinormal[2]	= currBinormal[2];

		//printf("******** \n",i);
		/////////////////////////////////////////////////////
		// build the Quaternions corresponding to Frames
		/////////////////////////////////////////////////////
		for(i=start;i<(mo2->numPoints-2);i++) {

			//printf("[%d]\t",i);
			// obtain tangent
			t[0] = mo2->TangList[i].p3.d[0];
			t[1] = mo2->TangList[i].p3.d[1];
			t[2] = mo2->TangList[i].p3.d[2];
			//printf("t	= %f, %f, %f\n",t[0],t[1],t[2]);
			
			// obtain first-derivative of tangent
			tt[0] =	mo2->dTangList[i].p3.d[0];
			tt[1] =	mo2->dTangList[i].p3.d[1];
			tt[2] =	mo2->dTangList[i].p3.d[2];

			//----------------------------------------------->
			//@	obtain cross-product of	Tangent	and	First
			//	Derivative of tangent (Binormal	vector)
			cross3D(t, tt, currBinormal);
			len	= length3D(currBinormal);
			currBinormal[0]	= currBinormal[0]/len;
			currBinormal[1]	= currBinormal[1]/len;
			currBinormal[2]	= currBinormal[2]/len;
			//printf("b	= %f, %f, %f\n",currBinormal[0],currBinormal[1],currBinormal[2]);
			// <----------------------------------------------

			//----------------------------------------------->
			//@	obtain the third orthonormal component
			//	from cross-product of Tangent and Binormal
			/*cross3D(currBinormal,t, currNormal);
			len	= length3D(currNormal);
			currNormal[0] =	currNormal[0]/len;
			currNormal[1] =	currNormal[1]/len;
			currNormal[2] =	currNormal[2]/len;*/
			//printf("n	= %f, %f, %f\n",currNormal[0],currNormal[1],currNormal[2]);
			// <----------------------------------------------
			
			//----------------------------------------------->
			//@	check the dot products for orthonormality
			/*printf("dotProd(t,B) = %f\n",dot3D(t,currBinormal));
			printf("dotProd(t,S) = %f\n",dot3D(t,currNormal));
			printf("dotProd(S,B) = %f\n",dot3D(currNormal,currBinormal));*/
			// <----------------------------------------------

			//----------------------------------------------->
			//@	not	doing this right now but this is 
			//	one	of the steps. Not sure about this.
			dotPrevCurrBinormal	= dot3D(prevBinormal, currBinormal);
			//printf("dot =	%f\n",dotPrevCurrBinormal);
			if(dotPrevCurrBinormal<0.9)	{
				mo2->markFrame[mo2->markFrameCounter]=i;
				mo2->markFrameCounter++;
			}
			// <----------------------------------------------

			// copy	currBinormal into prevBinormal
			prevBinormal[0]	= currBinormal[0];
			prevBinormal[1]	= currBinormal[1];
			prevBinormal[2]	= currBinormal[2];

			//----------------------------------------------->
			//------------ Frenet to Quaternion	------------->
			
			//@	check on Quaternion	4D inner-product
			/*
			mat[0] = t[0];
			mat[1] = t[1];
			mat[2] = t[2];

			mat[4] = currBinormal[0];
			mat[5] = currBinormal[1];
			mat[6] = currBinormal[2];

			mat[8] = currNormal[0];
			mat[9] = currNormal[1];
			mat[10]= currNormal[2];		  
			*/
			////mat[0] = t[0];
			////mat[1] = t[1];
			////mat[2] = t[2];

			////mat[4] = currNormal[0];
			////mat[5] = currNormal[1];
			////mat[6] = currNormal[2];

			////mat[8] = currBinormal[0];
			////mat[9] = currBinormal[1];
			////mat[10]= currBinormal[2];		

	  ////		mat[3]	= mat[7] = mat[11] = mat[12] = mat[13] = mat[14] = 0;
			////mat[15]	= 1.0;

			/*
			printf("[%d] \n",i);
			
			printf("t =	%f,	%f,	%f\n",t[0],t[1],t[2]);
			printf("n =	%f,	%f,	%f\n",currNormal[0],currNormal[1],currNormal[2]);
			printf("b =	%f,	%f,	%f\n",currBinormal[0],currBinormal[1],currBinormal[2]);
			*/
			////ExtractQuaternionFromMat(mat,q);
			////quatNormalize(q);

			//	check if dot product of	neighboring	quaternions
			//	is near	negative one
			////dot_PQ = quatQuatDotProduct(p,q);
			//printf("[%d] dot_PQ =	%f\n",i,dot_PQ);
			////if(dot_PQ <= 0.0) {

			////	// flip	one	of the quaternions
			////	r->r = (q->r);
			////	r->i = -(q->i);
			////	r->j = -(q->j);
			////	r->k = -(q->k);
			////	quatCopy(r,q);
			////	
			////	//printf("	   ... dotProd after flipping =	%f\n",quatQuatDotProduct(p,q));
			////}
			////if(	(dot_PQ>0.0) &&	(dot_PQ<0.75)) {
			////	mo2->markFrame[mo2->markFrameCounter]=i;
			////	mo2->markFrameCounter++;
			////	//printf("[%d] ...NEAR ZERO	marking	this (+)\n",i);
			////}
			
			////quatAngle =	2*acos(q->r);
			////mo2->markQuatAngle[i] =	quatAngle;
			// <----------------------------------------------
			// <----------------------------------------------

			// save	out	resulting quaternion
			////mo2->quatList[i].r = q->r;
			////mo2->quatList[i].i = q->i;
			////mo2->quatList[i].j = q->j;
			////mo2->quatList[i].k = q->k;				  

			// save	current	quaternion as 'previous-quaternion'
			// for next	iteration
			////quatCopy(q,p);
			//printf("norm = %f\n",qNorm(q));

			
		}
		////mo2->markFrame[mo2->markFrameCounter] =	mo2->numPoints-2;
		////mo2->markFrameCounter++;

		
		/*printf("***\n");
		for(i=0;i<mo2->markFrameCounter;i++) {
			printf("%d ",mo2->markFrame[i]);
		}
		printf("\n");*/

		// Calculate quaternions for smoothly
		// varying portions	of curve only

		start =	0;
		if(mo2->markFrameCounter ==	0)
			end	= mo2->numPoints-1;
		else
			end	= mo2->markFrame[0];
		glPushAttrib(GL_LINE_BIT | GL_LIGHTING_BIT | GL_CURRENT_BIT);
		glDisable(GL_LIGHTING);
		glLineWidth(2.0);
		for(i=0;i<=mo2->markFrameCounter;i++) {
			//if(i==(mo2->markFrameCounter)-1) {
			//	end	= mo2->numPoints-2;
			//	fprintf(stderr,"!start = %d, end = %d\n",start,end);
			//}
			//
			//if(i<(mo2->markFrameCounter-1)) {
			//	start =	mo2->markFrame[i];
			//	end	= mo2->markFrame[i+1]-1;
			//	fprintf(stderr,"start =	%d,	end	= %d\n",start,end);
			//}
			//fprintf(stderr,"start	= %d, end =	%d\n",start,end);
			glColor3f(0.8f,0.1f,0.1f);
			for(j=start; j<end;	j++) {
				//printf("j= %d\n",j);
				v[0] = mo2->pntList[j].p3.d[0];
				v[1] = mo2->pntList[j].p3.d[1];
				v[2] = mo2->pntList[j].p3.d[2];
				glPushMatrix();
				glTranslatef(v[0],v[1],v[2]);
				glutSolidSphere(0.025,8,8);
				glPopMatrix();
				
			}
			glColor3f(0.1f,0.1f,0.1f);
			glBegin(GL_LINE_STRIP);
				for(j=start; j<end;	j++) {
				v[0] = mo2->pntList[j].p3.d[0];
				v[1] = mo2->pntList[j].p3.d[1];
				v[2] = mo2->pntList[j].p3.d[2];
				glVertex3f(v[0],v[1],v[2]);
			}
			glEnd();
			start =	mo2->markFrame[i]+1;
			end	= mo2->markFrame[i+1];
			if(i==(mo2->markFrameCounter)-1) {
				end	= mo2->numPoints-1;
			}
		}
		glPopAttrib();
		
	}
}

void computeFrenetQuaternionMap3 ( void	)
{
	int		i		= 0;			// index into points array
	int		j		= 0;
//	int		limit	= 0;			// max number of frames	to be calculated
	int		start	= 0;			// start index for the for-loop
	int		end		= 0;
	double	t[3];
	double	tt[3];
	double	len		= 0.0;

	double	prevBinormal[3];	// vector for previous Binormal	vector
	double	currBinormal[3];	// vector for current Binormal vector
	double	currNormal[3];
	double	dotPrevCurrBinormal;// scalar, dot-product of previous and current Binormal
	
	double	mat[16];			// rotation	matrix
	quaternion *p, *q, *r;		//
	double	dot_PQ;				// quaternion inner-product

	//initialize 
	mo2->markFrameCounter =	0;
	mo2->markQuatCurveCounter =	0;

	//family>
	double	quatAngle =	0.0;
//	double	v[3];
//	double	z[3];
//	double	crossZV[3];
//	double	dotZV=0.0;
//	double	theta =	0.0;
//	double	dTheta = 4*M_PI/20.0;	
	//<family

	double ccolor[3];
	ccolor[0] =	0.0;
	ccolor[1] =	0.0;
	ccolor[2] =	0.0;

	//----------------------------------------------->
	//@	INIT section
	p =	(quaternion*) my_malloc(sizeof(quaternion));
	q =	(quaternion*) my_malloc(sizeof(quaternion));
	r =	(quaternion*) my_malloc(sizeof(quaternion));

	p->r = 0;
	p->i = 0;
	p->j = 0;
	p->k = 0;
	
	dotPrevCurrBinormal	= 0.0;
	// <----------------------------------------------


	if(mo2->numPoints >	2) {

		t[0] = mo2->TangList[i].p3.d[0];
		t[1] = mo2->TangList[i].p3.d[1];
		t[2] = mo2->TangList[i].p3.d[2];
		
		// obtain first-derivative of tangent
		tt[0] =	mo2->dTangList[i].p3.d[0];
		tt[1] =	mo2->dTangList[i].p3.d[1];
		tt[2] =	mo2->dTangList[i].p3.d[2];
		//----------------------------------------------->

		//@	obtain cross-product of	Tangent	and	First
		//	Derivative of tangent (Binormal	vector)
		cross3D(t, tt, currBinormal);
		len	= length3D(currBinormal);
		currBinormal[0]	= currBinormal[0]/len;
		currBinormal[1]	= currBinormal[1]/len;
		currBinormal[2]	= currBinormal[2]/len;
		// <----------------------------------------------

		//----------------------------------------------->
		//@	obtain the third orthonormal component
		//	from cross-product of Tangent and Binormal
		cross3D(currBinormal,t,	currNormal);
		len	= length3D(currNormal);
		currNormal[0] =	currNormal[0]/len;
		currNormal[1] =	currNormal[1]/len;
		currNormal[2] =	currNormal[2]/len;
		// <----------------------------------------------

		//----------------------------------------------->
		//@	Frenet-Serret frame	is given by	the	set	of 
		//	coordinate axes:
		//			_  _  _	
		//		  [	T  N  B	]
		//
		// <----------------------------------------------

		//----------------------------------------------->
		//@	build the 4x4 Rotation matris with the
		//	columns	as the orthonormal basis constructed
		//	above using: Tangent, Binormal and Third component
		mat[0] = t[0];
		mat[1] = t[1];
		mat[2] = t[2];

		mat[4] = currNormal[0];
		mat[5] = currNormal[1];
		mat[6] = currNormal[2];

		mat[8] = currBinormal[0];
		mat[9] = currBinormal[1];
		mat[10]= currBinormal[2];		

		mat[3]	= mat[7] = mat[11] = mat[12] = mat[13] = mat[14] = 0.0;
		mat[15]	= 1.0;
		// <----------------------------------------------

		//----------------------------------------------->
		//@	obtain the quaternion from rotation	matrix
		//	formed by the orthonormal basis	of the frame
		//	that have three	basis-vectors at a point p:
		//	Tangent, Binormal and Third	component
		ExtractQuaternionFromMat(mat,p);
		quatNormalize(p);	
		// <----------------------------------------------
		
		mo2->quatList[i].r = p->r;
		mo2->quatList[i].i = p->i;
		mo2->quatList[i].j = p->j;
		mo2->quatList[i].k = p->k;				  

		quatAngle =	2*acos(p->r);
		mo2->markQuatAngle[i] =	quatAngle;

		start =	1;

		prevBinormal[0]	= currBinormal[0];
		prevBinormal[1]	= currBinormal[1];
		prevBinormal[2]	= currBinormal[2];

		for(i=start;i<(mo2->numPoints-2);i++)	{

			// obtain tangent
			t[0] = mo2->TangList[i].p3.d[0];
			t[1] = mo2->TangList[i].p3.d[1];
			t[2] = mo2->TangList[i].p3.d[2];
			
			// obtain first-derivative of tangent
			tt[0] =	mo2->dTangList[i].p3.d[0];
			tt[1] =	mo2->dTangList[i].p3.d[1];
			tt[2] =	mo2->dTangList[i].p3.d[2];

			//----------------------------------------------->
			//@	obtain cross-product of	Tangent	and	First
			//	Derivative of tangent (Binormal	vector)
			cross3D(t, tt, currBinormal);
			len	= length3D(currBinormal);
			currBinormal[0]	= currBinormal[0]/len;
			currBinormal[1]	= currBinormal[1]/len;
			currBinormal[2]	= currBinormal[2]/len;
			// <----------------------------------------------

			//----------------------------------------------->
			//@	not	doing this right now but this is 
			//	one	of the steps. Not sure about this.
			dotPrevCurrBinormal	= dot3D(prevBinormal, currBinormal);
			printf("[%d] dotBirnormals = %f\n",i,dotPrevCurrBinormal);
			if(dotPrevCurrBinormal<0.10) {
				mo2->markFrame[mo2->markFrameCounter]=i+1;
				printf("marking... %d\n",i+1);
				mo2->markFrameCounter++;
			}
			// <----------------------------------------------

			// copy	currBinormal into prevBinormal
			prevBinormal[0]	= currBinormal[0];
			prevBinormal[1]	= currBinormal[1];
			prevBinormal[2]	= currBinormal[2];
		}

		// Calculate quaternions for smoothly
		// varying portions	of curve only
		start =	0;
		if(mo2->markFrameCounter ==	0)
			end	= mo2->numPoints-2;
		else
			end	= mo2->markFrame[0];
		glPushAttrib(GL_LINE_BIT | GL_LIGHTING_BIT | GL_CURRENT_BIT);
		glDisable(GL_LIGHTING);
		glLineWidth(2.0);
		int	check=0;
		int	colnum = 0;
		for(i=0;i<=mo2->markFrameCounter;i++) {
			check=0;
			for(j=start; j<end;	j++) {
				check++;

				// obtain tangent
				t[0] = mo2->TangList[j].p3.d[0];
				t[1] = mo2->TangList[j].p3.d[1];
				t[2] = mo2->TangList[j].p3.d[2];
			
				// obtain first-derivative of tangent
				tt[0] =	mo2->dTangList[j].p3.d[0];
				tt[1] =	mo2->dTangList[j].p3.d[1];
				tt[2] =	mo2->dTangList[j].p3.d[2];

				//----------------------------------------------->
				//@	obtain cross-product of	Tangent	and	First
				//	Derivative of tangent (Binormal	vector)
				cross3D(t, tt, currBinormal);
				len	= length3D(currBinormal);
				currBinormal[0]	= currBinormal[0]/len;
				currBinormal[1]	= currBinormal[1]/len;
				currBinormal[2]	= currBinormal[2]/len;
				// <----------------------------------------------

				//----------------------------------------------->
				//@	obtain the third orthonormal component
				//	from cross-product of Tangent and Binormal
				cross3D(currBinormal,t,	currNormal);
				len	= length3D(currNormal);
				currNormal[0] =	currNormal[0]/len;
				currNormal[1] =	currNormal[1]/len;
				currNormal[2] =	currNormal[2]/len;
				// <----------------------------------------------

				mat[0] = t[0];
				mat[1] = t[1];
				mat[2] = t[2];

				mat[4] = currNormal[0];
				mat[5] = currNormal[1];
				mat[6] = currNormal[2];

				mat[8] = currBinormal[0];
				mat[9] = currBinormal[1];
				mat[10]= currBinormal[2];		

				mat[3]	= mat[7] = mat[11] = mat[12] = mat[13] = mat[14] = 0;
				mat[15]	= 1.0;

				ExtractQuaternionFromMat(mat,q);
				quatNormalize(q);

				//	check if dot product of	neighboring	quaternions
				//	is near	negative one
				dot_PQ = quatQuatDotProduct(p,q);
				//if(dot_PQ<0.9	&& dot_PQ>0.0) {
				//	fprintf(stderr,"!!dot(prev,curr) = %f\n",dot_PQ);
				//	glPushMatrix();
				//	glTranslatef(q->r,q->j,q->k);
				//	glutSolidSphere(0.035,16,16);
				//	glPopMatrix();
				//}
				//else if(dot_PQ <0.0 && dot_PQ>=-0.9)
				//	fprintf(stderr,"->dot(prev,curr) = %f\n",dot_PQ);
				//else if(dot_PQ<-0.9)
				//	fprintf(stderr,"<<dot(prev,curr) = %f\n",dot_PQ);
				//else

				if(dot_PQ <	0.0) {
					// flip	one	of the quaternions
					//fprintf(stderr,"# [%d] dot(prev,curr) = %f\n",j,dot_PQ);
					r->r = (q->r);
					r->i = -(q->i);
					r->j = -(q->j);
					r->k = -(q->k);
					//quatCopy(r,q);
					dot_PQ = quatQuatDotProduct(p,q);
				}
				else {
					//fprintf(stderr,"> [%d] dot(prev,curr) = %f\n",j,dot_PQ);
				}

				if(dot_PQ<-0.9)
				{
					mo2->markQuatCurve[mo2->markQuatCurveCounter] =	j;
					mo2->markQuatCurveCounter++;
				}

				//save out resulting quaternion
				mo2->quatList[j].r = q->r;
				mo2->quatList[j].i = q->i;
				mo2->quatList[j].j = q->j;
				mo2->quatList[j].k = q->k;				  

				// save	current	quaternion as 'previous-quaternion'
				// for next	iteration
				quatCopy(q,p);
			}

			if(check>0)
				colnum = i;
			if(i==mo2->markFrameCounter)
				colnum = 0;
			getSegmentColor(colnum,	ccolor);
			drawQuaternionCurve(start,end,ccolor);

			//!	Compute	range for next section of smooth-curve
			start =	mo2->markFrame[i]+1;
			end	= mo2->markFrame[i+1];
			if(i==(mo2->markFrameCounter)-1) {
				end	= mo2->numPoints-2;
			}
		}
		//printf(">	Mark Quaternion	Curve\n");
		//for(i=0;i<mo2->markQuatCurveCounter;i++) {
		//	printf("%d ",mo2->markQuatCurve[i]);
		//}

		//printf("\n");		
		//printf("> Mark QUAT Frames\n");
		//glColor3f(0.0f,1.0f,0.0f);
		//for(i=0;i<mo2->markQuatCurveCounter;i++) {
		//	printf("flipping quaternion.... %d \n",mo2->markQuatCurve[i]);
		//}

		glPopAttrib();
		
	}
}

void computeFrenetQuaternionMap4 ( void	)
{
	int		h		= 0;
	int		i		= 0;			// index into points array
//	int		j		= 0;
//	int		limit	= 0;			// max number of frames	to be calculated
	int		start	= 0;			// start index for the for-loop
	int		end		= 0;
	double	t[3];
	double	tt[3];
	double	len		= 0.0;

	double	prevBinormal[3];	// vector for previous Binormal	vector
	double	currBinormal[3];	// vector for current Binormal vector
	double	currNormal[3];
	double	dotPrevCurrBinormal;// scalar, dot-product of previous and current Binormal
	
	double	mat[16];			// rotation	matrix
	quaternion *p, *q, *r;		//
	double	dot_PQ;				// quaternion inner-product

	//initialize 
	mo2->markFrameCounter =	0;
	mo2->markQuatCurveCounter =	0;

	//family>
//	double	quatAngle =	0.0;
//	double	v[3];
//	double	z[3];
//	double	crossZV[3];
//	double	dotZV=0.0;
//	double	theta =	0.0;
//	double	dTheta = 4*M_PI/20.0;	
	//<family

	double ccolor[3];
	ccolor[0] =	0.0;
	ccolor[1] =	0.0;
	ccolor[2] =	0.0;

	//----------------------------------------------->
	//@	INIT section
	p =	(quaternion*) my_malloc(sizeof(quaternion));
	q =	(quaternion*) my_malloc(sizeof(quaternion));
	r =	(quaternion*) my_malloc(sizeof(quaternion));

	p->r = 0;
	p->i = 0;
	p->j = 0;
	p->k = 0;
	
	dotPrevCurrBinormal	= 0.0;
	// <----------------------------------------------


	if(mo2->numPoints >	2) {
		t[0] = mo2->TangList[i].p3.d[0];
		t[1] = mo2->TangList[i].p3.d[1];
		t[2] = mo2->TangList[i].p3.d[2];
		
		// obtain first-derivative of tangent
		tt[0] =	mo2->dTangList[i].p3.d[0];
		tt[1] =	mo2->dTangList[i].p3.d[1];
		tt[2] =	mo2->dTangList[i].p3.d[2];
		//----------------------------------------------->

		//@	obtain cross-product of	Tangent	and	First
		//	Derivative of tangent (Binormal	vector)
		cross3D(t, tt, currBinormal);
		len	= length3D(currBinormal);
		currBinormal[0]	= currBinormal[0]/len;
		currBinormal[1]	= currBinormal[1]/len;
		currBinormal[2]	= currBinormal[2]/len;
		// <----------------------------------------------

		//----------------------------------------------->
		//@	obtain the third orthonormal component
		//	from cross-product of Tangent and Binormal
		cross3D(currBinormal,t,	currNormal);
		len	= length3D(currNormal);
		currNormal[0] =	currNormal[0]/len;
		currNormal[1] =	currNormal[1]/len;
		currNormal[2] =	currNormal[2]/len;
		// <----------------------------------------------

		//----------------------------------------------->
		//@	build the 4x4 Rotation matris with the
		//	columns	as the orthonormal basis constructed
		//	above using: Tangent, Binormal and Third component
		mat[0] = t[0];
		mat[1] = t[1];
		mat[2] = t[2];

		mat[4] = currNormal[0];
		mat[5] = currNormal[1];
		mat[6] = currNormal[2];

		mat[8] = currBinormal[0];
		mat[9] = currBinormal[1];
		mat[10]= currBinormal[2];		

		mat[3]	= mat[7] = mat[11] = mat[12] = mat[13] = mat[14] = 0.0;
		mat[15]	= 1.0;
		// <----------------------------------------------

		//----------------------------------------------->
		//@	obtain the quaternion from rotation	matrix
		//	formed by the orthonormal basis	of the frame
		//	that have three	basis-vectors at a point p:
		//	Tangent, Binormal and Third	component
		ExtractQuaternionFromMat(mat,p);
		quatNormalize(p);	
		// <----------------------------------------------
		
		mo2->quatList[i].r = p->r;
		mo2->quatList[i].i = p->i;
		mo2->quatList[i].j = p->j;
		mo2->quatList[i].k = p->k;				  

		prevBinormal[0]	= currBinormal[0];
		prevBinormal[1]	= currBinormal[1];
		prevBinormal[2]	= currBinormal[2];

		start =	1;
		for(i=start;i<(mo2->numPoints-2);i++)	{

			// obtain tangent
			t[0] = mo2->TangList[i].p3.d[0];
			t[1] = mo2->TangList[i].p3.d[1];
			t[2] = mo2->TangList[i].p3.d[2];
			
			// obtain first-derivative of tangent
			tt[0] =	mo2->dTangList[i].p3.d[0];
			tt[1] =	mo2->dTangList[i].p3.d[1];
			tt[2] =	mo2->dTangList[i].p3.d[2];

			//----------------------------------------------->
			//@	obtain cross-product of	Tangent	and	First
			//	Derivative of tangent (Binormal	vector)
			cross3D(t, tt, currBinormal);
			len	= length3D(currBinormal);
			currBinormal[0]	= currBinormal[0]/len;
			currBinormal[1]	= currBinormal[1]/len;
			currBinormal[2]	= currBinormal[2]/len;
			// <----------------------------------------------

			//----------------------------------------------->
			//@	not	doing this right now but this is 
			//	one	of the steps. Not sure about this.
			dotPrevCurrBinormal	= dot3D(prevBinormal, currBinormal);
			if(dotPrevCurrBinormal<0.90) {
				printf("[%d] binormal dot = %f\n",i,dotPrevCurrBinormal);
				mo2->markFrame[mo2->markFrameCounter]=i;
				mo2->markFrameCounter++;
			}
			// <----------------------------------------------

			// copy	currBinormal into prevBinormal
			prevBinormal[0]	= currBinormal[0];
			prevBinormal[1]	= currBinormal[1];
			prevBinormal[2]	= currBinormal[2];

			////----------------------------------------------->
			////@	obtain the third orthonormal component
			////	from cross-product of Tangent and Binormal
			//cross3D(currBinormal,t,	currNormal);
			//len	= length3D(currNormal);
			//currNormal[0] =	currNormal[0]/len;
			//currNormal[1] =	currNormal[1]/len;
			//currNormal[2] =	currNormal[2]/len;
			//// <----------------------------------------------

			//mat[0] = t[0];
			//mat[1] = t[1];
			//mat[2] = t[2];

			//mat[4] = currNormal[0];
			//mat[5] = currNormal[1];
			//mat[6] = currNormal[2];

			//mat[8] = currBinormal[0];
			//mat[9] = currBinormal[1];
			//mat[10]= currBinormal[2];		

			//mat[3]	= mat[7] = mat[11] = mat[12] = mat[13] = mat[14] = 0;
			//mat[15]	= 1.0;

			//ExtractQuaternionFromMat(mat,q);
			//quatNormalize(q);

			////save out resulting quaternion
			//mo2->quatList[i].r = q->r;
			//mo2->quatList[i].i = q->i;
			//mo2->quatList[i].j = q->j;
			//mo2->quatList[i].k = q->k;
		}

		printf("# Marked Frames : Binormal mismatch#\n");
		//for(i=0; i<mo2->markFrameCounter; i++) {
		//	printf("%d \n",mo2->markFrame[i]);
		//}

		start =	0;
		if(mo2->markFrameCounter ==	0)
			end	= mo2->numPoints-2;
		else
			end	= mo2->markFrame[0]-2;
		for(h=0;h<=mo2->markFrameCounter;h++) {

			printf("start = %d, end = %d\n",start,end);
			
			//------------------------------------------------->
			for(i=start; i<end; i++) {
				//! compute equivalent quaternions
				// obtain tangent
				t[0] = mo2->TangList[i].p3.d[0];
				t[1] = mo2->TangList[i].p3.d[1];
				t[2] = mo2->TangList[i].p3.d[2];
				
				// obtain first-derivative of tangent
				tt[0] =	mo2->dTangList[i].p3.d[0];
				tt[1] =	mo2->dTangList[i].p3.d[1];
				tt[2] =	mo2->dTangList[i].p3.d[2];

				//@	obtain Binormal	vector
				cross3D(t, tt, currBinormal);
				len	= length3D(currBinormal);
				currBinormal[0]	= currBinormal[0]/len;
				currBinormal[1]	= currBinormal[1]/len;
				currBinormal[2]	= currBinormal[2]/len;

				//@	obtain Normal
				cross3D(currBinormal,t,	currNormal);
				len	= length3D(currNormal);
				currNormal[0] =	currNormal[0]/len;
				currNormal[1] =	currNormal[1]/len;
				currNormal[2] =	currNormal[2]/len;

				mat[0] = t[0]; mat[1] = t[1]; mat[2] = t[2];

				mat[4] = currNormal[0]; mat[5] = currNormal[1]; mat[6] = currNormal[2];

				mat[8] = currBinormal[0]; mat[9] = currBinormal[1]; mat[10]= currBinormal[2];		

				mat[3]	= mat[7] = mat[11] = mat[12] = mat[13] = mat[14] = 0; 
				mat[15]	= 1.0;

				ExtractQuaternionFromMat(mat,q);
				quatNormalize(q);

				//save out resulting quaternion
				mo2->quatList[i].r = q->r;
				mo2->quatList[i].i = q->i;
				mo2->quatList[i].j = q->j;
				mo2->quatList[i].k = q->k;			
			}

			//! check quaternions
			if(start==0)
				start=1;
			for(i=start; i<end; i++) {
				p->r = mo2->quatList[i-1].r; p->i = mo2->quatList[i-1].i; p->j = mo2->quatList[i-1].j; p->k = mo2->quatList[i-1].k; 
				q->r = mo2->quatList[i].r;	 q->i = mo2->quatList[i].i;   q->j = mo2->quatList[i].j;   q->k = mo2->quatList[i].k; 
				dot_PQ = quatQuatDotProduct(p,q);
				//printf("[%d] dot = %f\n",i,dot_PQ);
				//! flip quaternion
				if(dot_PQ <0.9) {
					printf("[%d] dot = %f\n",i,dot_PQ);
				
				}
				//if(dot_PQ<-0.5) {
				//	mo2->quatList[i].r = q->r;
				//	mo2->quatList[i].i = -q->i;
				//	mo2->quatList[i].j = -q->j;
				//	mo2->quatList[i].k = -q->k;
				//	dot_PQ = quatQuatDotProduct(p,q);
				//	//printf("#[%d] dot = %f\n",i,dot_PQ);
				//}
			}
			getSegmentColor(h,ccolor);	
			drawQuaternionCurve(start,end,ccolor);
			//<-------------------------------------------------

			//!	Compute	range for next section of smooth-curve
			start =	mo2->markFrame[h]+1;
			end	= mo2->markFrame[h+1]-2;
			if(h==(mo2->markFrameCounter)-1) {
				end	= mo2->numPoints-2;
			}
		}

		//glDisable(GL_LIGHTING);
		//dot_PQ = 0.0;
		//mo2->markQuatCurveCounter = 0;
		//for(i=1; i<mo2->numPoints-2; i++) {
		//	p->r = mo2->quatList[i-1].r; p->i = mo2->quatList[i-1].i; p->j = mo2->quatList[i-1].j; p->k = mo2->quatList[i-1].k; 
		//	q->r = mo2->quatList[i].r;	 q->i = mo2->quatList[i].i;   q->j = mo2->quatList[i].j;   q->k = mo2->quatList[i].k; 
		//	dot_PQ = quatQuatDotProduct(p,q);
		//	//printf("[%d] dot = %f\n",i,dot_PQ);
		//	if(dot_PQ < -0.9) {
		//		mo2->markQuatCurve[mo2->markQuatCurveCounter] = i;
		//		mo2->markQuatCurveCounter++;
		//		glPushMatrix();
		//		glColor3f(0.4f,0.4f,0.4f);
		//		glTranslatef(q->i,q->j,q->k);
		//		glutSolidSphere(0.08,16,16);
		//		glPopMatrix();
		//	}
		//}
		//printf("----->\n");
		//printf("# Marked Quats : Quaternion mismatch#\n");
		//for(i=0; i<mo2->markQuatCurveCounter; i++) {
		//	printf("%d \n",mo2->markQuatCurve[i]);
		//}

		//int qindex = 0;
		//for(i=0; i<mo2->markQuatCurveCounter; i++) {
		//	qindex = mo2->markQuatCurve[i];
		//	printf("Flipping quaternion... %d\n",qindex);
		//	q->r = mo2->quatList[qindex].r;	 q->i = mo2->quatList[qindex].i;   q->j = mo2->quatList[qindex].j;   q->k = mo2->quatList[qindex].k; 
		//	r->r = (q->r);
		//	r->i = -(q->i);
		//	r->j = -(q->j);
		//	r->k = -(q->k);

		//	mo2->quatList[qindex].r = r->r;
		//	mo2->quatList[qindex].i = r->i;
		//	mo2->quatList[qindex].j = r->j;
		//	mo2->quatList[qindex].k = r->k;
		//	glColor3f(0.0f,0.7f,0.0f);
		//	glPushMatrix();
		//		glTranslatef(r->i,r->j,r->k);
		//		glutSolidSphere(0.08,16,16);
		//	glPopMatrix();
		//}
		//for(i=1; i<mo2->numPoints-2; i++) {
		//	p->r = mo2->quatList[i-1].r; p->i = mo2->quatList[i-1].i; p->j = mo2->quatList[i-1].j; p->k = mo2->quatList[i-1].k; 
		//	q->r = mo2->quatList[i].r;	 q->i = mo2->quatList[i].i;   q->j = mo2->quatList[i].j;   q->k = mo2->quatList[i].k; 
		//	dot_PQ = quatQuatDotProduct(p,q);
		//	printf("[%d] dot = %f\n",i,dot_PQ);
		//}
		//printf("----->\n");		
		//*************************************************************
		//start =	0;
		//if(mo2->markFrameCounter ==	0)
		//	end	= mo2->numPoints-2;
		//else
		//	end	= mo2->markFrame[0];
		//glPushAttrib(GL_LINE_BIT | GL_LIGHTING_BIT | GL_CURRENT_BIT);
		//glDisable(GL_LIGHTING);
		//glLineWidth(2.0);
		//int	check=0;
		//int	colnum = 0;
		//for(i=0;i<=mo2->markFrameCounter;i++) {
		//	printf("start = %d, end = %d\n",start,end);
		//	colnum=i;
		//	if(i==mo2->markFrameCounter)
		//		colnum = 0;
		//	getSegmentColor(colnum,	ccolor);
		//	drawQuaternionCurve(start,end,ccolor);

		//	//!	Compute	range for next section of smooth-curve
		//	start =	mo2->markFrame[i]+1;
		//	end	= mo2->markFrame[i+1];
		//	if(i==(mo2->markFrameCounter)-1) {
		//		end	= mo2->numPoints-2;
		//	}
		//}
		

		//// Calculate quaternions for smoothly varying portions	of curve only
		//start =	0;
		//if(mo2->markFrameCounter ==	0)
		//	end	= mo2->numPoints-2;
		//else
		//	end	= mo2->markFrame[0];
		//glPushAttrib(GL_LINE_BIT | GL_LIGHTING_BIT | GL_CURRENT_BIT);
		//glDisable(GL_LIGHTING);
		//glLineWidth(2.0);
		//int	check=0;
		//int	colnum = 0;
		//for(i=0;i<=mo2->markFrameCounter;i++) {
		//	check=0;
		//	for(j=start; j<end;	j++) {
		//		check++;

		//		// obtain tangent
		//		t[0] = mo2->TangList[j].p3.d[0];
		//		t[1] = mo2->TangList[j].p3.d[1];
		//		t[2] = mo2->TangList[j].p3.d[2];
		//	
		//		// obtain first-derivative of tangent
		//		tt[0] =	mo2->dTangList[j].p3.d[0];
		//		tt[1] =	mo2->dTangList[j].p3.d[1];
		//		tt[2] =	mo2->dTangList[j].p3.d[2];

		//		//----------------------------------------------->
		//		//@	obtain cross-product of	Tangent	and	First
		//		//	Derivative of tangent (Binormal	vector)
		//		cross3D(t, tt, currBinormal);
		//		len	= length3D(currBinormal);
		//		currBinormal[0]	= currBinormal[0]/len;
		//		currBinormal[1]	= currBinormal[1]/len;
		//		currBinormal[2]	= currBinormal[2]/len;
		//		// <----------------------------------------------

		//		//----------------------------------------------->
		//		//@	obtain the third orthonormal component
		//		//	from cross-product of Tangent and Binormal
		//		cross3D(currBinormal,t,	currNormal);
		//		len	= length3D(currNormal);
		//		currNormal[0] =	currNormal[0]/len;
		//		currNormal[1] =	currNormal[1]/len;
		//		currNormal[2] =	currNormal[2]/len;
		//		// <----------------------------------------------

		//		mat[0] = t[0];
		//		mat[1] = t[1];
		//		mat[2] = t[2];

		//		mat[4] = currNormal[0];
		//		mat[5] = currNormal[1];
		//		mat[6] = currNormal[2];

		//		mat[8] = currBinormal[0];
		//		mat[9] = currBinormal[1];
		//		mat[10]= currBinormal[2];		

		//		mat[3]	= mat[7] = mat[11] = mat[12] = mat[13] = mat[14] = 0;
		//		mat[15]	= 1.0;

		//		ExtractQuaternionFromMat(mat,q);
		//		quatNormalize(q);

		//		//	check if dot product of	neighboring	quaternions
		//		//	is near	negative one
		//		dot_PQ = quatQuatDotProduct(p,q);
		//		//if(dot_PQ<0.9	&& dot_PQ>0.0) {
		//		//	fprintf(stderr,"!!dot(prev,curr) = %f\n",dot_PQ);
		//		//	glPushMatrix();
		//		//	glTranslatef(q->r,q->j,q->k);
		//		//	glutSolidSphere(0.035,16,16);
		//		//	glPopMatrix();
		//		//}
		//		//else if(dot_PQ <0.0 && dot_PQ>=-0.9)
		//		//	fprintf(stderr,"->dot(prev,curr) = %f\n",dot_PQ);
		//		//else if(dot_PQ<-0.9)
		//		//	fprintf(stderr,"<<dot(prev,curr) = %f\n",dot_PQ);
		//		//else
		//		//if(dot_PQ <	0.0) {
		//		//	// flip	one	of the quaternions
		//		//	fprintf(stderr,"# [%d] dot(prev,curr) = %f\n",j,dot_PQ);
		//		//	//r->r = (q->r);
		//		//	//r->i = -(q->i);
		//		//	//r->j = -(q->j);
		//		//	//r->k = -(q->k);
		//		//	//quatCopy(r,q);
		//		//	//dot_PQ = quatQuatDotProduct(p,q);
		//		//}
		//		//else
		//		//	fprintf(stderr,"> [%d] dot(prev,curr) = %f\n",j,dot_PQ);

		//		if(dot_PQ<-0.9)
		//		{
		//			mo2->markQuatCurve[mo2->markQuatCurveCounter] =	j;
		//			mo2->markQuatCurveCounter++;
		//		}

		//		//save out resulting quaternion
		//		mo2->quatList[j].r = q->r;
		//		mo2->quatList[j].i = q->i;
		//		mo2->quatList[j].j = q->j;
		//		mo2->quatList[j].k = q->k;				  

		//		// save	current	quaternion as 'previous-quaternion'
		//		// for next	iteration
		//		quatCopy(q,p);
		//	}

		//	if(check>0)
		//		colnum = i;
		//	if(i==mo2->markFrameCounter)
		//		colnum = 0;
		//	getSegmentColor(colnum,	ccolor);
		//	drawQuaternionCurve(start,end,ccolor);

		//	//!	Compute	range for next section of smooth-curve
		//	start =	mo2->markFrame[i]+1;
		//	end	= mo2->markFrame[i+1];
		//	if(i==(mo2->markFrameCounter)-1) {
		//		end	= mo2->numPoints-2;
		//	}
		//}

		//printf("\n");		
		//printf("> Mark QUAT Frames\n");
		//glColor3f(0.0f,1.0f,0.0f);
		//for(i=0;i<mo2->markQuatCurveCounter;i++) {
		//	printf("flipping quaternion.... %d \n",mo2->markQuatCurve[i]);
		//	mo2->quatList[i].i = -(mo2->quatList[i].i);
		//	mo2->quatList[i].j = -(mo2->quatList[i].j);
		//	mo2->quatList[i].k = -(mo2->quatList[i].k);
		//}
		//dot_PQ = 0.0;

		glPopAttrib();
		
	}
}

void computeFrenetQuaternionMap5 ( void	)
{
	int		i		= 0;			// index into points array
//	int		j		= 0;
//	int		limit	= 0;			// max number of frames	to be calculated
	int		start	= 0;			// start index for the for-loop
//	int		end		= 0;
	double	t[3];
	double	tt[3];
	double	len		= 0.0;

	double	prevBinormal[3];	// vector for previous Binormal	vector
	double	currBinormal[3];	// vector for current Binormal vector
	double	currNormal[3];
	double	dotPrevCurrBinormal;// scalar, dot-product of previous and current Binormal
	
	double	mat[16];			// rotation	matrix
	quaternion *p, *q, *r;		//
	double	dot_PQ;				// quaternion inner-product

	//initialize 
	mo2->markFrameCounter =	0;
	mo2->markQuatCurveCounter =	0;

	//family>
	double	quatAngle =	0.0;
//	double	v[3];
//	double	z[3];
//	double	crossZV[3];
//	double	dotZV=0.0;
//	double	theta =	0.0;
//	double	dTheta = 4*M_PI/20.0;	
	//<family

	double ccolor[3];
	ccolor[0] =	0.0;
	ccolor[1] =	0.0;
	ccolor[2] =	0.0;

	//----------------------------------------------->
	//@	INIT section
	p =	(quaternion*) my_malloc(sizeof(quaternion));
	q =	(quaternion*) my_malloc(sizeof(quaternion));
	r =	(quaternion*) my_malloc(sizeof(quaternion));

	p->r = 0;
	p->i = 0;
	p->j = 0;
	p->k = 0;
	
	dotPrevCurrBinormal	= 0.0;
	// <----------------------------------------------


	if(mo2->numPoints >	2) {
		printf("\n>>>>>>>\n");
		t[0] = mo2->TangList[i].p3.d[0];
		t[1] = mo2->TangList[i].p3.d[1];
		t[2] = mo2->TangList[i].p3.d[2];
		
		// obtain first-derivative of tangent
		tt[0] =	mo2->dTangList[i].p3.d[0];
		tt[1] =	mo2->dTangList[i].p3.d[1];
		tt[2] =	mo2->dTangList[i].p3.d[2];
		//----------------------------------------------->

		//@	obtain cross-product of	Tangent	and	First
		//	Derivative of tangent (Binormal	vector)
		cross3D(t, tt, currBinormal);
		len	= length3D(currBinormal);
		currBinormal[0]	= currBinormal[0]/len;
		currBinormal[1]	= currBinormal[1]/len;
		currBinormal[2]	= currBinormal[2]/len;
		// <----------------------------------------------

		//----------------------------------------------->
		//@	obtain the third orthonormal component
		//	from cross-product of Tangent and Binormal
		cross3D(currBinormal,t,	currNormal);
		len	= length3D(currNormal);
		currNormal[0] =	currNormal[0]/len;
		currNormal[1] =	currNormal[1]/len;
		currNormal[2] =	currNormal[2]/len;
		// <----------------------------------------------

		//----------------------------------------------->
		//@	Frenet-Serret frame	is given by	the	set	of 
		//	coordinate axes:
		//			_  _  _	
		//		  [	T  N  B	]
		//
		// <----------------------------------------------

		//----------------------------------------------->
		//@	build the 4x4 Rotation matris with the
		//	columns	as the orthonormal basis constructed
		//	above using: Tangent, Binormal and Third component
		mat[0] = t[0];
		mat[1] = t[1];
		mat[2] = t[2];

		mat[4] = currNormal[0];
		mat[5] = currNormal[1];
		mat[6] = currNormal[2];

		mat[8] = currBinormal[0];
		mat[9] = currBinormal[1];
		mat[10]= currBinormal[2];		

		mat[3]	= mat[7] = mat[11] = mat[12] = mat[13] = mat[14] = 0.0;
		mat[15]	= 1.0;
		// <----------------------------------------------

		//----------------------------------------------->
		//@	obtain the quaternion from rotation	matrix
		//	formed by the orthonormal basis	of the frame
		//	that have three	basis-vectors at a point p:
		//	Tangent, Binormal and Third	component
		ExtractQuaternionFromMat(mat,p);
		quatNormalize(p);	
		// <----------------------------------------------
		
		mo2->quatList[i].r = p->r;
		mo2->quatList[i].i = p->i;
		mo2->quatList[i].j = p->j;
		mo2->quatList[i].k = p->k;				  

		quatAngle =	2*acos(p->r);
		mo2->markQuatAngle[i] =	quatAngle;

		start =	1;

		prevBinormal[0]	= currBinormal[0];
		prevBinormal[1]	= currBinormal[1];
		prevBinormal[2]	= currBinormal[2];
		
		for(i=start;i<(mo2->numPoints-2);i++)	{

			// obtain tangent
			t[0] = mo2->TangList[i].p3.d[0];
			t[1] = mo2->TangList[i].p3.d[1];
			t[2] = mo2->TangList[i].p3.d[2];
			
			// obtain first-derivative of tangent
			tt[0] =	mo2->dTangList[i].p3.d[0];
			tt[1] =	mo2->dTangList[i].p3.d[1];
			tt[2] =	mo2->dTangList[i].p3.d[2];

			//----------------------------------------------->
			//@	obtain cross-product of	Tangent	and	First
			//	Derivative of tangent (Binormal	vector)
			cross3D(t, tt, currBinormal);
			len	= length3D(currBinormal);
			currBinormal[0]	= currBinormal[0]/len;
			currBinormal[1]	= currBinormal[1]/len;
			currBinormal[2]	= currBinormal[2]/len;
			// <----------------------------------------------

			//----------------------------------------------->
			//@	not	doing this right now but this is 
			//	one	of the steps. Not sure about this.
			dotPrevCurrBinormal	= dot3D(prevBinormal, currBinormal);
			
			if(dotPrevCurrBinormal<0.10) {
				//printf("# [%d] dotBirnormals = %f\n",i,dotPrevCurrBinormal);
				//mo2->markFrame[mo2->markFrameCounter]=i+1;
				//mo2->markFrameCounter++;
			}
			else {
				//printf("  [%d] dotBirnormals = %f\n",i,dotPrevCurrBinormal);
			}
			// <----------------------------------------------

			// copy	currBinormal into prevBinormal
			prevBinormal[0]	= currBinormal[0];
			prevBinormal[1]	= currBinormal[1];
			prevBinormal[2]	= currBinormal[2];
		}

		//for(i=0;i<mo2->markFrameCounter; i++){ 
		//	printf("%d \n",mo2->markFrame[i]);
		//}
		for(i=0; i<mo2->numPoints-2;i++) {
			// obtain tangent
			t[0] = mo2->TangList[i].p3.d[0];
			t[1] = mo2->TangList[i].p3.d[1];
			t[2] = mo2->TangList[i].p3.d[2];
			
			// obtain first-derivative of tangent
			tt[0] =	mo2->dTangList[i].p3.d[0];
			tt[1] =	mo2->dTangList[i].p3.d[1];
			tt[2] =	mo2->dTangList[i].p3.d[2];

			cross3D(t, tt, currBinormal);
			len	= length3D(currBinormal);
			currBinormal[0]	= currBinormal[0]/len;
			currBinormal[1]	= currBinormal[1]/len;
			currBinormal[2]	= currBinormal[2]/len;		

			cross3D(currBinormal,t,	currNormal);
			len	= length3D(currNormal);
			currNormal[0] =	currNormal[0]/len;
			currNormal[1] =	currNormal[1]/len;
			currNormal[2] =	currNormal[2]/len;

			mat[0] = t[0];
			mat[1] = t[1];
			mat[2] = t[2];

			mat[4] = currNormal[0];
			mat[5] = currNormal[1];
			mat[6] = currNormal[2];

			mat[8] = currBinormal[0];
			mat[9] = currBinormal[1];
			mat[10]= currBinormal[2];		

			mat[3]	= mat[7] = mat[11] = mat[12] = mat[13] = mat[14] = 0;
			mat[15]	= 1.0;

			ExtractQuaternionFromMat(mat,q);
			quatNormalize(q);

			//save out resulting quaternion
			mo2->quatList[i].r = q->r;
			mo2->quatList[i].i = q->i;
			mo2->quatList[i].j = q->j;
			mo2->quatList[i].k = q->k;
		}
		
		for(i=1;i<mo2->numPoints-2; i++) {
			
			p->r = mo2->quatList[i-1].r;
			p->i = mo2->quatList[i-1].i;
			p->j = mo2->quatList[i-1].j;
			p->k = mo2->quatList[i-1].k;

			q->r = mo2->quatList[i].r;
			q->i = mo2->quatList[i].i;
			q->j = mo2->quatList[i].j;
			q->k = mo2->quatList[i].k;

			dot_PQ = quatQuatDotProduct(p,q);
			if(dot_PQ < -0.5) {
				//printf("# %d> %f\n",i,dot_PQ);
				mo2->markQuatCurve[mo2->markQuatCurveCounter] = i;
				mo2->markQuatCurveCounter++;
				
				mo2->quatList[i].r = q->r;
				mo2->quatList[i].i = -q->i;
				mo2->quatList[i].j = -q->j;
				mo2->quatList[i].k = -q->k;
			}
			else {
				//printf("%d> %f\n",i,dot_PQ);
			}

		}
		mo2->markQuatCurve[mo2->markQuatCurveCounter] = mo2->numPoints;
		drawQuaternionCurve(ccolor);
		glPushAttrib(GL_LIGHTING_BIT);
		glDisable(GL_LIGHTING);
		for(i=1;i<mo2->numPoints-2; i++) {
			p->r = mo2->quatList[i-1].r;
			p->i = mo2->quatList[i-1].i;
			p->j = mo2->quatList[i-1].j;
			p->k = mo2->quatList[i-1].k;

			q->r = mo2->quatList[i].r;
			q->i = mo2->quatList[i].i;
			q->j = mo2->quatList[i].j;
			q->k = mo2->quatList[i].k;

			dot_PQ = quatQuatDotProduct(p,q);
			if(dot_PQ < 0.9 && dot_PQ > -0.9) {
				printf("# %d> %f\n",i,dot_PQ);
				glPushMatrix();
				glTranslatef(p->i,p->j,p->k);
				glColor3f(1.0f,0.0f,0.0f);
				glutSolidSphere(0.08,16,16);
				glPopMatrix();
				glPushMatrix();
				glTranslatef(q->i,q->j,q->k);
				glColor3f(0.0f,1.0f,0.0f);
				glutSolidSphere(0.05,16,16);
				glPopMatrix();
			}
		}
		glPopAttrib();
		/*

		start =	0;
		if(mo2->markQuatCurveCounter==	0)
			end	= mo2->numPoints-2;
		else
			end	= mo2->markQuatCurve[0];
		for(i=0; i<=mo2->markQuatCurveCounter; i++) {

			//fprintf(stderr,"start = %d, end = %d\n",start,end);
			getSegmentColor(i,ccolor);
			drawQuaternionCurve(start,end,ccolor);

			//!	Compute	range for next section of smooth-curve
			start =	mo2->markQuatCurve[i]+1;
			end	= mo2->markQuatCurve[i+1];
			if(i==(mo2->markQuatCurveCounter)-1) {
				end	= mo2->numPoints-2;
			}
		}
		*/
		/*
		// Calculate quaternions for smoothly
		// varying portions	of curve only
		start =	0;
		if(mo2->markFrameCounter ==	0)
			end	= mo2->numPoints-2;
		else
			end	= mo2->markFrame[0];
		glPushAttrib(GL_LINE_BIT | GL_LIGHTING_BIT | GL_CURRENT_BIT);
		glDisable(GL_LIGHTING);
		glLineWidth(2.0);
		int	check=0;
		int	colnum = 0;
		for(i=0;i<=mo2->markFrameCounter;i++) {
			check=0;
			for(j=start; j<end;	j++) {
				check++;

				// obtain tangent
				t[0] = mo2->TangList[j].p3.d[0];
				t[1] = mo2->TangList[j].p3.d[1];
				t[2] = mo2->TangList[j].p3.d[2];
			
				// obtain first-derivative of tangent
				tt[0] =	mo2->dTangList[j].p3.d[0];
				tt[1] =	mo2->dTangList[j].p3.d[1];
				tt[2] =	mo2->dTangList[j].p3.d[2];

				//----------------------------------------------->
				//@	obtain cross-product of	Tangent	and	First
				//	Derivative of tangent (Binormal	vector)
				cross3D(t, tt, currBinormal);
				len	= length3D(currBinormal);
				currBinormal[0]	= currBinormal[0]/len;
				currBinormal[1]	= currBinormal[1]/len;
				currBinormal[2]	= currBinormal[2]/len;
				// <----------------------------------------------

				//----------------------------------------------->
				//@	obtain the third orthonormal component
				//	from cross-product of Tangent and Binormal
				cross3D(currBinormal,t,	currNormal);
				len	= length3D(currNormal);
				currNormal[0] =	currNormal[0]/len;
				currNormal[1] =	currNormal[1]/len;
				currNormal[2] =	currNormal[2]/len;
				// <----------------------------------------------

				mat[0] = t[0];
				mat[1] = t[1];
				mat[2] = t[2];

				mat[4] = currNormal[0];
				mat[5] = currNormal[1];
				mat[6] = currNormal[2];

				mat[8] = currBinormal[0];
				mat[9] = currBinormal[1];
				mat[10]= currBinormal[2];		

				mat[3]	= mat[7] = mat[11] = mat[12] = mat[13] = mat[14] = 0;
				mat[15]	= 1.0;

				ExtractQuaternionFromMat(mat,q);
				quatNormalize(q);

				//	check if dot product of	neighboring	quaternions
				//	is near	negative one
				dot_PQ = quatQuatDotProduct(p,q);
				//if(dot_PQ<0.9	&& dot_PQ>0.0) {
				//	fprintf(stderr,"!!dot(prev,curr) = %f\n",dot_PQ);
				//	glPushMatrix();
				//	glTranslatef(q->r,q->j,q->k);
				//	glutSolidSphere(0.035,16,16);
				//	glPopMatrix();
				//}
				//else if(dot_PQ <0.0 && dot_PQ>=-0.9)
				//	fprintf(stderr,"->dot(prev,curr) = %f\n",dot_PQ);
				//else if(dot_PQ<-0.9)
				//	fprintf(stderr,"<<dot(prev,curr) = %f\n",dot_PQ);
				//else

				if(dot_PQ <	0.0) {
					// flip	one	of the quaternions
					//fprintf(stderr,"# [%d] dot(prev,curr) = %f\n",j,dot_PQ);
					r->r = (q->r);
					r->i = -(q->i);
					r->j = -(q->j);
					r->k = -(q->k);
					//quatCopy(r,q);
					dot_PQ = quatQuatDotProduct(p,q);
				}
				else {
					//fprintf(stderr,"> [%d] dot(prev,curr) = %f\n",j,dot_PQ);
				}

				if(dot_PQ<-0.9)
				{
					mo2->markQuatCurve[mo2->markQuatCurveCounter] =	j;
					mo2->markQuatCurveCounter++;
				}

				//save out resulting quaternion
				mo2->quatList[j].r = q->r;
				mo2->quatList[j].i = q->i;
				mo2->quatList[j].j = q->j;
				mo2->quatList[j].k = q->k;				  

				// save	current	quaternion as 'previous-quaternion'
				// for next	iteration
				quatCopy(q,p);
			}

			if(check>0)
				colnum = i;
			if(i==mo2->markFrameCounter)
				colnum = 0;
			getSegmentColor(colnum,	ccolor);
			drawQuaternionCurve(start,end,ccolor);

			//!	Compute	range for next section of smooth-curve
			start =	mo2->markFrame[i]+1;
			end	= mo2->markFrame[i+1];
			if(i==(mo2->markFrameCounter)-1) {
				end	= mo2->numPoints-2;
			}
		}
		//printf(">	Mark Quaternion	Curve\n");
		//for(i=0;i<mo2->markQuatCurveCounter;i++) {
		//	printf("%d ",mo2->markQuatCurve[i]);
		//}

		//printf("\n");		
		//printf("> Mark QUAT Frames\n");
		//glColor3f(0.0f,1.0f,0.0f);
		//for(i=0;i<mo2->markQuatCurveCounter;i++) {
		//	printf("flipping quaternion.... %d \n",mo2->markQuatCurve[i]);
		//}
		*/
		glPopAttrib();

	}
}

void computeFrenetQuaternionMap6 ( void	)
{
	int		i			= 0;	// index into points array
	//int		start		= 0;	// start index for the for-loop
	double	len			= 0.0;	// vector length
	double	dot_PQ		= 0.0;	// quaternion inner-product
	//double	quatAngle	= 0.0;

	double	t[3];
	double	tt[3];
	double	prevBinormal[3];	// vector for previous Binormal	vector
	double	currBinormal[3];	// vector for current Binormal vector
	double	currNormal[3];
	double	dotPrevCurrBinormal;// scalar, dot-product of previous and current Binormal
	//double	qvec[3];			// quaternion components
	double	mat[16];			// rotation	matrix
	quaternion *p, *q, *r;		//
	double	quatProd[1000];

	double ccolor[3];
	ccolor[0] =	0.0;
	ccolor[1] =	0.0;
	ccolor[2] =	0.0;

	//initialize 
	mo2->markFrameCounter =	0;
	mo2->markQuatCurveCounter =	0;
	currNormal[0] = currNormal[1] = currNormal[2] = 0.0;

	//----------------------------------------------->
	//@	INIT section
	p =	(quaternion*) my_malloc(sizeof(quaternion));
	q =	(quaternion*) my_malloc(sizeof(quaternion));
	r =	(quaternion*) my_malloc(sizeof(quaternion));
	// <----------------------------------------------
	
	//----------------------------------------------->

	glPushAttrib( GL_ALL_ATTRIB_BITS );
	glDisable( GL_LIGHTING );
	glLineWidth(0.5f);

	//# 1. Initial calculation
	t[0] = mo2->TangList[i].p3.d[0];
	t[1] = mo2->TangList[i].p3.d[1];
	t[2] = mo2->TangList[i].p3.d[2];

	tt[0] =	mo2->dTangList[i].p3.d[0];
	tt[1] =	mo2->dTangList[i].p3.d[1];
	tt[2] =	mo2->dTangList[i].p3.d[2];

	cross3D(t, tt, currBinormal);
	len	= length3D(currBinormal);
	prevBinormal[0]	= currBinormal[0]/len;
	prevBinormal[1]	= currBinormal[1]/len;
	prevBinormal[2]	= currBinormal[2]/len;	

	mat[0] = t[0];
	mat[1] = t[1];
	mat[2] = t[2];

	mat[4] = currNormal[0];
	mat[5] = currNormal[1];
	mat[6] = currNormal[2];

	mat[8] = currBinormal[0];
	mat[9] = currBinormal[1];
	mat[10]= currBinormal[2];		

	mat[3]	= mat[7] = mat[11] = mat[12] = mat[13] = mat[14] = 0.0;
	mat[15]	= 1.0;

	ExtractQuaternionFromMat(mat,p);
	quatNormalize(p);	

	//# 2. Calculate all vectors
	//printf(">>>\n");
	for(i=1; i<mo2->numPoints-2; i++) {
		t[0] = mo2->TangList[i].p3.d[0];
		t[1] = mo2->TangList[i].p3.d[1];
		t[2] = mo2->TangList[i].p3.d[2];

		tt[0] =	mo2->dTangList[i].p3.d[0];
		tt[1] =	mo2->dTangList[i].p3.d[1];
		tt[2] =	mo2->dTangList[i].p3.d[2];

		cross3D(t, tt, currBinormal);
		len	= length3D(currBinormal);
		currBinormal[0]	= currBinormal[0]/len;
		currBinormal[1]	= currBinormal[1]/len;
		currBinormal[2]	= currBinormal[2]/len;

		cross3D(currBinormal,t,	currNormal);
		len	= length3D(currNormal);
		currNormal[0] =	currNormal[0]/len;
		currNormal[1] =	currNormal[1]/len;
		currNormal[2] =	currNormal[2]/len;

		dotPrevCurrBinormal = dot3D(currBinormal, prevBinormal);

		if(dotPrevCurrBinormal<=0.6) {
			mo2->markFrame[mo2->markFrameCounter] = i-1;
			mo2->markFrameCounter++;
		}

		//#2.a Extract quaternions
		mat[0] = t[0];
		mat[1] = t[1];
		mat[2] = t[2];

		mat[4] = currNormal[0];
		mat[5] = currNormal[1];
		mat[6] = currNormal[2];

		mat[8] = currBinormal[0];
		mat[9] = currBinormal[1];
		mat[10]= currBinormal[2];		

		mat[3]	= mat[7] = mat[11] = mat[12] = mat[13] = mat[14] = 0.0;
		mat[15]	= 1.0;

		ExtractQuaternionFromMat(mat,q);
		quatNormalize(q);	

		dot_PQ = quatQuatDotProduct( p, q );
		quatProd[i] = dot_PQ;
		//printf("[%d] dot(Qp,Qq) = %f\n",i,dot_PQ);

		if(mo2->kflip) {
			if(dot_PQ < -0.9) {
				r->r = -q->r;
				r->i = -q->i;
				r->j = -q->j;
				r->k = -q->k;

				quatCopy(r, q);
				//printf("[%d] dot(Qp,Qq) = %f\n",i,quatQuatDotProduct( p, q ));
				//quatProd[i] = quatQuatDotProduct(p,q);
			}

		}
		mo2->quatList[i].r = q->r;
		mo2->quatList[i].i = q->i;
		mo2->quatList[i].j = q->j;
		mo2->quatList[i].k = q->k;

		prevBinormal[0]	= currBinormal[0];
		prevBinormal[1]	= currBinormal[1];
		prevBinormal[2]	= currBinormal[2];	
		quatCopy(q,p);

		//glPushAttrib(GL_ALL_ATTRIB_BITS);
		//glEnable(GL_COLOR_MATERIAL);
		//glEnable(GL_LIGHTING);

		//double delta = 0.075;
		//getQuatComp(mo2->qwhichQuatComponent, qorig[i], qvec);
		//glColor4f(0.6,0.0,0.0,0.5f);
		//glPushMatrix();
		//	glTranslatef(qvec[0], qvec[1], qvec[2]);
		//	glutSolidSphere(0.03,16,16);
		//glPopMatrix();
		///*glPushMatrix();
		//	glTranslatef(qvec[0]+delta, qvec[1]+delta, qvec[2]+delta);
		//	gluOrtho2D(-0.1,0.1,-0.1,0.1);
		//	glDisable(GL_LIGHTING);
		//	glColor3f(0.2f,0.2f,0.2f);
		//	sprintf(obj_qstr,"%d",i);
		//	glRasterPos2f(0.0,0.0);
		//	stext(obj_qstr);
		//glPopMatrix();*/

		//getQuatComp(mo2->qwhichQuatComponent, qfliped[i], qvec);
		//glColor4f(0.0,0.6,0.0,1.0);
		//glEnable(GL_LIGHTING);
		//glPushMatrix();
		//	glTranslatef(qvec[0], qvec[1], qvec[2]);
		//	glutSolidSphere(0.03,16,16);
		//glPopMatrix();
		///*glPushMatrix();
		//	glTranslatef(qvec[0]+delta, qvec[1]+delta, qvec[2]+delta);
		//	gluOrtho2D(-0.1,0.1,-0.1,0.1);
		//	glDisable(GL_LIGHTING);
		//	glColor3f(0.2f,0.2f,0.2f);
		//	sprintf(obj_qstr,"%d",i);
		//	glRasterPos2f(0.0,0.0);
		//	stext(obj_qstr);
		//glPopMatrix();*/
		//glPopAttrib();

		//glColor3f(0.1f,0.9f,0.9f);
		//glPushMatrix();
		//	glTranslatef(mo2->pntList[i].p3.d[0],mo2->pntList[i].p3.d[1],mo2->pntList[i].p3.d[2]);
		//	glScalef(0.5,0.5,0.5);
		//	glBegin(GL_LINES);
		//	glVertex3f(0,0,0);
		//	glVertex3f(t[0], t[1], t[2]);
		//	glEnd();
		//	glPushMatrix();
		//		glTranslatef(t[0],t[1],t[2]);
		//		gluOrtho2D(-0.1,0.1,-0.1,0.1);
		//		glDisable(GL_LIGHTING);
		//		glColor3f(0.2f,0.2f,0.2f);
		//		sprintf(obj_qstr,"T");
		//		glRasterPos2f(0.0,0.0);
		//		stext(obj_qstr);
		//	glPopMatrix();
		//glPopMatrix();	

		//glColor3f(0.9f,0.1f,0.9f);
		//glPushMatrix();
		//	glTranslatef(mo2->pntList[i].p3.d[0],mo2->pntList[i].p3.d[1],mo2->pntList[i].p3.d[2]);
		//	glScalef(0.5,0.5,0.5);
		//	glBegin(GL_LINES);
		//	glVertex3f(0,0,0);
		//	glVertex3f(tt[0], tt[1], tt[2]);
		//	glEnd();
		//	glPushMatrix();
		//		glTranslatef(tt[0],tt[1],tt[2]);
		//		gluOrtho2D(-0.1,0.1,-0.1,0.1);
		//		glDisable(GL_LIGHTING);
		//		glColor3f(0.2f,0.2f,0.2f);
		//		sprintf(obj_qstr,"dT");
		//		glRasterPos2f(0.0,0.0);
		//		stext(obj_qstr);
		//	glPopMatrix();
		//glPopMatrix();	

		//glColor3f(0.1,0.1,0.1);
		//glDisable(GL_LIGHTING);
		//glPushMatrix();
		//	glTranslatef(mo2->pntList[i].p3.d[0],mo2->pntList[i].p3.d[1],mo2->pntList[i].p3.d[2]);
		//	glScalef(0.5,0.5,0.5);
		//	glBegin(GL_LINES);
		//	glVertex3f(0,0,0);
		//	glVertex3f(currBinormal[0], currBinormal[1], currBinormal[2]);
		//	glEnd();
		//	glPushMatrix();
		//		glTranslatef(currBinormal[0], currBinormal[1], currBinormal[2]);
		//		gluOrtho2D(-0.1,0.1,-0.1,0.1);
		//		glDisable(GL_LIGHTING);
		//		glColor3f(0.2f,0.2f,0.2f);
		//		sprintf(obj_qstr,"B%d",i);
		//		glRasterPos2f(0.0,0.0);
		//		stext(obj_qstr);
		//	glPopMatrix();
		//glPopMatrix();	

		//glColor3f(1.0,1.0,0.0);
		//glPushMatrix();
		//	glTranslatef(mo2->pntList[i].p3.d[0],mo2->pntList[i].p3.d[1],mo2->pntList[i].p3.d[2]);
		//	glScalef(0.5,0.5,0.5);
		//	glBegin(GL_LINES);
		//	glVertex3f(0,0,0);
		//	glVertex3f(currNormal[0], currNormal[1], currNormal[2]);
		//	glEnd();
		//	glPushMatrix();
		//		glTranslatef(currNormal[0], currNormal[1], currNormal[2]);
		//		gluOrtho2D(-0.1,0.1,-0.1,0.1);
		//		glDisable(GL_LIGHTING);
		//		glColor3f(0.2f,0.2f,0.2f);
		//		sprintf(obj_qstr,"N");
		//		glRasterPos2f(0.0,0.0);
		//		stext(obj_qstr);
		//	glPopMatrix();
		//glPopMatrix();	
	}
	//for(i=0;i<mo2->markFrameCounter; i++){ 
	//	printf("%d \n",mo2->markFrame[i]);
	//}
	//for(i=1;i<mo2->numPoints-2; i++){ 
	//	printf("%d newDotprod = %f\n",i,quatProd[i]);
	//}
	glPopAttrib();

	my_free(p);
	my_free(q);
	my_free(r);
}

void computeFrenetQuaternionMap7 ( void	)
{
	int		i			= 0;	// index into points array
	double	len			= 0.0;	// vector length
	double	dot_PQ		= 0.0;	// quaternion inner-product
	int		next		= 0;

	double	t[3];
	double	tt[3];
	double	prevBinormal[3];	// vector for previous Binormal	vector
	double	currBinormal[3];	// vector for current Binormal vector
	double	currNormal[3];
	double	dotPrevCurrBinormal;// scalar, dot-product of previous and current Binormal
	//double	qvec[3];			// quaternion components
	double	mat[16];			// rotation	matrix
	quaternion *p, *q, *r;		//
	double	quatProd[1000];
	
	double ccolor[3];
	ccolor[0] =	0.0;
	ccolor[1] =	0.0;
	ccolor[2] =	0.0;

	//@	INIT section
	mo2->markFrameCounter =	0;
	mo2->markQuatCurveCounter =	0;
	mo2->markQuatCurveCounter2 = 0;

	p =	(quaternion*) my_malloc(sizeof(quaternion));
	q =	(quaternion*) my_malloc(sizeof(quaternion));
	r =	(quaternion*) my_malloc(sizeof(quaternion));
	
	
	//printf(">>>\n");
	// calculate quaternions
	for(i=0; i<mo2->numPoints; i++) {

		if(i==(mo2->numPoints-1)) {
			t[0] = mo2->TangList[next].p3.d[0];
			t[1] = mo2->TangList[next].p3.d[1];
			t[2] = mo2->TangList[next].p3.d[2];

			tt[0] =	mo2->dTangList[next].p3.d[0];
			tt[1] =	mo2->dTangList[next].p3.d[1];
			tt[2] =	mo2->dTangList[next].p3.d[2];
		}
		else {
			t[0] = mo2->TangList[i].p3.d[0];
			t[1] = mo2->TangList[i].p3.d[1];
			t[2] = mo2->TangList[i].p3.d[2];

			tt[0] =	mo2->dTangList[i].p3.d[0];
			tt[1] =	mo2->dTangList[i].p3.d[1];
			tt[2] =	mo2->dTangList[i].p3.d[2];
		}

		cross3D(t, tt, currBinormal);
		len	= length3D(currBinormal);
		currBinormal[0]	= currBinormal[0]/len;
		currBinormal[1]	= currBinormal[1]/len;
		currBinormal[2]	= currBinormal[2]/len;

		cross3D(currBinormal,t,	currNormal);
		len	= length3D(currNormal);
		currNormal[0] =	currNormal[0]/len;
		currNormal[1] =	currNormal[1]/len;
		currNormal[2] =	currNormal[2]/len;

		//#2.a Extract quaternions
		mat[0] = t[0];
		mat[1] = t[1];
		mat[2] = t[2];

		mat[4] = currNormal[0];
		mat[5] = currNormal[1];
		mat[6] = currNormal[2];

		mat[8] = currBinormal[0];
		mat[9] = currBinormal[1];
		mat[10]= currBinormal[2];		

		mat[3]	= mat[7] = mat[11] = mat[12] = mat[13] = mat[14] = 0.0;
		mat[15]	= 1.0;

		ExtractQuaternionFromMat(mat,q);
		quatNormalize(q);	

		if(i>=1) {
			dotPrevCurrBinormal = dot3D(currBinormal, prevBinormal);
			if(dotPrevCurrBinormal<=0.6) {
				mo2->markFrame[mo2->markFrameCounter] = i-1;
				mo2->markFrameCounter++;
				mo2->markQuatCurve[mo2->markQuatCurveCounter] = i-1;
				mo2->markQuatCurveCounter++;

			}
			
			dot_PQ = quatQuatDotProduct( p, q );
			quatProd[i] = dot_PQ;
			//printf("[%d] dot(Qp,Qq) = %f\n",i,dot_PQ);
			if(mo2->kflip) {
				if(dot_PQ < -0.9) {
					//mo2->markQuatCurve2[mo2->markQuatCurveCounter2] = i-1;
					//mo2->markQuatCurveCounter2++;
					r->r = -q->r;
					r->i = -q->i;
					r->j = -q->j;
					r->k = -q->k;
					quatCopy(r, q);
					//printf("[%d] dot(Qp,Qq) = %f\n",i,quatQuatDotProduct( p, q ));
					//quatProd[i] = quatQuatDotProduct(p,q);
				}
			}
			else {
				if(dot_PQ < -0.9) {
					mo2->markQuatCurve[mo2->markQuatCurveCounter] = i-1;
					mo2->markQuatCurveCounter++;
				}
			}
		}
		mo2->quatList[i].r = q->r;
		mo2->quatList[i].i = q->i;
		mo2->quatList[i].j = q->j;
		mo2->quatList[i].k = q->k;
		//printf("[%d] q = %f, %f, %f, %f\n",i,q->r,q->i,q->j,q->k);

		prevBinormal[0]	= currBinormal[0];
		prevBinormal[1]	= currBinormal[1];
		prevBinormal[2]	= currBinormal[2];	
		quatCopy(q,p);
	}

	//for(i=0;i<mo2->markFrameCounter; i++){ 
	//	printf("%d \n",mo2->markFrame[i]);
	//}
	//printf("Mark Quat counter> \n");
	//for(i=0;i<mo2->markQuatCurveCounter; i++){ 
	//	printf("%d \n",mo2->markQuatCurve[i]);
	//}
	my_free(p);
	my_free(q);
	my_free(r);
 }


void computeFrenetQuaternionMap8 ( void )
{
	int		i			= 0;	// index into points array
	double	len			= 0.0;	// vector length
	double	dot_PQ		= 0.0;	// quaternion inner-product
	int		next		= 0;

	double	t[3];
	double	tt[3];
	double	prevBinormal[3];	// vector for previous Binormal	vector
	double	currBinormal[3];	// vector for current Binormal vector
	double	currNormal[3];
	double	dotPrevCurrBinormal;// scalar, dot-product of previous and current Binormal
	//double	qvec[3];			// quaternion components
	double	mat[16];			// rotation	matrix
	quaternion *p, *q, *r, *qf;	//
	double	quatProd[1000];
	
	double ccolor[3];
	ccolor[0] =	0.0;
	ccolor[1] =	0.0;
	ccolor[2] =	0.0;

	//# Quat rings
	double	theta = 0.0;
	double	qvec[3];
	double	scalar=0.0;
	double	angle = 0.0;
	double  vecpart = 0.0;
	double	qvecpart[3];
	int		flipedkya = 0;
	int		flipedbackkya = 0;
	int		flipcounter = 0;


	//@	INIT section
	mo2->markFrameCounter =	0;
	mo2->markQuatCurveCounter =	0;
	mo2->markQuatCurveCounter2 = 0;

	p =	(quaternion*) my_malloc(sizeof(quaternion));
	q =	(quaternion*) my_malloc(sizeof(quaternion));
	r =	(quaternion*) my_malloc(sizeof(quaternion));
	qf=	(quaternion*) my_malloc(sizeof(quaternion));
	
	//printf(">>>\n");
	// calculate quaternions
	for(i=0; i<mo2->numPoints; i++) {

		if(i==(mo2->numPoints-1)) {
			t[0] = mo2->TangList[next].p3.d[0];
			t[1] = mo2->TangList[next].p3.d[1];
			t[2] = mo2->TangList[next].p3.d[2];

			tt[0] =	mo2->dTangList[next].p3.d[0];
			tt[1] =	mo2->dTangList[next].p3.d[1];
			tt[2] =	mo2->dTangList[next].p3.d[2];
		}
		else {
			t[0] = mo2->TangList[i].p3.d[0];
			t[1] = mo2->TangList[i].p3.d[1];
			t[2] = mo2->TangList[i].p3.d[2];

			tt[0] =	mo2->dTangList[i].p3.d[0];
			tt[1] =	mo2->dTangList[i].p3.d[1];
			tt[2] =	mo2->dTangList[i].p3.d[2];
		}
		//printf("[%d]\t t = %f, %f, %f\n",i,t[0],t[1],t[2]);

		

		cross3D(t, tt, currBinormal);
		len	= length3D(currBinormal);
		currBinormal[0]	= currBinormal[0]/len;
		currBinormal[1]	= currBinormal[1]/len;
		currBinormal[2]	= currBinormal[2]/len;

		cross3D(currBinormal,t,	currNormal);
		len	= length3D(currNormal);
		currNormal[0] =	currNormal[0]/len;
		currNormal[1] =	currNormal[1]/len;
		currNormal[2] =	currNormal[2]/len;

		//#2.a Extract quaternions
		mat[0] = t[0];
		mat[1] = t[1];
		mat[2] = t[2];

		mat[4] = currNormal[0];
		mat[5] = currNormal[1];
		mat[6] = currNormal[2];

		mat[8] = currBinormal[0];
		mat[9] = currBinormal[1];
		mat[10]= currBinormal[2];		

		mat[3]	= mat[7] = mat[11] = mat[12] = mat[13] = mat[14] = 0.0;
		mat[15]	= 1.0;

		ExtractQuaternionFromMat(mat,q);
		quatNormalize(q);	
		quatCopy(q,mo2->qpntOrig[i]);
		//if(i==0)
		quatCopy(q,mo2->qpntFliped[i]);
		if(i>=1) {
			dotPrevCurrBinormal = dot3D(currBinormal, prevBinormal);
			if(dotPrevCurrBinormal<=0.6) {
				mo2->markFrame[mo2->markFrameCounter] = i-1;
				mo2->markFrameCounter++;
				mo2->markQuatCurve[mo2->markQuatCurveCounter] = i-1;
				mo2->markQuatCurveCounter++;

			}
			quatCopy(q,mo2->qpntOrig[i]);
			dot_PQ = quatQuatDotProduct( p, q );
			quatProd[i] = dot_PQ;
			//printf("[%d] dot(Qp,Qq) = %f\n",i,dot_PQ);
			if(dot_PQ > 0.9 && flipedbackkya || (flipcounter==1 && i==mo2->numPoints-1) ){
				mo2->qflipindex[flipcounter] = i;
				flipcounter++;
				flipedbackkya = 0;
			}
			
			if(dot_PQ > 0.9)
				flipedkya = 0;

			if(mo2->kflip) {
				if(dot_PQ < -0.9) {
					//printf("flipping... %d\n",i);
					if(flipedkya==0) {
						mo2->qflipindex[flipcounter] = i;
						flipcounter++;
						flipedbackkya = 1;
					}
					flipedkya = 1;

					//mo2->markQuatCurve[mo2->markQuatCurveCounter] = i-1;
					//mo2->markQuatCurveCounter++;
					r->r = -q->r;
					r->i = -q->i;
					r->j = -q->j;
					r->k = -q->k;
					quatCopy(r, q);
					quatCopy(q,mo2->qpntFliped[i]);
					//printf("[%d] dot(Qp,Qq) = %f\n",i,quatQuatDotProduct( p, q ));
					//quatProd[i] = quatQuatDotProduct(p,q);
				}
			}
			else {
				if(dot_PQ < -0.9) {
					mo2->markQuatCurve[mo2->markQuatCurveCounter] = i-1;
					mo2->markQuatCurveCounter++;
				}
			}
		}
		mo2->qflipcounter = flipcounter;

		//## QUATERNION RINGS
		if(mo2->qshowQuatRings) {
			glPushAttrib(GL_ALL_ATTRIB_BITS | GL_LIGHTING_BIT);
			glDisable(GL_LIGHTING);
			glLineWidth(1.25f);
			scalar = q->r;
			angle  = 2.0*acos(scalar);
			vecpart= sin(angle/2.0);
			qvecpart[0] = q->i/vecpart;
			qvecpart[1] = q->j/vecpart;
			qvecpart[2] = q->k/vecpart;
			if(i%10==0) {
				theta=0.0;
				glColor3f(0.0f,0.0f,0.0f);
				glBegin(GL_LINE_STRIP);
				while(theta<=16.0) {
					makeFamilyOfQuaternions(theta,qvecpart,qf);
					getQuatComp(mo2->qwhichQuatComponent,qf,qvec);
					glVertex3f(qvec[0],qvec[1],qvec[2]);
					theta = theta + 0.05;
				}		
				glEnd();
			}
			//printf("\tv = %f, %f, %f\n", qvecpart[0], qvecpart[1], qvecpart[2]);
			glPopAttrib();
		}

		//# Copy quaternion into quatList[]
		mo2->quatList[i].r = q->r;
		mo2->quatList[i].i = q->i;
		mo2->quatList[i].j = q->j;
		mo2->quatList[i].k = q->k;
		//printf("[%d] q = %f, %f, %f, %f\n",i,q->r,q->i,q->j,q->k);

		prevBinormal[0]	= currBinormal[0];
		prevBinormal[1]	= currBinormal[1];
		prevBinormal[2]	= currBinormal[2];	
		quatCopy(q,p);
	}
	//printf("flipcounter = %d\n",flipcounter);
	if( flipcounter>1 && flipcounter%2==1) {
		mo2->qflipindex[flipcounter] = mo2->numPoints-1 ;
		flipcounter++;			
	}

	my_free(p);
	my_free(q);
	my_free(r);
 }

void makeFamilyOfQuaternions	( double theta, double *v, quaternion* qf ) 
{
	double		z[] = {0,0,1}; 
	double		crossZV[3];
	double		len;
	double		phi = 0.3;
	//double		mat[16];
	//double		t[3],r[3],s[3];
	quaternion	*qq,*qb;

	qq	 = (quaternion*) my_malloc(sizeof(quaternion));
	qb	 = (quaternion*) my_malloc(sizeof(quaternion));

	//////////////////////////////////////////////////////////////////

	////////
	// qq //
	////////
	//printf("theta = %f\n",theta);
	//printf("v = %f, %f, %f\n",v[0],v[1],v[2]);

	quatBuildQuaternion(theta,v,qq);
	//printf("qq = %f, %f, %f %f\n",qq->r,qq->i,qq->j,qq->k);

	////////
	// qb //
	////////
	cross3D(z,v,crossZV);
	phi = acos(dot3D(z,v));
	len = length3D(crossZV);
	crossZV[0] /= len;
	crossZV[1] /= len;
	crossZV[2] /= len;
	quatBuildQuaternion(phi,crossZV,qb);
	//printf("qb = %f, %f, %f %f\n",qb->r,qb->i,qb->j,qb->k);

	////////
	// qf //
	////////
	qf->r = qq->r*qb->r - qq->i*qb->i - qq->j*qb->j - qq->k*qb->k;
	qf->i = qq->i*qb->r + qq->r*qb->i + qq->j*qb->k - qq->k*qb->j;
	qf->j = qq->j*qb->r + qq->r*qb->j + qq->k*qb->i - qq->i*qb->k;
	qf->k = qq->k*qb->r + qq->r*qb->k + qq->i*qb->j - qq->j*qb->i;
	//quatConjugate(qf);
	//printf("qf = %f, %f, %f %f, qnorm = %f\n",qf->r,qf->i,qf->j,qf->k,qNorm(qf));
	//////////////////////////////////////////////////////////////////
	my_free(qq);
	my_free(qb);

}


void drawRings ( void )
{
	int		i = 0;
	int		j = 0;
	double	t1[3];
	//double	t2[3], t[3];
	double	v[3];
	//double	s[3];
	//double	v0[3];
	
	////double len,	dot_vt;
	//double cross_vv0[3];
	//double cross_T1T2[3];
	//double theta = 0.0;
	//double phi	 = 0.0;

	//# Quat rings
	double	qtheta = 0.0;
	//double	qvec[3];
	//double	scalar=0.0;
	//double	angle = 0.0;
	//double  vecpart = 0.0;
	double	qvecpart[3];
	int		counter = 0;
	//double	normvec[3];
	double	delta = 0.2;
	int		renderPoints = mo2->numPoints-1;

	quaternion *p,*q,*u, *qf;
	p =	(quaternion*) malloc(sizeof(quaternion));
	q =	(quaternion*) malloc(sizeof(quaternion));
	u =	(quaternion*) malloc(sizeof(quaternion));
	qf = (quaternion*) malloc(sizeof(quaternion));

	if(mo2->numPoints >	2) {

		constructKnot();
		makeTangent();

		v[0] = mo2->TangList[0].p3.d[0];
		v[1] = mo2->TangList[0].p3.d[1];
		v[2] = mo2->TangList[0].p3.d[2];

		if( mo2->rshowFace || mo2->rshowCurve || mo2->rshowEdge ) {

			//PRELIMINARY
			qtheta = 0.0;
			j=0;
			while(qtheta<(4.0*M_PI)) {
				makeFamilyOfQuaternions(qtheta, v, qf);
				getQuatComp(mo2->qwhichQuatComponent, qf, qvecpart);
				prevList[j]->p3.d[0] = qvecpart[0];
				prevList[j]->p3.d[1] = qvecpart[1];
				prevList[j]->p3.d[2] = qvecpart[2];

				firstList[j]->p3.d[0] = qvecpart[0];
				firstList[j]->p3.d[1] = qvecpart[1];
				firstList[j]->p3.d[2] = qvecpart[2];
				qtheta = qtheta + delta;
				j++;
			}
			counter = j;

			// RENDER ALL RINGS
			for(i=0;i<(renderPoints);i++)	{
				
				t1[0] =	mo2->TangList[i].p3.d[0];
				t1[1] =	mo2->TangList[i].p3.d[1];
				t1[2] =	mo2->TangList[i].p3.d[2];

				if(i>=0) {

					qtheta = 0.0;
					j = 0;

					while(qtheta<(4.0*M_PI)) {
						makeFamilyOfQuaternions(qtheta, t1, qf);
						getQuatComp(mo2->qwhichQuatComponent, qf, qvecpart);
						currList[j]->p3.d[0] = qvecpart[0];
						currList[j]->p3.d[1] = qvecpart[1];
						currList[j]->p3.d[2] = qvecpart[2];
						qtheta = qtheta + delta;
						j++;
					}
					counter = j;
				
					// face
					if(mo2->rshowFace) {
						glPushAttrib(GL_ALL_ATTRIB_BITS);
						glEnable(GL_LIGHTING);
						glColor3f(0.0f,0.0f,0.0f);
						glLineWidth(0.5f);

						glEnable(GL_COLOR_MATERIAL);
						qtheta = 0.0;
						j=0;
						for(j=0;j<counter;j++) {
							glBegin(GL_QUADS);
								//if(j==0)
								//	glColor3f(0.1f,0.5f,0.7f);
								//else
									glColor3f(0.65f,0.3f,0.5f);
								if(j<(counter-1)) {
									glNormal3f(prevList[j]->p3.d[0],	prevList[j]->p3.d[1],	prevList[j]->p3.d[2]);
									glVertex3f(prevList[j]->p3.d[0],	prevList[j]->p3.d[1],	prevList[j]->p3.d[2]); 
									glVertex3f(prevList[j+1]->p3.d[0],	prevList[j+1]->p3.d[1],	prevList[j+1]->p3.d[2]); 
									glVertex3f(currList[j+1]->p3.d[0],	currList[j+1]->p3.d[1],	currList[j+1]->p3.d[2]); 
									glVertex3f(currList[j]->p3.d[0],	currList[j]->p3.d[1],	currList[j]->p3.d[2]); 
								}
								else {
									glNormal3f(prevList[0]->p3.d[0],	prevList[0]->p3.d[1],	prevList[0]->p3.d[2]);
									glVertex3f(prevList[0]->p3.d[0],	prevList[0]->p3.d[1],	prevList[0]->p3.d[2]); 
									glVertex3f(prevList[j]->p3.d[0],	prevList[j]->p3.d[1],	prevList[j]->p3.d[2]); 
									glVertex3f(currList[j]->p3.d[0],	currList[j]->p3.d[1],	currList[j]->p3.d[2]); 
									glVertex3f(currList[0]->p3.d[0],	currList[0]->p3.d[1],	currList[0]->p3.d[2]); 
								}
							glEnd();
						}
						if(i==(mo2->numPoints-1)) {
							qtheta = 0.0;
							j=0;
							for(j=0;j<counter;j++) {
								glBegin(GL_QUADS);
								if(j<(counter-1)) {
									glNormal3f(firstList[j]->p3.d[0],	firstList[j]->p3.d[1],	firstList[j]->p3.d[2]);
									glVertex3f(firstList[j]->p3.d[0],	firstList[j]->p3.d[1],	firstList[j]->p3.d[2]); 
									glVertex3f(firstList[j+1]->p3.d[0],	firstList[j+1]->p3.d[1],firstList[j+1]->p3.d[2]); 
									glVertex3f(currList[j+1]->p3.d[0],	currList[j+1]->p3.d[1],	currList[j+1]->p3.d[2]); 
									glVertex3f(currList[j]->p3.d[0],	currList[j]->p3.d[1],	currList[j]->p3.d[2]); 
								}
								else {
									glNormal3f(prevList[0]->p3.d[0],	prevList[0]->p3.d[1],	prevList[0]->p3.d[2]);
									glVertex3f(prevList[0]->p3.d[0],	prevList[0]->p3.d[1],	prevList[0]->p3.d[2]); 
									glVertex3f(prevList[j]->p3.d[0],	prevList[j]->p3.d[1],	prevList[j]->p3.d[2]); 
									glVertex3f(currList[j]->p3.d[0],	currList[j]->p3.d[1],	currList[j]->p3.d[2]); 
									glVertex3f(currList[0]->p3.d[0],	currList[0]->p3.d[1],	currList[0]->p3.d[2]); 
								}
								glEnd();
							}
						}
						glPopAttrib();
					}

					// edges
					if(mo2->rshowEdge) {
						glPushAttrib(GL_ALL_ATTRIB_BITS);
						glDisable(GL_LIGHTING);

                                                glEnable(GL_BLEND);
                                                glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
                                                glEnable(GL_LINE_SMOOTH);

						//polygon offset
						glEnable( GL_POLYGON_OFFSET_FILL );
						glPolygonOffset ( -3.0, -3.0 );
						glPolygonMode( GL_FRONT_AND_BACK, GL_LINE );
						glLineWidth(1.5f);
						if(mo2->rshowFace)
							glLineWidth(2.5f);
						else
							glLineWidth(0.5f);

						glColor3f(0.1f,0.1f,0.1f);
						for(j=0; j<counter; j++) {
							glBegin(GL_LINE_STRIP);
								if(j<(counter-1)) {
									glVertex3f(prevList[j]->p3.d[0],	prevList[j]->p3.d[1],	prevList[j]->p3.d[2]); 
									glVertex3f(prevList[j+1]->p3.d[0],	prevList[j+1]->p3.d[1],	prevList[j+1]->p3.d[2]); 
									glVertex3f(currList[j+1]->p3.d[0],	currList[j+1]->p3.d[1],	currList[j+1]->p3.d[2]); 
									glVertex3f(currList[j]->p3.d[0],	currList[j]->p3.d[1],	currList[j]->p3.d[2]); 
								}
								else {
									glVertex3f(prevList[0]->p3.d[0],	prevList[0]->p3.d[1],	prevList[0]->p3.d[2]); 
									glVertex3f(prevList[j]->p3.d[0],	prevList[j]->p3.d[1],	prevList[j]->p3.d[2]);
									glVertex3f(currList[j]->p3.d[0],	currList[j]->p3.d[1],	currList[j]->p3.d[2]); 
									glVertex3f(currList[0]->p3.d[0],	currList[0]->p3.d[1],	currList[0]->p3.d[2]); 
								}

							glEnd();
						}
						if(i==(mo2->numPoints-1)) {
							for(j=0;j<counter;j++) {
								glBegin(GL_LINE_STRIP);
								if(j<(counter-1)) {
									glNormal3f(firstList[j]->p3.d[0],	firstList[j]->p3.d[1],	firstList[j]->p3.d[2]);
									glVertex3f(firstList[j]->p3.d[0],	firstList[j]->p3.d[1],	firstList[j]->p3.d[2]); 
									glVertex3f(firstList[j+1]->p3.d[0],	firstList[j+1]->p3.d[1],firstList[j+1]->p3.d[2]); 
									glVertex3f(currList[j+1]->p3.d[0],	currList[j+1]->p3.d[1],	currList[j+1]->p3.d[2]); 
									glVertex3f(currList[j]->p3.d[0],	currList[j]->p3.d[1],	currList[j]->p3.d[2]); 
								}
								else {
									glNormal3f(prevList[0]->p3.d[0],	prevList[0]->p3.d[1],	prevList[0]->p3.d[2]);
									glVertex3f(prevList[0]->p3.d[0],	prevList[0]->p3.d[1],	prevList[0]->p3.d[2]); 
									glVertex3f(prevList[j]->p3.d[0],	prevList[j]->p3.d[1],	prevList[j]->p3.d[2]); 
									glVertex3f(currList[j]->p3.d[0],	currList[j]->p3.d[1],	currList[j]->p3.d[2]); 
									glVertex3f(currList[0]->p3.d[0],	currList[0]->p3.d[1],	currList[0]->p3.d[2]); 
								}
								glEnd();
							}
						}
						glPopAttrib();
					}

					// rings
					if(mo2->rshowCurve) {
						glPushAttrib(GL_ALL_ATTRIB_BITS);
						glDisable(GL_LIGHTING);
                                                
                                                glEnable(GL_BLEND);
                                                glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
                                                glEnable(GL_LINE_SMOOTH);

						
						//polygon offset
						glEnable( GL_POLYGON_OFFSET_FILL );
						glPolygonOffset ( -5.0, -5.0 );
						glPolygonMode( GL_FRONT_AND_BACK, GL_LINE );
						if(mo2->rshowFace)
							glLineWidth(2.5f);
						else
							glLineWidth(0.75f);

						if(i==0)
							glColor3f(0.9f, 0.9f, 0.0f);
						else
							glColor3f(0.2f,0.2f,0.2f);

						glBegin(GL_LINE_STRIP);
						for(j=0; j<=counter; j++) {

							if(j<counter)
								glVertex3f(currList[j]->p3.d[0],currList[j]->p3.d[1],currList[j]->p3.d[2]); 
							else
								glVertex3f(currList[0]->p3.d[0],currList[0]->p3.d[1],currList[0]->p3.d[2]); 
						}
						glEnd();
						glPopAttrib();
					}

					if(i>=0) { //COPY currList into prevList
						qtheta=0.0;
						j=0;
						while(qtheta<(4.0*M_PI)) {
							prevList[j]->p3.d[0] = currList[j]->p3.d[0];
							prevList[j]->p3.d[1] = currList[j]->p3.d[1];
							prevList[j]->p3.d[2] = currList[j]->p3.d[2];
							qtheta = qtheta + delta;
							j++;
						}	
					}

				}
			}
		}
	}
}








void drawQuaternionRingSpace ( void ) {
	//# Quat rings
	int		i = 0;
	int		j = 0;
	double	t[3];
	double	t1[3],t2[3];
	double	v[3],v0[3];
	
	double len,	dot_vt;

	double cross_T1T2[3];
	double theta = 0.0;
	double phi	 = 0.0;
	double	qtheta = 0.0;
	double	qvec[3];

	double	delta = 0.0;
	double	temp[3];
	int		counter = 0;
	quaternion *p,*q,*u, *qf, *b, *base;
	b = (quaternion*) my_malloc(sizeof(quaternion));
	p =	(quaternion*) malloc(sizeof(quaternion));
	q =	(quaternion*) malloc(sizeof(quaternion));
	u =	(quaternion*) malloc(sizeof(quaternion));
	qf = (quaternion*) malloc(sizeof(quaternion));
	base =	(quaternion*) malloc(sizeof(quaternion));

	constructKnot();
	makeTangent();

		glPushAttrib(GL_ALL_ATTRIB_BITS);
		glDisable(GL_LIGHTING);
		glColor3f(0,0,0);

		// Initialize
		t[0] = 0.0; 
		t[1] = 0.0;
		t[2] = 1.0;
		qvec[0] = qvec[1] = qvec[2] = 0.0;

		v[0] = mo2->TangList[0].p3.d[0];
		v[1] = mo2->TangList[0].p3.d[1];
		v[2] = mo2->TangList[0].p3.d[2];

		dot_vt = dot3D(t,v);
		phi	   = acos(dot_vt);
		
		cross3D(v,t,v0);
		len=length3D(v0);
		v0[0] /= len; v0[1] /= len; v0[2] /= len;
		
		p->r = 0.0;  
		p->i = 0.0; 
		p->j = 1.0; 
		p->k = 0.0;
		
		q->r =		cos(mo2->startTheta/2.0); 
		q->i = v[0]*sin(mo2->startTheta/2.0); 
		q->j = v[1]*sin(mo2->startTheta/2.0); 
		q->k = v[2]*sin(mo2->startTheta/2.0);

		quatQuatProduct(q, p, base);

		delta = 0.2;

		//qtheta = 0.0;
		//while ( qtheta <=4.0*M_PI+delta) {

		//	q->r =		cos(qtheta/2.0); 
		//	q->i = v[0]*sin(qtheta/2.0); 
		//	q->j = v[1]*sin(qtheta/2.0); 
		//	q->k = v[2]*sin(qtheta/2.0);
		//	quatQuatProduct(q, p, u);
		//	getQuatComp(mo2->qwhichQuatComponent,u,qvec);
		//	//makeFamilyOfQuaternions(qtheta, v, qf);
		//	//getQuatComp(mo2->qwhichQuatComponent,qf,qvec);
		//	prevList[counter]->p3.d[0] = qvec[0]; 
		//	prevList[counter]->p3.d[1] = qvec[1]; 
		//	prevList[counter]->p3.d[2] = qvec[2];

		//	firstList[counter]->p3.d[0] = qvec[0];
		//	firstList[counter]->p3.d[1] = qvec[1];
		//	firstList[counter]->p3.d[2] = qvec[2];

		//	qtheta += delta;
		//	counter++;
		//}//#end WHILE

		for(i=0; i<mo2->numPoints; i++) {
			qtheta	= 0.0;
			counter	= 0;
			j		= 0;
			v[0] = mo2->TangList[i].p3.d[0];
			v[1] = mo2->TangList[i].p3.d[1];
			v[2] = mo2->TangList[i].p3.d[2];

			if(i==0)
				glColor3f(1.0f,1.0f,0.0f);
			else
				glColor3f(0.0f,0.0f,0.0f);

			while ( qtheta <=4.0*M_PI+delta) {

				q->r =		cos(qtheta/2.0); 
				q->i = v[0]*sin(qtheta/2.0); 
				q->j = v[1]*sin(qtheta/2.0); 
				q->k = v[2]*sin(qtheta/2.0);
				quatQuatProduct(q, p, u);
				getQuatComp(mo2->qwhichQuatComponent,u,qvec);
				//makeFamilyOfQuaternions(qtheta, v, qf);
				//getQuatComp(mo2->qwhichQuatComponent,qf,qvec);

				if(qtheta==0) {
					temp[0] = qvec[0]; temp[1] = qvec[1]; temp[2] = qvec[2];
				}

				if(i>=0){
					currList[j]->p3.d[0] = qvec[0]; 
					currList[j]->p3.d[1] = qvec[1]; 
					currList[j]->p3.d[2] = qvec[2];
				}
				if(i==0) {
					prevList[j]->p3.d[0] = qvec[0]; 
					prevList[j]->p3.d[1] = qvec[1]; 
					prevList[j]->p3.d[2] = qvec[2];

					firstList[j]->p3.d[0] = qvec[0];
					firstList[j]->p3.d[1] = qvec[1];
					firstList[j]->p3.d[2] = qvec[2];
				}

				qtheta += delta;
				counter++;
				j++;
			}//# end WHILE

			glPushAttrib(GL_ALL_ATTRIB_BITS);
			glPushMatrix();
			glColor3f(0.2f,0.8f,0.0f);
			glTranslatef(temp[0],temp[1],temp[2]);
			glutSolidSphere(0.01,8,8);
			glPopMatrix();
			glPopAttrib();

			if(i>=0) {
				if(mo2->showLabels) {
					glPushAttrib(GL_ALL_ATTRIB_BITS);
					//glEnable(GL_LIGHTING);
                    glEnable(GL_BLEND);
                    glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
                    glEnable(GL_LINE_SMOOTH);
					if(i==0 && mo2->showLabels) {
						j=0;
						glColor3f(1.0f,1.0f,0.0f);
						glPushMatrix();
						glTranslatef(firstList[j]->p3.d[0],	firstList[j]->p3.d[1],	firstList[j]->p3.d[2]); 
						glutSolidSphere(0.015,16,16);
						sprintf(obj_qstr,"FIRST Ring Frame: %d",j);
						glTranslatef(0.05,0.05,0.05);
						glRasterPos2f(0.0,0.0);
						glDisable(GL_LIGHTING);
						stext(obj_qstr);
						glPopMatrix();
					}
					else if(i==mo2->numPoints-1 && mo2->showLabels) {
						j=0;
						glColor3f(0.0f,1.0f,1.0f);
						glPushMatrix();
						glTranslatef(currList[j]->p3.d[0],	currList[j]->p3.d[1],	currList[j]->p3.d[2]); 
						glutSolidSphere(0.015,16,16);
						sprintf(obj_qstr,"LAST Ring Frame: %d",j);
						glTranslatef(0.05,0.05,0.05);
						glRasterPos2f(0.0,0.0);
						glDisable(GL_LIGHTING);
						stext(obj_qstr);
						glPopMatrix();
					}
					glPopAttrib();
				}
				if( mo2->rshowFace || mo2->rshowCurve || mo2->rshowEdge ) {

					//@ Render Rings
					if(mo2->rshowCurve) {
						glPushAttrib(GL_ALL_ATTRIB_BITS);
						glDisable(GL_LIGHTING);
                                                
						if(mo2->rshowFace)
							glDisable(GL_BLEND);
						else {
							glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
							glEnable(GL_LINE_SMOOTH);
							glEnable(GL_BLEND);
						}
						
						//polygon offset
						glEnable( GL_POLYGON_OFFSET_FILL );
						glPolygonOffset ( -5.0, -5.0 );
						glPolygonMode( GL_FRONT_AND_BACK, GL_LINE );
						if(mo2->rshowFace)
							glLineWidth(2.5f);
						else
							glLineWidth(1.0f);

						if(i==0) {
							glColor3f(1.0f,1.0f,0.0f);
						}
						else if(i==mo2->numPoints-1) {
							glColor3f(0.0f,1.0f,1.0f);
						}
						else {
							glColor3f(0.2f,0.2f,0.2f);
						}
					
						if(mo2->numOfRings == 0) {
							if( (i==0 || i==mo2->numPoints-1)) {
							glBegin(GL_LINE_STRIP);
							for(j=0; j<=counter; j++) {

								if(j<counter)
									glVertex3f(currList[j]->p3.d[0],currList[j]->p3.d[1],currList[j]->p3.d[2]); 
								else
									glVertex3f(currList[0]->p3.d[0],currList[0]->p3.d[1],currList[0]->p3.d[2]); 
							}
							glEnd();
							}
						}
						else if(i%( 5-mo2->numOfRings) == 0 || (i==0 || i==mo2->numPoints-1)) {
							glBegin(GL_LINE_STRIP);
							for(j=0; j<=counter; j++) {

								if(j<counter)
									glVertex3f(currList[j]->p3.d[0],currList[j]->p3.d[1],currList[j]->p3.d[2]); 
								else
									glVertex3f(currList[0]->p3.d[0],currList[0]->p3.d[1],currList[0]->p3.d[2]); 
							}
							glEnd();
						}
						glPopAttrib();
					}

					//@ Render Faces
					if(mo2->rshowFace) {
						glPushAttrib(GL_ALL_ATTRIB_BITS);
						glEnable(GL_LIGHTING);
						glColor3f(0.0f,0.0f,0.0f);
						glLineWidth(0.5f);

						glEnable(GL_COLOR_MATERIAL);
						qtheta = 0.0;
							
						if(i<(mo2->numPoints-1))
							glColor3f(0.65f,0.3f,0.5f);
						else
							glColor3f(0.1f,0.4f,0.8f);
	
						if(i<(mo2->numPoints)) {
							for(j=0;j<counter;j++) {
								glColor3f(0.65f,0.3f,0.5f);
								glBegin(GL_QUADS);
									if(j<(counter-1)) {
										glNormal3f(prevList[j]->p3.d[0],	prevList[j]->p3.d[1],	prevList[j]->p3.d[2]);
										glVertex3f(prevList[j]->p3.d[0],	prevList[j]->p3.d[1],	prevList[j]->p3.d[2]); 
										glVertex3f(prevList[j+1]->p3.d[0],	prevList[j+1]->p3.d[1],	prevList[j+1]->p3.d[2]); 
										glVertex3f(currList[j+1]->p3.d[0],	currList[j+1]->p3.d[1],	currList[j+1]->p3.d[2]); 
										glVertex3f(currList[j]->p3.d[0],	currList[j]->p3.d[1],	currList[j]->p3.d[2]); 
									}
									else {
										glNormal3f(prevList[0]->p3.d[0],	prevList[0]->p3.d[1],	prevList[0]->p3.d[2]);
										glVertex3f(prevList[0]->p3.d[0],	prevList[0]->p3.d[1],	prevList[0]->p3.d[2]); 
										glVertex3f(prevList[j]->p3.d[0],	prevList[j]->p3.d[1],	prevList[j]->p3.d[2]); 
										glVertex3f(currList[j]->p3.d[0],	currList[j]->p3.d[1],	currList[j]->p3.d[2]); 
										glVertex3f(currList[0]->p3.d[0],	currList[0]->p3.d[1],	currList[0]->p3.d[2]); 
									}
								glEnd();
								//if(i==(mo2->numPoints-1) ){
								//	glPushMatrix();
								//	glTranslatef(currList[j]->p3.d[0],	currList[j]->p3.d[1],	currList[j]->p3.d[2]); 
								//	glColor3f(0.9f,0.0f,0.0f);
								//	glutSolidSphere(0.01,16,16);
								//	glPopMatrix();
								//	glPushMatrix();
								//	glTranslatef(firstList[j]->p3.d[0],	firstList[j]->p3.d[1],	firstList[j]->p3.d[2]); 
								//	glColor3f(0.0f,0.9f,0.0f);
								//	glutSolidSphere(0.01,16,16);
								//	glPopMatrix();
								//	if(j==(counter-1)) {
								//	glBegin(GL_QUADS);
								//		glNormal3f(firstList[0]->p3.d[0],	firstList[0]->p3.d[1],	firstList[0]->p3.d[2]);
								//		glVertex3f(currList[j]->p3.d[0],	currList[j]->p3.d[1],	currList[j]->p3.d[2]); 
								//		glVertex3f(currList[0]->p3.d[0],	currList[0]->p3.d[1],	currList[0]->p3.d[2]); 
								//		glVertex3f(firstList[0]->p3.d[0],	firstList[0]->p3.d[1],	firstList[0]->p3.d[2]); 
								//		glVertex3f(firstList[j]->p3.d[0],	firstList[j]->p3.d[1],	firstList[j]->p3.d[2]); 
								//	glEnd();
								//	}
								//}

								//if(i==(mo2->numPoints-1)) {
									//glBegin(GL_QUADS);
									//if(j<(counter-1)) {
										//glNormal3f(firstList[j]->p3.d[0],	firstList[j]->p3.d[1],	firstList[j]->p3.d[2]);
										//glVertex3f(firstList[j]->p3.d[0],	firstList[j]->p3.d[1],	firstList[j]->p3.d[2]); 
										//glVertex3f(firstList[j+1]->p3.d[0],	firstList[j+1]->p3.d[1],firstList[j+1]->p3.d[2]); 
										//glVertex3f(currList[j+1]->p3.d[0],	currList[j+1]->p3.d[1],	currList[j+1]->p3.d[2]); 
										//glVertex3f(currList[j]->p3.d[0],	currList[j]->p3.d[1],	currList[j]->p3.d[2]); 
									//}
									//else {
									//	glNormal3f(prevList[0]->p3.d[0],	prevList[0]->p3.d[1],	prevList[0]->p3.d[2]);
									//	glVertex3f(prevList[0]->p3.d[0],	prevList[0]->p3.d[1],	prevList[0]->p3.d[2]); 
									//	glVertex3f(prevList[j]->p3.d[0],	prevList[j]->p3.d[1],	prevList[j]->p3.d[2]); 
									//	glVertex3f(currList[j]->p3.d[0],	currList[j]->p3.d[1],	currList[j]->p3.d[2]); 
									//	glVertex3f(currList[0]->p3.d[0],	currList[0]->p3.d[1],	currList[0]->p3.d[2]); 
									//}
									//glEnd();	
								//}
							}

						}
						glPopAttrib();
					}
					//@ Render Edges or Mesh
					if(mo2->rshowEdge) {
						glPushAttrib(GL_ALL_ATTRIB_BITS);
						glDisable(GL_LIGHTING);

						//glEnable(GL_BLEND);
						//glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
						//glEnable(GL_LINE_SMOOTH);

						//polygon offset
						glEnable( GL_POLYGON_OFFSET_FILL );
						glPolygonOffset ( -3.0, -3.0 );
						glPolygonMode( GL_FRONT_AND_BACK, GL_LINE );
						glLineWidth(1.5f);
						if(mo2->rshowFace)
							glLineWidth(2.5f);
						else
							glLineWidth(0.5f);

						glColor3f(0.2f,0.2f,0.2f);
						for(j=0; j<counter; j++) {
							glBegin(GL_LINE_STRIP);
								if(j<(counter-1)) {
									glVertex3f(prevList[j]->p3.d[0],	prevList[j]->p3.d[1],	prevList[j]->p3.d[2]); 
									glVertex3f(prevList[j+1]->p3.d[0],	prevList[j+1]->p3.d[1],	prevList[j+1]->p3.d[2]); 
									glVertex3f(currList[j+1]->p3.d[0],	currList[j+1]->p3.d[1],	currList[j+1]->p3.d[2]); 
									glVertex3f(currList[j]->p3.d[0],	currList[j]->p3.d[1],	currList[j]->p3.d[2]); 
								}
								else {
									glVertex3f(prevList[0]->p3.d[0],	prevList[0]->p3.d[1],	prevList[0]->p3.d[2]); 
									glVertex3f(prevList[j]->p3.d[0],	prevList[j]->p3.d[1],	prevList[j]->p3.d[2]);
									glVertex3f(currList[j]->p3.d[0],	currList[j]->p3.d[1],	currList[j]->p3.d[2]); 
									glVertex3f(currList[0]->p3.d[0],	currList[0]->p3.d[1],	currList[0]->p3.d[2]); 
								}

							glEnd();
						}
						//if(i==(mo2->numPoints-1)) {
						//	for(j=0;j<counter;j++) {
						//		glBegin(GL_LINE_STRIP);
						//		if(j<(counter-1)) {
						//			glNormal3f(firstList[j]->p3.d[0],	firstList[j]->p3.d[1],	firstList[j]->p3.d[2]);
						//			glVertex3f(firstList[j]->p3.d[0],	firstList[j]->p3.d[1],	firstList[j]->p3.d[2]); 
						//			glVertex3f(firstList[j+1]->p3.d[0],	firstList[j+1]->p3.d[1],firstList[j+1]->p3.d[2]); 
						//			glVertex3f(currList[j+1]->p3.d[0],	currList[j+1]->p3.d[1],	currList[j+1]->p3.d[2]); 
						//			glVertex3f(currList[j]->p3.d[0],	currList[j]->p3.d[1],	currList[j]->p3.d[2]); 
						//		}
						//		else {
						//			glNormal3f(prevList[0]->p3.d[0],	prevList[0]->p3.d[1],	prevList[0]->p3.d[2]);
						//			glVertex3f(prevList[0]->p3.d[0],	prevList[0]->p3.d[1],	prevList[0]->p3.d[2]); 
						//			glVertex3f(prevList[j]->p3.d[0],	prevList[j]->p3.d[1],	prevList[j]->p3.d[2]); 
						//			glVertex3f(currList[j]->p3.d[0],	currList[j]->p3.d[1],	currList[j]->p3.d[2]); 
						//			glVertex3f(currList[0]->p3.d[0],	currList[0]->p3.d[1],	currList[0]->p3.d[2]); 
						//		}
						//		glEnd();
						//	}
						//}
						glPopAttrib();
					}
				}//#end IF(mo2->rshowface || mo2->rshowCurve || mo2->rshowEdge)

			}//# IF(i>=0)

			for(j=0;j<counter;j++) {
				prevList[j]->p3.d[0] = currList[j]->p3.d[0];
				prevList[j]->p3.d[1] = currList[j]->p3.d[1];
				prevList[j]->p3.d[2] = currList[j]->p3.d[2];
			}
			//#(DEBUG ONLY) displays beads on the rings 
			//glEnable(GL_LIGHTING);
			//qtheta = 0.0;
			//while ( qtheta <=4.0*M_PI+0.2) {
			//	//makeFamilyOfQuaternions(qtheta,v,qf);
			//	q->r =		cos(qtheta/2.0); 
			//	q->i = v[0]*sin(qtheta/2.0); 
			//	q->j = v[1]*sin(qtheta/2.0); 
			//	q->k = v[2]*sin(qtheta/2.0);
			//	quatQuatProduct(q, p, u);

			//	getQuatComp(mo2->qwhichQuatComponent,u,qvec);

			//	glPushMatrix();
			//	glTranslatef(qvec[0],qvec[1],qvec[2]);
			//	glutSolidSphere(0.01,8,8);
			//	glPopMatrix();
			//	qtheta += 0.2;
			//}//#end WHILE

			//update base ref frame
			t1[0] =	mo2->TangList[i].p3.d[0];	t1[1] =	mo2->TangList[i].p3.d[1];	t1[2] =	mo2->TangList[i].p3.d[2];
			t2[0] =	mo2->TangList[i+1].p3.d[0]; t2[1] =	mo2->TangList[i+1].p3.d[1]; t2[2] =	mo2->TangList[i+1].p3.d[2];
			cross3D(t1,t2, cross_T1T2);
			len	= length3D(cross_T1T2);
			if(len!=0) {
				cross_T1T2[0] /= len;
				cross_T1T2[1] /= len;
				cross_T1T2[2] /= len;
				theta =	acos(dot3D(t1,t2));
				
				base->r = p->r;
				base->i = p->i;
				base->j = p->j;
				base->k = p->k;
				q->r = cos(theta/2.0);
				q->i = cross_T1T2[0]*sin(theta/2.0);
				q->j = cross_T1T2[1]*sin(theta/2.0);
				q->k = cross_T1T2[2]*sin(theta/2.0);
				quatQuatProduct ( q, base, p);
				quatNormalize(p);
			}

		}//#end FOR

		glPopAttrib();
}